function unno,vfin,a=a,eta0mg=eta0,eta0nm=eta0nm,doppler=dopplerl, $
                bin=bin,glande=glande,x1=x1, $
                b0mg=b0,b0nm=b0nm,b1=b1, $
                Btotal=Btotal,gamma=gammain,chi=chiin,mu=mu, $
                wave=wavein,fill=fill,noanomalous=noanomalous

;+
;NAME:
;     UNNO
;PURPOSE:
;     Computes unno stokes profiles
;CATEGORY:
;CALLING SEQUENCE:
;     result = unno(v)
;INPUTS:
;     v = wave from line center divided by doppler width
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     wave = wavelength of line center in A (def = 6301.5)
;     a = voigt parameter (def = 0.4)
;     doppler = doppler width in mA (def = 30.0)
;     eta0mg = absorption coefficient (def = 10.)
;     eta0nm = abs oceff for non-magnetic line (def = eta0)
;     b0mg,b1 = coeff of linear source function (def = 1000.,5000.)
;     b0nm = b0 for non-magnetic line (def = b0)
;     glande = magnetic g for the line (def=1.667)
;     Btotal = total magnetic field in G (def=0)
;     gamma = field inclination (degrees, def=0)
;     chi = field azimuth (degrees, def=0)
;     mu = cosine of position angle on the sun (def = 1)
;     fill = filling factor (def = 1)
;     bin = rebin the data by this factor before output (def = 1)
;     /noanomalous = do not include faraday rotation etc.
;OUTPUTS:
;     result = structure holding the Stokes profile, absorption coeff etc.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
;MODIFICATION HISTORY:
;     T. Metcalf 01-Feb-1994
;-


   vf = vfin
   if n_elements(bin) NE 1 then BIN = 1  & BIN = fix(BIN)


   NPOINTS = n_elements(vf)

   if n_elements(B0) NE 1 then B0 = 1000.
   if n_elements(B1) NE 1 then B1 = 5000.
   if n_elements(b0nm) NE 1 then b0nm = b0
   if n_elements(a) NE 1 then a = 0.4
   if n_elements(dopplerl) NE 1 then dopplerl = 30.
   if N_elements(eta0) NE 1 then eta0 = 10.
   if n_elements(eta0nm) NE 1 then eta0nm = eta0

   if n_elements(Btotal) NE 1 then Btotal = 0.
   if n_elements(gammain) NE 1 then gammain = 0.
   if n_elements(chiin) NE 1 then chiin = 0.
   if n_elements(mu) NE 1 then mu = 1.
   if n_elements(fill) NE 1 then fill = 1.

   gamma = gammain*!dtor
   chi = chiin*!dtor

   ;print,'Input Btotal gamma chi mu fill'
   ;read,Btotal,gamma,chi,mu,fill

   ;print,B0,B1,a,dopplerl,Btotal,gamma,chi,mu,eta0,fill

;  Convert Doppler width (mA) to frequency units (Hz):

   clight = 2.9979e10   ; cm/s
   elec = 4.803e-10     ; esu
   me = 9.110e-28       ; g

   if n_elements(wavein) NE 1 then wave = 6301.5e-8 $     ; cm
   else wave = wavein*1.0e-8
   if n_elements(glande) NE 1 then glande = 1.667    ;  6301=1.667   6302=2.5

   doppler = (clight/wave^2)*dopplerl*1.0e-11 ; Hz

   larmor = (glande*elec*Btotal)/(4.0*(!pi)*me*clight)

   ;print,'vb=',larmor/doppler

   etaIQUV = eta(eta0, vf, a, larmor/doppler, gamma, chi)
   save_eta = etaIQUV

   n = n_elements(vf)

   etaI = etaIQUV.etaI
   etaQ = etaIQUV.etaQ
   etaU = etaIQUV.etaU
   etaV = etaIQUV.etaV

   if keyword_set(noanomalous) then begin 
      rhoQ = 0.
      rhoU = 0.
      rhoV = 0.
   endif else begin
      rhoQ = etaIQUV.rhoQ
      rhoU = etaIQUV.rhoU
      rhoV = etaIQUV.rhoV
   endelse

   delta = etaIQUV.Delta

   Im = B0 + (mu*B1/delta)*(1+etaI)*((1+etaI)^2+rhoQ^2+rhoU^2+rhoV^2)

   Q = -fill*(mu*B1/delta)*((1+etaI)^2*etaQ+(1+etaI)*(etaV*rhoU-etaU*rhoV) + $
         rhoQ*(etaQ*rhoQ + etaU*rhoU + etaV*rhoV))

   U = -fill*(mu*B1/delta)*((1+etaI)^2*etaU+(1+etaI)*(etaQ*rhoV-etaV*rhoQ) + $
         rhoU*(etaQ*rhoQ + etaU*rhoU + etaV*rhoV))

   V = -fill*(mu*B1/delta)*((1+etaI)^2*etaV + $
              rhoV*(etaQ*rhoQ+etaU*rhoU+etaV*rhoV))

   ; Get non-magnetic I profile

   etaIQUV = eta(eta0nm, vf-x1, a, 0., 0., 0.)  ; No field
   etaI = etaIQUV.etaI
   Inm = B0nm + mu*B1/(1+etaI)

   I = fill*Im + (1.0-fill)*Inm

   ; Rebin

   vf = rebin(vf,n/BIN)
   Im = rebin(Im,n/BIN)
   Inm = rebin(Inm,n/BIN)
   Icont = rebin(I(n_elements(I)-1-9:n_elements(I)-1),1)
   Icont = Icont(0)
   I = rebin(I,n/BIN)
   Q = rebin(Q,n/BIN)
   U = rebin(U,n/BIN)
   V = rebin(V,n/BIN)

   Rv = fltarr(n_elements(V))
   Rq = Rv
   G = Rv
   dIdv = deriv(vf,I)
   vbcosg = fill*(larmor/doppler)*cos(gamma)
   voigt,a,vf,h,f
   dhdv = 2.0*(a*f-vf*h)
   d2hdv2 = 4.0*[(vf^2-a^2)*h - 2.0*a*vf*f - h/2.0 + a*0.56418958]
   good = where(dhdv NE 0.)
   ratio = fltarr(n_elements(V))
   ratio(good) = d2hdv2(good)/dhdv(good)
   vbsing = fill*((larmor/doppler)*sin(gamma)/2.0)^2*ratio
   good = where(dIdv NE 0.)
   if vbcosg NE 0. then begin
      if good(0) GE 0 then $
         Rv(good) = V(good)/(vbcosg*dIdv(good))
   endif
   good = where(dIdv NE 0. and vbsing NE 0.)
   if good(0) GE 0 then $
      Rq(good) = sqrt(Q(good)^2+U(good)^2)/abs(vbsing(good)*dIdv(good))
   G = ratio*vf

   save_eta.etaI = rebin(save_eta.etaI,n/BIN)
   save_eta.etaQ = rebin(save_eta.etaQ,n/BIN)
   save_eta.etaU = rebin(save_eta.etaU,n/BIN)
   save_eta.etaV = rebin(save_eta.etaV,n/BIN)
   save_eta.rhoQ = rebin(save_eta.rhoQ,n/BIN)
   save_eta.rhoU = rebin(save_eta.rhoU,n/BIN)
   save_eta.rhoV = rebin(save_eta.rhoV,n/BIN)
   save_eta.Delta = rebin(save_eta.Delta,n/BIN)

   return,add2str(dummy,['I','Q','U','V', $
                  'vf','fill','Btotal','gamma','chi','mu','vb', $
                  'B_obs_long','B_obs_trans', $
                  'wave','glande','voigt_p','doppler', $
                  'Im','Inm','Icont','eta','B0','B1','Rv','Rq','G'], $
                  i,q,u,v, $
                  vf,fill,btotal,gamma,chi,mu,larmor/doppler, $
                  Btotal*cos(gamma)*fill,Btotal*sin(gamma)*sqrt(fill), $
                  wave*1.0e8,glande,a,dopplerl, $
                  Im,Inm,Icont,save_eta,B0,B1,Rv,Rq,g)

end
