
function bparallel,length, current
   common bparallel_private,x,y,xh,yh,dxidX,dxidY,dyidX,dyidY,s,j,alpha,noalpha, $
                             count,p,q,idbpxd,idbpyd,debug,nx,ny

   if count EQ 0 then reset=1 else reset=0
   count = count + 1L

   ; Get heliographic and image coordinates of this point.

   xxh = interpol(xh,s,length)  ; heliographic, must use *linear* interpolation!
   yyh = interpol(yh,s,length)

   ; Get the image coordinates for this xxh and yyh.  This
   ; is required so that we know where to interpolate the 
   ; magnetic field to.
   if n_elements(dxidX) GT 1 then begin
      xx1 = (interpol(x,s,length)>0)<(nx-1)  ; initial guess
      yy1 = (interpol(y,s,length)>0)<(ny-1)
      ; now improve the guess with the transform matrices for
      ; xx1,yy1
      xx = dxidX[xx1,yy1]*xxh + dxidY[xx1,yy1]*yyh
      yy = dyidX[xx1,yy1]*xxh + dyidY[xx1,yy1]*yyh
  endif else begin
      xx = dxidX*xxh + dxidY*yyh
      yy = dyidX*xxh + dyidY*yyh
  endelse

   ; Plot location of this point
   if keyword_set(debug) then begin
      pq2xy,xxh,yyh,p,q,xxp,yyp
      plots,xxp-min(idbpxd),yyp-min(idbpyd),psym=1,/device
      empty
   endif

   ; Get B*dl
   answer = dblarr(n_elements(length))
   for i=0,n_elements(length)-1 do begin
      ; get unit vector along the contour of integration
      diff = min(abs(s-length[i]),ss)
      if s[ss] GT length[i] OR ss GE n_elements(s)-1 then ss = (ss-1) > 0
      dx = xh[ss+1]-xh[ss]
      dy = yh[ss+1]-yh[ss]
      lunit = [dx,dy]/sqrt(dx^2+dy^2)
      b = bicubic2d(j,xx[i],yy[i],alpha,reset=reset,noalpha=noalpha,/quiet)
      if keyword_set(debug) then $
         print,strcompress('Bz('+string(length)+')='+string(b[2]))
      answer[i] = b[0]*lunit[0]+b[1]*lunit[1]
   endfor

   return, answer 

end

;+

function ampere, jin, level=levelin, ss=ssin, alpha=alphain, $
                 verbose=verbosein,error=error, area=carea

;NAME:
;     AMPERE.PRO
;PURPOSE:
;     Compute the current flowing through a closed contour over a
;     vector magnetogram by applying Ampere's Law.
;CATEGORY:
;CALLING SEQUENCE:
;     current = ampere(magnetogram)
;INPUTS:
;     magnetogram = magnetogram structure with at least BX, BY, BZ,
;                   and a point structure.  the point structure must 
;                   have at least the pix_size array giving the pixel
;                   sizes in arcsec, b0, p, lat and cmd.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS:
;     ss = array of indices which fill the region of the
;          magnetogram to be considered.  If keyword "level" (below) is *not*
;          specified, then the contour of integration is the boundary
;          of the region specified by ss.  If ss is not set, then the
;          user will be asked to set a region using wdefroi.
;     level = Magnetic field level which the contour should follow.
;             The integration contour will follow this field level
;             only within the region specified by ss.  This is a
;             *signed* number.  So if you want a contour at -500 G,
;             level should be -500.
;     error = returns with an estimate of the error on the current in
;             Amps.
;     alpha = Alpha value to use in interpolating the vector field
;             between pixels.  If this is not set, then alpha is not
;             used in the interpolation (which is different from
;             setting alpha to zero).
;     area = returns with the area inside the closed contour (sq. meters)
;     /verbose = output some diagnostics and plots along the way.
;OUTPUTS:
;     current = the net current flowing through the contour of
;               integration in Amps.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The integration path is defined by points in the image plane,
;     but these points are connected by straight lines in heliographic
;     coordinates.  So, for long segments, there may be a bit of
;     distortion.
;
;     The pix_size in the magnetogram point structure must have units
;     of arcsec in order for the units of the result to be correct.
;PROCEDURE:
;     Uses the IDL RK4 Runge-Kutta integration with the adaptive
;     stepsize described in Numerical Recipes.
;MODIFICATION HISTORY:
;     T. Metcalf 2004-Sep-20
;     T. Metcalf 2004-Dec-03 Corrected factor of 10 in units (!). Added
;                            area keyword.  Corrected sign (!) of
;                            current using signed area.  Before this
;                            change, the current had the wrong sign
;                            and was 10 times too large.  Duh.
;-

   common bparallel_private,x,y,xh,yh,dxidX,dxidY,dyidX,dyidY,s,j,alpha,noalpha, $
                             count,p,q,idbpxd,idbpyd,debug,nx,ny

   debug = 0

   j = jin
   noalpha = 0
   if n_elements(alphain) GT 0 then alpha=float(alphain) $
   else noalpha = 1
   if keyword_set(verbosein) or keyword_set(debug) then verbose=1 else verbose=0
   if n_elements(ssin) LE 0 then begin
      print, "Select integration region."
      ssa = wdefroi(j.bz)
      ssin = ssa
   endif else ssa =ssin

   ; initialization

   count = 0L
   sbz = size(j.bz)
   nx = sbz[1]
   ny = sbz[2]

   ; Get coordinate transformations

   b0 = double(j.point.b0)
   p = double(j.point.p)
   if tag_index(j,'CMD') GE 0 then lc = double(j.cmd)*!dtor $
   else lc = double(j.point.cmd)
   if tag_index(j,'LATITUDE') GE 0 then bc = double(j.latitude)*!dtor $
   else bc = 	double(j.point.lat)

   cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
   cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
         - COS(bc)*COS(b0)*SIN(p) 
   cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
   cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
         + COS(bc)*COS(b0)*COS(p)

   dxidX = cxx   ; At pixel corners, not centers
   dyidX = cyx
   dxidY = cxy
   dyidY = cyy

   determ = cxx*cyy-cyx*cxy  ; Get image to helio transformation
   icxx = cyy/determ
   icxy = -cxy/determ
   icyx = -cyx/determ
   icyy = cxx/determ

   dXdxi = icxx
   dYdxi = icyx
   dXdyi = icxy
   dYdyi = icyy

   pix_size = double(j.point.pix_size)
   if n_elements(pix_size) EQ 1 then $
      pix_size = [pix_size(0),pix_size(0)]
   mpix_size = min(pix_size)

   x = (lindgen(nx,ny) MOD nx)*(pix_size(0)/mpix_size); image pixel coordinates
   y = (lindgen(nx,ny) / nx)*(pix_size(1)/mpix_size) 
   bpxd = dXdxi*x + dXdyi*y           ; Heliographic pixel coordinates
   bpyd = dYdxi*x + dYdyi*y

   ; Find the edges of the selected region

   if n_elements(levelin) GT 0 then begin
      mask = fltarr(nx,ny)
      if n_elements(ssa) GT 0 then mask[ssa] = 1.0 else mask[*]=1.0
      select = j.bz * mask
      level = levelin[0]
      lsign = 1
      if level LT 0 then begin
         lsign = -1
         level = abs(level)  ; to make contour work right need positive level
         select = -select
      endif
   endif else begin
      select = bytarr(nx,ny)
      select[ssa] = 1
      level = 1
   endelse
   contour,select,level=level,path_xy=path_xy, $
                            path_info=path_info, $
                            /path_data_coords
   if n_elements(path_xy) LE 0 then message,'ERROR: no '+string(level*lsign)+' contour found'
   if n_elements(path_info) GT 1 then begin
      ssclosed = where(path_info.type EQ 1,ncontours) ; close contours
      if ncontours LE 0 then message,'ERROR: no closed contour found'
      if ncontours EQ 1 then sscontour = ssclosed[0]
      if ncontours GT 1 then begin
         message,/info,'WARNING: there are multiple closed contours.  ' + $
                       'Using the longest contour.'
         clength = 0.0
         sscontour = ssclosed[0]
         for i=0,ncontours-1 do begin ; find longest contour
            ssc = ssclosed[i]
            length = 0.0d0
            n = path_info[ssc].n
            o = path_info[ssc].offset
            for t=1L,n-1L do begin 
               dx = path_xy[0,o+t]-path_xy[0,o+t-1]
               dy = path_xy[1,o+t]-path_xy[1,o+t-1]
               length = length + sqrt(dx^2+dy^2)
            endfor
            ;tvscl,j.bz
            ;print,i,length,ssc,sscontour
            ;plots,path_xy[*,o:o+n-1],/device,color=128
            if length GE clength then begin
               sscontour = ssc
               clength = length
            endif
         endfor
      endif
      o = path_info[sscontour].offset
      n = path_info[sscontour].n
      path_xy = path_xy[*,o:o+n-1]
   endif
   path_xy = [[path_xy],[path_xy(*,0)]] ; close the contour
   if keyword_set(verbose) then begin
      tvscl,j.bz
      plots,path_xy,/device
      empty
   endif

   spath = size(path_xy)
   n = spath[2]

   ; Transformation from image to helio

   x = reform(path_xy[0,*])  ; in image coordinates
   y = reform(path_xy[1,*])

   ; Get the heliographic coordinates of the integration path

   xip = reform(path_xy[0,*])
   yip = reform(path_xy[1,*])
   xh = dXdxi[xip,yip]*xip + dXdyi[xip,yip]*yip
   yh = dYdxi[xip,yip]*xip + dYdyi[xip,yip]*yip

   if keyword_set(verbose) then begin
      ; Show integration path in heliographic coordinates
      tvscl,ideproject(j.bz,j.point,b_pxd=idbpxd,b_pyd=idbpyd)
      polywarp,idbpxd,idbpyd,bpxd,bpyd,2,p,q
      pq2xy,xh,yh,p,q,xp,yp
      ; Straight lines from point to point in heliographic
      ; coordinates - same as integration.
      plots,xp-min(idbpxd),yp-min(idbpyd),/device
      empty
   endif

   ; The path is defined here as straight lines connecting each point
   ; *in heliographic coordinates*.  They'd appear a bit curved then
   ; in image coordinates.  We need to get the path length since the
   ; path length is the itegration variable.

   s = dblarr(n)  ; path length
   carea = poly_area(xh,yh,/double,/sign)  ; area inside contour
   if carea NE 0.0 then sarea = carea/abs(carea) else sarea=1.0d0 ; sign of area
   carea = abs(carea)
   for i=1L,n-1L do begin
      dx = xh[i]-xh[i-1]
      dy = yh[i]-yh[i-1]
      s[i] = s[i-1] + sqrt(dx^2+dy^2)
   endfor

   if keyword_set(verbose) then begin
      print,'Path length = ',s[n-1],' pixels'
   endif

   ; Integrate along the selected path using Runge-Kutta and
   ; the variable step size described in Numerical Recipes.

   current = 0.0d0
   length = 0.0d0
   step = s[n-1] ; start with entire path and let the adaptive stepsize do its thing
   done = 0

   REPEAT begin
      hstep = step/2.0
      Bp = bparallel(length)
      Bp2 = bparallel(length+hstep)
      Delta0 = abs(1.e-4*max(abs([Bp,Bp2]))*step)
      ; rk4 takes the value of the current and adds the next step
      ; in the integral to it.
      c1 = rk4(current,Bp,length,step,'bparallel',/double)
      c2 = rk4(current,Bp,length,hstep,'bparallel',/double)
      c3 = rk4(c2,Bp2,length+hstep,hstep,'bparallel',/double)
      Delta1 = abs(c3-c1)
      if Delta1 EQ 0. then Delta1=Delta0
      ; This is the adaptive stepsize from Numerical Recipes
      if Delta0 GE Delta1 OR step LE 0.01 then begin
         length = length + step  ; accept this step
         if length GE s[n-1] then done=1
         step = 0.9*step*abs(Delta0/Delta1)^0.20
         if length+step GT s[n-1] and NOT done then begin
            step = s[n-1]-length         
         endif
         current = c3 ; save current and move to next step
      endif else begin
         ; step size too large, do not accept this step
         step = 0.9*step*abs(Delta0/Delta1)^0.25
      endelse
      if step LT 0.01 and NOT done then step = 0.01
   endrep UNTIL done
 
   ; the units are computed as follows:
   ; current = integral(B*dl), so Gauss*pixels is the output
   ;           of rk4.
   ; 721.9e3*pxlength,  pixels->meters
   ; 10e-4, Gauss -> Tesla
   ; mu = 4*pi*10^(-7)
   ; current is thus in Amps

   units = 721.9e3*sqrt(j.point.pix_size[0]*j.point.pix_size[1]) * $
           1.e-4 / (4.*!pi*1.0e-7)

   current = current * units * sarea ; sign from contour direction
   ; carea in units of pixels so convert to sq. meters
   carea = carea * 721.9e3*j.point.pix_size[0]*721.9e3*j.point.pix_size[1]
   catch,error_status
   if error_status NE 0 then begin
      message,/info,'WARNING: error in get_berrs, assuming 100 G errors on Bx,By'
      berr = {sig_bxh:100.0,sig_byh:100.0}  ; Guess 100 G errors, on error
   endif else begin
      berr = get_berrs(j)
   endelse
   catch,/cancel
   ; error = dB*L where L is the total path length and dB is the
   ; error on the horizontal field along the integration path, 
   ; dB = Berr/sqrt(n) where n is the number of data points along 
   ; the path.  Since L=s[n-1] is in pixel units, n=s[n-1] and thus
   ; L/sqrt(n) is sqrt(L).
   error = sqrt(berr.sig_bxh*berr.sig_byh*s[n-1])*units

   if keyword_set(verbose) then begin
      print,'Current (Amps): '+string(current)+' +/- '+string(error)
      print,'Lenth of contour: ',length*721.9*sqrt(j.point.pix_size[0]*j.point.pix_size[1]),' km'
   endif

   return,current

end
