;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;
;  lowandlou.pro - Solves equation (8) of Lou & Low (1990 ApJ 352 343), a
;                  second-order nonlinear ODE, using a numerical relaxation
;                  algorithm.
;
;  usage: lowandlou,eigenfn,eigenva,n,mugrid,trialfn,trialva,eps=eps,
;                   maxit=maxit,eigenfnpr=eigenfnpr,/plot,/quiet
;       where  eigenfn = on output, solution to the equation
;              eigenva = on output, corresponding eigenvalue (variable a^2)
;              n = on input, chosen value of variable n
;              mugrid = on input, listing of mu gridpoints
;              trialfn = on input, trial function evaluated on the mu grid
;              trialva = on input, trial eigenvalue
;              eps = convergence parameter, default=1e-10
;              maxit = maximum number of iterations, default=100
;              eigenfnpr = on output, first derivative of eigenfunction
;              plot = if set, plots eigenfunction as it converges (or not)
;              quiet = if set, suppress all screen output, except on error
;
;  Notes:  - Numerical relaxation method used here essentially follows section
;            7.2.5 of Hansen and Kawaler's Stellar Interiors textbook.
;          - Boundary conditions are P=0 at mu=-1,1 (as in equation (9) of the
;            paper) and dP/dmu=10 at mu=-1 (as stated in text following
;            equation (12) of paper).
;
;  M.DeRosa -  2 Nov 2004 - created
;             10 Nov 2004 - fixed formulae for n>1 (but still odd)
;
;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

@lowandlouauxiliaries

pro lowandlou,eigenfn,eigenva,n,mugrid,trialfn,trialva,eps=eps,maxit=maxit,$
  eigenfnpr=eigenfnpr,plot=plot,quiet=quiet

;  usage message
if n_params() lt 6 then begin
  print,'  lowandlou,eigenfn,eigenva,n,mugrid,trialfn,trialva,eps=eps,'+ $
    'maxit=maxit,eigenfnpr=eigenfnpr,/plot,/quiet'
  return
endif

;  some error checking
nmu=n_elements(mugrid)
if nmu lt 5 then begin
  print,'  ERROR:  grid must be at least 5 elements'
  return
endif
if n_elements(trialfn) ne nmu then begin
  print,'  ERROR:  mugrid and trialfn must be same length'
  return
endif
if n_elements(n) ne 1 then begin
  print,'  ERROR:  n must be a scalar'
  return
endif
if n_elements(trialva) ne 1 then begin
  print,'  ERROR:  trialva must be a scalar'
  return
endif
if n_elements(eps) ne 1 then eps=1d-10 else eps=double(eps)
if n_elements(maxit) ne 1 then maxit=100 else maxit=round(maxit)
if maxit lt 2 then maxit=100

;  initialize trial functions
xguess=double(trialfn)
xpguess=deriv(mugrid,xguess)
laguess=double(trialva)

;  plot, if requested
if keyword_set(plot) then plot,mugrid,xguess,yr=[-10,10],thick=3

;  iterate
flag=0b
for i=1,maxit do begin

  ;  fill in M and b from matrix equation M.x = b
  bvec=getrhs(mugrid,xguess,xpguess,laguess,_extra={n:n})
  mmat=getlhs(mugrid,xguess,xpguess,laguess,_extra={n:n})

  ;  solve M.x = b
  ludc,mmat,index
  delta=lusol(mmat,index,bvec)

  ;  update solution
  evenix=2*indgen(nmu)
  xguess=xguess+delta(evenix)
  oddix=evenix+1
  xpguess=xpguess+delta(oddix)
  laguess=laguess+delta(nmu*2)

  ;  plot
  if keyword_set(plot) then oplot,mugrid,xguess

  ;  print update
  error=sqrt(total(delta^2))
  if not keyword_set(quiet) then $
    print,'  iteration ',i,', convergence error = ',error,f='(a,i6,a,e16.8)'
  if error lt eps then begin
    i=maxit
    flag=1b
    if not keyword_set(quiet) then print,'  eigenvalue = ',laguess
  endif

endfor

;  nonconvergence error
if flag then begin
  if not keyword_set(quiet) then print,'  HUZZAH!  The solution converged!'
endif else print,'  ERROR:  Solution did not converge.'

;  set outputs
eigenfn=xguess
eigenva=laguess
eigenfnpr=xpguess

end
