
;  script to obtain a Low & Lou (1990 ApJ 352 343) solution
;  MLD -  2 Nov 2004 - created
;        10 Nov 2004 - fixed transpose formulae for n>1 (but still odd)

;  needs gaussquad_legendre, linrange, get_interpolation_index from my
;  personal IDL library (included in the auxiliary routines)

;  if doing accuracy check, needs minmax from the IDL astromony users library

;  include the auxiliary routines
@lowandlouauxiliaries

;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;  first solve nonlinear eigenvalue problem, eq.(8) of paper:
;
;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

;  choose a value for n (must be an odd positive integer)
n=1.

;  set up mu (can't go all the way to -1 and 1 due to singularity in equation)
npoints=64
;mu=linrange(npoints,-1,1)*0.9999  ;  linear grid
gaussquad_legendre,npoints,mu  ;  stretched grid (more points near boundaries)

;  desired number of nodes in eigenfunction (no guarantees, of course)
nodes=1

;  trial function = sine wave with P=0 at mu=-1,1 and dP/dmu=10 at mu=-1 
;  (may need to play around with this)
trialfn=20/(!dpi*(nodes+1))*sin((mu+1)*!dpi/2*(nodes+1))

;  trial eigenvalue (may need to play around with this too)
trialla=2

;  run the solver... hopefully it converges...
lowandlou,eigenfn,eigenva,n,mu,trialfn,trialla,eigenfnpr=eigenfnpr
if eigenva lt 0 then begin
  print,'  STOPPED:  negative eigenvalue'
  stop
endif
print

;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;  now set up desired grid (x,y,z), and get alpha (variable alpha), components
;  of magnetic field (bx,by,bz), and current (jx,jy,jz) at each point
;
;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

;  set parameter l (depth below surface of source)
dl=0.3

;  set parameter phi (rotates axis of symmetry away from the vertical axis)
phi=!pi/2  ;  in radians

;  set number of x,y,z points (keep in mind what you set npoints to above)
nx=64
ny=64
nz=64

;  set up index
xix=linrange(nx,-1,1)
yix=linrange(ny,-1,1)
zix=linrange(nz,0,2)

;  set up grids
x=fltarr(nx,ny,nz,/noz)  &  for i=0,nz-1 do x(*,*,i)=xix#replicate(1,ny)
y=fltarr(nx,ny,nz,/noz)  &  for i=0,nz-1 do y(*,*,i)=replicate(1,nx)#yix
z=fltarr(nx,ny,nz,/noz)  &  for i=0,nz-1 do z(*,*,i)=zix(i)

;  determine gridpoint indices in (X,Y,Z) coordinates using eq.(13) of paper
print,'  setting up grid'
cphi=cos(phi)
sphi=sin(phi)
bigx=x*cphi-(z+dl)*sphi
bigy=y
bigz=x*sphi+(z+dl)*cphi

;  determine gridpoints in spherical coordinates using usual transformation
r=sqrt(bigx*bigx+bigy*bigy+bigz*bigz)
th=atan(sqrt(bigx*bigx+bigy*bigy),bigz)
ph=atan(bigy,bigx)

;  compute Br,Bth,Bph and alpha using eq.(3) of paper
print,'  computing Br, Bth, Bph'
mugrid=get_interpolation_index(mu,cos(th))
eigint=interpolate(eigenfn,mugrid)
rfac=r^(-2-n)
sth=sin(th)  &  cth=cos(th)
sph=sin(ph)  &  cph=cos(ph)
br=-rfac*interpolate(eigenfnpr,mugrid)
bth=n*rfac*eigint/sth
bph=sqrt(eigenva)*rfac*abs(eigint)^(1+1/n)/sth
alpha=reform(sqrt(eigenva)*(1+1/n)*sign_mld(eigint)*abs(eigint)^(1/n)/r,nx,ny,nz,/ov)

;  convert vector components to cartesian (X,Y,Z) using eq.(14) of paper
print,'  transforming Br, Bth, Bph to Bx, By, Bz'
bbigx=br*sth*cph+bth*cth*cph-bph*sph
bbigy=br*sth*sph+bth*cth*sph+bph*cph
bbigz=br*cth-bth*sth

;  rotate components from (X,Y,Z) to (x,y,z) using eq.(15) of paper
bx=bbigx*cphi+bbigz*sphi
by=bbigy
bz=-bbigx*sphi+bbigz*cphi

;  compute current
jx=alpha*bx
jy=alpha*by
jz=alpha*bz

;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
;  some checks (see how close del.B is to 0, and curl B is to alpha*B)
;
;^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

print,'****hit spacebar to perform accuracy check, any other key to end'
if get_kbrd(1) eq ' ' then begin

  d1=fltarr(nx,ny,nz,/noz)  ;  temporary variables for derivatives
  d2=fltarr(nx,ny,nz,/noz)
  d3=fltarr(nx,ny,nz,/noz)

  ;  compute del.B
  print,'  computing div B from finite differences'  
  for i=0,nz-1 do for j=0,ny-1 do d1(*,j,i)=deriv(xix,bx(*,j,i))  ;  dBx/dx
  for i=0,nz-1 do for j=0,nx-1 do d2(j,*,i)=deriv(yix,by(j,*,i))  ;  dBy/dy
  for i=0,ny-1 do for j=0,nx-1 do d3(j,i,*)=deriv(zix,bz(j,i,*))  ;  dBz/dz
  divb=d1+d2+d3
  print,'    min/max of div B =',minmax(divb)
  bzmax=max(bz(*,*,0),whmax)
  bzmin=min(bz(*,*,0),whmin)
  dist=sqrt((xix(whmax mod nx)-xix(whmin mod nx))^2 $
               +(yix(whmax/nx)-yix(whmin/nx))^2)
  print,'    ... compared with estimated slope of Bz pattern at bottom = ', $
    (bzmax-bzmin)/dist

  ;  compute curl B
  print,'  computing curl B from finite differences'
  for i=0,nz-1 do for j=0,ny-1 do d1(*,j,i)=deriv(xix,by(*,j,i))  ;  dBy/dx
  for i=0,nz-1 do for j=0,nx-1 do d2(j,*,i)=deriv(yix,bx(j,*,i))  ;  dBx/dy
  jzc=d1-d2
  for i=0,nz-1 do for j=0,nx-1 do d1(j,*,i)=deriv(yix,bz(j,*,i))  ;  dBz/dy
  for i=0,ny-1 do for j=0,nx-1 do d2(j,i,*)=deriv(zix,by(j,i,*))  ;  dBy/dz
  jxc=d1-d2
  for i=0,ny-1 do for j=0,nx-1 do d1(j,i,*)=deriv(zix,bx(j,i,*))  ;  dBx/dz
  for i=0,nz-1 do for j=0,ny-1 do d2(*,j,i)=deriv(xix,bz(*,j,i))  ;  dBz/dx
  jyc=d1-d2

  ;  print out (curl B)-(alpha.B), can also do point-by-point comparisons
  print,'    min/max of computed x-component of (curl B)-(alpha*B)= ', $
    minmax(jxc-jx)
  print,'    ... compared with min/max of alpha*Bx = ',minmax(jx)
  print,'    min/max of computed y-component of (curl B)-(alpha*B) = ', $
    minmax(jyc-jy)
  print,'    ... compared with min/max of alpha*By = ',minmax(jy)
  print,'    min/max of computed z-component of (curl B)-(alpha*B) = ', $
    minmax(jzc-jz)
  print,'    ... compared with min/max of alpha*Bz = ',minmax(jz)

endif

end
