
;+
;NAME:
;     bff_movie
;PURPOSE:
;     Makes an animation sequence for the 3D force free field plots.
;CATEGORY:
;CALLING SEQUENCE:
;     bff_movie,vertices,xrot,yrot,zrot
;INPUTS:
;     vertices = vertices structure output from blines.pro
;OPTIONAL INPUT PARAMETERS:
;     xrot = rotation step size around x-axis (degrees)
;     yrot = rotation step size around y-axis (degrees)
;     zrot = rotation step size around z-axis (degrees)
;KEYWORD PARAMETERS
;     nframes = number of frames in the movie
;     noreturn = If set, make the rotation in one direction only
;     noreset = if set, don't set up the 3D transformation (use current value)
;     image = plot an image under the contours (image = 2D array)
;     position = plot window specs for contour, oplot etc., [x0,y0,x1,y1,z0,z1]
;     quiet = if set, don't print all the information
;     az = initial orientation of the plot (same as surface)
;     ax = initial orientation of the plot (same as surface)
;     t3d:  accepts all the t3d keywords
;     zbuffer = use z-buffering (slow!!)
;OUTPUTS:
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     To get a good speed in the movie you may have to run bff_movie with 
;     a small window.
;
;     The image doesn't plot right when perspective is included in the 3D 
;     transformation ... I don't understand why.
;
;     The maximum number of frames seems to be 127.  Unless the noreturn 
;     keyword is set, the rotation uses two frames for each image and hence
;     the limit is more like 64 frames
;PROCEDURE:
;MODIFICATION HISTORY:
;     T. Metcalf   Feb. 1992
;-

pro bff_movie, v, xrot, yrot, zrot, nframes=nframes, noreturn=noreturn, $
               noreset=noreset, image=image, az=az, ax=ax, position=position, $
               translate=translate, scale=scale, rotate=rotate, $
               perspective=perspective, oblique=oblique, xyexch=xyexch, $
               xzexch=xzexch, yzexch=yzexch, zbuffer=zbuffer,quiet=quiet

  if n_elements(v) LE 0 then begin
     print,'ERROR: bff_movie: You must specify the vertices structure.'
     return
  endif

  if n_elements(nframes) LE 0 then nframes = 40

  if n_elements(xrot) LE 0 OR n_elements(yrot) LE 0 OR $
     n_elements(zrot) LE 0 then begin
    xrot = 0.
    if nframes GT 40 AND keyword_set(noreturn) then yrot = 360./nframes $
    else yrot = 1.
    zrot = 0.
  endif

  if n_elements(position) NE 6 then position = [0.1,0.1,.9,.9,0.,1.] $
  else position = float(position)

  !X.S = [position(0),(position(2)-position(0))/(v.nx-1.-(0.))]
  !Y.S = [position(1),(position(3)-position(1))/(v.ny-1.-(0.))]
  ;!Z.S = [position(4),(position(5)-position(4))/(v.zscale*(v.nz-1.)-(0.))]
  ; Set z-scale to same as x,y scales
  ;nzeff = ((1./!X.S(1)) > (1./!Y.S(1)))*(position(5)-position(4))
  !Z.S = [position(4),!X.S(1)<!Y.S(1)]

  if NOT keyword_set(noreset) then begin
     if n_elements(az) LE 0 then az=25.
     if n_elements(ax) LE 0 then ax=60.
     t3d,/reset,translate=[-0.5,-0.5,-0.5]  ; Copied from surfr.pro
     t3d, rotate=[-90,az,0.]
     t3d,rotate=[ax,0,0.]
     ; For some reason I don't understand, images don't plot right with 
     ; perspective
     if n_elements(image) LE 0 then t3d,perspective=4.0
     scale3d
  endif

  if n_elements(translate) NE 0 OR $
     n_elements(scale) NE 0 OR $
     n_elements(rotate) NE 0 OR $
     n_elements(perspective) NE 0 OR $
     n_elements(oblique) NE 0 OR $
     n_elements(xyexch) NE 0 OR $
     n_elements(xzexch) NE 0 OR $
     n_elements(yzexch) NE 0 then begin
      t3d,translate=translate,scale=scale,rotate=rotate, $
          perspective=perspective,oblique=oblique,xyexch=xyexch,xzexch=xzexch, $
          yzexch=yzexch
      scale3d
  endif 
  
  if NOT keyword_set(noreturn) then xinteranimate,set=[!D.X_SIZE,!D.Y_SIZE, $
     nframes*2-2], title='Bff Animation: Alpha=' $
     + strcompress(string(v.alpha,format="(f15.2)"),/remove_all) + '  Zscale=' $
     + strcompress(string(v.zscale,format="(f15.3)"),/remove_all) $
  else xinteranimate,set=[!D.X_SIZE,!D.Y_SIZE,nframes], $
     title='Bff Animation: Alpha=' $
     + strcompress(string(v.alpha,format="(f15.2)"),/remove_all) + '  Zscale=' $
     + strcompress(string(v.zscale,format="(f15.2)"),/remove_all)

  total_rot = [0,0,0]

  ; Get the 2D position of the magnetogram center.  This point will remain
  ; fixed in the rotation

  scenter = cvt_to_2d(v.nx/2.,v.ny/2.,0.) 

  for frame=0,nframes-1 do begin

     if NOT keyword_set(quiet) then begin
        print
        print,frame,total_rot, $
              format="('Plotting frame ',i3,'  Cumulative Rotation = ',3f6.1)"
        print
     endif

     ; For some reason zbuffering doesn't like the image keyword (even it it's
     ; not set!).
     if keyword_set(zbuffer) then begin
        v=blines(v,/vertices,title=strcompress(string(frame),/remove_all), $
                 position=position,charsize=2,/noreset3d, $
                 zbuffer=zbuffer,quiet=quiet)
     endif $
     else begin
        v=blines(v,/vertices,title=strcompress(string(frame),/remove_all), $
                 position=position,charsize=2,/noreset3d,image=image, $
                 zbuffer=zbuffer,quiet=quiet)
     endelse

     t3d,rot=[xrot,yrot,zrot]   ; One step of the rotation

     ; The following translation insures that the center of the magnetogram
     ; does not move during the rotation

     center = cvt_to_2d(v.nx/2.,v.ny/2.,0.)
     t3d,translate=[scenter(0)-center(0),scenter(1)-center(1),0]

     xinteranimate,window=!D.WINDOW,frame=frame  ; store the plot

     ; do the reverse direction
     if (NOT keyword_set(noreturn)) AND (frame NE (nframes-1)) $
        AND (frame NE 0) then $
        xinteranimate,window=!D.WINDOW,frame=nframes*2-2-frame

     total_rot = total_rot + [xrot,yrot,zrot]

  endfor

  xinteranimate,100

end
