;+
;NAME:
;     BHELIO
;PURPOSE:
;     Transforms magnetic field structure to heliographic coordinates by
;     interpolating the field data onto a uniform heliographic grid.
;CATEGORY:
;CALLING SEQUENCE:
;     newmag = bhelio(mag,bx,by,bz,i_cont,p,q,pinv,qinv)
;INPUTS:
;     mag = magnetic field structure (from bfits.pro)
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     bin = bins up by this factor to a finer grid before interpolation 
;           (default = 1)
;     alpha = force-free alpha to use in the bicubic interpolation to a 
;             finer grid.  If not set, alpha is not used in the 
;             interpolation.  Only used if bin NE 1.
;     /plot = Plot the new magnetogram
;     offlimb = array of 0's or 1's.  A 1 indicates that the pixel is off the
;               limb and should not be transformed.  Usually a pixel with R
;               over Rsun > 0.98 should be avoided, otherwise the arrays get
;               too huge after the transformation.
;OUTPUTS:
;     newmag = new magnetic field structure
;     bx,by,bz = the x,y,z field on the heliographic grid
;     i_cont = rebinned continuum image
;     p,q = the poly_2d transformation used to go to heliographic
;     pinv,qinv = the poly_2d transformation which goes back to image coords
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The bx, by, bz, b_pxd, and b_pyd fields must be present in mag.
;
;     The pointing info is preserved, but is now inconsistent due to
;     the warping.  Hence if the output of this program is passed to
;     some other code that uses the pointing info, the results will be
;     incorrect.  For example, the output of this code cannot be
;     passed to jxyz.pro or get_latlon.pro.
;PROCEDURE:    degree=2

;MODIFICATION HISTORY:
;     T. Metcalf 1999-Aug-17  Added offlimb keyword.
;     TRM 2005-Mar-10 Added more fields to the output structure.
;     TRM 2005-Jul-07 Fill in lat/cmd in missing areas.
;-

function bhelio,magin,bx2,by2,bz2,i_cont,p,q,pinv,qinv,plot=plot,bin=bin, $
           alpha=alpha,offlimb=offlimb

   if n_elements(bin) NE 1 then bin=1

   if bin GT 1 then begin
      mag = binup_b(magin,fix(bin),alpha,noalpha=(n_elements(alpha) EQ 0)) 
      nx = n_elements(magin.bz(*,0))
      ny = n_elements(magin.bz(0,*))
      nnx = fix(nx * bin)
      nny = fix(ny * bin)
      binupx = float(nnx)/float(nx)
      binupy = float(nny)/float(ny)
   endif $
   else begin
      mag = magin
      binupx = 1.
      binupy = 1.
   endelse

   ; Put the magnetogram on a uniform grid in heliographic coordinates
   ; Uses the Bx, By, Bz variables.

   bnx = n_elements(mag.bz(*,0))
   bny = n_elements(mag.bz(0,*))
   bnxy = bnx*bny

   pix_size = double(mag.point.pix_size)
   if n_elements(pix_size) EQ 1 then pix_size = [pix_size(0),pix_size(0)]
   mpix_size = min(pix_size)

   x = (lindgen(bnx,bny) MOD bnx)
   y = (long(lindgen(bnx,bny) / bnx))

   ; p,q will be the forward poly_2d transformation and pinv,qinv will be for
   ; the inverse transform which puts the grid back to the original at the
   ; end of the computations.

   bpxd = mag.b_pxd*binupx
   bpyd = mag.b_pyd*binupy

   if n_elements(offlimb) EQ n_elements(bpxd) then begin
      good = where(offlimb EQ 0,ngood)
      if ngood LE 0 then message,'All points are masked out with offlimb keyword'
   endif else begin
      if n_elements(offlimb) NE 0 then message,'offlimb has the wrong size'
      good = lindgen(n_elements(bpxd))
      ngood = n_elements(bpxd)
   endelse

   interp = 2 & cubic=-0.5 & degree=3

   if ngood LT (degree+1)^2 then message,'Too many points are off the limb.'

   polywarp,double(x[good]),double(y[good]),bpxd[good]-min(bpxd[good]),bpyd[good]-min(bpyd[good]),degree,p,q
   polywarp,bpxd[good]-min(bpxd[good]),bpyd[good]-min(bpyd[good]),double(x[good]),double(y[good]),degree,pinv,qinv

   nx = (long(max(bpxd[good]) - min(bpxd[good])) + 1)  ; Make sure that pixel scale 
   ny = (long(max(bpyd[good]) - min(bpyd[good])) + 1)  ; does not change.

   bx2 = poly_2d(mag.bx,p,q,interp,nx,ny,missing=0.0,cubic=cubic)   ; Field on heliographic grid
   by2 = poly_2d(mag.by,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   bz2 = poly_2d(mag.bz,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   bln = poly_2d(mag.b_long,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   btr = poly_2d(mag.b_trans,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   baz = poly_2d(mag.b_azim,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   okdata = poly_2d(1+intarr(bnx,bny),p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   if tag_index(magin,'JZ') GE 0 then $
      bjz = poly_2d(mag.jz,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   if tag_index(magin,'VELOCITY') GE 0 then $
      vel = poly_2d(mag.velocity,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   if tag_index(magin,'LATITUDE') GE 0 AND $
      tag_index(magin,'CMD') GE 0 then begin
      lat = poly_2d(mag.latitude,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
      cmd = poly_2d(mag.cmd,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   endif else begin
      ll = get_latlon(mag)
      lat = poly_2d(ll.latitude,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
      cmd = poly_2d(ll.cmd,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   endelse

    ; Fill in the lat and cmd in the missing areas by extrapolation
    good = where(okdata NE 0,ngood)
    if ngood GE (degree+1)^2 then begin
       xy = lonlat2xy(transpose([[cmd[good]],[lat[good]]]),b0=mag.point.b0,radius=mag.point.radius)
       x = double(lindgen(nx,ny) MOD nx)
       y = double(lindgen(nx,ny) / nx)
       polywarp,double(xy[0,*]),double(xy[1,*]),x[good],y[good],degree,pp,qq  ; get coordinate transform
       pq2xy,x,y,pp,qq,xh,yh   ; transform uniform grid to helio coords
       ll = xy2lonlat( $
                     reform( $
                             transpose([[[xh]],[[yh]]],[2,0,1]), $
                             2,nx*ny), $
                     b0=mag.point.b0,radius=mag.point.radius)
       bad = where(okdata EQ 0,nbad)
       if nbad GT 0 then begin
          cmd[bad] = ll[0,bad]
          lat[bad] = ll[1,bad]
       endif
    endif else message,/info,'WARNING: Not enough points to fix missing latitude and cmd values.'

   bpxd = (lindgen(nx,ny) MOD nx)  ; we have transformed to a uniform grid
   bpyd = (long(lindgen(nx,ny) / nx))
   ;xtest = poly_2d(mag.b_pxd*binupx,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   ;ytest = poly_2d(mag.b_pyd*binupx,p,q,interp,nx,ny,missing=0.0,cubic=cubic)
   ;pmm,xtest[good]-bpxd[good] - min(xtest[good]) + min(bpxd[good])
   ;pmm,ytest[good]-bpyd[good] - min(ytest[good]) + min(bpyd[good])

   if tag_index(magin,'I_CONT') GE 0 then begin
      i_cont = poly_2d(rebin(magin.i_cont,n_elements(magin.i_cont(*,0))*bin, $
                                          n_elements(magin.i_cont(0,*))*bin), $
                       p,q,interp,nx,ny,missing=0,cubic=cubic)
   endif else if n_elements(i_cont) EQ bnxy then begin
      i_cont = poly_2d(rebin(i_cont,n_elements(i_cont(*,0))*bin, $
                                    n_elements(i_cont(0,*))*bin), $
                       p,q,interp,nx,ny,missing=0,cubic=cubic)
   endif

   point = mag.point
   point.pix_size = point.pix_size/float(bin)
   point.p = 0.0  ; solar coordiantes now
   point.lat = bilin(lat,(nx-1.)/2.,(ny-1.)/2.)*!dtor
   point.cmd = bilin(cmd,(nx-1.)/2.,(ny-1.)/2.)*!dtor
   bpxd = bpxd / point.pix_size[0]  ; convert to pixel units
   bpyd = bpyd / point.pix_size[1]

   mag2 = {b_long:bln,b_trans:btr,b_azim:baz,bx:bx2,by:by2,bz:bz2, $
           b_pxd:bpxd,b_pyd:bpyd,latitude:lat,cmd:cmd,okdata:okdata, $
           point:point}
   if n_elements(i_cont) GT 0 then mag2=add2str(mag2,'I_CONT',i_cont)
   if n_elements(bjz) GT 0 then $
      mag2=add2str(mag2,['JZ','J_PXD','J_PYD'],bjz,bpxd+0.5,bpyd+0.5)
   if n_elements(vel) GT 0 then mag2=add2str(mag2,'VELOCITY',vel)
   ; The following is wrong:  because of the warping, get_latlon
   ; will not give the right answer.
   ;;ll = get_latlon(mag2,/add) ; update latitude and longitude

   if keyword_set(plot) then begin
      if n_elements(i_cont) EQ nx*ny then $
         vmgram,mag2,/inc,/xyz,image=i_cont,/helio,/arrows $
      else $
         vmgram,mag2,/inc,/xyz,/helio,/arrows
   endif

   return,mag2

end
