
;+
;NAME:
;     BICUBIC3D
;PURPOSE:
;     Does a 3D bicubic interpolation of the 3D magnetic field
;CATEGORY:
;CALLING SEQUENCE:
;     binterp = bicubic3d(mag,x,y,z,alpha)
;INPUTS:
;     mag = magnetic field structure containing at least BXP,BYP,BZP
;     x,y,z = the coordinates of the field point
;     alpha = force free alpha (used to compute the derivatives)
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     reset = force the computation of the derivatives.  Otherwise they are
;             computed only when the routine is called the first time.
;OUTPUTS:
;     binterp = a singe vector containing the interpolated field [Bx,By,Bz]
;COMMON BLOCKS:
;     Xbicubic3d_derivs = Makes the derivatives into static variables
;     Ybicubic3d_derivs = Makes the derivatives into static variables
;     Zbicubic3d_derivs = Makes the derivatives into static variables
;     times_called3d = contains static variable ntimes
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
; Performs several 2D bicubic interpolations to build up the 3D interpolation.
; The derivatives are computed numerically and the divB=0 and the force-free
; conditions are used where possible for consistency.
;MODIFICATION HISTORY:
;
;       Feb 4, 1992  T. Metcalf
;       July 20, 2000 T. Metcalf  Changed the way the force-free and divB=0
;                                 conditions are applied to beat down the
;                                 noise slightly.  No big deal.
;-

function bicubic3d,mag,x,y,z,alpha,reset=reset

   ; The common blocks make the derivatives into static variables

   common Xbicubic3d_derivs,dBxdx,dBxdy,dBxdz,d2Bxdzdy,d2Bxdxdy
   common Ybicubic3d_derivs,dBydx,dBydy,dBydz,d2Bydzdy,d2Bydxdy
   common Zbicubic3d_derivs,dBzdx,dBzdy,dBzdz,d2Bzdzdy,d2Bzdxdy
   common times_called3d,ntimes

   nx = n_elements(mag.bxp(*,0,0))
   ny = n_elements(mag.bxp(0,*,0))
   nz = n_elements(mag.bxp(0,0,*))

   i = (fix(x) > 0) < (nx-2)
   j = (fix(y) > 0) < (ny-2)
   k = (fix(z) > 0) < (nz-2)


;   bx = mag.bxp
;   by = mag.byp
;   bz = mag.bzp

   if n_elements(ntimes) LE 0 then ntimes = 0L $
   else ntimes = 1L

   ; compute derivatives on the first pass or if reset keyword is present

   if keyword_set(reset) or (NOT ntimes) then begin

      dx = 2.0
      dy = 2.0
      dz = 2.0

      ; Compute numerical derivatives

      dBxdx = (shift(mag.bxp,-1,0,0) - shift(mag.bxp,1,0,0))/dx
      dBxdy = (shift(mag.bxp,0,-1,0) - shift(mag.bxp,0,1,0))/dy
      dBxdz = (shift(mag.bxp,0,0,-1) - shift(mag.bxp,0,0,1))/dz

      dBydx = (shift(mag.byp,-1,0,0) - shift(mag.byp,1,0,0))/dx
      dBydy = (shift(mag.byp,0,-1,0) - shift(mag.byp,0,1,0))/dy
      dBydz = (shift(mag.byp,0,0,-1) - shift(mag.byp,0,0,1))/dz

      dBzdx = (shift(mag.bzp,-1,0,0) - shift(mag.bzp,1,0,0))/dx
      dBzdy = (shift(mag.bzp,0,-1,0) - shift(mag.bzp,0,1,0))/dy
      dBzdz = (shift(mag.bzp,0,0,-1) - shift(mag.bzp,0,0,1))/dz

      ; Avoid edge problems using divB=0

      dBxdx[0,*,*] = -dBydy[0,*,*] - dBzdz[0,*,*]  
      dBydy[*,0,*] = -dBxdx[*,0,*] - dBzdz[*,0,*]
      dBzdz[*,*,0] = -dBxdx[*,*,0] - dBydy[*,*,0]
      dBxdx[nx-1,*,*] = -dBydy[nx-1,*,*] - dBzdz[nx-1,*,*]
      dBydy[*,ny-1,*] = -dBxdx[*,ny-1,*] - dBzdz[*,ny-1,*]
      dBzdz[*,*,nz-1] = -dBxdx[*,*,nz-1] - dBydy[*,*,nz-1]

      ; Avoid edge problems using FF conditions

      dBxdy[*,0,*] = dBydx[*,0,*] - alpha*mag.bzp[*,0,*] 
      dBydx[0,*,*] = dBxdy[0,*,*] + alpha*mag.bzp[0,*,*]
      dBxdy[*,ny-1,*] = dBydx[*,ny-1,*] - alpha*mag.bzp[*,ny-1,*]
      dBydx[nx-1,*,*] = dBxdy[nx-1,*,*] + alpha*mag.bzp[nx-1,*,*]

      dBxdz[*,*,0] = dBzdx[*,*,0] + alpha*mag.byp[*,*,0]
      dBzdx[0,*,*] = dBxdz[0,*,*] - alpha*mag.byp[0,*,*]
      dBxdz[*,*,nz-1] = dBzdx[*,*,nz-1] + alpha*mag.byp[*,*,nz-1]
      dBzdx[nx-1,*,*] = dBxdz[nx-1,*,*] - alpha*mag.byp[nx-1,*,*]

      dBydz[*,*,0] = dBzdy[*,*,0] - alpha*mag.bxp[*,*,0]
      dBzdy[*,0,*] = dBydz[*,0,*] + alpha*mag.bxp[*,0,*]
      dBydz[*,*,nz-1] = dBzdy[*,*,nz-1] - alpha*mag.bxp[*,*,nz-1]
      dBzdy[*,ny-1,*] = dBydz[*,ny-1,*] + alpha*mag.bxp[*,ny-1,*]

      ; Implement the force-free conditions

      dBydx_temp = (dBydx + dBxdy + alpha*mag.bzp)/2.0 ; Average numerical & FF deriv
      dBxdy_temp = (dBxdy + dBydx - alpha*mag.bzp)/2.0
      dBxdz_temp = (dBxdz + dBzdx + alpha*mag.byp)/2.0
      dBzdx_temp = (dBzdx + dBxdz - alpha*mag.byp)/2.0
      dBzdy_temp = (dBzdy + dBydz + alpha*mag.bxp)/2.0
      dBydz_temp = (dBydz + dBzdy - alpha*mag.bxp)/2.0

      ; Implement the divB = 0 condition

      dBzdz_temp = (dBzdz - dBydy - dBxdx)/2.0
      dBydy_temp = (dBydy - dBxdx - dBzdz)/2.0
      dBxdx_temp = (dBxdx - dBydy - dBzdz)/2.0

      ; Clean up temp variables

      dBydx = temporary(dBydx_temp)
      dBxdy = temporary(dBxdy_temp)
      dBxdz = temporary(dBxdz_temp)
      dBzdx = temporary(dBzdx_temp)
      dBzdy = temporary(dBzdy_temp)
      dBydz = temporary(dBydz_temp)
      dBxdx = temporary(dBxdx_temp)
      dBydy = temporary(dBydy_temp)
      dBzdz = temporary(dBzdz_temp)

      ; Calculate 2nd derivatives numerically

      d2Bxdzdy = (shift(dBxdz,0,-1,0) - shift(dBxdz,0,1,0))/dy
      d2Bxdxdy = (shift(dBxdx,0,-1,0) - shift(dBxdx,0,1,0))/dy
      d2Bydzdy = (shift(dBydz,0,-1,0) - shift(dBydz,0,1,0))/dy
      d2Bydxdy = (shift(dBydx,0,-1,0) - shift(dBydx,0,1,0))/dy
      d2Bzdzdy = (shift(dBzdz,0,-1,0) - shift(dBzdz,0,1,0))/dy
      d2Bzdxdy = (shift(dBzdx,0,-1,0) - shift(dBzdx,0,1,0))/dy
   ;   d2Bydydy = (shift(dBydy,0,-1,0) - shift(dBydy,0,1,0))/dy  ; Extras
   ;   d2Bzdydy = (shift(dBzdy,0,-1,0) - shift(dBzdy,0,1,0))/dy  ; Extras
   ;   d2Bxdydy = (shift(dBxdy,0,-1,0) - shift(dBxdy,0,1,0))/dy  ; Extras

      ; Can't avoid boundary problems using force free conditions for
      ; the 2nd derivatives, so just use the adjacent values

      d2Bxdzdy[*,0,*] = d2Bxdzdy[*,1,*]
      d2Bxdxdy[*,0,*] = d2Bxdxdy[*,1,*]
      d2Bydzdy[*,0,*] = d2Bydzdy[*,1,*]
      d2Bydxdy[*,0,*] = d2Bydxdy[*,1,*] 
      d2Bzdzdy[*,0,*] = d2Bzdzdy[*,1,*]
      d2Bzdxdy[*,0,*] = d2Bzdxdy[*,1,*]
    ;  d2Bydydy[*,0,*] = d2Bydydy[*,1,*]  ; Extras
    ;  d2Bzdydy[*,0,*] = d2Bzdydy[*,1,*]  ; Extras
    ;  d2Bxdydy[*,0,*] = d2Bxdydy[*,1,*]  ; Extras

      d2Bxdzdy[*,ny-1,*] = d2Bxdzdy[*,ny-2,*] 
      d2Bxdxdy[*,ny-1,*] = d2Bxdxdy[*,ny-2,*]
      d2Bydzdy[*,ny-1,*] = d2Bydzdy[*,ny-2,*]
      d2Bydxdy[*,ny-1,*] = d2Bydxdy[*,ny-2,*] 
      d2Bzdzdy[*,ny-1,*] = d2Bzdzdy[*,ny-2,*]
      d2Bzdxdy[*,ny-1,*] = d2Bzdxdy[*,ny-2,*]
    ;  d2Bydydy[*,ny-1,*] = d2Bydydy[*,ny-2,*]  ; Extras
    ;  d2Bzdydy[*,ny-1,*] = d2Bzdydy[*,ny-2,*]  ; Extras
    ;  d2Bxdydy[*,ny-1,*] = d2Bxdydy[*,ny-2,*]  ; Extras

      ; Apply Force Free conditions to 2nd derivatives

    ;  d2Bxdzdy_temp = (d2Bxdzdy + d2Bzdxdy + alpha*dBydy)/2.0
    ;  d2Bxdxdy_temp = (d2Bxdxdy - d2Bydydy - d2Bzdzdy)/2.0
    ;  d2Bydzdy = (temporary(d2Bydzdy) + temporary(d2Bzdydy) - alpha*dBxdy )/2.0
    ;  d2Bydxdy = (temporary(d2Bydxdy) + temporary(d2Bxdydy) + alpha*dBzdy)/2.0
    ;  d2Bzdzdy_temp = (d2Bzdzdy - temporary(d2Bydydy) - d2Bxdxdy)/2.0
    ;  d2Bzdxdy_temp = (d2Bzdxdy + d2Bxdzdy - alpha*dBydy)/2.0

    ;  d2Bxdzdy = temporary(d2Bxdzdy_temp)
    ;  d2Bxdxdy = temporary(d2Bxdxdy_temp)
    ;  d2Bzdzdy = temporary(d2Bzdzdy_temp)
    ;  d2Bzdxdy = temporary(d2Bzdxdy_temp)

      print,'INFO: bicubic3d: Derivatives computed'

   endif

   zmk = float(z-k)
   zmk1 = 1.0-zmk

   i1 = i+1
   j1 = j+1
   k1 = k+1

;
; Bx
;

   ; B(i,j,z)
   f = [mag.bxp(i,j,k),mag.bxp(i,j1,k),mag.bxp(i,j1,k1),mag.bxp(i,j,k1)]
   f1 = [dBxdz(i,j,k),dBxdz(i,j1,k),dBxdz(i,j1,k1),dBxdz(i,j,k1)]
   f2 = [dBxdy(i,j,k),dBxdy(i,j1,k),dBxdy(i,j1,k1),dBxdy(i,j,k1)]
   f12 = [d2Bxdzdy(i,j,k),d2Bxdzdy(i,j1,k),d2Bxdzdy(i,j1,k1),d2Bxdzdy(i,j,k1)]
   ;f = [mag.bxp(i,j,k),mag.bxp(i,j,k1),mag.bxp(i,j1,k1),mag.bxp(i,j1,k)]
   ;f1 = [dBxdz(i,j,k),dBxdz(i,j,k1),dBxdz(i,j1,k1),dBxdz(i,j1,k)]
   ;f2 = [dBxdy(i,j,k),dBxdy(i,j,k1),dBxdy(i,j1,k1),dBxdy(i,j1,k)]
   ;f12 = [d2Bxdzdy(i,j,k),d2Bxdzdy(i,j,k1),d2Bxdzdy(i,j1,k1),d2Bxdzdy(i,j1,k)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf00z,ansf100z,ansf200z
   ; B(i,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf01z,ansf101z,ansf201z
   ; B(i+1,j,z)
   f = [mag.bxp(i1,j,k),mag.bxp(i1,j1,k),mag.bxp(i1,j1,k1),mag.bxp(i1,j,k1)]
   f1 = [dBxdz(i1,j,k),dBxdz(i1,j1,k),dBxdz(i1,j1,k1),dBxdz(i1,j,k1)]
   f2 = [dBxdy(i1,j,k),dBxdy(i1,j1,k),dBxdy(i1,j1,k1),dBxdy(i1,j,k1)]
   f12=[d2Bxdzdy(i1,j,k),d2Bxdzdy(i1,j1,k),d2Bxdzdy(i1,j1,k1),d2Bxdzdy(i1,j,k1)]
   ;f = [mag.bxp(i1,j,k),mag.bxp(i1,j,k1),mag.bxp(i1,j1,k1),mag.bxp(i1,j1,k)]
   ;f1 = [dBxdz(i1,j,k),dBxdz(i1,j,k1),dBxdz(i1,j1,k1),dBxdz(i1,j1,k)]
   ;f2 = [dBxdy(i1,j,k),dBxdy(i1,j,k1),dBxdy(i1,j1,k1),dBxdy(i1,j1,k)]
   ;f12=[d2Bxdzdy(i1,j,k),d2Bxdzdy(i1,j,k1),d2Bxdzdy(i1,j1,k1),d2Bxdzdy(i1,j1,k)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf10z,ansf110z,ansf210z
   ; B(i+1,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf11z,ansf111z,ansf211z
   f = [ansf00z,ansf01z,ansf11z,ansf10z]
   f1 =   [ dBxdx(i,j,k)*zmk1+zmk*(dBxdx(i,j,k1)), $ ;dBx/dx @ z=z (lin interp)
            dBxdx(i,j1,k)*zmk1+zmk*(dBxdx(i,j1,k1)), $
            dBxdx(i1,j1,k)*zmk1+zmk*(dBxdx(i1,j1,k1)), $
            dBxdx(i1,j,k)*zmk1+zmk*(dBxdx(i1,j,k1)) ]
   f2 =   [ dBxdy(i,j,k)*zmk1+zmk*(dBxdy(i,j,k1)), $  ; dBx/dy at z=z
            dBxdy(i,j1,k)*zmk1+zmk*(dBxdy(i,j1,k1)), $
            dBxdy(i1,j1,k)*zmk1+zmk*(dBxdy(i1,j1,k1)), $
            dBxdy(i1,j,k)*zmk1+zmk*(dBxdy(i1,j,k1)) ]
   f12 =   [ d2Bxdxdy(i,j,k)*zmk1+zmk*(d2Bxdxdy(i,j,k1)), $
             d2Bxdxdy(i,j1,k)*zmk1+zmk*(d2Bxdxdy(i,j1,k1)), $
             d2Bxdxdy(i1,j1,k)*zmk1+zmk*(d2Bxdxdy(i1,j1,k1)), $
             d2Bxdxdy(i1,j,k)*zmk1+zmk*(d2Bxdxdy(i1,j,k1)) ]
   ;f = [ansf00z,ansf10z,ansf11z,ansf01z]
   ;f1 =   [ dBxdx(i,j,k)*zmk1+zmk*(dBxdx(i,j,k1)), $ ;dBx/dx @ z=z (lin interp)
   ;         dBxdx(i1,j,k)*zmk1+zmk*(dBxdx(i1,j,k1)), $
   ;         dBxdx(i1,j1,k)*zmk1+zmk*(dBxdx(i1,j1,k1)), $
   ;         dBxdx(i,j1,k)*zmk1+zmk*(dBxdx(i,j1,k1)) ]
   ;f2 =   [ dBxdy(i,j,k)*zmk1+zmk*(dBxdy(i,j,k1)), $  ; dBx/dy at z=z
   ;         dBxdy(i1,j,k)*zmk1+zmk*(dBxdy(i1,j,k1)), $
   ;         dBxdy(i1,j1,k)*zmk1+zmk*(dBxdy(i1,j1,k1)), $
   ;         dBxdy(i,j1,k)*zmk1+zmk*(dBxdy(i,j1,k1)) ]
   ;f12 =   [ d2Bxdxdy(i,j,k)*zmk1+zmk*(d2Bxdxdy(i,j,k1)), $
   ;          d2Bxdxdy(i1,j,k)*zmk1+zmk*(d2Bxdxdy(i1,j,k1)), $
   ;          d2Bxdxdy(i1,j1,k)*zmk1+zmk*(d2Bxdxdy(i1,j1,k1)), $
   ;          d2Bxdxdy(i,j1,k)*zmk1+zmk*(d2Bxdxdy(i,j1,k1)) ]
   bicubic,f,f1,f2,f12,i,i1,j,j1,x,y,bxout,ansf1,ansf2

;
; By
;

   ; B(i,j,z)
   f = [mag.byp(i,j,k),mag.byp(i,j1,k),mag.byp(i,j1,k1),mag.byp(i,j,k1)]
   f1 = [dBydz(i,j,k),dBydz(i,j1,k),dBydz(i,j1,k1),dBydz(i,j,k1)]
   f2 = [dBydy(i,j,k),dBydy(i,j1,k),dBydy(i,j1,k1),dBydy(i,j,k1)]
   f12 = [d2Bydzdy(i,j,k),d2Bydzdy(i,j1,k),d2Bydzdy(i,j1,k1),d2Bydzdy(i,j,k1)]
   ;f = [mag.byp(i,j,k),mag.byp(i,j,k1),mag.byp(i,j1,k1),mag.byp(i,j1,k)]
   ;f1 = [dBydz(i,j,k),dBydz(i,j,k1),dBydz(i,j1,k1),dBydz(i,j1,k)]
   ;f2 = [dBydy(i,j,k),dBydy(i,j,k1),dBydy(i,j1,k1),dBydy(i,j1,k)]
   ;f12 = [d2Bydzdy(i,j,k),d2Bydzdy(i,j,k1),d2Bydzdy(i,j1,k1),d2Bydzdy(i,j1,k)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf00z,ansf100z,ansf200z
   ; B(i,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf01z,ansf101z,ansf201z
   ; B(i+1,j,z)
   f = [mag.byp(i1,j,k),mag.byp(i1,j1,k),mag.byp(i1,j1,k1),mag.byp(i1,j,k1)]
   f1 = [dBydz(i1,j,k),dBydz(i1,j1,k),dBydz(i1,j1,k1),dBydz(i1,j,k1)]
   f2 = [dBydy(i1,j,k),dBydy(i1,j1,k),dBydy(i1,j1,k1),dBydy(i1,j,k1)]
   f12 = [d2Bydzdy(i1,j,k),d2Bydzdy(i1,j1,k),d2Bydzdy(i1,j1,k1),d2Bydzdy(i1,j,k1)]
   ;f = [mag.byp(i1,j,k),mag.byp(i1,j,k1),mag.byp(i1,j1,k1),mag.byp(i1,j1,k)]
   ;f1 = [dBydz(i1,j,k),dBydz(i1,j,k1),dBydz(i1,j1,k1),dBydz(i1,j1,k)]
   ;f2 = [dBydy(i1,j,k),dBydy(i1,j,k1),dBydy(i1,j1,k1),dBydy(i1,j1,k)]
   ;f12 = [d2Bydzdy(i1,j,k),d2Bydzdy(i1,j,k1),d2Bydzdy(i1,j1,k1),d2Bydzdy(i1,j1,k)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf10z,ansf110z,ansf210z
   ; B(i+1,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf11z,ansf111z,ansf211z
   f = [ansf00z,ansf01z,ansf11z,ansf10z]
   f1 =   [ dBydx(i,j,k)*zmk1+zmk*(dBydx(i,j,k1)), $  ; dBy/dx at z=z
            dBydx(i,j1,k)*zmk1+zmk*(dBydx(i,j1,k1)), $
            dBydx(i1,j1,k)*zmk1+zmk*(dBydx(i1,j1,k1)), $
            dBydx(i1,j,k)*zmk1+zmk*(dBydx(i1,j,k1)) ]
   f2 =   [ dBydy(i,j,k)*zmk1+zmk*(dBydy(i,j,k1)), $  ; dBy/dy at z=z
            dBydy(i,j1,k)*zmk1+zmk*(dBydy(i,j1,k1)), $
            dBydy(i1,j1,k)*zmk1+zmk*(dBydy(i1,j1,k1)), $
            dBydy(i1,j,k)*zmk1+zmk*(dBydy(i1,j,k1)) ]
   f12 =   [ d2Bydxdy(i,j,k)*zmk1+zmk*(d2Bydxdy(i,j,k1)), $
             d2Bydxdy(i,j1,k)*zmk1+zmk*(d2Bydxdy(i,j1,k1)), $
             d2Bydxdy(i1,j1,k)*zmk1+zmk*(d2Bydxdy(i1,j1,k1)), $
             d2Bydxdy(i1,j,k)*zmk1+zmk*(d2Bydxdy(i1,j,k1)) ]
   ;f = [ansf00z,ansf10z,ansf11z,ansf01z]
   ;f1 =   [ dBydx(i,j,k)*zmk1+zmk*(dBydx(i,j,k1)), $  ; dBy/dx at z=z
   ;         dBydx(i1,j,k)*zmk1+zmk*(dBydx(i1,j,k1)), $
   ;         dBydx(i1,j1,k)*zmk1+zmk*(dBydx(i1,j1,k1)), $
   ;         dBydx(i,j1,k)*zmk1+zmk*(dBydx(i,j1,k1)) ]
   ;f2 =   [ dBydy(i,j,k)*zmk1+zmk*(dBydy(i,j,k1)), $  ; dBy/dy at z=z
   ;         dBydy(i1,j,k)*zmk1+zmk*(dBydy(i1,j,k1)), $
   ;         dBydy(i1,j1,k)*zmk1+zmk*(dBydy(i1,j1,k1)), $
   ;         dBydy(i,j1,k)*zmk1+zmk*(dBydy(i,j1,k1)) ]
   ;f12 =   [ d2Bydxdy(i,j,k)*zmk1+zmk*(d2Bydxdy(i,j,k1)), $
   ;          d2Bydxdy(i1,j,k)*zmk1+zmk*(d2Bydxdy(i1,j,k1)), $
   ;          d2Bydxdy(i1,j1,k)*zmk1+zmk*(d2Bydxdy(i1,j1,k1)), $
   ;          d2Bydxdy(i,j1,k)*zmk1+zmk*(d2Bydxdy(i,j1,k1)) ]
   bicubic,f,f1,f2,f12,i,i1,j,j1,x,y,byout,ansf1,ansf2


;
;Bz
;

   ; B(i,j,z)
   f = [mag.bzp(i,j,k),mag.bzp(i,j1,k),mag.bzp(i,j1,k1),mag.bzp(i,j,k1)]
   f1 = [dBzdz(i,j,k),dBzdz(i,j1,k),dBzdz(i,j1,k1),dBzdz(i,j,k1)]
   f2 = [dBzdy(i,j,k),dBzdy(i,j1,k),dBzdy(i,j1,k1),dBzdy(i,j,k1)]
   f12 = [d2Bzdzdy(i,j,k),d2Bzdzdy(i,j1,k),d2Bzdzdy(i,j1,k1),d2Bzdzdy(i,j,k1)]
   ;f = [mag.bzp(i,j,k),mag.bzp(i,j,k1),mag.bzp(i,j1,k1),mag.bzp(i,j1,k)]
   ;f1 = [dBzdz(i,j,k),dBzdz(i,j,k1),dBzdz(i,j1,k1),dBzdz(i,j1,k)]
   ;f2 = [dBzdy(i,j,k),dBzdy(i,j,k1),dBzdy(i,j1,k1),dBzdy(i,j1,k)]
   ;f12 = [d2Bzdzdy(i,j,k),d2Bzdzdy(i,j,k1),d2Bzdzdy(i,j1,k1),d2Bzdzdy(i,j1,k)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf00z,ansf100z,ansf200z
   ; B(i,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf01z,ansf101z,ansf201z
   ; B(i+1,j,z)
   f = [mag.bzp(i1,j,k),mag.bzp(i1,j1,k),mag.bzp(i1,j1,k1),mag.bzp(i1,j,k1)]
   f1 = [dBzdz(i1,j,k),dBzdz(i1,j1,k),dBzdz(i1,j1,k1),dBzdz(i1,j,k1)]
   f2 = [dBzdy(i1,j,k),dBzdy(i1,j1,k),dBzdy(i1,j1,k1),dBzdy(i1,j,k1)]
   f12 = [d2Bzdzdy(i1,j,k),d2Bzdzdy(i1,j1,k),d2Bzdzdy(i1,j1,k1),d2Bzdzdy(i1,j,k1)]
   ;f = [mag.bzp(i1,j,k),mag.bzp(i1,j,k1),mag.bzp(i1,j1,k1),mag.bzp(i1,j1,k)]
   ;f1 = [dBzdz(i1,j,k),dBzdz(i1,j,k1),dBzdz(i1,j1,k1),dBzdz(i1,j1,k)]
   ;f2 = [dBzdy(i1,j,k),dBzdy(i1,j,k1),dBzdy(i1,j1,k1),dBzdy(i1,j1,k)]
   ;f12 = [d2Bzdzdy(i1,j,k),d2Bzdzdy(i1,j,k1),d2Bzdzdy(i1,j1,k1),d2Bzdzdy(i1,j1,k)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf10z,ansf110z,ansf210z
   ; B(i+1,j1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf11z,ansf111z,ansf211z
   f = [ansf00z,ansf01z,ansf11z,ansf10z]
   f1 =   [ dBzdx(i,j,k)*zmk1+zmk*(dBzdx(i,j,k1)), $  ; dBz/dx at z=z
            dBzdx(i,j1,k)*zmk1+zmk*(dBzdx(i,j1,k1)), $
            dBzdx(i1,j1,k)*zmk1+zmk*(dBzdx(i1,j1,k1)), $
            dBzdx(i1,j,k)*zmk1+zmk*(dBzdx(i1,j,k1)) ]
   f2 =   [ dBzdy(i,j,k)*zmk1+zmk*(dBzdy(i,j,k1)), $  ; dBz/dy at z=z
            dBzdy(i,j1,k)*zmk1+zmk*(dBzdy(i,j1,k1)), $
            dBzdy(i1,j1,k)*zmk1+zmk*(dBzdy(i1,j1,k1)), $
            dBzdy(i1,j,k)*zmk1+zmk*(dBzdy(i1,j,k1)) ]
   f12 =   [ d2Bzdxdy(i,j,k)*zmk1+zmk*(d2Bzdxdy(i,j,k1)), $
             d2Bzdxdy(i,j1,k)*zmk1+zmk*(d2Bzdxdy(i,j1,k1)), $
             d2Bzdxdy(i1,j1,k)*zmk1+zmk*(d2Bzdxdy(i1,j1,k1)), $
             d2Bzdxdy(i1,j,k)*zmk1+zmk*(d2Bzdxdy(i1,j,k1)) ]
   ;f = [ansf00z,ansf10z,ansf11z,ansf01z]
   ;f1 =   [ dBzdx(i,j,k)*zmk1+zmk*(dBzdx(i,j,k1)), $  ; dBz/dx at z=z
   ;         dBzdx(i1,j,k)*zmk1+zmk*(dBzdx(i1,j,k1)), $
   ;         dBzdx(i1,j1,k)*zmk1+zmk*(dBzdx(i1,j1,k1)), $
   ;         dBzdx(i,j1,k)*zmk1+zmk*(dBzdx(i,j1,k1)) ]
   ;f2 =   [ dBzdy(i,j,k)*zmk1+zmk*(dBzdy(i,j,k1)), $  ; dBz/dy at z=z
   ;         dBzdy(i1,j,k)*zmk1+zmk*(dBzdy(i1,j,k1)), $
   ;         dBzdy(i1,j1,k)*zmk1+zmk*(dBzdy(i1,j1,k1)), $
   ;         dBzdy(i,j1,k)*zmk1+zmk*(dBzdy(i,j1,k1)) ]
   ;f12 =   [ d2Bzdxdy(i,j,k)*zmk1+zmk*(d2Bzdxdy(i,j,k1)), $
   ;          d2Bzdxdy(i1,j,k)*zmk1+zmk*(d2Bzdxdy(i1,j,k1)), $
   ;          d2Bzdxdy(i1,j1,k)*zmk1+zmk*(d2Bzdxdy(i1,j1,k1)), $
   ;          d2Bzdxdy(i,j1,k)*zmk1+zmk*(d2Bzdxdy(i,j1,k1)) ]
   bicubic,f,f1,f2,f12,i,i1,j,j1,x,y,bzout,ansf1,ansf2

;   print,mag.bxp(i,j,k),mag.byp(i,j,k),mag.bzp(i,j,k)
;   print,mag.bxp(i1,j1,k1),mag.byp(i1,j1,k1),mag.bzp(i1,j1,k1)
   return,[bxout,byout,bzout]

end
