
;+
;NAME:
;     BICUBIC3D
;PURPOSE:
;     Does a 3D bicubic interpolation of the 3D magnetic field
;CATEGORY:
;CALLING SEQUENCE:
;     binterp = bicubic3d(mag,x,y,z,alpha)
;INPUTS:
;     mag = magnetic field structure containing at least BXP,BYP,BZP
;     x,y,z = the coordinates of the field point
;     alpha = force free alpha (used to compute the derivatives)
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     reset = force the computation of the derivatives.  Otherwise they are
;             computed only when the routine is called the first time.
;OUTPUTS:
;     binterp = a singe vector containing the interpolated field [Bx,By,Bz]
;COMMON BLOCKS:
;     Xbicubic3d_derivs = Makes the derivatives into static variables
;     Ybicubic3d_derivs = Makes the derivatives into static variables
;     Zbicubic3d_derivs = Makes the derivatives into static variables
;     times_called3d = contains static variable ntimes
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
; Performs several 2D bicubic interpolations to build up the 3D interpolation.
; The derivatives are computed numerically and the divB=0 and the force-free
; conditions are used where possible for consistency.
;MODIFICATION HISTORY:
;
;       Feb 4, 1992  T. Metcalf
;-

function bicubic3d,mag,x,y,z,alpha,reset=reset

   ; The common blocks make the derivatives into static variables

   common Xbicubic3d_derivs,dBxdx,dBxdy,dBxdz,d2Bxdzdy,d2Bxdxdy
   common Ybicubic3d_derivs,dBydx,dBydy,dBydz,d2Bydzdy,d2Bydxdy
   common Zbicubic3d_derivs,dBzdx,dBzdy,dBzdz,d2Bzdzdy,d2Bzdxdy
   common times_called3d,ntimes

   nx = n_elements(mag.bxp(*,0,0))
   ny = n_elements(mag.bxp(0,*,0))
   nz = n_elements(mag.bxp(0,0,*))

   i = (fix(x) > 0) < (nx-2)
   j = (fix(y) > 0) < (ny-2)
   k = (fix(z) > 0) < (nz-2)


;   bx = mag.bxp
;   by = mag.byp
;   bz = mag.bzp

   if n_elements(ntimes) LE 0 then ntimes = 0L $
   else ntimes = 1L

   ; compute derivatives on the first pass or if reset keyword is present

   if keyword_set(reset) or (NOT ntimes) then begin

      dx = 2.0
      dy = 2.0
      dz = 2.0

      dBxdx = (shift(mag.bxp,-1,0,0) - shift(mag.bxp,1,0,0))/dx
      dBxdy = (shift(mag.bxp,0,-1,0) - shift(mag.bxp,0,1,0))/dy
      dBxdz = (shift(mag.bxp,0,0,-1) - shift(mag.bxp,0,0,1))/dz
      d2Bxdzdy = (shift(dBxdz,0,-1,0) - shift(dBxdz,0,1,0))/dy
      d2Bxdxdy = (shift(dBxdx,0,-1,0) - shift(dBxdx,0,1,0))/dy

;      dBydx = (shift(mag.byp,-1,0,0) - shift(mag.byp,1,0,0))/dx
      dBydx = dBxdy + alpha*mag.bzp      ; Force free condition
      dBydy = (shift(mag.byp,0,-1,0) - shift(mag.byp,0,1,0))/dy
      dBydz = (shift(mag.byp,0,0,-1) - shift(mag.byp,0,0,1))/dz
      d2Bydzdy = (shift(dBydz,0,-1,0) - shift(dBydz,0,1,0))/dy
      d2Bydxdy = (shift(dBydx,0,-1,0) - shift(dBydx,0,1,0))/dy

;      dBzdx = (shift(mag.bzp,-1,0,0) - shift(mag.bzp,1,0,0))/dx
      dBzdx = dBxdz - alpha*mag.byp      ; Force free condition
;      dBzdy = (shift(mag.bzp,0,-1,0) - shift(mag.bzp,0,1,0))/dy
      dBzdy = dBydz + alpha*mag.bxp      ; Force free condition
;      dBzdz = (shift(mag.bzp,0,0,-1) - shift(mag.bzp,0,0,1))/dz
      dBzdz = -dBydy - dBxdx        ; divB = 0
      d2Bzdzdy = (shift(dBzdz,0,-1,0) - shift(dBzdz,0,1,0))/dy
      d2Bzdxdy = (shift(dBzdx,0,-1,0) - shift(dBzdx,0,1,0))/dy

      print,'INFO: bicubic3d: Derivatives computed'

   endif

   zmk = float(z-k)
   zmk1 = 1.0-zmk

   i1 = i+1
   j1 = j+1
   k1 = k+1

;
; Bx
;

   ; B(i,j,z)
   f = [mag.bxp(i,j,k),mag.bxp(i,j1,k),mag.bxp(i,j1,k1),mag.bxp(i,j,k1)]
   f1 = [dBxdz(i,j,k),dBxdz(i,j1,k),dBxdz(i,j1,k1),dBxdz(i,j,k1)]
   f2 = [dBxdy(i,j,k),dBxdy(i,j1,k),dBxdy(i,j1,k1),dBxdy(i,j,k1)]
   f12 = [d2Bxdzdy(i,j,k),d2Bxdzdy(i,j1,k),d2Bxdzdy(i,j1,k1),d2Bxdzdy(i,j,k1)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf00z,ansf100z,ansf200z
   ; B(i,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf01z,ansf101z,ansf201z
   ; B(i+1,j,z)
   f = [mag.bxp(i1,j,k),mag.bxp(i1,j1,k),mag.bxp(i1,j1,k1),mag.bxp(i1,j,k1)]
   f1 = [dBxdz(i1,j,k),dBxdz(i1,j1,k),dBxdz(i1,j1,k1),dBxdz(i1,j,k1)]
   f2 = [dBxdy(i1,j,k),dBxdy(i1,j1,k),dBxdy(i1,j1,k1),dBxdy(i1,j,k1)]
   f12=[d2Bxdzdy(i1,j,k),d2Bxdzdy(i1,j1,k),d2Bxdzdy(i1,j1,k1),d2Bxdzdy(i1,j,k1)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf10z,ansf110z,ansf210z
   ; B(i+1,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf11z,ansf111z,ansf211z
   f = [ansf00z,ansf01z,ansf11z,ansf10z]
   f1 =   [ dBxdx(i,j,k)*zmk1+zmk*(dBxdx(i,j,k1)), $ ;dBx/dx @ z=z (lin interp)
            dBxdx(i,j1,k)*zmk1+zmk*(dBxdx(i,j1,k1)), $
            dBxdx(i1,j1,k)*zmk1+zmk*(dBxdx(i1,j1,k1)), $
            dBxdx(i1,j,k)*zmk1+zmk*(dBxdx(i1,j,k1)) ]
   f2 =   [ dBxdy(i,j,k)*zmk1+zmk*(dBxdy(i,j,k1)), $  ; dBx/dy at z=z
            dBxdy(i,j1,k)*zmk1+zmk*(dBxdy(i,j1,k1)), $
            dBxdy(i1,j1,k)*zmk1+zmk*(dBxdy(i1,j1,k1)), $
            dBxdy(i1,j,k)*zmk1+zmk*(dBxdy(i1,j,k1)) ]
   f12 =   [ d2Bxdxdy(i,j,k)*zmk1+zmk*(d2Bxdxdy(i,j,k1)), $
             d2Bxdxdy(i,j1,k)*zmk1+zmk*(d2Bxdxdy(i,j1,k1)), $
             d2Bxdxdy(i1,j1,k)*zmk1+zmk*(d2Bxdxdy(i1,j1,k1)), $
             d2Bxdxdy(i1,j,k)*zmk1+zmk*(d2Bxdxdy(i1,j,k1)) ]
   bicubic,f,f1,f2,f12,i,i1,j,j1,x,y,bxout,ansf1,ansf2

;
; By
;

   ; B(i,j,z)
   f = [mag.byp(i,j,k),mag.byp(i,j1,k),mag.byp(i,j1,k1),mag.byp(i,j,k1)]
   f1 = [dBydz(i,j,k),dBydz(i,j1,k),dBydz(i,j1,k1),dBydz(i,j,k1)]
   f2 = [dBydy(i,j,k),dBydy(i,j1,k),dBydy(i,j1,k1),dBydy(i,j,k1)]
   f12 = [d2Bydzdy(i,j,k),d2Bydzdy(i,j1,k),d2Bydzdy(i,j1,k1),d2Bydzdy(i,j,k1)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf00z,ansf100z,ansf200z
   ; B(i,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf01z,ansf101z,ansf201z
   ; B(i+1,j,z)
   f = [mag.byp(i1,j,k),mag.byp(i1,j1,k),mag.byp(i1,j1,k1),mag.byp(i1,j,k1)]
   f1 = [dBydz(i1,j,k),dBydz(i1,j1,k),dBydz(i1,j1,k1),dBydz(i1,j,k1)]
   f2 = [dBydy(i1,j,k),dBydy(i1,j1,k),dBydy(i1,j1,k1),dBydy(i1,j,k1)]
   f12 = [d2Bydzdy(i1,j,k),d2Bydzdy(i1,j1,k),d2Bydzdy(i1,j1,k1),d2Bydzdy(i1,j,k1)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf10z,ansf110z,ansf210z
   ; B(i+1,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf11z,ansf111z,ansf211z
   f = [ansf00z,ansf01z,ansf11z,ansf10z]
   f1 =   [ dBydx(i,j,k)*zmk1+zmk*(dBydx(i,j,k1)), $  ; dBy/dx at z=z
            dBydx(i,j1,k)*zmk1+zmk*(dBydx(i,j1,k1)), $
            dBydx(i1,j1,k)*zmk1+zmk*(dBydx(i1,j1,k1)), $
            dBydx(i1,j,k)*zmk1+zmk*(dBydx(i1,j,k1)) ]
   f2 =   [ dBydy(i,j,k)*zmk1+zmk*(dBydy(i,j,k1)), $  ; dBy/dy at z=z
            dBydy(i,j1,k)*zmk1+zmk*(dBydy(i,j1,k1)), $
            dBydy(i1,j1,k)*zmk1+zmk*(dBydy(i1,j1,k1)), $
            dBydy(i1,j,k)*zmk1+zmk*(dBydy(i1,j,k1)) ]
   f12 =   [ d2Bydxdy(i,j,k)*zmk1+zmk*(d2Bydxdy(i,j,k1)), $
             d2Bydxdy(i,j1,k)*zmk1+zmk*(d2Bydxdy(i,j1,k1)), $
             d2Bydxdy(i1,j1,k)*zmk1+zmk*(d2Bydxdy(i1,j1,k1)), $
             d2Bydxdy(i1,j,k)*zmk1+zmk*(d2Bydxdy(i1,j,k1)) ]
   bicubic,f,f1,f2,f12,i,i1,j,j1,x,y,byout,ansf1,ansf2


;
;Bz
;

   ; B(i,j,z)
   f = [mag.bzp(i,j,k),mag.bzp(i,j1,k),mag.bzp(i,j1,k1),mag.bzp(i,j,k1)]
   f1 = [dBzdz(i,j,k),dBzdz(i,j1,k),dBzdz(i,j1,k1),dBzdz(i,j,k1)]
   f2 = [dBzdy(i,j,k),dBzdy(i,j1,k),dBzdy(i,j1,k1),dBzdy(i,j,k1)]
   f12 = [d2Bzdzdy(i,j,k),d2Bzdzdy(i,j1,k),d2Bzdzdy(i,j1,k1),d2Bzdzdy(i,j,k1)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf00z,ansf100z,ansf200z
   ; B(i,j+1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf01z,ansf101z,ansf201z
   ; B(i+1,j,z)
   f = [mag.bzp(i1,j,k),mag.bzp(i1,j1,k),mag.bzp(i1,j1,k1),mag.bzp(i1,j,k1)]
   f1 = [dBzdz(i1,j,k),dBzdz(i1,j1,k),dBzdz(i1,j1,k1),dBzdz(i1,j,k1)]
   f2 = [dBzdy(i1,j,k),dBzdy(i1,j1,k),dBzdy(i1,j1,k1),dBzdy(i1,j,k1)]
   f12 = [d2Bzdzdy(i1,j,k),d2Bzdzdy(i1,j1,k),d2Bzdzdy(i1,j1,k1),d2Bzdzdy(i1,j,k1)]
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j,ansf10z,ansf110z,ansf210z
   ; B(i+1,j1,z)
   bicubic,f,f1,f2,f12,k,k1,j,j1,z,j1,ansf11z,ansf111z,ansf211z
   f = [ansf00z,ansf01z,ansf11z,ansf10z]
   f1 =   [ dBzdx(i,j,k)*zmk1+zmk*(dBzdx(i,j,k1)), $  ; dBz/dx at z=z
            dBzdx(i,j1,k)*zmk1+zmk*(dBzdx(i,j1,k1)), $
            dBzdx(i1,j1,k)*zmk1+zmk*(dBzdx(i1,j1,k1)), $
            dBzdx(i1,j,k)*zmk1+zmk*(dBzdx(i1,j,k1)) ]
   f2 =   [ dBzdy(i,j,k)*zmk1+zmk*(dBzdy(i,j,k1)), $  ; dBz/dy at z=z
            dBzdy(i,j1,k)*zmk1+zmk*(dBzdy(i,j1,k1)), $
            dBzdy(i1,j1,k)*zmk1+zmk*(dBzdy(i1,j1,k1)), $
            dBzdy(i1,j,k)*zmk1+zmk*(dBzdy(i1,j,k1)) ]
   f12 =   [ d2Bzdxdy(i,j,k)*zmk1+zmk*(d2Bzdxdy(i,j,k1)), $
             d2Bzdxdy(i,j1,k)*zmk1+zmk*(d2Bzdxdy(i,j1,k1)), $
             d2Bzdxdy(i1,j1,k)*zmk1+zmk*(d2Bzdxdy(i1,j1,k1)), $
             d2Bzdxdy(i1,j,k)*zmk1+zmk*(d2Bzdxdy(i1,j,k1)) ]
   bicubic,f,f1,f2,f12,i,i1,j,j1,x,y,bzout,ansf1,ansf2

;   print,mag.bxp(i,j,k),mag.byp(i,j,k),mag.bzp(i,j,k)
;   print,mag.bxp(i1,j1,k1),mag.byp(i1,j1,k1),mag.bzp(i1,j1,k1)
   return,[bxout,byout,bzout]

end
