;+
;NAME:
;     binterp
;PURPOSE:
;     Linearly interpolates the 3D magnetic field between grid points
;CATEGORY:
;CALLING SEQUENCE:
;     b = binterp(mag,x,y,z)
;INPUTS:
;     mag = magnetic field structure containing at least BXP,BYP,BZP
;     x,y,z = the coordinates of the field point
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;OUTPUTS:
;     b = a singe vector containing the interpolated field [Bx,By,Bz]
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
;     Very straightforward
;MODIFICATION HISTORY:
;
;       Feb 4, 1992  T. Metcalf
;       March 31, 1993  T. Metcalf  fixed a slight bug
;-

function binterp,mag,x,y,z

; Bilinear interpolation in 3-D

   nx = n_elements(mag.bxp(*,0,0))
   ny = n_elements(mag.bxp(0,*,0))
   nz = n_elements(mag.bxp(0,0,*))

   i = (fix(x) > 0) < (nx-2)
   j = (fix(y) > 0) < (ny-2)
   k = (fix(z) > 0) < (nz-2)

   t = x-i
   u = y-j
   v = z-k

   Bx =   (1.-t)*(1.-u)*(1.-v)*mag.bxp(i,j,k) $         ;  0 0 0 
        + (1.-t)*(1.-u)*v*mag.bxp(i,j,k+1) $            ;  0 0 1
        + (1.-t)*u*(1.-v)*mag.bxp(i,j+1,k) $            ;  0 1 0 
        + (1.-t)*u*v*mag.bxp(i,j+1,k+1) $               ;  0 1 1
        + t*(1.-u)*(1.-v)*mag.bxp(i+1,j,k) $            ;  1 0 0 
        + t*(1.-u)*v*mag.bxp(i+1,j,k+1) $               ;  1 0 1
        + t*u*(1.-v)*mag.bxp(i+1,j+1,k) $               ;  1 1 0
        + t*u*v*mag.bxp(i+1,j+1,k+1)                    ;  1 1 1
        
   By =   (1.-t)*(1.-u)*(1.-v)*mag.byp(i,j,k) $         ;  0 0 0 
        + (1.-t)*(1.-u)*v*mag.byp(i,j,k+1) $            ;  0 0 1
        + (1.-t)*u*(1.-v)*mag.byp(i,j+1,k) $            ;  0 1 0 
        + (1.-t)*u*v*mag.byp(i,j+1,k+1) $               ;  0 1 1
        + t*(1.-u)*(1.-v)*mag.byp(i+1,j,k) $            ;  1 0 0 
        + t*(1.-u)*v*mag.byp(i+1,j,k+1) $               ;  1 0 1
        + t*u*(1.-v)*mag.byp(i+1,j+1,k) $               ;  1 1 0
        + t*u*v*mag.byp(i+1,j+1,k+1)                    ;  1 1 1
        
   Bz =   (1.-t)*(1.-u)*(1.-v)*mag.bzp(i,j,k) $         ;  0 0 0 
        + (1.-t)*(1.-u)*v*mag.bzp(i,j,k+1) $            ;  0 0 1
        + (1.-t)*u*(1.-v)*mag.bzp(i,j+1,k) $            ;  0 1 0 
        + (1.-t)*u*v*mag.bzp(i,j+1,k+1) $               ;  0 1 1
        + t*(1.-u)*(1.-v)*mag.bzp(i+1,j,k) $            ;  1 0 0 
        + t*(1.-u)*v*mag.bzp(i+1,j,k+1) $               ;  1 0 1
        + t*u*(1.-v)*mag.bzp(i+1,j+1,k) $               ;  1 1 0
        + t*u*v*mag.bzp(i+1,j+1,k+1)                    ;  1 1 1

   return,[Bx,By,Bz]
        
end
