
;+
;NAME:
;     bline
;PURPOSE:
;     Plots a single line of force for a 3D magnetic field
;CATEGORY:
;CALLING SEQUENCE:
;     bline,mag,xpix,ypix,zscale,alpha
;INPUTS:
;     mag = magnetic field structure computed by bff.pro.  Must contain at least
;           BXP,BYP,BZP
;     xpix = the x pixel number to start the line at
;     ypix = the y pixel number to start the line at
;     zscale = sets zscale (1 z pixel = zscale*xy pixel)
;     alpha = the force-free alpha
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     /backwards = plot the line backwards (useful for Bz(xpix,ypix)<0)
;     Bcolor = color to draw the line with
;     /reset = Used to reset the derivative computation in bicubic3d
;     /fast = Use the fast interpolation (slightly less accurate?)
;     position = plot window specs for contour, oplot etc., [x0,y0,x1,y1,z0,z1]
;     /nofilldot = Don't fill dot which marks the footpoint. This causes the 
;             zbuffer to crash and you can turn the filling off with nofilldot.
;     /nospot = do not plot footpoint spots
;     spot_color = color index for footpoint spots
;     /quiet = work quietly
;OUTPUTS:
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Not very useful by itself.  Use blines.pro as a front end.
;PROCEDURE:
;     Uses Runge-Kutta diffeq solver to integrate the field line.
;     A spot is plotted on z=0 where the field line
;     hits z=0.  This helps with interpreting the 3D plot.
;MODIFICATION HISTORY:
;
;       Feb 4, 1992  T. Metcalf
;       2001 Nov 2 T. Metcalf, Fixed bug in fld_interp which did not set the
;                              flag variable.  This eliminated the long pauses
;                              for some field lines.
;-

; Plot line of force

pro bline, magin, xpix, ypix, zscalein, alphain, vertices, $
           backwards=backwardsin, Bcolor=Bcolor,resetin=resetin, $
           fastin=fastin,position=position,bthick=inbthick, $
           nofilldot=nofilldot,nospot=nospot,spot_color=spot_color,quiet=quiet

COMMON fld_interp_comm,mag,zscale,backwards,fast,alpha,reset,count

   count = 0L
   mag = magin
   if n_elements(zscalein) LE 0 then zscale = 1.0 else zscale=zscalein
   if n_elements(backwardsin) LE 0 then backwards=0 else backwards=backwardsin
   if n_elements(fastin) LE 0 then fast = 0 else fast = fastin
   if n_elements(alphain) LE 0 then alpha=0. else alpha = alphain
   if n_elements(resetin) LE 0 then reset=0 else reset=resetin

   if n_elements(inbthick) NE 1 then bthick=1 else bthick=inbthick

   if n_elements(spot_color) LE 0 then spot_color=1

   if keyword_set(nofilldot) then $
      usersym,[0.,0.707107,1.,0.707107,0.,-0.707107,-1.,-0.707107], $
              [1.,0.707107,0.,-0.707107,-1.,-0.707107,0.,0.707107] $
   else $
      usersym,[0.,0.707107,1.,0.707107,0.,-0.707107,-1.,-0.707107], $
              [1.,0.707107,0.,-0.707107,-1.,-0.707107,0.,0.707107],/fill

   nx = n_elements(mag.bxp(*,0,0))
   ny = n_elements(mag.bxp(0,*,0))
   nz = n_elements(mag.bxp(0,0,*))

   ; Set up cube for 3D plotting
   !X.S = [position(0),(position(2)-position(0))/(nx-1.-(0.))]
   !Y.S = [position(1),(position(3)-position(1))/(ny-1.-(0.))]
   ;!Z.S = [position(4),(position(5)-position(4))/(zscale*(nz-1.)-(0.))]
   ; Set z-scale to same as x,y scales
   ;nzeff = ((1./!X.S(1)) > (1./!Y.S(1)))*(position(5)-position(4))
   !Z.S = [position(4),!X.S(1)<!Y.S(1)]
   
   x = float(xpix)
   y = float(ypix)
   z = 0.

   limit = long(5.*sqrt(nx^2+ny^2+nz^2))

   if NOT keyword_set(nospot) then $
      oplot,[x],[y],zvalue=0.0,psym=8,/noclip,/t3d, $
            symsize=1.,color=spot_color  ; Plot spot at z=0

   ; Begin storing the computed vertices

   vertices = [-1000,1,Bcolor,x,y,z]   ; -1000 = START   (-2000 = end-of-line)
                                       ; 1 = DOT         (0 = no dot)
                                       ; Bcolor=color    (stores current color)

   diffeq_tfl,'fld_interp',[x,y,z],0.0,limit,ss,rr, $
             TOL=1e-4,DEPVAR=['X','Y','Z'], $
             ylbound=0.,yubound=[nx-1,ny-1,(nz-1)*zscale]

   Npt = n_elements(rr(0,*))
   maxss = max(ss)
   if maxss GE limit then message,/info,'Maximum field line length exceeded'

   if Npt GT 2 then begin  ; Avoid short lines

      xnew = rr(0,Npt-1)
      ynew = rr(1,Npt-1)
      znew = rr(2,Npt-1)
      plot_dot = (znew LE 0.)
      ;print,xnew,ynew,znew,plot_dot

      Npt = Npt-1   ; drop last pt (thru wall?)
      ss = ss(0:Npt-1) & rr = rr(*,0:Npt-1) 

      x = rr(0,Npt-1)
      y = rr(1,Npt-1)
      z = rr(2,Npt-1)
      B = binterp(mag,x,y,z/zscale)
      Bn = B/sqrt(B(0)^2+B(1)^2+B(2)^2)
      if keyword_set(backwards) then Bn = -Bn
      ;print,x,y,z,Bn

      ; The line is done - finish off with a line segment reaching to the 
      ; boundary
      case 1 of 
        (xnew LT 0.): tstep = abs((x)/Bn(0))
        (ynew LT 0.): tstep = abs((y)/Bn(1))
        (znew LT 0.): tstep = abs((z)/Bn(2))
        (xnew GE (nx-1)): tstep = abs((x-nx+1)/Bn(0))
        (ynew GE (ny-1)): tstep = abs((y-ny+1)/Bn(1))
        (znew GE (nz-1)*zscale): tstep = abs((z-(nz-1)*zscale)/(Bn(2)))
        else: tstep=0.0
      endcase
      if maxss LT limit then begin
         xnew = x + Bn(0)*tstep
         ynew = y + Bn(1)*tstep
         znew = z + Bn(2)*tstep
      endif

      ;print,xnew,ynew,znew,tstep

      if plot_dot AND NOT keyword_set(nospot) then $
         oplot,[xnew],[ynew],zvalue=znew, $
                             psym=8, $
                             /noclip,/t3d, $
                             symsize=1., $
                             color = spot_color

      for i=0,Npt-1 do begin
         vertices = [vertices,rr(0,i),rr(1,i),rr(2,i)]
      endfor
      vertices = [vertices,xnew,ynew,znew]

      plots,[reform(rr(0,*)),xnew], $
            [reform(rr(1,*)),ynew], $
            [reform(rr(2,*)),znew], $
            /T3D,/DATA,color=Bcolor,/noclip,thick=bthick

      if NOT keyword_set(quiet) then begin
         print,strcompress('Field line length is '+string(ss(n_elements(ss)-1))+' in x-y pixels')
         print,strcompress('Field line height is '+string(max(rr(2,*)))+' in x-y pixels')
      endif


   endif $  ; end of avoid short lines
   else begin
      if NOT keyword_set(quiet) then print,'Short line ... no plot'
      plot_dot=0
   endelse

   if plot_dot then vertices = [vertices,-2000,1,Bcolor] $
   else vertices = [vertices,-2000,0,Bcolor] 

 end


FUNCTION fld_interp,s,r
COMMON fld_interp_comm,mag,zscale,backwards,fast,alpha,reset,count
COMMON DIFFEQCOM,flag
   if n_elements(count) LE 0 then count=0L
   count = count+1L
   ; Interpolate x,y,z/zscale.  Note: z/zscale = grid point number
   if fast then $
      B = binterp(mag,r(0),r(1),r(2)/zscale) $
   else $
      B = bicubic3d(mag,r(0),r(1),r(2)/zscale,alpha, $
          reset=(reset AND count EQ 1L))
   Bmag = sqrt(B(0)^2+B(1)^2+B(2)^2)
   if Bmag NE 0.0 then begin
      Bn = B/Bmag 
      flag = 0
   endif else begin 
      Bn = B
      flag = 1
   endelse
   if keyword_set(backwards) then Bn = -Bn
   return,Bn
END
