
;+
;NAME:
;     blines
;PURPOSE:
;     Plots magnetic lines of force for the input 3D magnetic field.  The lines
;     originating in different polarities of the vertical field have different
;     colors.  Footpoints (z=0) are marked with a dot at the end of the field
;     line.
;CATEGORY:
;CALLING SEQUENCE:
;     vertices=blines(mag)
;INPUTS:
;     mag = magnetic field structure computed by bff.pro. Must contain at least
;           BXP,BYP,BZP,alpha,zscale:
;              BXP = 3D x component of the magnetic field
;              BYP = 3D y component of the magnetic field
;              BZP = 3D z component of the magnetic field
;              ALPHA = force-free parameter
;              ZSCALE = scale in the z axis, e.g. .01 expands the z-axis by 100
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     t3d:  Accepts all the transformation keywords from t3d
;     /noreset3d = Do not reset the 3D transformation (good transformation must
;                 already be defined)
;     /noerase = do not erase the display before plotting
;     zscale = sets zscale (1 z pixel = zscale*xy pixel)
;     /flux = Select field lines so that the density of the field lines
;             reflects the magnetic flux.
;     Bcut = only plots lines for fields above Bcut total field strength.
;            Default is infinity if /flux is not set, but 800 G if /flux 
;            is set.
;     Jcut = only plots lines for fields above Jcut total current
;     /nocolor = does not mess with the color table
;     /nodash = do not use dashed line for negative contours with /nocolor
;     /print_colors = uses colors which work better on the color printer
;     title = title for the plot
;     xtitle = title for x-axis (default = alpha and zscale)
;     /positive = only plot lines originating in positive Bz
;     /negative = only plot lines originating in negative Bz
;     /slow = Uses slow interpolation of field lines (slightly more accurate??)
;     /nocontours = do not plot the any contours (Bz and neutral line)
;     /nobz = do not plot Bz contours (but do plot neutral line if /neutral)
;     skip = give the number of grid points to skip between each plotted line
;     /vertices = input/output of computed vertices.  If the input variable
;                exists, the lines will be plotted without the slow
;                computations.  If you're inputting precomputed vetices, most
;                of the keywords controlling which filed lines are plotted will
;                not have any effect.
;     position = plot window specs for contour, oplot etc., [x0,y0,x1,y1,z0,z1]
;                in normalized coordinates (default = [0,0,1,1,0,1])
;     edge = a structure with tags top, bottom, left, right which give
;            the size of the plot borders in centimeters.
;            e.g. edge={top:0.0,bottom:1.0,left:0.5,right:0.5}
;     charsize = size of character on the plot
;     size = scale of the whole plot
;     xstep = The size of the x-pixels relative to the y-pixels (default = 1.)
;     xyscale = scales the x-y direction by this factor
;     image = 2D array plotted as an image underneath the contours
;     missing = value to use in interpolated image if no data is available
;     /quiet = if set, don't print informational messages
;     /zbuffer = Use z-buffering (slow!!)
;     /rot2image = rotate the 3D Transform to the image plane.  Only works if
;                 /vertices is set.
;     bthick = thickness of the field lines on the plot
;     /reselect = Reselect the field lines.  Only works if zbuffer is off, 
;                /vertices is selected, and magfield contains the 3D field.
;     magfield = 3-D magnetic field to be used with reselect
;     pixsize = pixel size in arc seconds.  If set, the axes will show 
;               arcseconds rather than pixel number.  This doesn't work very
;               well if the /rot2image keyword is set since the axes do not
;               fall along the contours in this case.
;     levels = vector, specifies the Bz contour levels
;     sub = [x0,x1,y0,y1] specifying a sub-range of the plot to show
;     outposition = keyword used to return plotting position
;     /nospot = do not plot footpoint spots
;     /neutral = plot the magnetic neutral line
;     nthick = the thickness of the neutral line
;     /sample = do not interpolate in poly_2d when warping the image
;     /noaxes = do not draw axes
;     spot_color = color index of footpoint spot 
;     use_colorP = color index for positive field contours
;     use_colorM = color index for negative field contours
;     use_colorN = color index for the neutral line
;     use_colorlinP = color index for field lines which start in positive field
;     use_colorlinM = color index for field lines which start in negative field
;     pspixelsize = controls the size of postscript pixels.  Larger number is
;                   smaller pixels (i.e., larger number is more pixels).  
;                   Default = 1.  pspixelsize=10 gives a HUGE postscript file.
;     contourdata = Data for contours.  Must be the same size as bz.
;     /noselect = do not ask for interactive selection of field lines.
;OUTPUTS:
;     vertices = a structure of lines segment coordinates for further plotting
;COMMON BLOCKS:
;SIDE EFFECTS:
;     Alters the color table.
;RESTRICTIONS:
;     Z-buffering doesn't work with PostScript.
;PROCEDURE:
;     Makes repeated calls to bline.pro (one call for each line).  The aspect
;     ratio of the window is taken into account so that the pixels are the 
;     correct shape (default = square).
;MODIFICATION HISTORY:
;
;       Feb 4, 1992  T. Metcalf
;-

function blines,inmag,translate=translate,scale=scale,reset=reset,rotate=rotate,perspective=perspective,oblique=oblique,xyexch=xyexch,xzexch=xzexch,yzexch=yzexch,noreset3d=noreset3d,zscale=zscale,Bcut=Bcut,nocolor=nocolor,nodash=nodash, $
print_colors=print_colors,title=title,positive=positive,negative=negative,nocontours=nocontours,skip=skip,vertices=vertices, $
position=inposition,xstep=xstep,xyscale=xyscale,charsize=charsize,image=image,quiet=quiet,zbuffer=zbuffer,rot2image=rot2image,size=allsize,Jcut=Jcut,bthick=bthick,reselect=reselect, $
magfield=magfield,pixsize=pixsize,levels=levels,sub=sub,missing=missing,slow=slow,outposition=outposition,nospot=nospot,spot_color=spot_color,neutral=neutral,nthick=nthick,nobz=nobz,sample=insample,noaxes=noaxes, $
use_colorP=use_colorP,use_colorM=use_colorM,use_colorN=use_colorN,use_colorlinP=use_colorPlin,use_colorlinM=use_colorMlin,pspixelsize=pspixelsize,contourdata=contourdata,noselect=noselect, $
noerase=noerase,edge=edge,flux=flux,xtitle=xtitle


   if n_elements(insample) GT 0 then begin
      if insample EQ 0 then sample = 1 else sample=0
   endif $
   else sample=1 ; Interpolate by default

; protect the input from change
   if keyword_set(sub) then begin
      mag = subff(inmag,sub)
      if n_elements(image) GT 0 then $
         dimage = float(image(sub(0):sub(1),sub(2):sub(3)))
   endif $
   else begin
      mag = inmag
      if n_elements(image) GT 0 then dimage = float(image)
   endelse

; Start with  surfr & scale3d
;    or with  t3d,/reset & scale3d

   save_P = !P
   save_X = !X
   save_Y = !Y

   if keyword_set(reselect) and n_elements(magfield) LE 0 then $
      message,'Use the magfield keyword to pass the 3D magnetic field'

;  Set up all the defaults:

   if n_elements(nthick) LE 0 then nthick = 5
   if n_elements(slow) LE 0 then slow = 0
   if n_elements(charsize) LE 0 then charsize=1.
   if n_elements(xstep) LE 0 then xstep=1.
   if n_elements(xyscale) LE 0 then xyscale=1.
   if NOT keyword_set(vertices) then begin
      if n_elements(Bcut) LE 0 then begin
         if keyword_set(flux) then Bcut = 800. else Bcut = 1.0e30 
      endif $
      else Bcut = abs(float(Bcut))
      if NOT keyword_set(quiet) and NOT keyword_set(Jcut) then $
       print,Bcut,format="('INFO: blines: Lines plotted where B_total > ',f7.1)"
   
      if n_elements(zscale) LE 0 then zscale = mag.zscale $
      else begin
         if zscale NE mag.zscale then $
            print,'WARNING: Z Axis will not be properly scaled.',zscale, $
                  mag.zscale
      endelse
   
      ;if n_elements(step) LE 0 then begin
      ;   step = 0.1
      ;   ;if keyword_set(fast) then step = 0.1 $
      ;   ;else step = 0.1;*(0.01/float(zscale))
      ;   if NOT keyword_set(quiet) then $
      ;      print,step,format="('INFO: blines: Step size set to ',f3.1)"
      ;endif
   
      if n_elements(skip) LE 0 then skip = 1 $
      else skip = fix(skip>0)+1
   endif $
   else begin
      if tag_index(mag,'BCUT',/quiet) GE 0 then $
         if NOT keyword_set(quiet) then print,'Bcut is ',mag.Bcut
      ;if tag_index(mag,'STEP',/quiet) GE 0 then $
      ;   if NOT keyword_set(quiet) then print,'Step size is ',mag.step
      if tag_index(mag,'SKIP',/quiet) GE 0 then $
         if NOT keyword_set(quiet) then print,'Skip is ',mag.skip
      if n_elements(zscale) LE 0 then begin
         if tag_index(mag,'ZSCALE',/quiet) GE 0 then begin
            zscale = mag.zscale
         endif $
         else zscale = .01
      endif
      if NOT keyword_set(quiet) then print,'zscale is set to ',zscale
      if n_elements(title) LE 0 and $
         tag_index(mag,'TITLE',/quiet) GE 0 then title=mag.title
   endelse
   if tag_index(mag,'PIX_SIZE',/quiet) GE 0 then $
      if NOT keyword_set(quiet) then $
      print,min(mag.pix_size),format="('INFO: blines: Pixel size is',f6.2,' arcsec')"

   if NOT keyword_set(vertices) then begin
      nx = n_elements(mag.bxp(*,0,0))
      ny = n_elements(mag.bxp(0,*,0))
      nz = n_elements(mag.bxp(0,0,*))
   endif $
   else begin
      nx = mag.nx
      ny = mag.ny
      nz = mag.nz
   endelse

   xpix = lindgen(nx,ny) MOD nx
   ypix = lindgen(nx,ny) / long(nx)
   if keyword_set(pixsize) then begin
      xpix = xpix * pixsize
      ypix = ypix * pixsize
      if tag_index(mag,'PIX_SIZE') GE 0 then begin
        if pixsize NE mag.pix_size then $
            message,/info,'Resetting pix size to '+strcompress(string(pixsize))
      endif $
      else if tag_index(mag,'POINT') GE 0 then begin
        if tag_index(mag.point,'PIX_SIZE') GE 0 then begin
           if pixsize NE min(mag.point.pix_size) then $
               message,/info,'Resetting pix size to '+strcompress(string(pixsize))
        endif
      endif
   endif $
   else pixsize = 1.0


   if keyword_set(nocolor) then begin
      cplinestyle = [0]
      if keyword_set(nodash) then cmlinestyle=[0] else cmlinestyle = [2]
   endif $
   else begin 
     cplinestyle = [0]
     cmlinestyle = [0]
   endelse


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   ; Set up color table

   default_color = !d.table_size-1

  if (!d.table_size GT 2) then begin
    if n_elements(use_colorP) LE 0 then use_colorP = 10
    if n_elements(use_colorM) LE 0 then use_colorM = 11
    if n_elements(use_colorN) LE 0 then use_colorN = 12
    if n_elements(use_colorPlin) LE 0 then use_colorPlin = 8
    if n_elements(use_colorMlin) LE 0 then use_colorMlin = 9
    if NOT (keyword_set(nocolor)) then begin
      if keyword_set(print_colors) then begin
         ; Use only values of 255 here so that the printed line is solid
         specialct,use_colorP,255,0,0           ; Red contours (positive field)
         ;specialct,use_colorM,0,0,255           ; Blue contours (negative field)
         specialct,use_colorM,255,255,0        ; Yellow contours (negative field)
         specialct,use_colorN,0,0,0
         specialct,use_colorPlin,255,255,0       ; Yellow contours (positive B)
         ;specialct,use_colorMlin,0,255,0       ; Green contours (negative B)
         specialct,use_colorMlin,255,0,255       ; Purple contours (negative B)
         specialct,!d.table_size-1,0,0,0  ; Set the foreground to black
         specialct,0,255,255,255        ; Set the background to white
      endif $
      else begin
         specialct,use_colorP,255,0,0           ; Red contours (positive field)
         specialct,use_colorM,0,125,255         ; Blue contours (negative field)
         specialct,use_colorN,0,0,0
         ;specialct,use_colorPlin,255,255,0      ; Yellow contours (positive B)
         specialct,use_colorPlin,255,0,255       ; Purple contours (positive B)
         specialct,use_colorMlin,0,255,0         ; Green contours (negative B)
         ;specialct,use_colorMlin,255,0,255      ; Purple contours (negative B)
         specialct,!d.table_size-1,0,0,0  ; Set the foreground to black
         specialct,0,255,255,255        ; Set the background to white
      endelse
    endif 
    ;else begin
    ;  if !d.name EQ 'PS' then begin
    ;     specialct,0,255,255,255        ; Set the background to white
    ;     specialct,default_color,255,255,255
    ;     specialct,use_colorP,0,0,0
    ;     specialct,use_colorM,0,0,0
    ;     specialct,use_colorN,0,0,0
    ;     specialct,use_colorPlin,0,0,0
    ;     specialct,use_colorMlin,0,0,0
    ;  endif else begin
    ;     specialct,0,0,0,0        ; Set the background to black
    ;     ;default_color = (default_color < !d.table_size-1)
    ;     specialct,default_color,255,255,255
    ;     specialct,use_colorP,255,255,255
    ;     specialct,use_colorM,255,255,255
    ;     specialct,use_colorN,255,255,255
    ;     specialct,use_colorPlin,255,255,255
    ;     specialct,use_colorMlin,255,255,255
    ;  endelse
    ;endelse
  endif $
  else begin
   use_colorP = 1  ; Default in case not on a color monitor
   use_colorM = 1
   use_colorN = 1
   use_colorPlin = 1
   use_colorMlin = 1
  endelse

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

   if n_elements(xtitle) LE 0 then $
      xtitle = "Alpha = "+strcompress(string(mag.alpha,format="(f12.3)"), $
               /remove_all)+" Zscale = "+strcompress(string(zscale, $
                                         format="(f12.3)"),/remove_all)

   if n_elements(title) LE 0 then title=' '

   if n_elements(inposition) NE 6 then position = [0.,0.,1.,1.,0.,1.] $
   else position = float(inposition)

;; Make the pixels square:
;
;   !P.POSITION = position(0:3)
;
;   deltax = (!P.POSITION(2)-!P.POSITION(0))*float(xyscale)  ; size of plot
;   deltay = (!P.POSITION(3)-!P.POSITION(1))*float(xyscale)
;
;   edge = 0.00 ; if there is a one pixel border around the plot, nx and ny are
;               ; effectively increased by 2
;
;   ; Adjust for aspect ratios
; 
;   if (nx LT ny) then begin   ; Adjust for FOV and step size aspect ratios
;     if (xstep GT ystep) then begin
;       deltax = deltax*(float(nx+edge)/float(ny+edge))
;       deltay = deltay*(float(ystep)/float(xstep))
;     endif $
;     else deltax = deltax*(float(nx+edge)/float(ny+edge))*(float(xstep)/float(ystep))
;   endif else begin
;     if (xstep GE ystep) then $
;       deltay =deltay*(float(ny+edge)/float(nx+edge))*(float(ystep)/float(xstep)) $
;     else begin
;       deltax = deltax*(float(xstep)/float(ystep))
;       deltay = deltay*(float(ny+edge)/float(nx+edge))
;     endelse
;   endelse
; 
;   ; adjust for the physical aspect ratio of the window
; 
;   cm_aspect = (float(!D.X_SIZE)/float(!D.X_PX_CM)) / $
;               (float(!D.Y_SIZE)/float(!D.Y_PX_CM))
;   deltay = deltay*cm_aspect
; 
;   ; Scale both dimensions so that they are as large as possible
; 
;   cm_x_scale = float(xyscale)*(!P.POSITION(2)-!P.POSITION(0))/deltax
;   cm_y_scale = float(xyscale)*(!P.POSITION(3)-!P.POSITION(1))/deltay
;   deltax = deltax * min([cm_x_scale,cm_y_scale])
;   deltay = deltay * min([cm_x_scale,cm_y_scale])
; 
;   ; center the plot with offsets:
;   xoffset = (0.00 > ((!P.POSITION(2)-!P.POSITION(0)-deltax)/2.00));no offsets<0
;   yoffset = (0.00 > ((!P.POSITION(3)-!P.POSITION(1)-deltay)/2.00))
; 
;   !P.POSITION = [!P.POSITION(0)+xoffset,!P.POSITION(1)+yoffset, $
;                  !P.POSITION(0)+deltax+xoffset,!P.POSITION(1)+deltay+yoffset]

   !P.POSITION = pixsquare(nx,ny,position=position(0:3),xyscale=xyscale, $
                           xstep=xstep,edge=edge)

   position(0:3) = !P.POSITION

   ; Set up cube for 3D plotting
   !X.S = [position(0),(position(2)-position(0))/(nx-1.-(0.))]
   !Y.S = [position(1),(position(3)-position(1))/(ny-1.-(0.))]
   ;!Z.S = [position(4),(position(5)-position(4))/(zscale*(nz-1.)-(0.))]
   ; Set z-scale to same as x,y scales
   ;nzeff = ((1./!X.S(1)) > (1./!Y.S(1)))*(position(5)-position(4))
   !Z.S = [position(4),!X.S(1)<!Y.S(1)]

   if n_elements(levels) LE 0 then $
      levels = [100.,200.,400.,800.,1600.,3200.]

   if NOT keyword_set(noerase) then erase

   if NOT keyword_set(noreset3d) then begin
      t3d,reset=reset,translate=translate,scale=scale,rotate=rotate, $
          perspective=perspective,oblique=oblique,xyexch=xyexch,xzexch=xzexch, $
          yzexch=yzexch
      scale3d
   endif
   if keyword_set(vertices) GT 0 then begin
      if keyword_set(rot2image) then begin
         t3d,/reset & scale3d
         scenter = cvt_to_2d(nx/2.,ny/2.,0.)  ; Intitial center

         ;xf = set_xform(mag.point.b0,mag.point.p,mag.point.lat,mag.point.cmd)
         ;ixform = fltarr(4,4)
         ;ixform(0:2,0:2) = transpose(xf.helio2image)
         ;ixform(3,3) = 1.0
         ;!P.T = ixform # !P.T 
         ;print,!P.T

         ; Rotation to image coordinates:

         t3d,rotate=[-mag.point.lat,mag.point.cmd,0.0]*!radeg
         t3d,rotate=[mag.point.b0,0.0,0.0]*!radeg
         t3d,rotate=[0.0,0.0,mag.point.p]*!radeg 

         if n_elements(allsize) EQ 1 then !P.T(3,3)=allsize

         ; The following translation insures that the magnetogram is centered
         center = cvt_to_2d(nx/2.,ny/2.,0.)
         t3d,translate=[scenter(0)-center(0),scenter(1)-center(1),0]

         ;print,!P.T


      endif
    endif $ 
    else begin
         t3d,/reset & scale3d  ; force a reset the first time through
         if n_elements(allsize) EQ 1 then !P.T(3,3)=allsize
    endelse


; Plot an image, if requested

  if n_elements(dimage) GT 0  then begin

     ; Here I assume that the image is the same size as the original magnetogram
     ; and transform the image to the heliographic coordinates.  This will
     ; distort the image unless the "rot2image keyword is set.

     nx_image = n_elements(dimage(*,0))
     ny_image = n_elements(dimage(0,*))
     x = lindgen(nx_image,ny_image) MOD nx_image
     y = long(lindgen(nx_image,ny_image) / nx_image)
     ; p,q will be the forward poly_2d transformation 
     polywarp,x,y,mag.b_pxd-min(mag.b_pxd),mag.b_pyd-min(mag.b_pyd),3,p,q
     dimage = poly_2d(dimage,p,q,sample,nx,ny,missing=-255)
     ;background = where(dimage EQ -1., bcount)
     ;if n_elements(missing) LE 0 then missing = (max(dimage)+min(dimage))/2.
     ;if bcount GT 0 then dimage(background)=missing

     ; Set up the data coordinates using a "fake" call to contour which
     ; matches the "real" call below.
     contour, mag.bz,xpix,ypix, /T3D, ZVALUE=0.0,levels=levels,  $
              title=title, $
              xst=1+4*keyword_set(noaxes), $
              yst=1+4*keyword_set(noaxes), $
              position=position, charsize=charsize, $
              c_colors=[use_colorP],color=default_color, $
              /noerase,/nodata,/noclip

     ; Set up cube for 3D plotting
     !X.S = [position(0),(position(2)-position(0))/(nx-1.-(0.))]
     !Y.S = [position(1),(position(3)-position(1))/(ny-1.-(0.))]
     ;!Z.S = [position(4),(position(5)-position(4))/(zscale*(nz-1.)-(0.))]
     ; Set z-scale to same as x,y scales
     ;nzeff = ((1./!X.S(1)) > (1./!Y.S(1)))*(position(5)-position(4))
     !Z.S = [position(4),!X.S(1)<!Y.S(1)]

     xtv = [0.,0,nx-1,nx-1]
     ytv = [0.,ny-1,ny-1,0]
     ztv = [0.,0,0,0]

     ; Get device coord of the image corners after the 3D projection
     ;dtv=convert_coord(xtv,ytv,ztv,/TO_DEVICE,/T3D,/DATA)
     ;utv = reform(dtv(0,*))
     ;vtv = reform(dtv(1,*))

     dtv = fltarr(3,4)
     dtv(*,0) = convert_coord(cvt_to_2d(0,0,0),/normal,/to_device)
     dtv(*,1) = convert_coord(cvt_to_2d(0,ny-1,0),/normal,/to_device)
     dtv(*,2) = convert_coord(cvt_to_2d(nx-1,ny-1,0),/normal,/to_device)
     dtv(*,3) = convert_coord(cvt_to_2d(nx-1,0,0),/normal,/to_device)
     utv = reform(dtv(0,*))
     vtv = reform(dtv(1,*))
  
     if !D.NAME EQ 'PS' then begin
        factor = long(max(utv)-min(utv)+1)/nx
        if n_elements(pspixelsize) LE 0 then pspixelsize=1.
        factor = factor/float(pspixelsize)
        utv = utv/factor
        vtv = vtv/factor
     endif else factor=1
     ; Warp the image to the proper shape
     polywarp,xtv,ytv,utv-min(utv),vtv-min(vtv),1,kx,ky
     newimage =poly_2d(dimage,kx,ky,sample,long(max(utv)-min(utv)+1), $
                                      long(max(vtv)-min(vtv)+1),missing=-255)
     background = where(newimage LT 0., bcount)
     foreground = where(newimage GE 0., fcount)
     if fcount LE 0 then foreground = background  ; this should never happen
     image_max = max(newimage(foreground))
     image_min = min(newimage(foreground))

     ; Rescale the image.  The bytscl call ensures that the image will not use
     ; the first 15 colors which will be reserved for contours.
     sclimage = (15b+bytscl(newimage,max=image_max,min=image_min, $
                           top=(!d.table_size-17b)))
     if keyword_set(negative) then begin
        sclimage = max(sclimage)+min(sclimage)-sclimage
     endif
     if n_elements(missing) LE 0 then missing = 15b 
     if bcount GT 0 then sclimage(background)=missing

     tv,sclimage,min(utv)*factor,min(vtv)*factor,/device,z=0, $
        xsize=(max(utv)-min(utv))*factor, $
        ysize=(max(vtv)-min(vtv))*factor

  endif

   outposition = position

;  Plot the field lines

   if n_elements(contourdata) NE n_elements(mag.bz) then contourdata = mag.bz

   if keyword_set(vertices) GT 0 then begin
      if keyword_set(zbuffer) then begin
         device_name = !D.NAME
         xsize = !D.X_SIZE & if n_elements(xsize) NE 1 then xsize=640
         ysize = !D.Y_SIZE & if n_elements(ysize) NE 1 then ysize=512
         flush = tvrd() ; Grab anything already drawn
         set_plot,'Z'
         device,set_resolution=[xsize,ysize]
         tv,flush,z=0
      endif
; Plot the contours of the vertical field
      if NOT keyword_set(nocontours) then begin
         if NOT keyword_set(nobz) then begin
            contour,contourdata,xpix,ypix,/T3D,/NOERASE,ZVALUE=0.00, $
               LEVELS=levels,c_colors=[use_colorP],color=default_color, $
               /noclip, $
               xst=1+4*keyword_set(noaxes),yst=1+4*keyword_set(noaxes), $
               position=position,charsize=charsize,c_linestyle=cplinestyle
            contour,contourdata,xpix,ypix,/T3D,/NOERASE,ZVALUE=0.00, $
               LEVELS=-1*rotate(levels,2),c_colors=[use_colorM], $
               color=default_color,/noclip,xst=1+4*keyword_set(noaxes), $
               yst=1+4*keyword_set(noaxes),position=position,xtitle=xtitle, $
               title=title,charsize=charsize,c_linestyle=cmlinestyle
         endif
         if keyword_set(neutral) then $
            contour,get_neutral(mag.bz),xpix,ypix,/T3D,/NOERASE,ZVALUE=0.00, $
               LEVELS=[0],c_colors=[use_colorN],color=default_color,/noclip, $
               xst=1+4*keyword_set(noaxes),yst=1+4*keyword_set(noaxes), $
               position=position,xtitle=xtitle,title=title,charsize=charsize, $
               c_linestyle=cplinestyle,thick=nthick
      endif
      bplot,mag,position=position,zscale=zscale,quiet=quiet,bthick=bthick, $
            nofilldot=zbuffer,spot_color=spot_color,nospot=nospot
      if keyword_set(reselect) then begin
         if keyword_set(zbuffer) then  $
            message,/info,'/reselect cannot work with zbuffer' $
         else begin
            avertices = mag.vertices
            bselect,magfield,zscale,mag.alpha,avertices,use_colorPlin, $
                    use_colorMlin,fast=(NOT slow),position=position, $
                    /compute_deriv,bthick=bthick,nofilldot=zbuffer, $
                    nospot=nospot,spot_color=spot_color
            ; delete the old vertices and add the new vertices
            mag = add2str(mag,['vertices'], $
                          reform(avertices,3,n_elements(avertices)/3), $
                          delete=['vertices'])
         endelse
      endif
      if NOT keyword_set(noreset3d) then begin
         !P = save_P
         !X = save_X
         !Y = save_Y
      endif
      if keyword_set(zbuffer) then begin 
         image = tvrd()
         device,/close
         case strupcase(device_name) of 
           'CGM': set_plot,'CGM'
           'NULL': set_plot,'NULL'
           'PCL': set_plot,'PCL'
           'PS': set_plot,'PS'
           'REGIS': set_plot,'REGIS'
           'SUN': set_plot,'SUN'
           'TEK': set_plot,'TEK'
           'X': set_plot,'X'
           'Z': set_plot,'Z'
         endcase
         tv,image
      endif
      return,mag
   endif $
   else begin
      t3d,/reset & scale3d
      ; Plot the contours of the vertical field
      if NOT keyword_set(nocontours) then begin
         if NOT keyword_set(nobz) then begin
            contour,contourdata,xpix,ypix,/T3D,/NOERASE,ZVALUE=0.00,LEVELS=levels,c_colors=[use_colorP],color=default_color,/noclip,xst=1+4*keyword_set(noaxes),yst=1+4*keyword_set(noaxes),position=position,charsize=charsize,c_linestyle=cplinestyle
            contour,contourdata,xpix,ypix,/T3D,/NOERASE,ZVALUE=0.00,LEVELS=-1*rotate(levels,2),c_colors=[use_colorM],color=default_color,/noclip,xst=1+4*keyword_set(noaxes),yst=1+4*keyword_set(noaxes),position=position,xtitle=xtitle,title=title,charsize=charsize,c_linestyle=cmlinestyle
         endif
         if keyword_set(neutral) then $
            contour,get_neutral(mag.bz),xpix,ypix,/T3D,/NOERASE,ZVALUE=0.00,LEVELS=[0],c_colors=[use_colorN],color=default_color,/noclip,xst=1+4*keyword_set(noaxes),yst=1+4*keyword_set(noaxes),position=position,xtitle=xtitle,title=title,charsize=charsize,c_linestyle=cplinestyle,thick=nthick
      endif
      Btotal = reform(sqrt(mag.bxp(*,*,0)^2+mag.byp(*,*,0)^2+mag.bzp(*,*,0)^2))
      if keyword_set(Jcut) then Jtotal = mag.j_total
      Bz = reform(mag.bzp(*,*,0))

      if keyword_set(positive) then begin
        if NOT keyword_set(Jcut) then $
           pixel = where((Btotal GT Bcut) and (Bz GE 0.), count) $
        else $
           pixel = where((Jtotal GT Jcut) and (Bz GE 0.), count) 
      endif $
      else if keyword_set(negative) then begin
        if NOT keyword_set(Jcut) then $
           pixel = where((Btotal GT Bcut) and (Bz LT 0.), count) $
        else $
           pixel = where((Jtotal GT Jcut) and (Bz LT 0.), count) 
      endif $
      else begin
        if NOT keyword_set(Jcut) then $
           pixel = where(Btotal GT Bcut, count) $
        else $
           pixel = where(Jtotal GT Jcut, count)
      endelse

      if skip GT 1 then begin
         pixel = pixel(long(skip)*lindgen(n_elements(pixel)/long(skip)))
         count = n_elements(pixel)
      endif

      if keyword_set(flux) and pixel(0) GE 0 then begin
         ; Restrict whatever is in pixel to make sure that flux is correctly
         ; represented by the density of field lines.  Take out points from
         ; pixel until the density is reasonable.  Won't be perfect.
         meansize = 3
         density = make_array(size=size(Btotal))
         density(pixel) = 1.0    ; Set selected pixels for density computation
         mdensity = runmean(density,meansize,pixel=pixel)
         bdensity = (abs(Btotal)-Bcut) > 1.0
         bdensity = runmean(bdensity,meansize,pixel=pixel)
         bdensity = bdensity/max(bdensity)
         skips = mdensity/bdensity
         print,format="('Flux: ',$)"
         ; The 1.1 is somewhat arbitrary
         WHILE (max(skips(pixel),maxplace) GT 1.1) AND (count GT 0) DO begin
            density(pixel(maxplace)) = 0.
            pixel = kill_index(pixel,pixel(maxplace))
            if pixel(0) LT 0 then message,'Problem with density computation'
            count = n_elements(pixel)
            mdensity = runmean(density,meansize,pixel=pixel)
            skips(pixel) = mdensity(pixel)/bdensity(pixel)
            if count MOD 10 EQ 0 then print,format="('.',$)"
         endwhile
         print
      endif

      for pix=0L,count-1L do begin
         xpix = pixel(pix) MOD nx
         ypix = fix(pixel(pix)/nx)
         if (Btotal(xpix,ypix) GT Bcut) then begin
            backwards = 0
            Bcolor = use_colorPlin
            if Bz(xpix,ypix) LT 0. then begin
               backwards = 1
               Bcolor = use_colorMlin
            endif
            bline,mag,fix(xpix),fix(ypix),zscale,mag.alpha,lvertices, $
                  backwards=backwards,Bcolor=Bcolor,reset=(pix EQ 0),fast=(NOT slow),$
                  position=position,bthick=bthick,nofilldot=zbuffer, $
                  nospot=nospot,spot_color=spot_color,quiet=quiet
            if pix EQ 0 then vertices = lvertices $
            else vertices = [vertices,lvertices]
         endif
      endfor
      if NOT keyword_set(noselect) then $
         bselect,mag,zscale,mag.alpha,vertices,use_colorPlin,use_colorMlin, $
                 fast=(NOT slow),position=position, $
                 compute_deriv=(count LE 0),bthick=bthick,nofilldot=zbuffer, $
                 nospot=nospot,spot_color=spot_color

      if NOT keyword_set(noreset3d) then begin
         !P = save_P
         !X = save_X
         !Y = save_Y
      endif
      if n_elements(vertices) LE 0 then begin
         message,/info,'WARNING: No field lines.'
         return,-1
      endif
      return,add2str(dummy_str,['nx','ny','nz','alpha','zscale','title', $
                                'bcut','skip','vertices','bz','point', $
                                'b_pxd','b_pyd'],nx,ny,nz, $
                                mag.alpha,zscale,title,bcut,skip-1, $
                                reform(vertices,3,n_elements(vertices)/3), $
                                mag.bz,mag.point,mag.b_pxd,mag.b_pyd)
   endelse

end
