

function color_blines, vertices, colors=colors, print=print, hrange=hrange

;+
;NAME:
;     COLOR_BLINES
;PURPOSE:
;     Sets the color of magnetic field lines based on their maximum height.
;CATEGORY:
;CALLING SEQUENCE:
;     v = color_blines(v)
;INPUTS:
;     v = Vertices array or structure
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     colors = [lowest color index, highest color index]
;     /print = print the height vs. color
;     hrange = vector giving the ranges of heights of the colors.  e.g.
;              hrange=[10.] sets the first color to LT 10 and the second
;              color to GT 10.
;OUTPUTS:
;     v = Vertices array or structure
;COMMON BLOCKS:
;SIDE EFFECTS:
;     Reorders the field lines in vertices:  lowest to highest
;RESTRICTIONS:
;PROCEDURE:
;     The color index is contained in the vertices array.  This routine simply
;     sets the value of the color index by byte scaling the height.
;MODIFICATION HISTORY:
;     T. Metcalf  May 25, 1993
;-


   if n_elements(power) NE 1 then power = 1.0

   v_is_structure = 0
   vs = size(vertices)
   if vs(vs(0)+1) EQ 8 then begin
      v_is_structure = 1
      if tag_index(vertices,'VERTICES') LT 0 then $
         message,'Vertices structure does not include tag "vertices"'
      v = vertices.vertices
   endif

   if n_elements(v) LE 9 or n_elements(v) MOD 3 NE 0 then $
      message,'Vertices has the wrong format'

   v = reform(v,3,n_elements(v)/3)

   lstart = where(v(0,*) EQ -1000., nlines)  ; -1000 indicates start of line
   lend = where(v(0,*) EQ -2000., nelines)   ; -2000 indicates end of line

   if (nelines NE nlines) then $
      message,/info,'WARNING: # of line ends NE # of line starts'

   if (nlines LE 0) OR (nelines NE nlines) then return,-1

   ht_line = -1
   for line = 0,nlines-1 do begin
      ht_line = [ht_line,max(v(2,lstart(line)+1:lend(line)-1))]
   endfor
   ht_line = ht_line(1:n_elements(ht_line)-1)
   
   if n_elements(colors) NE 2 then colors = [15b,!d.table_size-2]
   lo_color = byte(colors(0))
   hi_color = byte(colors(1))

   color = bytscl(ht_line,top=hi_color-lo_color) + lo_color 

   if keyword_set(hrange) then begin
      if hrange(0) GT 0. then hrange=[0.0,hrange]
      color = intarr(n_elements(ht_line)) & color(*)=-1
      color_list = indgen(hi_color-lo_color+1)+lo_color
      nclist = n_elements(color_list)
      for i=0,n_elements(hrange)-2 do begin
         c = where(ht_line GE hrange(i) AND ht_line LE hrange(i+1))
         if c(0) GE 0 then color(c)=color_list(i<(nclist-1))
      endfor
      c = where(color LT 0)
      if c(0) GE 0 then color(c)=hi_color
   endif

   if keyword_set(print) then begin
      for c = fix(lo_color),fix(hi_color) do begin
         i=where(color EQ c,cnum)
         if i(0) GE 0 then $
            print,strcompress('Color '+string(c)+': z = ' + $
                  string(min(ht_line(i))) + ' to ' + $
                  string(max(ht_line(i))) + ' ... ' + $
                  string(cnum)+' lines')
      endfor
   endif

   v(2,lstart) = color
   v(2,lend) = color

   hsort = sort(ht_line)
   vout = v
   pointer = 0L
   for line = 0,nlines-1 do begin
      aline = v(*,lstart(hsort(line)):lend(hsort(line)))
      naline = n_elements(aline(0,*))
      vout(*,pointer:pointer+naline-1)=aline
      pointer = pointer + naline
   endfor
   v = temporary(vout)

   if v_is_structure then begin
      vout = vertices
      vout.vertices = v
      v = temporary(vout)
   endif
   
   return,v
   
end
