
pro diffeq_tfl, funct, init, start, finish, times, yvalues, tol=tol, $
               report = report, Params= params, Listname = listname, $
               Depvar = depvar, ylbound=ylbound, yubound=yubound
;+
;NAME:
;       DIFFEQ_TFL
;PURPOSE:
;       Solve a system of first order ordinary differential equations
;                    yi' = fi(t, y1(t),...yn(y)), i = 1,..., n
;                    ai  = yi(start),             i = 1,..., n         
;       using Runge-Kutta method of order 2 and 3. Stepsize is selected
;       automatically and hence is variable. 
;         
;CATEGORY:
;       Mathematical Functions, General
;CALLING SEQUENCE:
;       diffeq_tfl, funct, init,  start, finish, times, yvalues, $
;                  tol = tol, Params = params, report = report, $
;                  Listname = listname, Depvar = depvar
;
;INPUTS:
;     
;      funct = character string containing the name of the user suppled
;              function implementing f = [f1, ...., fn]. This function
;              should be written in IDL and have two arguments, namely
;              the scalar valued  time argument t and the vector argument
;              [y1(t),..., yn(t)].  Additional constant parameters may be
;              supplied through the keyword Params.
;      init  = the vector [a1, ..., an] of initial values.
;      start = the initial value of t.
;      finish = the final value of t. 
;  
;  
;KEYWORD PARAMETERS:
;      tol =    the error tolerance. Default is .001.
;      Params = a keyword to be passed to the function f. A purpose
;               for Params may be to specify constant parameter values in the
;               case that f is a parametric family of functions. See
;               example below. If the IDL function to compute f does
;               not have a keyword Params, then Params should not be
;               set in the call to diffeq_tfl.
;      report = a flag, if set, to signal that at each step the time
;               value, step size, and dependent variable values should
;               be written to the screen or to a file specified by keyword
;               Listname.    
;      Listname = the name of the file to receive any output. Default is
;                 to the screen. 
;      Depvar = a string array of the names of the dependent variables to
;               be used in the output. Depvar(i) = name of variable i.
;      yubound = upper bound to y values (aborts if reached).  Can be a scalar
;                or it can have the same dimensions as a singel yvalue.
;                for example, if the yvalues are x,y,z, the bounds can be
;                scalar or be a 3-element vector.
;      ylbound = lower bound to y values (aborts if reached).
;OUTPUT PARAMETERS:	
;      times = vector of times at which f is computed.
;      yvalues = Array of y values. If ti = times(i),
;
;                 Yvalues(*, i) = f(ti,y1(ti),..., yn(ti))
;                               
;                               =  [f1(ti,y1(ti),..., yn(ti)),  ..., 
;                                              fn(ti,y1(ti), ...,yn(ti))].
;COMMON BLOCKS:
;      None
;SIDE EFFECTS:
;      None
;RESTRICTIONS:
;      None
;EXAMPLE:
;       Solve 
;          y1' = -.1 * y1, 
;          y2' = .1*y1 - .05*y2,
;          y3' = .05*y2
; 
;          y1(0) = 1000, y2(0) = 0, y3(0) = 0
;
;       on the interval [0, 5].
;
;       First, we define the function radio by
;
;           function radio, t, y, params = params
;            k = params(0) & kp = params(1)
;            return, [-k*y(0), k*y(0) - kp*y(1), kp* y(1)]
;            end
;
;      Next, we call
;            diffeq_23, "radio", [1000, 0, 0], 0, 5., times, yvalues, $
;                        params = [.1, .05], /report
;
;      The result may be plotted by
;
;          plot, times, yvalues(0,*)
;          for i = 1,2 do oplot, times, yvalues(i,*)
;
;MODIFICATION HISTORY:
;                CAB, Sept., 1991.
;	920515 ANM Add DIFFEQCOM
;	920520 ANM Allow DEPVAR (bad name choice) to rename "Time" too.
;		   [NO] Add ability to integrate backwards (h < 0) [NO]
;       930514 TRM Modified ANM's program to work with blines.pro
;-
COMMON DIFFEQCOM,flag	; funct returns 0 => OK
flag=0			; funct returns 1 => Nan, retry with h/2

On_Error,2

; Check parameters
IF N_PARAMS(0) lt 4 THEN MESSAGE, "Missing parameters"

IF KEYWORD_SET(listname) THEN OPENW, unit, /GET, listname ELSE unit = -1

IF KEYWORD_SET(tol) eq 0 THEN tol =.001
IF KEYWORD_SET(Params) eq 0 THEN Paramset = 0 ELSE Paramset = 1

IF KEYWORD_SET(report) eq 0 THEN report = 0 $
ELSE BEGIN
   SN = size( depvar)
   n = n_elements(init)
   IF (SN(1) eq 0) THEN BEGIN
      I = lindgen(n)
      Names = ['Var' + StrTrim(I, 2)]
   ENDIF ELSE IF SN(1) lt n THEN BEGIN
      I = lIndgen(N)
      Names = [DepVar, 'Var' + StrTrim(I(SN(1) : N-1),2)]
   ENDIF ELSE Names = depvar
   IF sn(1) NE n+2 THEN names = ['Times','Stepsize',names]
   FOR i = 0, n DO PRINTF, unit, FORMAT='(A13,2x,$)', Names(i)
   PRINTF, unit, FORMAT = '(A13)',Names(n+1)
   PRINTF,unit, " "
ENDELSE

if n_elements(ylbound) and n_elements(yubound) then dobound=1 else dobound=0

; Initialize
h = (finish - start)   ;h = stepsize
minh  = h/20000.        
;maxh  = h/5.
maxh  = h/50.  ; XXXX KDL playing with over-shooting horiz. traces.
times = start
h = h/100.
t = start
v = init
yvalues = v

IF report ne 0 THEN BEGIN
   printf,unit, format ='(G13.6, 2x, G13.6, 2x, $)', t, h
   for i = 0, n-2 do printf,unit, format ='(G13.6, 2x,$)', v(i)
   printf,unit,format = '(G13.6)', v(n-1)
ENDIF


errbound = tol * max([sqrt(total(v^2)), 1])

; compute yvalues for variable step sizes

again:
WHILE t lt finish and h ge minh DO BEGIN

   IF t+h gt finish THEN h = finish - t
     
   IF Paramset eq 0 THEN BEGIN
      k1 = CALL_FUNCTION(funct,t, v)
	IF flag EQ 1 THEN BEGIN h = 0.5*h & GOTO,again & END
      k2 = CALL_FUNCTION(funct, t+h, v + h * k1)
	IF flag EQ 1 THEN BEGIN h = 0.5*h & GOTO,again & END 
      k3 = CALL_FUNCTION(funct, t + h/2, v + h*(k1 + k2)/4)
	IF flag EQ 1 THEN BEGIN h = 0.5*h & GOTO,again & END
   ENDIF ELSE BEGIN
      k1 = CALL_FUNCTION(funct,t, v, Params = Params)
	IF flag EQ 1 THEN BEGIN h = 0.5*h & GOTO,again & END
      k2 = CALL_FUNCTION(funct, t+h, v + h * k1, Params = Params)
	IF flag EQ 1 THEN BEGIN h = 0.5*h & GOTO,again & END
      k3 = CALL_FUNCTION(funct, t + h/2, v + h*(k1 + k2)/4, Params = Params)
	IF flag EQ 1 THEN BEGIN h = 0.5*h & GOTO,again & END
   ENDELSE

   err = (h*(k1 - 2*k3 + k2)/3)^2
   err = sqrt(total(err))
   err_bound = tol * max([sqrt(total(v^2)), 1])

   IF err le err_bound THEN BEGIN
      t = t + h
      v = v + h*(k1 + 4*k3 + k2)/6
      times   = [times,t]
      yvalues = [[yvalues], [v]]
      IF report  ne 0 THEN BEGIN
         PRINTF, unit, format ='(G13.6, 2x, G13.6, 2x, $)', t, h
         FOR i = 0, n-2 DO PRINTF, unit, format ='(G13.6, 2x,$)', v(i)
         PRINTF, unit,format = '(G13.6)', v(n-1)
      ENDIF
      if dobound then begin
        if min(v-ylbound) LT 0 OR max(v-yubound) GT 0 then goto,done
      endif
   ENDIF
   IF err ne 0 THEN h = MIN([maxh, .9*h*(err_bound / err)^(1/3.0)])

ENDWHILE


IF t lt finish THEN PRINT, " Beware of singularity"

done:
IF unit ne -1 THEN Free_lun, unit

RETURN
END

