
;+

function fff2bff,fffout,jin,b0=b0,pangle=pangle,radius=radius,lat=lat,cmd=cmd,pixel_size=pixel, $
                 alpha=alpha,xh=xh,yh=yh,zh=zh,plot=plot,i_cont=i_cont,offlimb=offlimb

;NAME:
;     fff2bff
;PURPOSE:
;     Converts the output from fff.pro to the format of the bff.pro program
;     so that the fff program can be used with blines.pro, etc.
;CATEGORY:
;CALLING SEQUENCE:
;     f=fff2bff(fffout,j,b0=b0,pangle=p,lat=lat,cmd=cmd,pixel_size=pixel_size, $
;                      alpha=alpha,xh=xh,yh=yh,zh=zh,i_cont=i_cont)
;INPUTS:
;     fffout = output from fff.pro, fltarr(nx,ny,nz,3)
;     b0 = b0 in radians
;     pangle = p in radians
;     radius = solar radius in arcsec
;     lat = latitude of center of mgram in radians.  Scalar or fltarr(nx,ny).
;     cmd = cmd of center of mgram in radians.  Scalar or fltarr(nx,ny).
;     pixel_size = pixel size in arc sec.  A single number give x and y sizes 
;                  while a 2-element vector gives [xsize,ysize].
;     alpha = force-free alpha (1/pixels)
;     xh = x coordinates in Mm (output from fff.pro): fltarr(nx,ny)
;     yh = y coordinates in Mm (output from fff.pro): fltarr(nx,ny)
;     zh = height of computed mgrams in Mm: fltarr(nz)
;OPTIONAL INPUT PARAMETERS:
;     j = magnetic field structure.  If present, the pointing parameters
;         will be taken from the structure and b0, p , lat, cmd, pixel_size
;         need not be specified.
;     i_cont = continuum image
;KEYWORD PARAMETERS
;     /plot = plot the field at each height.
;     offlimb = array of 0's or 1's.  A 1 indicates that the pixel is off the
;               limb and should not be transformed.  Usually a pixel with R
;               over Rsun > 0.98 should be avoided.
;OUTPUTS:
;     f = magneitc field structure which can be passed directly to blines.pro.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;    The z-coordinate must be *uniformly* increasing
;PROCEDURE:
;MODIFICATION HISTORY:
;    1996-02-13  T. Metcalf
;    1999-Aug-17 T. Metcalf  Added offlimb keyword
;-

   ssjin = size(jin)
   
   if ssjin(n_elements(ssjin)-2) EQ 8 then begin  ; Is it a structure?
      if NOT keyword_set(b0) then b0 = jin.point.b0
      if NOT keyword_set(pangle) then pangle = jin.point.p
      if tag_index(jin,'LATITUDE') GE 0 then begin
         if (NOT keyword_set(lat)) AND (n_elements(jin.latitude) EQ n_elements(fffout(*,*,0,0))) then $
            lat = jin.latitude*!dtor
      endif
      if tag_index(jin,'CMD') GE 0 then begin
         if (NOT keyword_set(cmd)) AND (n_elements(jin.cmd) EQ n_elements(fffout(*,*,0,0))) then $
         cmd = jin.cmd*!dtor
      endif
      if NOT keyword_set(lat) then lat = jin.point.lat
      if NOT keyword_set(cmd) then cmd = jin.point.cmd
      if NOT keyword_set(pixel) then pixel = jin.point.pix_size
      if NOT keyword_set(radius) then radius = jin.point.radius
   endif

   if n_elements(alpha) NE 1 then message,'You must specify alpha'
   if n_elements(zh) LT 3 then message,'You must specify zh'
   if n_elements(xh) LT 3 then message,'You must specify xh'
   if n_elements(yh) LT 3 then message,'You must specify yh'
   if (n_elements(b0) NE 1) OR (n_elements(pangle) NE 1) OR (n_elements(radius) NE 1) OR  $
      (NOT (n_elements(lat) EQ 1 OR n_elements(lat) EQ n_elements(fffout(*,*,0,0)))) OR $
      (NOT (n_elements(cmd) EQ 1 OR n_elements(cmd) EQ n_elements(fffout(*,*,0,0)))) $
      then message,'You must specify b0, p, lat, and cmd'
   if n_elements(pixel) LT 1 then message,'You must specify pixel_size'
   if n_elements(pixel) EQ 1 then $
      pix_size = double([pixel(0),pixel(0)]) $
   else $
      pix_size=double(pixel)
   mpix_size = min(pix_size)

   nz = n_elements(zh)
   dz = (zh(1)-zh(0))/mpix_size/.7219

   point = add2str(junk,['b0','p','cmd','lat','pix_size','radius'],b0,pangle,mean(cmd),mean(lat),pix_size,radius)
   b_pxd = xh/mpix_size/.7219
   b_pyd = yh/mpix_size/.7219 

   for z=0L,nz-1L do begin

      bxh = fffout(*,*,z,0)
      byh = fffout(*,*,z,1)
      bzh = fffout(*,*,z,2)

      bxyz,bxi,byi,bzi,bxh,byh,bzh,p=pangle,b0=b0,lat=lat,cmd=cmd,/helio2image
      bl = bzi
      bt = sqrt(bxi^2+byi^2)
      az = atan(-bxi,byi)/!dtor

      fmag = {b_long:bl,b_trans:bt,b_azim:az,bx:bxh,by:byh,bz:bzh, $
              b_pxd:b_pxd,b_pyd:b_pyd,point:point}
      hmag = bhelio(fmag,bx,by,bz,i_cont,p,q,pinv,qinv,alpha=alpha,plot=plot,offlimb=offlimb)

      if z EQ 0 then begin
         nx = n_elements(bx(*,0))
         ny = n_elements(by(0,*))
         newmag = {bxp:fltarr(nx,ny,nz), $
                   byp:fltarr(nx,ny,nz), $
                   bzp:fltarr(nx,ny,nz), $
                   zscale:dz, $
                   alpha:alpha, $
                   bx:bx, $
                   by:by, $
                   bz:bz, $
                   b_pxd:xh/mpix_size/.7219, $
                   b_pyd:yh/mpix_size/.7219, $
                   pix_size:mpix_size, $
                   point:point, $
                   bx_obs:fffout(*,*,0,0), $
                   by_obs:fffout(*,*,0,1), $
                   bz_obs:fffout(*,*,0,2)}
         if n_elements(i_cont) EQ nx*ny then $
            newmag=add2str(newmag,['I_CONT'],i_cont)
      endif

      newmag.bxp(*,*,z) = bx
      newmag.byp(*,*,z) = by
      newmag.bzp(*,*,z) = bz

   endfor

   return,newmag

end
