;+
;NAME:
;     FIND_ALPHA
;PURPOSE:
;     Determines the best value of alpha for a force-free extrapolation by
;     checking the resulting error in the photospheric transverse field over 
;     the entire FOV.
;CATEGORY:
;CALLING SEQUENCE:
;     best_alpha = find_alpha(mag)
;INPUTS:
;     mag = magnetic field structure
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     alimit = the max alpha (in abs) to check
;     astep = the step size in alpha for the search
;     finite = passed to bff.pro
;     avals = (Output) The alpha values checked
;     evals = (Output) The errors at each of the alpha values
;     /noplot = do not plot the error vs. alpha
;     bzcut = Compute the error only at those points with bz GE bzcut
;     xrange = Compute the error only within xrange [x1,x2]
;     yrange = Compute the error only within yrange [y1,y2]
;OUTPUTS:
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
;     Computes the force-free field, then computes 
;     total(sqrt((ff.by-ff.by_obs)^2.0d0+(ff.bx-ff.bx_obs)^2.0d0))
;     as the error.  Computes best alpha from a parabolic fit to the errors
;     with line_cent.pro.
;MODIFICATION HISTORY:
;     T. Metcalf  December 1992
;-


function find_alpha,mag,alimit=alimit,astep=astep,finite=finite, $
                    avals=avals,evals=evals,noplot=noplot,bzcut=bzcut, $
                    xrange=xrange,yrange=yrange

   nx = n_elements(mag.b_long(*,0))
   ny = n_elements(mag.b_long(0,*))

   if n_elements(alimit) NE 1 then begin
      if NOT keyword_set(finite) then alimit=1.0 $
      else alimit = 2.5
   endif
   alimit=abs(float(alimit))
   if n_elements(astep) NE 1 then begin
      if NOT keyword_set(finite) then astep=0.02 $
      else astep = 0.1
   endif
   astep=abs(float(astep))

   if n_elements(bzcut) NE 1 then bzcut=0.0

   if n_elements(xrange) NE 2 then xrange=[0,nx-1]
   xrange = abs(fix(xrange))
   if n_elements(yrange) NE 2 then yrange=[0,ny-1]
   yrange = abs(fix(yrange))
   x = lindgen(nx,ny)
   y = x / nx
   x = x MOD nx

   alpha = -alimit
   avals = [0.0]
   evals = [0.0d0]
   REPEAT begin
      message,/info,'Alpha = '+strcompress(string(alpha))
      ff=bff(mag,alpha=alpha,finite=finite,/no_heliographic,nz=1,/quiet)
      good = where(abs(ff.bz) GE bzcut AND x GE xrange(0) and x LE xrange(1) $
                                       AND y GE yrange(0) and y LE yrange(1))
      if good(0) GE 0 then begin
         error = total(sqrt((ff.by(good)-ff.by_obs(good))^2.0d0 + $
                            (ff.bx(good)-ff.bx_obs(good))^2.0d0)) $
                       / n_elements(good)
         avals = [avals,alpha]
         evals = [evals,error]
      endif
      alpha = alpha + astep
   endrep UNTIL (alpha GT alimit)

   if n_elements(avals) EQ 1 then $
      message,'Data cuts were too strict ... no fit'
   avals = avals(1:n_elements(avals)-1)
   evals = evals(1:n_elements(evals)-1)

   emin = min(evals,eplace)
   amin = line_cent(evals,avals,n_p=((0.1/astep)>3))
;   print,'Minimum error at alpha = '+ $
;          strcompress(string(amin))

   if NOT keyword_set(noplot) then $
      plot,avals,evals,/ynozero,xtitle='Alpha',ytitle='Error (G/pix)', $
           title='Error in Computed Force-Free Horizontal Field vs. Alpha'

   return,amin

end