;+
;NAME:
;     GET_ALPHA
;PURPOSE:
;     Determines the best value of alpha for a force-free extrapolation by
;     checking the resulting error in the photospheric transverse field.
;CATEGORY:
;CALLING SEQUENCE:
;     best_alpha = get_alpha(mag)
;INPUTS:
;     mag = magnetic field structure
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     /shear = Use field weighted shear rather than transverse field as
;              the measure of the error in the FF field.
;     alimit = the max alpha (in abs) to check.
;     astep = the step size in alpha for the search.
;     avals = (Output) The alpha values checked.
;     evals = (Output) The errors at each of the alpha values.
;     /resolve = resolve 180 degree ambiguity at every alpha.  This is the
;                default if tag BX, BY, and BZ are NOT present in the mag
;                structure (if BX, BY, and BZ are present, get_alpha assumes
;                the ambiguity has been resolved already).
;     btcut = Compute the error only at those points with b_transverse GE btcut
;     xrange = Compute the error only within xrange [x1,x2].
;     yrange = Compute the error only within yrange [y1,y2].
;     /xyz = Use heliographic field for comparison.
;     /noplot = do not plot the error vs. alpha.
;     /quiet = work quietly.
;     /verbose = work loudly.
;OUTPUTS:
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
;     Computes the force-free field from LOS field, then computes 
;     total(sqrt((ff.byi-byi)^2.0d0+(ff.bxi-ff.bxi)^2))
;     as the error.  Computes best alpha from a parabolic fit to the errors
;     with line_cent.pro.
;MODIFICATION HISTORY:
;     T. Metcalf  December 1992
;     TRM 15-Apr-94 Converted find_alpha (using bff.pro) to get_alpha (using
;                   fff.pro).
;     TRM 04-Apr-03 Added /shear keyword.
;-


function get_alpha,mag,alimit=alimit,astep=astep,resolve=resolve,xyz=xyz, $
                    avals=avals,evals=evals,noplot=noplot,btcut=btcut, $
                    xrange=xrange,yrange=yrange,quiet=quiet,verbose=verbose, $
                    shear=shear,window=win

   common get_alpha_private,plotwindow

   if n_elements(win) EQ 1 then plotwindow = win

   nx = n_elements(mag.b_long(*,0))
   ny = n_elements(mag.b_long(0,*))

   if n_elements(alimit) NE 1 then alimit=1.0d0
   alimit=abs(double(alimit))
   if n_elements(astep) NE 1 then astep=0.02d0
   astep=abs(double(astep))

   if n_elements(btcut) NE 1 then btcut=0.0

   if n_elements(xrange) NE 2 then xrange=[0,nx-1]
   xrange = abs(fix(xrange))
   if n_elements(yrange) NE 2 then yrange=[0,ny-1]
   yrange = abs(fix(yrange))
   x = lindgen(nx,ny)
   y = x / nx
   x = x MOD nx

   if (keyword_set(resolve) AND (NOT keyword_set(xyz))) OR $
         (tag_index(mag,'BX') LT 0 OR $
          tag_index(mag,'BY') LT 0 OR $
          tag_index(mag,'BZ') LT 0) then qresolve=1b $
   else qresolve=0b

   alpha = -alimit
   avals = [0.0d0]
   evals = [0.0d0]

   p = mag.point.p
   b0 = mag.point.b0
   if tag_index(mag,'CMD',/quiet) GE 0 then lc = mag.cmd*!dtor $
   else lc = mag.point.cmd
   if tag_index(mag,'LATITUDE',/quiet) GE 0 then bc = mag.latitude*!dtor $
   else bc = mag.point.lat
   pixsize = mag.point.pix_size

   if keyword_set(xyz) then begin
      good = where(sqrt(mag.bx^2+mag.by^2+mag.bz^2) GE abs(btcut) $
                   AND x GE xrange(0) and x LE xrange(1) $
                   AND y GE yrange(0) and y LE yrange(1))
   endif else begin
      good = where(abs(mag.b_trans) GE abs(btcut) $
                   AND x GE xrange(0) and x LE xrange(1) $
                   AND y GE yrange(0) and y LE yrange(1))
   endelse
   nbtrans = n_elements(good)
   if good(0) LT 0 then begin
      message,/info,'All data points eliminated in field cut'
      return,0.0
   endif

   if keyword_set(xyz) then begin
      azim = atan(-mag.bx,mag.by)/!dtor
      bxi = mag.bx
      byi = mag.by
   endif else begin
      azim = (mag.b_azim MOD 360.)
      bxi = -mag.b_trans*sin(azim*!dtor) 
      byi = mag.b_trans*cos(azim*!dtor)
   endelse

   REPEAT begin

      if keyword_set(xyz) then begin
         pot=fff(mag.bz,/normal,/image,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pixsize, $
                 /quiet,alpha=alpha,bx=mag.bx,by=mag.by)
      endif else begin
         pot=fff(mag.b_long,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pixsize,/quiet, $
                 alpha=alpha,/image,btrans=mag.b_trans,azim=azim)
      endelse
      pxi = pot(*,*,0)  ; These are image coordinates
      pyi = pot(*,*,1)

      ; Fix the azimuthal angle from the force-free field.

      check = check_math(0,1)
         pa_azimuth = atan(-pxi,pyi)/!dtor  ; Az is CCW from North
      check = check_math(1,0)

      azim = mag.b_azim

      if qresolve then begin

         fixa = where(abs(arange(azim,pa_azimuth)-pa_azimuth) GT $
                      90.,count)
         if (count gt 0) then azim(fixa) = azim(fixa) + 180.
         azim = (azim MOD 360.)
         bxi = -mag.b_trans*sin(azim*!dtor) 
         byi = +mag.b_trans*cos(azim*!dtor)

      endif

      if good(0) GE 0 then begin
         if keyword_set(shear) then begin
            ;Check the error from the field weighted shear angle
            error = total(mag.b_trans[good]*abs(arange(azim(good),pa_azimuth(good))-pa_azimuth(good)))/total(mag.b_trans[good])
         endif else begin
            ; Check the error the transverse field
            error = total(sqrt((pxi(good)-bxi(good))^2+(pyi(good)-byi(good))^2)) $
                    / nbtrans
         endelse 
         avals = [avals,alpha]
         evals = [evals,error]
      endif
      if keyword_set(verbose) then $
         message,/info,strcompress('Alpha = '+string(alpha) + $
                                   ', Error = '+string(error) + 'G')
      alpha = alpha + astep
   endrep UNTIL (alpha GT alimit)

   if n_elements(avals) EQ 1 then begin
      message,/info,'Data cuts were too strict ... no fit'
      return,0.
   endif
   avals = avals(1:n_elements(avals)-1)
   evals = evals(1:n_elements(evals)-1)

   emin = min(evals,eplace)
   ; For the shear, expect 45 deg for random orientations.
   if keyword_set(shear) AND emin GT 42. then $
      message,/info,'Warning the minimum shear is rather large.'
   amin = line_cent(evals,avals,n_points=3,p_fit=pfit)
   if NOT keyword_set(quiet) then $
      message,/info,strcompress('Minimum error at alpha = '+ string(amin))

   if NOT keyword_set(noplot) then begin
      win = !d.window
      if n_elements(plotwindow) LE 0 then begin
         window,/free
         plotwindow = !d.window
      endif else begin
         device,window=w
         if plotwindow LT n_elements(w) then begin
            if w[plotwindow] then wset,plotwindow $
            else window,plotwindow
         endif else begin
            window,/free
            plotwindow = !d.window
         endelse
      endelse
      if keyword_set(shear) then begin
         ytitle='Field Weighted Shear Angle (Deg)' 
         title = 'Computed Field-Weighted Shear Angle vs. Alpha'
      endif else begin
         ytitle='Error (G/pix)'
         title = 'Error in Computed Force-Free Transverse Field vs. Alpha'
      endelse
      plot,avals,evals,/ynozero,ytitle=ytitle,title=title, $
           xtitle=strcompress('Alpha: Best alpha is '+string(amin))
      oplot,avals,pfit(0)+pfit(1)*avals+pfit(2)*avals^2,line=1
      if win GE 0 then wset,win;; else window,/free
   endif

   return,amin

end
