
;+
;NAME:
;     v3d_rotate
;PURPOSE:
;     Rotates a 3D vector by an angle phi about the axis specified by the 
;     vector n.
;CATEGORY:
;CALLING SEQUENCE:
;     result = v3d_rotate(v,n,phi,tmatrix=tmatrix)
;INPUTS:
;     v = 3D vector to be rotated
;     n = 3D vector specifying the axis of rotation
;     phi = magnitude of the rotation (RADIANS).  Positive is a CCW rotation 
;           looking down the axis to the origin.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     degrees = if set phi is in degrees instead of radians
;OUTPUTS:
;     result = rotated 3D vector
;     tmatrix = transformation matrix
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     v and n must be cartesian (x,y,z) vectors.
;     tmatrix is defined such that result=v#ixform
;PROCEDURE:
;     From Goldstein "Calssical Mechanics", 2nd edition, page 165
;MODIFICATION HISTORY:
;     T. Metcalf  Aug 12, 1992
;-

function v3d_rotate,v,n,phi,tmatrix=ixform,degrees=degrees

      if n_elements(v) NE 3 then message,'ERROR: input vector is not 3D'
      if n_elements(n) NE 3 then message,'ERROR: input axis vector is not 3D'
       
      if keyword_set(degrees) then convert = !dtor $
      else convert = 1.0

      e0 = cos(convert*phi/2.0d0)
      e = (n/sqrt(total(n^2)))*sin(convert*phi/2.0d0)  ; n is normalized here

      ixform = dblarr(3,3)
      ixform(0,0) = e0^2 + e(0)^2 - e(1)^2 - e(2)^2
      ixform(0,1) = 2.*(e(0)*e(1) + e0*e(2))
      ixform(0,2) = 2.*(e(0)*e(2) - e0*e(1))

      ixform(1,0) = 2.*(e(1)*e(0) - e(2)*e0)
      ixform(1,1) = e0^2 - e(0)^2 + e(1)^2 - e(2)^2
      ixform(1,2) = 2.*(e(1)*e(2) + e(0)*e0)

      ixform(2,0) = 2.*(e(2)*e(0) + e(1)*e0)
      ixform(2,1) = 2.*(e(2)*e(1) - e(0)*e0)
      ixform(2,2) = e0^2 - e(0)^2 - e(1)^2 + e(2)^2

      return, v#ixform

end