
function bfits,filnam,index=fnum,date_obs=date,time_obs=time,header=header, $
               barray=barray,help=help,dec2sun=dec2sun,sun2dec=sun2dec, $
               vax2sun=vax2sun,sun2vax=sun2vax

common ubfitsstr, unique

;+
; NAME:
;        BFITS
; PURPOSE:
;        Reads a FITS magnetic field data file into a structure.
; CATEGORY:
;        Input/Output.
; CALLING SEQUENCE:
;        result = bfits(filename)
; INPUTS:
;        filename = string containing the file name.
; OPTIONAL (KEYWORD) INPUT PARAMETERS:
;        index = nonnegative integer. If this parameter is present, a period
;              and the index number are appended to the filename (e.g., '.34').
;              This option makes handling of data in the MCCD file naming
;              convention easier.
; KEYWORD PARAMETERS:
;        dec2sun = Convert DEC IEEE formatted data file to Sun IEEE format
;        sun2dec = Convert Sun IEEE formatted data file to DEC IEEE format
;        vax2sun = Convert VAX formatted data file to Sun IEEE format
;        sun2vax = Convert Sun IEEE formatted data file to VAX format
; OUTPUTS:
;        result = Structure containing the FITS images.
; OPTIONAL (KEYWORD) OUTPUT PARAMETERS:
;        date_obs = date of observation (string).
;        time_obs = time of observation (string).
;        header = string vector, containing the full FITS header (each element
;                 of the vector contains one FITS keyword parameter).
;	 barray = returns the 3D array from rfits if desired
; COMMON BLOCKS:
;        ubfitsstr: unique = Keeps track of number of calls to BFITS in order
;                            to make a unique structure name each time BFITS
;                            is called.
; SIDE EFFECTS:
;        None.
; RESTRICTIONS:
;        o Only magnetic field FITS files are read. 
;	 o FITS extensions (e.g., groups and tables) are not supported.
; MODIFICATION HISTORY:
;        T. Metcalf: Nov, 1990
;
;        10/24/91 Using strstat to check whether structure name is ok.
;
;        1995-07-11 Fixed a problem with multiple tag names which appeared
;                   with IDL 4.0
;-

if keyword_set(help) then begin
  print,'function bfits,filnam,index=fnum,date_obs=date,time_obs=time,'
  print,'               header=header,barray=barray,dec2sun=dec2sun,help=help'
  return,0
endif

; Set up the counter which keeps track of the number of times BFITS is called

  if (n_elements(unique) LE 0) then unique=0L $  ; Defines "unique" if undefined
  else unique = unique+1L   ; Increments "unique" if already defined

; Call JPW's FITS reader to get the FITS file into a big array

barray = rfits(filnam,index=fnum,date_obs=date,time_obs=time,header=header)
if keyword_set(dec2sun) OR keyword_set(sun2dec) then dec2sun,barray $
else if keyword_set(vax2sun) then vax2sun,barray $
else if keyword_set(sun2vax) then sun2vax,barray

blank = rkey(header,'BLANK','INTEGER')  ; This number represents missing data
bsiz = size(blank)
if bsiz(bsiz(0)+1) EQ 7 then blank = -32768  ; If string, blank was undefined

; Check that the file is a valid magnetic field FITS file

if ((rkey(header,'NAXIS','INTEGER')) NE 3) then begin
  print,'ERROR: bfits: Naxis must be equal to 3'
  return,0
endif
if ((rkey(header,'IMAGE1','CHAR')) EQ 'undefined') then begin
  print,'ERROR: bfits: Must be at least one image in the FITS file.'
  return,0
endif

; Attempt to get pointing information.  If pointing info is not available, 
; get_point will be false.

get_point = 1
b0 = rkey(header,'B0','REAL')*!dtor
s=size(b0) & if s(n_elements(s)-2) NE 4 then get_point=0
p = rkey(header,'P','REAL')*!dtor
s=size(p) & if s(n_elements(s)-2) NE 4 then get_point=0
pos_cmd = rkey(header,'POS-CMD','REAL')*!dtor
s=size(pos_cmd) & if s(n_elements(s)-2) NE 4 then get_point=0
pos_lat = rkey(header,'POS-LAT','REAL')*!dtor
s=size(pos_lat) & if s(n_elements(s)-2) NE 4 then get_point=0
pix_size1= abs(rkey(header,'CDELT1','REAL'))
s=size(pix_size1) & if s(n_elements(s)-2) NE 4 then get_point=0
pix_size2= abs(rkey(header,'CDELT2','REAL'))
s=size(pix_size2) & if s(n_elements(s)-2) NE 4 then get_point=0

radius = rkey(header,'RADIUS','REAL')
s=size(radius) & if s(n_elements(s)-2) NE 4 then okradius=0b else okradius=1b

if get_point then begin
   sattempt = 0L
   REPEAT begin
      command ='point = {jzbfpoint'+strcompress(string(unique),/remove_all)+ $
               ',b0:b0,p:p,cmd:pos_cmd,lat:pos_lat,pix_size:[pix_size1,pix_size2]'
      if okradius then command=command+',radius:radius'
      command = command + '}'
      check = execute(command)
      unique = unique + 1L
      sattempt = sattempt + 1L 
   endrep UNTIL check OR (sattempt GT 100)
   if NOT check then begin
      print,'WARNING: bfits: Could not execute point structure.'
      point = 'Undefined'
   endif
endif $
else begin
   print,'WARNING: bfits: Could not get all the pointing information.'
   point = 'Undefined'
endelse

naxis1 = n_elements( barray(*,0,0) ) ; Get dimensions
naxis2 = n_elements( barray(0,*,0) )
naxis3 = n_elements( barray(0,0,*) )

; Set up the structure with tag names given by the IMAGE keywords.

def = strcompress('fltarr(' + string(naxis1) + ',' + string(naxis2) $
      + ')',/remove_all)

adv = long(0)
REPEAT begin
  check = strstat(strcompress('BFITS'+string(unique),/remove_all),/quiet)
  if check then begin
     unique=unique+1
     ;print,'Updating unique',unique
  endif
  adv = adv + 1
endrep UNTIL ((NOT check) OR (adv GT 2000)) 

tag_list = [""]
tag_numb = [-1L]
if NOT check then begin
   command = '{BFITS' + string(unique) ;"unique" forces unique name
   for i=1,naxis3 do begin
     keyword = strcompress('IMAGE' + string(i),/remove_all)
     tag = rkey(header,keyword,'CHAR')
     if (where(tag_list EQ tag))(0) GE 0 then $
        message,/info,'WARNING: Duplicate tag '+tag $ 
     else begin
        tag_list = [tag_list,tag]
        tag_numb = [tag_numb,i]
        command = command + ',' + tag + ':' + def
     endelse
   endfor
   command = strcompress(command,/remove_all)
   inst = rkey(header,'INSTRUME','CHAR')
   if inst NE 'undefined' then $
     command = command + ',' + 'instrument' + ':' + '"' + inst + '"'
   command = command + ',' + 'point:'+get_def(point)
   command = strcompress(command + '}')
   bstrut = makestr(command,status=check)
   bstrut.point = point
endif $
else check = 0

if (NOT check) then begin
  print,'ERROR: bfits: Could not create structure'
  return,0
endif

tag_list = tag_list(1:*)
tag_numb = tag_numb(1:*)
naxis3 = n_elements(tag_list)

; Fill the structure using scale values from the ISCALE keywords

names = tag_names(bstrut)
for i=0L,naxis3-1L do begin
  j = tag_numb(i)
  keyword = strcompress('ISCALE' + string(j),/remove_all)
  scale = rkey(header,keyword,'REAL')
  if (string(scale) EQ 'undefined') then begin
     scale = 1.0
     print,'WARNING: bfits: ',keyword,' not found.  Value set to 1.0.'
  endif
  bstrut.(i) = scale*barray(*,*,j-1)     ; Fill the structure
  ; Check for missing data
  bad = where( bstrut.(i) EQ long(blank/scale)*scale, count)
  if count GT 0 then begin
     print,count,names(i), $
           format="('WARNING: ',i4,' Missing data point(s) in ',a15)"
     if strupcase(strcompress(names(i),/remove_all)) EQ 'I_CONT' then $
        bstrut.(i)(bad) = 0
     if strupcase(strcompress(names(i),/remove_all)) EQ 'I_CENT' then $
        bstrut.(i)(bad) = 0
  endif
endfor

if tag_index(bstrut,'LATITUDE') LT 0 AND $
   tag_index(bstrut,'CMD') LT 0 AND okradius then $
   latlon=get_latlon(bstrut,/add,/quiet)

print,tag_names(bstrut)

return,bstrut

end

