
;+
;NAME:
;     bicubic
;PURPOSE:
;     Does a 2D bicubic interpolation
;CATEGORY:
;CALLING SEQUENCE:
;     bicubic,y,y1,y2,y12,x1l,x1u,x2l,x2u,x1,x2,ansy,ansy1,ansy2
;INPUTS:
;     y = The function values to be interpolated: 4 element vector with each
;         element the value at one corner of the grid box in which the point
;         to be interpolated falls (CW from the lower left).
;     y1 = The derivative of the function wrt x1 (4 element vector)
;     y2 = The derivative of the function wrt x2 (4 element vector)
;     y12 = the cross derivatives (4 element vector)
;     x1l = the lower coordinate of the grid square in the 1 direction
;     x1u = the upper coordinate of the grid square in the 1 direction
;     x2l = the lower coordinate of the grid square in the 2 direction
;     x2u = the upper coordinate of the grid square in the 2 direction
;     x1,x2 = the coordinates of the desired interpolation point
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;OUTPUTS:
;     ansy = the interpolated value of the function at x1,x2
;     ansy1 = the interpolated value of the derivative wrt x1 at x1,x2
;     ansy2 = the interpolated value of the derivative wrt x2 at x1,x2
;COMMON BLOCKS:
;     bicubic_wt_mat = makes the wt matrix and x vector static variables
;SIDE EFFECTS:
;RESTRICTIONS:
;EXAMPLE:
;
;  f is the values CW from the lower left
;  f1 is df/dx CW from the lower left
;  f2 is df/dy CW from the lower left
;  f12 is d2f/dxdy CW from the lower left
;
;       4 ---- 3        -1 ---- -1       3 ---- 1        0 ---- 0
;       |      |         |       |       |      |        |      |
;  f    |      |    f1   |       |   f2  |      |   f12  |      |
;       1 ---- 2        +1 ---- +1       3 ---- 1        0 ---- 0
; 
;   the call is bicubic,f,f1,f2,f12,xmin,xmax,ymin,ymax,x,y,answer
;
;   bicubic,[1,4,3,2],[1,-1,-1,1],[3,3,1,1],[0,0,0,0],0,1,0,1,0.,0.,ans
;     --> ans is 1.0
;   bicubic,[1,4,3,2],[1,-1,-1,1],[3,3,1,1],[0,0,0,0],0,1,0,1,1.,0.,ans
;     --> ans is 2.0
;PROCEDURE:
;     Translated from the C version in Numerical Recipes
;MODIFICATION HISTORY:
;
;       Feb 4, 1992  T. Metcalf
;-

pro bicubic,y,y1,y2,y12,x1l,x1u,x2l,x2u,x1,x2,ansy,ansy1,ansy2

common bicubic_wt_mat,wt,x

;
; Compute coeff
;

   d1=float(x1u-x1l)
   d2=float(x2u-x2l)

   if (x1u EQ x1l OR x2u EQ x2l) then begin
      print,'ERROR: Bad input to bicubic2D: boundaries cannot coincide'
      stop
   endif

   if n_elements(wt) LE 0 then begin
      wt= [ 1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0]
      wt= [wt,[-3,0,0,3,0,0,0,0,-2,0,0,-1,0,0,0,0,2,0,0,-2,0,0,0,0,1,0,0]]
      wt= [wt,[1,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]]
      wt= [wt,[0,0,0,0,1,0,0,0,0,0,0,0,-3,0,0,3,0,0,0,0,-2,0,0,-1,0,0,0,0]]
      wt= [wt,[2,0,0,-2,0,0,0,0,1,0,0,1,-3,3,0,0,-2,-1,0,0,0,0,0,0,0,0,0,0]]
      wt= [wt,[0,0,0,0,0,0,0,0,-3,3,0,0,-2,-1,0,0,9,-9,9,-9,6,3,-3,-6,6,-6]]
      wt= [wt,[-3,3,4,2,1,2,-6,6,-6,6,-4,-2,2,4,-3,3,3,-3,-2,-1,-1,-2,2,-2]]
      wt= [wt,[0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,-2,0,0,1,1,0,0]]
      wt= [wt,[-6,6,-6,6,-3,-3,3,3,-4,4,2,-2,-2,-2,-1,-1,4,-4,4,-4,2,2,-2,-2]]
      wt= [wt,[2,-2,-2,2,1,1,1,1]]
      wt=transpose(reform(wt,16,16))
      x=fltarr(16,/nozero)
   endif
   
   x(0:3) = y
   x(4:7) = y1*d1
   x(8:11) = y2*d2
   x(12:15) = y12*d1*d2

   c = reform(wt#x,4,4)

;
; Do the interpolation
;


   t=(x1-x1l)/d1
   u=(x2-x2l)/d2

   ansy=0.0
   ansy2=0.0
   ansy1=0.0

   for i=3,0,-1 DO Begin
      ansy=t*ansy+((c(i,3)*u+c(i,2))*u+c(i,1))*u+c(i,0)
      ansy2=t*ansy2+(3.0*c(i,3)*u+2.0*c(i,2))*u+c(i,1)
      ansy1=u*ansy1+(3.0*c(3,i)*t+2.0*c(2,i))*t+c(1,i)
   endfor
   ansy1 = ansy1 / d1
   ansy2 = ansy2 / d2

end
