
;+
;NAME:
;     BICUBIC2D
;PURPOSE:
;     Does a 2D bicubic interpolation of the magnetic field from a
;     vector magnetogram.
;CATEGORY:
;CALLING SEQUENCE:
;     binterp = bicubic2d(mag,x,y,alpha)
;INPUTS:
;     mag = magnetic field structure containing at least BX, BY, BZ,
;           b_pxd, b_pyd.  You won't have BX,BY,BZ unless the 180 degree
;           ambiguity has been resolved!
;     x,y = the coordinates of the field point in pixels where 0,0 is
;           the LL corner.
;     alpha = force free alpha (used to compute the derivatives)
;             Default is 0.0 which is not the same as setting the
;             /noalpha keyword.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     /noalpha = Do not use the value of alpha in the computation
;     /reset = force the computation of the derivatives.  Otherwise they are
;              computed only when the routine is called the first time.
;OUTPUTS:
;     binterp = a singe vector containing the interpolated field
;               [Bx,By,Bz] at the point x,y.
;COMMON BLOCKS:
;     Xbicubic2d_derivs = Makes the derivatives into static variables
;     Ybicubic2d_derivs = Makes the derivatives into static variables
;     Zbicubic2d_derivs = Makes the derivatives into static variables
;     times_called2d = contains static variable ntimes
;SIDE EFFECTS:
;RESTRICTIONS:
;     b_pxd, b_pyd are in pixels and alpha is in 1/pixels.
;     Be sure to set /reset when starting a new gram!!!!
;PROCEDURE:
; Performs 2D bicubic interpolation on a magnetic field.
; The derivatives are computed numerically and the force-free
; conditions are used where possible for consistency.
;MODIFICATION HISTORY:
;
;       Feb 4, 1992  T. Metcalf
;       2003-apr-18 TRM added FFF checks to make this routine
;                       consistent with bicubic3d.pro.  Fixed bug in
;                       the order of the parameters sent to bicubic.
;                       Set dx,dy to 2 rather than shifts of b_pxd and
;                       b_pyd.  Since the interpolation is done in the
;                       image plane, using b_pxd and b_pyd does not
;                       make sense.
;-

function bicubic2d,mag,x,y,alpha,reset=reset,noalpha=noalpha,quiet=quiet

   ; The common blocks make the derivatives into static variables

   common Xbicubic2d_derivs,dBxdx,dBxdy,dBxdz,d2Bxdzdy,d2Bxdxdy
   common Ybicubic2d_derivs,dBydx,dBydy,dBydz,d2Bydzdy,d2Bydxdy
   common Zbicubic2d_derivs,dBzdx,dBzdy,dBzdz,d2Bzdzdy,d2Bzdxdy
   common times_called2d,ntimes

   nx = n_elements(mag.bx(*,0))
   ny = n_elements(mag.bx(0,*))

   i = (fix(x) > 0) < (nx-2)
   j = (fix(y) > 0) < (ny-2)

   if n_elements(ntimes) LE 0 then ntimes = 0L $
   else ntimes = 1L

   ; compute derivatives on the first pass or if reset keyword is present

   if keyword_set(reset) or (NOT ntimes) then begin

      ;dx = shift(mag.b_pxd,-1,0) - shift(mag.b_pxd,1,0)
      ;dy = shift(mag.b_pyd,0,-1) - shift(mag.b_pyd,0,1)

      dx = 2.0
      dy = 2.0

      dBxdx = (shift(mag.bx,-1,0) - shift(mag.bx,1,0))/dx
      dBxdy = (shift(mag.bx,0,-1) - shift(mag.bx,0,1))/dy

      dBydx = (shift(mag.by,-1,0) - shift(mag.by,1,0))/dx
      dBydy = (shift(mag.by,0,-1) - shift(mag.by,0,1))/dy

      dBzdx = (shift(mag.bz,-1,0) - shift(mag.bz,1,0))/dx
      dBzdy = (shift(mag.bz,0,-1) - shift(mag.bz,0,1))/dy


      ; Avoid edge problems by using adjacent values

      dBydx[0,*] = dBydx[1,*]
      dBydx[nx-1,*] = dBydx[nx-2,*]
      dBxdy[*,0] = dBxdy[*,1]
      dBxdy[*,ny-1] = dBxdy[*,ny-2]
      dBxdx[0,*] = dBxdx[1,*]
      dBxdx[nx-1,*] = dBxdx[nx-2,*]
      dBydy[*,0] = dBydy[*,1]
      dBydy[*,ny-1] = dBydy[*,ny-2]
      dBzdx[0,*] = dBzdx[1,*]
      dBzdx[nx-1,*] = dBzdx[nx-2,*]
      dBzdy[*,0] = dBzdy[*,1]
      dBzdy[*,ny-1] = dBzdy[*,ny-2]
      
      if NOT keyword_set(noalpha) then begin

         ; Can not apply FF conditions to derivatives in the image plane.
         ; Compute the transformation between
         ; the image and heliographic coordinates

         ; xs = heliographic x (from b_pxd)
         ; ys = heliographic y (from b_pyd)
         ; xi = image x (unifirm pixel grid)
         ; yi = image y (unifirm pixel grid)


         ; These could be derived exactly from the pointing, but this should
         ; be good enough for the current purpose.  They are used self
         ; consistently and the b_pxd and b_pyd values vary slowly.

         DxsDxi = (shift(mag.b_pxd,-1,0) - shift(mag.b_pxd,1,0))/dx
         DysDxi = (shift(mag.b_pyd,-1,0) - shift(mag.b_pyd,1,0))/dx
         DxsDyi = (shift(mag.b_pxd,0,-1) - shift(mag.b_pxd,0,1))/dy
         DysDyi = (shift(mag.b_pyd,0,-1) - shift(mag.b_pyd,0,1))/dy

         ; Avoid edge problems by using adjacent values

         DxsDxi[0,*] = DxsDxi[1,*]
         DxsDxi[nx-1,*] = DxsDxi[nx-2,*]
         DysDxi[0,*] = DysDxi[1,*]
         DysDxi[nx-1,*] = DysDxi[nx-2,*]
         DxsDyi[*,0] = DxsDyi[*,1]
         DxsDyi[*,ny-1] = DxsDyi[*,ny-2]
         DysDyi[*,0] = DysDyi[*,1]
         DysDyi[*,ny-1] = DysDyi[*,ny-2]

         ; Invert the matrix

         determ = DxsDxi*DysDyi - DysDxi*DxsDyi
         DxiDxs = DysDyi/determ
         DxiDys = -DxsDyi/determ
         DyiDxs = -DysDxi/determ
         DyiDys = DxsDxi/determ

         ; compute heliographic derivatives

         dBxdxs = dBxdx*DxiDxs + dBxdy*DyiDxs
         dBxdys = dBxdx*DxiDys + dBxdy*DyiDys
         dBydxs = dBydx*DxiDxs + DBydy*DyiDxs
         dBydys = dBydx*DxiDys + DBydy*DyiDys

         ; Make sure FF conditions hold by averaging

         dBydxs_temp = (dBydxs + dBxdys + alpha*mag.bz)/2.0  ;  Average numerical & FF deriv
         dBxdys_temp = (dBxdys + dBydxs - alpha*mag.bz)/2.0
         dBydxs = temporary(dBydxs_temp)
         dBxdys = temporary(dBxdys_temp)

         ; Transform back to the image plane derivatives

         dBxdx = dBxdxs*DxsDxi + dBxdys*DysDxi
         dBxdy = dBxdxs*DxsDyi + dBxdys*DysDyi
         dBydx = dBydxs*DxsDxi + dBydys*DysDxi
         dBydy = dBydxs*DxsDyi + dBydys*DysDyi

      endif 

      ; Compute second derivatives

      d2Bxdxdy = (shift(dBxdx,0,-1) - shift(dBxdx,0,1))/dy
      d2Bydxdy = (shift(dBydx,0,-1) - shift(dBydx,0,1))/dy
      d2Bzdxdy = (shift(dBzdx,0,-1) - shift(dBzdx,0,1))/dy

      ; Avoid edge problems by using adjacent values

      d2Bxdxdy[*,0] = d2Bxdxdy[*,1]
      d2Bxdxdy[*,ny-1] = d2Bxdxdy[*,ny-2]
      d2Bydxdy[*,0] = d2Bydxdy[*,1] 
      d2Bydxdy[*,ny-1] = d2Bydxdy[*,ny-2] 
      d2Bzdxdy[*,0] = d2Bzdxdy[*,1]
      d2Bzdxdy[*,ny-1] = d2Bzdxdy[*,ny-2]

      if NOT keyword_set(quiet) then $
         print,'INFO: bicubic2d: Derivatives computed'

   endif
   
   i1 = i+1
   j1 = j+1

;
; Bx
;

   f = [mag.bx(i,j),mag.bx(i,j1),mag.bx(i1,j1),mag.bx(i1,j)]
   f1 = [dBxdx(i,j),dBxdx(i,j1),dBxdx(i1,j1),dBxdx(i1,j)]
   f2 = [dBxdy(i,j),dBxdy(i,j1),dBxdy(i1,j1),dBxdy(i1,j)]
   f12 = [d2Bxdxdy(i,j),d2Bxdxdy(i,j1),d2Bxdxdy(i1,j1),d2Bxdxdy(i1,j)]
   bicubic,f,f1,f2,f12,i,i+1,j,j+1,x,y,bxout,ansf1,ansf2

;
; By
;

   f = [mag.by(i,j),mag.by(i,j1),mag.by(i1,j1),mag.by(i1,j)]
   f1 = [dBydx(i,j),dBydx(i,j1),dBydx(i1,j1),dBydx(i1,j)]
   f2 = [dBydy(i,j),dBydy(i,j1),dBydy(i1,j1),dBydy(i1,j)]
   f12 = [d2Bydxdy(i,j),d2Bydxdy(i,j1),d2Bydxdy(i1,j1),d2Bydxdy(i1,j)]
   bicubic,f,f1,f2,f12,i,i+1,j,j+1,x,y,byout,ansf1,ansf2

;
;Bz
;

   f = [mag.bz(i,j),mag.bz(i,j1),mag.bz(i1,j1),mag.bz(i1,j)]
   f1 = [dBzdx(i,j),dBzdx(i,j1),dBzdx(i1,j1),dBzdx(i1,j)]
   f2 = [dBzdy(i,j),dBzdy(i,j1),dBzdy(i1,j1),dBzdy(i1,j)]
   f12 = [d2Bzdxdy(i,j),d2Bzdxdy(i,j1),d2Bzdxdy(i1,j1),d2Bzdxdy(i1,j)]
   bicubic,f,f1,f2,f12,i,i+1,j,j+1,x,y,bzout,ansf1,ansf2


;   print,mag.bx(i,j),mag.by(i,j),mag.bz(i,j)
;   print,mag.bx(i+1,j+1),mag.by(i+1,j+1),mag.bz(i+1,j+1)

   return,[bxout,byout,bzout]

end
