;+
;NAME:
;     bilinxy
;PURPOSE:
;     2-D bilinear interpolation on non-uniform grid
;CATEGORY:
;CALLING SEQUENCE:
;     b = bilin(array,x,y,xarr=xarr,yarr=yarr)
;INPUTS:
;     array = array to interpolate
;     x,y = the coordinates of the point(s) to be interpolated
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     xarr = array of x coordinates (same size as array), def = uniform grid
;     yarr = array of y coordinates (same size as array), def = uniform grid
;     degree = degree of polynomial used to convert between coordinate
;              systems.  The higher the degree, the more accurate the 
;              transformation, but the slower the program. Def = 3.
;OUTPUTS:
;     b = a single vector containing the interpolated values
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Uses polywarp to get a transformation between coordinate
;     systems.  May not work properly for very non-uniform grids.
;
;     The interpolation is done on a square grid regardless of how
;     non-uniform the x and y coordinates are.
;PROCEDURE:
;     Uses polywarp to get a transformation between coordinate systems.
;MODIFICATION HISTORY:
;
;       April 20, 1994  T. Metcalf
;-

function bilinxy,data,x,y,xarray=xarr,yarray=yarr,degree=degree

; Bilinear interpolation in 2-D on non-uniform grid

   nx = n_elements(data(*,0))
   ny = n_elements(data(0,*))
   nxvec = n_elements(x)
   nyvec = n_elements(y)
   if nxvec NE nyvec then message,'x and y must have the same size'

   xgrid = lindgen(nx,ny) MOD nx
   ygrid = lindgen(nx,ny) / nx

   if NOT keyword_set(xarr) then xarr=xgrid  ; default to uniform grid
   if NOT keyword_set(yarr) then yarr=ygrid

   if n_elements(degree) NE 1 then degree = 3 else degree = abs(fix(degree))
   polywarp,xgrid,ygrid,xarr,yarr,degree,p,q   ; transformation: x,y -> grid
   xi=make_array(size=size([x]),/double)  ; zeroed initially
   yi=make_array(size=size([y]),/double)
   for m=0,degree do begin      ; Convert x,y from xarr,yarr coordinates
      for n=0,degree do begin   ; to xgrid,ygrid coordinates.
         xm = x^m & yn = y^n
         xi = xi + p(n,m)*xm*yn
         yi = yi + q(n,m)*xm*yn
      endfor
   endfor

   i = (long(xi)>0)<(nx-2)  ; Coordinate of nearest pixel below x,y
   j = (long(yi)>0)<(ny-2)

   t = xi - i
   u = yi - j

   ; return bilinear interpolation

   return, (1.-t)*(1.-u)*data(i,j) $
         + t*(1.-u)*data(i+1,j) $
         + t*u*data(i+1,j+1) $
         + (1.-t)*u*data(i,j+1)
        
end
