;+
;NAME:
;     binup_b
;PURPOSE:
;     Increases the number of pixels in a magnetic structure by an integer
;     factor
;CATEGORY:
;CALLING SEQUENCE:
;     magout = binup_b(magin,bin,alpha)
;INPUTS:
;     magin = magnetic field structure
;     bin = integer factor for binning
;     alpha = force-free alpha used in interpolation
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     /noalpha = Do not use alpha in the interpolation
;OUTPUTS:
;     magout = new magnetic field structure 
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The magnetic field input must already have the 180 degree ambiguity 
;     resolved.
;
;     The vertical and horizontal fields must be present: BX, BY, BZ
;PROCEDURE:
;MODIFICATION HISTORY:
;      Written by T. Metcalf  29-Oct-93
;-

function binup_b,magin,binup,alpha,noalpha=noalpha

   if n_elements(alpha) NE 1 then alpha = 0.0

   ; Decrease grid spacing by interpolation. Also protects the input structure
   ; from change

   if n_elements(binup) GT 0 then begin  
      if binup GT 1 then begin
         nx = n_elements(magin.bz(*,0))
         ny = n_elements(magin.bz(0,*))
         nnx = fix(nx * binup) & binupx = float(nnx-1)/float(nx-1)
         nny = fix(ny * binup) & binupy = float(nny-1)/float(ny-1)
         print,nnx,nny,format="('INFO: binup_b: New dimensions: ',2i4)"
         bx2 = fltarr(nnx,nny)
         by2 = fltarr(nnx,nny)
         bz2 = fltarr(nnx,nny)
         for i=0,nnx-1 do begin
            for j=0,nny-1 do begin
               x = i/binupx
               y = j/binupy
               B = bicubic2d(magin,x,y,alpha,reset=(i EQ 0 AND j EQ 0), $
                             noalpha=noalpha)
               bx2(i,j) = B(0)
               by2(i,j) = B(1)
               bz2(i,j) = B(2)
            endfor
         endfor
         binupx = float(nnx)/float(nx)
         binupy = float(nny)/float(ny)
         p = [0,0,1.0/binupx,0]  ; Stretch X by a factor of binup
         q = [0,1.0/binupy,0,0]  ; Stretch Y by a factor of binup
         mag = add2str(mag,['bx','by','bz','b_pxd','b_pyd','point'], $
                       bx2,by2,bz2, $
                       poly_2d(magin.b_pxd,p,q,1,nnx,nny), $
                       poly_2d(magin.b_pyd,p,q,1,nnx,nny), $
                       magin.point,/quiet)
      endif $
      else mag = magin
   endif $
   else mag = magin

   return,mag


end
