
;+

function bmerge, ls_str, int_str, cutoff=cutoff, quiet=quiet, flux=flux, $
                 condition=condition

;NAME:
;     BMERGE
;PURPOSE:
; Merges two magnetic field structures.  LS_STR is the field computed
; with the Lites and Skumanich code, and INT_STR is the structure computed
; with the integral method.  The default cutoff is to use the integral 
; value when the LS values is less than 100 G total.  The cutoff keyword
; can be used to override this default.  INT_STR is merged into LS_STR.
;CATEGORY:
;CALLING SEQUENCE:
;     new_structure = bmerge(LS_STR,INT_STR,cutoff=cutoff,quiet=quiet)
;INPUTS:
;     LS_STR = Magnetic field structure from the Lites&Skumanich code
;     INT_STR = Magnetic field structure from the integral method
;OPTIONAL INPUT PARAMETERS:
;     none
;KEYWORD PARAMETERS
;     cutoff = The cutoff in total field strength below which the integral 
;              reduction field is used.
;     condition = Merge condition.  E.g. 'b_total LT 100.'  All points
;                 in the array which statisfy the condition will be set
;                 to the INTEGRAL structure.
;                 To form the condition, use the variables ls_str or int_str
;                 for the LS field structure and the INTEGRAL field structure
;                 respectively.  Also arrays x and y are available to choose
;                 particular regions.  The b_total array is the total field
;                 in the LS structure.  There should always be something like
;                 'b_total LT 100.' so that regions where the LS code did
;                 not compute the field will be set to the INTEGRAL arrays.
;                 For example, 'b_total LT 500 OR y LT 15'.
;     /flux = Multiply the integral transverse field by the sqrt of 
;               the filling factor from the unno code.  This assumes the 
;               filling factor in unfit regions means something.
;               (The transverse field from the L&S code should already be 
;               the flux).
;     quiet = If set, don't print anything.
;OUTPUTS:
;     new_structure = merged magnetic field structure.  INT_STR is used below
;                     the cutoff and LS_STR is used above the cutoff
;COMMON BLOCKS:
;     none
;SIDE EFFECTS:
;     none
;RESTRICTIONS:
;     none
;PROCEDURE:
;MODIFICATION HISTORY:
;     TRM  Oct, 1991
;     TRM  May 3, 1994  o Multiply integral transverse field by the sqrt of the
;                         LS filling factor to give the flux.
;                       o Add the velocity from INT to the output structure.
;                       o Add UNNO to structure to keep track of what pixels
;                         are from LS and what pixels are from INT.
;     TRM 1996-02-22    Don't merge the I_CONT array.  Use the integral version
;                       if available.
;-


   if (n_tags(ls_str) EQ 0 or n_tags(int_str) EQ 0) then begin
     print,'ERROR: bmerge: ls_str and int_str must be structures.'
     return, ls_str
   endif

; Check that the bare minimum of structure tags are present

   if tag_index(ls_str,'B_LONG') LT 0  OR $
      tag_index(ls_str,'B_TRANS') LT 0  OR $
      tag_index(ls_str,'B_AZIM') LT 0  then begin
      print,'ERROR: bmerge: LS structure is incomplete'
      return,ls_str
   endif
   if tag_index(int_str,'B_LONG') LT 0  OR $
      tag_index(int_str,'B_TRANS') LT 0  OR $
      tag_index(int_str,'B_AZIM') LT 0  then begin
      print,'ERROR: bmerge: INT structure is incomplete'
      return,ls_str
   endif

   mag = ls_str

   nx = n_elements(ls_str.b_long(*,0))
   ny = n_elements(ls_str.b_long(0,*))
   ntotal = nx*ny

   nxi = n_elements(int_str.b_long(*,0))
   nyi = n_elements(int_str.b_long(0,*))

   if (nxi NE nx) OR (nyi NE ny) then begin
      print,"ERROR: bmerge: Can't merge B structures: incompatible dimensions"
      return,ls_str
   endif

; Find all the tags in LS structure which are 2D arrays and have the same
; number of elements as b_long

   ntags = n_tags(ls_str)
   ls_names = tag_names(ls_str)   ; all the tag names
   tags = [-1]
   for i=0,ntags-1 do begin
      s = size(ls_str.(i))
      if s(0) EQ 2 then begin   ; If 2D array
         if s(n_elements(s)-1) EQ ntotal AND $
         strupcase(ls_names(i)) NE 'I_CONT' then $
         tags = [tags,i] ; If correct size and not continuum image
      endif
   endfor
   if n_elements(tags) LE 1 then begin
      print,'ERROR: bmerge: no array data in the LS structure!'
      return,ls_str
   endif $
   else begin
      tags = tags(1:(n_elements(tags)-1))   ; List of tag numbers to be output
   endelse
   ntags = n_elements(tags)

   b_total = sqrt(ls_str.b_long^2 + ls_str.b_trans^2)  ; Total field

   if n_elements(cutoff) EQ 0 then ucutoff = 100.0 $
   else ucutoff = cutoff(0)

   if NOT keyword_set(condition) then begin  ; Points to be changed
     if NOT keyword_set(quiet) then print,'Cutoff = ',string(ucutoff),' Gauss'
     condition = strtrim(strcompress('b_total LT '+string(ucutoff)),2)
   endif

   x = lindgen(nx,ny) MOD nx
   y = lindgen(nx,ny) / nx
   if NOT keyword_set(quiet) then print,'Condition = '+condition
   check=execute('not_computed = where ('+condition+', count)')
   if NOT check then message,'Condition failed'

   if count GT 0 then begin
     for i=0,ntags-1 do begin   ; loop through all the tags to be merged
        ;Get the tag index in the INT structure which has the same name as the 
        ;tag in the LS structure.
        int_tag = tag_index(int_str,ls_names(tags(i)))
        if int_tag GE 0 then begin
           ; Merge the structures
           mag.(tags(i))(not_computed) = int_str.(int_tag)(not_computed)
        endif $
        else begin
           ; The current tag is not in the INT structure
           if NOT keyword_set(quiet) then $
              print,'WARNING: bmerge: ' + ls_names(tags(i)) + $
                    ' not merged: used LS value since not in INT structure'
        endelse
     endfor
     if keyword_set(flux) then begin
        if tag_index(ls_str,'B_FILL') GE 0 then begin
           mag.b_trans(not_computed) = $
              mag.b_trans(not_computed)*sqrt(ls_str.b_fill) 
           message,/info,'Converted INT transverse field to flux using LS filling factor'
        endif $
        else print,'WARNING: bmerge: filling factor not in LS strucure.  ' + $
                   'Weak transverse field is not the flux'
     endif
   endif $
   else begin
      if NOT keyword_set(quiet) then begin
         print
         print,'No pixels with field less than ',ucutoff,' Gauss'
         print
      endif
   endelse

   ; Now find add useful integral arrays ...

   if tag_index(int_str,'VELOCITY') GE 0 AND $
      tag_index(ls_str,'VELOCITY') LT 0 then $
      mag=add2str(mag,'VELOCITY',int_str.velocity,/quiet)

   ; Use the integral i_cont if available

   if tag_index(int_str,'I_CONT') GE 0 AND $
      tag_index(ls_str,'I_CONT') GE 0 then $
      mag.i_cont=int_str.i_cont

   ls = bytarr(nx,ny)+1b
   ls(not_computed) = 0b
   mag = add2str(mag,'UNNO',ls,/quiet)

   mag = add2str(mag,'COMMENT',/quiet, $
                 strcompress('bmerge: '+strmid(!stime,0,17)+ $
                 ', ('+condition+')'))

   return,mag
   
end
