;+
;NAME:
;     brebin
;PURPOSE:
;     Resamples a magnetic structure to increase or decrease the size.
;CATEGORY:
;CALLING SEQUENCE:
;     magout = brebin(magin,nx,ny,alpha)
;INPUTS:
;     magin = magnetic field structure
;     nx,ny = new dimensions of arrays in the magin structure
;OPTIONAL INPUT PARAMETERS:
;     alpha = force-free alpha used in interpolation (def = 0.0)
;KEYWORD PARAMETERS
;     /sum = average rather than resample when reducing size of
;            arrays.  Highly reccommended, but slow.
;     /fast = just use poly_2d (fast) and don't worry about the
;             special field interpolations done with bicubic2d.
;             Probably less accurate.  If /fast is set, /sum is
;             ignored. 
;     /noalpha = Do not use alpha in the interpolation.  Note that if
;                alpha is not set, alpha=0.0 is assumed.  Using
;                /noalpha turns the whole thing off: it is not the
;                same as not specifying alpha.
;     /verbose = print progress.
;OUTPUTS:
;     magout = new magnetic field structure, resampled.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The magnetic field input must already have the 180 degree ambiguity 
;     resolved.  The vertical and horizontal fields must be present: BX, BY, BZ
;PROCEDURE:
;MODIFICATION HISTORY:
;      Written by T. Metcalf  01-Apr-2003
;      TRM 2004-07-20 Added sum keyword.  Fixed small offset in
;                     result. Added /fast keyword.
;-

function brebin,magin,nxin,nyin,alpha,noalpha=noalpha,sum=usesum,fast=fast,verbose=verbose

   if n_elements(alpha) NE 1 then alpha = 0.0

   nx = n_elements(magin.bz(*,0))
   ny = n_elements(magin.bz(0,*))
   nnx = nxin & binupx = float(nnx-1)/float(nx-1)
   nny = nyin & binupy = float(nny-1)/float(ny-1)
   bx2 = fltarr(nnx,nny)
   by2 = fltarr(nnx,nny)
   bz2 = fltarr(nnx,nny)
   dx = float(nx)/float(nnx)
   dy = float(ny)/float(nny)
   xshift = (1./dx-1.)/2.
   yshift = (1./dy-1.)/2.
   if NOT keyword_set(fast) then begin
      if keyword_set(verbose) then progress,0,label='brebin',/reset
      for i=0,nnx-1 do begin
        for j=0,nny-1 do begin
           x = i*dx   ; i is new coord, x is old
           y = j*dy
           if keyword_set(usesum) AND i NE 0 AND j NE 0 then begin
              xbin = fix(dx+0.999) > 1
              ybin = fix(dy+0.999) > 1
              B = [0.,0.,0.]
              ; Sum up B over new resolution element
              for k = 0,xbin-1 do begin
                 for l=0,ybin-1 do begin
                    x1 = x + dx*k/float(xbin)
                    y1 = y + dy*l/float(ybin)
                    B = B + bicubic2d(magin,x1,y1,alpha, $
                                      reset=(i EQ 0 AND j EQ 0 AND k EQ 0 and l EQ 0), $
                                      noalpha=noalpha,/quiet)   
                 endfor
              endfor
              B = B / float(xbin*ybin)
           endif else begin
              xbin=1
              ybin=1
              B = bicubic2d(magin,x,y,alpha,reset=(i EQ 0 AND j EQ 0), $
                            noalpha=noalpha,/quiet)
           endelse
           bx2(i,j) = B(0)
           by2(i,j) = B(1)
           bz2(i,j) = B(2)
        endfor
        if keyword_set(verbose) then progress,100.0*float(i+1.0)/nnx,frequency=5.
     endfor
     if keyword_set(verbose) then progress,100,/last
     p = [-xshift+(1./xbin-1.)/2.,0,1,0] ; shift x
     q = [-yshift+(1./ybin-1.)/2.,1,0,0] ; shift y
     bx2 = poly_2d(bx2,p,q,2,nnx,nny,cubic=-0.5)
     by2 = poly_2d(by2,p,q,2,nnx,nny,cubic=-0.5)
     bz2 = poly_2d(bz2,p,q,2,nnx,nny,cubic=-0.5)
   endif
   binupx = float(nnx)/float(nx)
   binupy = float(nny)/float(ny)
   p = [-xshift/2.,0,1.0/binupx,0]  ; Stretch X by a factor of binup
   q = [-yshift/2.,1.0/binupy,0,0]  ; Stretch Y by a factor of binup
   if keyword_set(fast) then begin
      bx2 = poly_2d(magin.bx,p,q,2,nnx,nny,cubic=-0.5)
      by2 = poly_2d(magin.by,p,q,2,nnx,nny,cubic=-0.5)
      bz2 = poly_2d(magin.bz,p,q,2,nnx,nny,cubic=-0.5)
   endif
   point = magin.point
   point.pix_size = point.pix_size / [binupx,binupy]
   mag = add2str(mag,['bx','by','bz','b_pxd','b_pyd','point'], $
                 bx2,by2,bz2, $
                 poly_2d(magin.b_pxd,p,q,1,nnx,nny), $
                 poly_2d(magin.b_pyd,p,q,1,nnx,nny), $
                 point,/quiet)
   if tag_index(magin,'I_CONT',/quiet) GE 0 then $
      mag = add2str(mag,['I_CONT'],poly_2d(magin.i_cont,p,q,2,nnx,nny,cubic=-0.5))
   if tag_index(magin,'CMD',/quiet) GE 0 then begin
      ;lc = double(magin.cmd)*!dtor 
      mag = add2str(mag,['CMD'],poly_2d(magin.cmd,p,q,1,nnx,nny))
      lc = double(mag.cmd)*!dtor 
   endif else lc = double(magin.point.cmd)
   if tag_index(magin,'LATITUDE',/quiet) GE 0 then begin
      ;bc = double(magin.latitude)*!dtor
      mag = add2str(mag,['LATITUDE'],poly_2d(magin.latitude,p,q,1,nnx,nny))
      bc = double(mag.latitude)*!dtor
   endif else bc = double(magin.point.lat)
   p = double(magin.point.p)
   b0 = double(magin.point.b0)

   bxyz,bxi,byi,bzi,bx2,by2,bz2,p=p,b0=b0,lat=bc,cmd=lc,/helio2image
   btrans = sqrt(bxi^2+byi^2)
   bazim = atan(-bxi,byi)*!radeg
   mag = add2str(mag,['B_LONG','B_TRANS','B_AZIM'],bzi,btrans,bazim)

   return,jxyz(mag,/quiet)


end
