;+
;NAME:
;	Bvector
;PURPOSE:
;	Plot a vector magnetogram
;CATEGORY:
;	Graphics
;CALLING SEQUENCE:
;	Bvector,magnetic_field
;INPUTS:
;	magnetic_field =  Structure of arrays of field parameters
;                         (Each tag is a (NX,NY) array)
;               3 tags Required:  *.B_LONG
;                                 *.B_TRANS
;                                 *.B_AZIM
;OPTIONAL KEYWORD INPUT PARAMETERS:
;       position = Defines plot window (like !p.position).
;       edge = structure with tage left,right,top, and bottom.  The tags
;              specify the border in cm.  Default is 0.0 on all sides.
;	length:  scales the length of the vectors (1.0 is the default)
;	missing: is the smallest transverse field which is plotted
;	maxT:    is the largest transverse field which is plotted
;	xval,yval:  the x and y coordinates of the raster points
;	            (default is integer pixel numbers)
;       levels:  specifies the contour levels to plot.  
;	         Note that only positve numbers should be used since 
;	         the routine calculates symmetric negative contours 
;       Jlevels: specifies the current contour levels
;	    automatically.
;       LineP_style:  specifies the line style for positive contours
;       LineM_style:  specifies the line style for negative contours
;       scale:   sets the scale of the plot. 
;	         1.0 (default) gives auto scaling
;                1.2 (for example) plots 20% larger than auto scaling
;       xstep:   x pixel size relative to y pixel size.  For non-square
;                pixels. 
;       charsize: Set the character size.  1.0 is "regular"
;KEYWORD PARAMETERS
;	/no_c_lab:  no labels are plotted on the contours
;	/incolor:   draws the contours in color if you are on a color display
;		      o Blue contours = positive field
;		      o Red contours = negative field
;		   Note: this changes color table entries 10,11,12!
;       /noct0: do not load colro table 0 before executing
;       /splot:  Leaves the plot parameters set (otherwise resets them)
;	/help:  Prints out procedure declaration and exits routine
;       reduce: Rebin the field to reduce the resolution and invoke the
;               magnifier to shoe the true field.
;       /magnify: Force invocation of the magnifier regardless of whether the
;                data were rebinned with the reduce keyword.
;       /no_magnify: Do not invoke the magnifier even if reduce is set
;       /novectors: Don't plot the transverse vectors
;       /noblong: Don't plot longitudinal contours
;       /current: Plot vertical current
;       /tcurrent: Plot total (force-free) current
;       /unproject: Plot the horizontal/vertical field rather than the
;                  line-of-sight/transverse field, if they are present in 
;                  the structure.
;       /solar: Do not add the solar P angle to the azimuth with the
;               /unproject keyword.  Normally the p-angle is added since the
;               Stokes data has terrestrial norht up, but the Bx, By, and Bz
;               are computed relative to solar north.  If the data actually
;               has solar north up, /solar will turn off the addition of the
;               P-angle.
;       /arrows: If set plot heads on the arrows.
;       /print_colors: Use alternate color table which looks ok on the printer
;       /negtive: make the image a negative image
;       /no_scale_image: Do not rescale the image intensities.  Image must be
;                       a byte array in this case.
;       /neutral_line: If set, plot a SMOOTHED neutral line on the plot
;       /nojinterp: If set, don't interpolate when putting J on the field grid.
;                  WARNING: If this keyword is set, the current contours may
;                  be incorrectly shifted by half a pixel or so.
;       bcutoff: The cutoff value of transverse field for defining low-field 
;                regions used to compute the error on Jz.  (default = 300 G)
;       lowsigma: starting sigma for Jz contours
;       thick: sets thickness of contours
;       /nowords: Print no titles, no information, nothing, on the plot
;OUTPUTS:
;	None
;COMMON BLOCKS:
;	None
;SIDE EFFECTS:
;	The incolor keyword will alter the color table.  In particular, if the
;	color table is not fixed with loadct, post script output will 
;	produce blank pages!
;RESTRICTIONS:
;	* Post-script printing is UNIX specific.
;       * The image underneath the field will not work with the printer
;       * The azimuth is assumed to be measured CCW from the +y axis
;       * The transverse field should be positive with the azimuth specifying
;         the direction of the transverse field.
;       * The neutral line is smoothed (if plotted) and may not be exact
;       * For the /unproject keyword, the X, Y, and Z fields are assumed to 
;         be solar west, solar north, and up repectively.
;PROCEDURE:
;	Plots line of sight magnetic field as contours and transverse field as 
;	vectors.  Trys very hard to keep the aspect ratio of the plot correct.
;MODIFICATION HISTORY:
;    T. Metcalf 3/1990
;
;	3/10/90  * Assume azimuth is measured CCW from the -y axis.
;	3/13/90  * Use trmvect instead of velovect.  This plots lines without
;	           arrowheads and puts the center of the line segment at the 
;	           grid point.  Makes nicer transverse fields.
;	         * Add ability to turn off contour labels
;	3/15/90  * Made the plots come out with the correct aspect ratio.
;	           They were close but now they're right
;	         * Added scale,xstep,ystep parameters for more general plotting
;	3/27/90  * Added maxT parameter to cut off Bt if it is too large
;	10/12/90 * Added colors into contours:
;	              o Blue contours = positive field
;	              o Red contours = negative field
;       11/19/90 * Changed input to field structure rather than field array.
;                  Azimuth is now measured CCW from the +y axis.
;                  Plot vectors *over* contours to make color plot look nicer.
;       08/29/91 * Added the reduce feature to rebin the data.  This will
;                  be particularly useful for IVM data.
;       10/17/91 * Added the capability of displaying images and currents
;                  with the magnetic field.
;       12/23/91 * Added the capability to put heads on the arrows.
;       9/16/93  * Fixed a bug in the azimuth computation when using /unproject
;                  the p-angle must be accounted for since the Bx, By, and Bz
;                  are computed in Solar coordinates.  The /solar keyword
;                  will override this and assumes that solar north is up 
;                  in the plot rather than terrestrial north.
;       10/5/93  * Took out the code which makes the pixels square.  Replaced
;                  it with a call to pixsquare.  This required removing the 
;                  ystep keyword.
;       10/6/93  * Fixed the plotting of the underlying image so that Bvector
;                  now works with device 'PS' as well as 'X'.
;-

pro Bvector, mag, output=out, title=title, length=length, $
             missing=missing, maxT=maxT, xval=x, yval=yin, $
             levels=levP, no_c_lab=noclab, help=help, $
             lineP_style=lPs, lineM_style=lMs,scale=scale,xstep=inxstep,$
             incolor=incolor,splot=splot,reduce=reduce, $
             image=simage, novectors=novectors, noblong=noblong, $
             magnify=magnify, current=current, unproject=unproject, $
             arrows=arrows, tcurrent=tcurrent, print_colors=print_colors, $
             negative=negative,charsize=charsize,neutral_line=neutral_line, $
             nojinterp=nojinterp,bcutoff=bcutoff,lowsigma=lowsigma,thick=thick,$
             Jlevels=levPjz,nowords=nowords,no_scale_image=no_scale_image, $
             no_magnify=no_magnify,solar=solar,noct0=noct0,position=position, $
             edge=cmedge

  if keyword_set(help) then begin
    print, 'pro Bvector, mag, output=out, title=title, length=length, $'
    print, 'missing=missing, maxT=maxT, xval=x, yval=yin, $'
    print, 'levels=levP, no_c_lab=noclab, help=help, $'
    print, 'lineP_style=lPs, lineM_style=lMs,scale=scale,xstep=xstep, $'
    print, 'incolor=incolor,slopt=splot,reduce=reduce $'
    print, 'image=simage, novectors=novectors, noblong=noblong, $'
    print, 'magnify=magnify, current=current,tcurrent=tcurrent,charsize=charsize'
    return
  endif

  if n_elements(position) NE 4 then position = [.1,.17,.9,.9] $
  else position = float(position)
  if n_elements(thick) NE 1 then thick = 1.
  if keyword_set(simage) then dimage = simage  ; Protect the input image

; Check that the data is a structure

  if (n_tags(mag) EQ 0) then begin
    print,'ERROR: bvector: bvector requires a magnetic field structure as input'
    return
  endif

; tcurrent has priority

   if keyword_set(tcurrent) then current = 0

   if n_elements(charsize) LE 0 then charsize=1.
   pcharsave = !P.charsize
   !P.charsize = charsize

;  Check that the required field data is present in the structure

  tnames = tag_names(mag)
  Blong = 0 & Btrans = 0 & azimuth = 0
  for i=0,n_tags(mag)-1 do begin
    if keyword_set(current) then begin
       if (tnames(i) EQ 'JZ') then Blong = 1
    endif $
    else if keyword_set(tcurrent) then begin
       if (tnames(i) EQ 'J_TOTAL') then Blong = 1
    endif $
    else if keyword_set(unproject) then begin
       if (tnames(i) EQ 'BZ') then Blong = 1
    endif $
    else begin
       if (tnames(i) EQ 'B_LONG') then Blong = 1
    endelse
    if keyword_set(unproject) then begin
      if (tnames(i) EQ 'BX') then Btrans = 1
      if (tnames(i) EQ 'BY') then azimuth = 1
    endif $
    else begin
      if (tnames(i) EQ 'B_TRANS') then Btrans = 1
      if (tnames(i) EQ 'B_AZIM') then azimuth = 1
    endelse
  endfor

  if (NOT Blong OR NOT Btrans or NOT azimuth) then begin
    if keyword_set(current) then $
       print,'ERROR: bvector: Structure tags JZ,B_TRANS,B_AZIM required' $
    else if keyword_set(tcurrent) then $
       print,'ERROR: bvector: Structure tags J_TOTAL,B_TRANS,B_AZIM required' $
    else if keyword_set(unproject) then $
       print,'ERROR: bvector: Structure tags BX,BY,BZ required' $
    else $
       print,'ERROR: bvector: Structure tags B_LONG,B_TRANS,B_AZIM required'
    return
  endif

  if keyword_set(unproject) AND NOT keyword_set(solar) then begin
     if tag_index(mag,'POINT') LT 0 then $
        message,'POINT structure required with /unproject'
     if tag_index(mag.point,'P') LT 0 then $
        message,'P-angle must be in the POINT structure with /unproject'
  endif

  if (n_elements(inxstep) EQ 0) then xstep=1.00  $
  else xstep = inxstep
  if xstep LT 0.0 then xstep=abs(xstep)
  if (n_elements(scale) EQ 0) then scale = 1.00  
  if scale LT 0.0 then scale=abs(scale)

; Set up field to be plotted

;  if keyword_set(current) then $ ; Shift the current by half a pixel
;     jz = poly_2d(mag.jz,[-0.5,0,1,0],[-0.5,1,0,0],1) 
  if keyword_set(current) then begin ; Put the current on the field grid
     degree = (sqrt(n_elements(mag.b_pxd))-1.0) < 1
     polywarp,mag.b_pxd,mag.b_pyd,mag.j_pxd,mag.j_pyd,long(degree),kx,ky
     if keyword_set(nojinterp) then jz = poly_2d(mag.jz,kx,ky,0) $
     else jz = poly_2d(mag.jz,kx,ky,1)
  endif
  if keyword_set(tcurrent) then begin ; Put the current on the field grid
     degree = (sqrt(n_elements(mag.b_pxd))-1.0) < 1
     polywarp,mag.b_pxd,mag.b_pyd,mag.j_pxd,mag.j_pyd,long(degree),kx,ky
     if keyword_set(nojinterp) then jz = poly_2d(mag.j_total,kx,ky,0) $
     else jz = poly_2d(mag.j_total,kx,ky,1)
  endif

  if keyword_set(tcurrent) then current = 1  ; from here on, the current is jtot

  if NOT keyword_set(unproject) then begin
     nx = n_elements(mag.b_long(*,0))
     ny = n_elements(mag.b_long(0,*))
  endif $
  else begin
     nx = n_elements(mag.bx(*,0))
     ny = n_elements(mag.bx(0,*))
  endelse

  if (n_elements(x) EQ 0) then x=findgen(nx) else begin  ; Get x or Check input
    if (nx NE n_elements(x)) then begin
      print,'ERROR: x values sent to Bvector are not of the correct dimension'
      stop
    endif
  endelse
  if (n_elements(yin) EQ 0) then y=findgen(ny) else begin
    y=yin
    if (ny NE n_elements(y)) then begin
      print,'ERROR: y values sent to Bvector are not of the correct dimension'
      stop
    endif
  endelse

  info_string=""
  if NOT keyword_set(noblong) OR NOT keyword_set(novectors) then begin
     if keyword_set(unproject) then begin
        Blong = mag.bz
        Btrans = sqrt(mag.bx^2+mag.by^2)
        junk = check_math(1,1)                ; Turn off math errors
        ; Must add the P angle to the azimuth since Bx, By, and Bz are
        ; relative to Solar North not terrestrial north
        azimuth = atan(-mag.bx,mag.by)
        if NOT keyword_set(solar) then azimuth = azimuth + mag.point.p
        junk = check_math(0,0)                ; Turn on math errors
        if keyword_set(noblong) then $
           info_string = "Horizontal Magnetic Field" $
        else if keyword_set(novectors) then $
           info_string = "Vertical Magnetic Field" $
        else $
           info_string = "Vertical-Horizontal Magnetic Fields"
     endif $
     else begin
        Blong = mag.b_long
        Btrans = mag.b_trans
        azimuth = mag.b_azim/(!radeg)       ; azimuth assumed to be in degrees
        if (NOT keyword_set(noblong)) AND (NOT keyword_set(novectors)) then $
           info_string = "Line-of-Sight\Transverse Magnetic Fields" $
        else if keyword_set(noblong) then $
           info_string = "Transverse Magnetic Field" $
        else if keyword_set(novectors) then $
           info_string = "Line-of-Sight Field"
     endelse
  endif $
  else Blong = mag.bz

  if keyword_set(current) then begin
     if keyword_set(tcurrent) then begin
        info_string = info_string + $
                      "!CJ!Dtotal!N (Force-Free)"
     ; = J!Dz!N/!M!!!3cos(!4h!3)!M!!!3   (!M!!!3cos(!4h!3)!M!!!3!Mb!30.25)"
     endif $
     else begin
        info_string = info_string + "!CVertical Electric Current"
     endelse
  endif

  nx = n_elements(Blong(*,0))
  ny = n_elements(Blong(0,*))

  if keyword_set(reduce) then begin

    ; Reduce the scale in the plot so the vectors can be seen

    reduce = fix(reduce)
    if (reduce EQ 1) then begin            ; Compute reduce if not specified
       reducex = fix(nx/25) > 1            ; Reduce array size by this factor
       reducey = fix(ny/25) > 1
    endif $
    else begin
       reducex = abs(reduce)
       reducey = abs(reduce)
    endelse

    pxd = findgen(nx)/float(reducex)    ; The x,y positions of pixels
    pyd = findgen(ny)/float(reducey)
    set_scale = fix(1.5*(reducex + reducey)/2.0) > 5  ; Magnify 1.5 pixels

    print,reducex,format="('ReduceX = ',i2)"
    print,reducey,format="('ReduceY = ',i2)"

    while (nx MOD reducex) NE 0 do nx=nx-1    ; Get even multiples by dropping
    while (ny MOD reducey) NE 0 do ny=ny-1    ; some data
;    if keyword_set(current) then $
;      jz = poly_2d(mag.jz(0:nx-1,0:ny-1),[-0.5,0,1,0],[-0.5,1,0,0],1) 
    if keyword_set(tcurrent) then $
       jz = poly_2d(mag.j_total(0:nx-1,0:ny-1),kx,ky,1) $
    else if keyword_set(current) then $ ; Put the current on the field grid
       jz = poly_2d(mag.jz(0:nx-1,0:ny-1),kx,ky,1)
    if keyword_set(simage) then dimage = dimage(0:nx-1,0:ny-1)

    ;Blong = mag.b_long(0:nx-1,0:ny-1)
    ;Btrans = mag.b_trans(0:nx-1,0:ny-1)
    ;azimuth = mag.b_azim(0:nx-1,0:ny-1)/(!radeg)
    Blong = Blong(0:nx-1,0:ny-1)
    Btrans = Btrans(0:nx-1,0:ny-1)
    azimuth = azimuth(0:nx-1,0:ny-1)

    nx = nx/reducex
    ny = ny/reducey
    xstep = xstep*float(reducex)/float(reducey)

    Bx = rebin(-Btrans * sin(azimuth),nx,ny)    ; Rebin the data
    By = rebin(Btrans * cos(azimuth),nx,ny)
    Blong = rebin(Blong,nx,ny)
    Btrans = sqrt(Bx^2 + By^2)
    if keyword_set(current) then jz = rebin(jz,nx,ny)
    if keyword_set(simage) then dimage = rebin(dimage,nx,ny)
    junk = check_math(1,1)                ; Turn off math errors
    azimuth = atan(-Bx,By)
    junk = check_math(0,0)                ; Turn on math errors

    x=findgen(nx)
    y=findgen(ny)

  end

  nx = n_elements(Blong(*,0))
  ny = n_elements(Blong(0,*))

  if (n_elements(x) EQ 0) then x=findgen(nx) else begin  ; Get x or Check input
    if (nx NE n_elements(x)) then begin
      print,'ERROR: x values sent to Bvector are not of the correct dimension'
      stop
    endif
  endelse
  if (n_elements(yin) EQ 0) then y=findgen(ny) else begin
    y=yin
    if (ny NE n_elements(y)) then begin
      print,'ERROR: y values sent to Bvector are not of the correct dimension'
      stop
    endif
  endelse

  if (keyword_set(magnify) AND (NOT keyword_set(reduce))) then begin
     pxd = x
     pyd = y
  endif

    ; u and v are the x and y components of the transverse field
  u = -Btrans*sin(azimuth) ; This arrangement assumes that zero azimuth is
  v = Btrans*cos(azimuth)  ; along the y axis, and is CCW!

  if (n_elements(out) EQ 0) then out = 'display'    ; Set up defaults
  if (n_elements(title) EQ 0) then begin
    if tag_index(mag,'INSTRUMENT') GE 0 then begin
        if strupcase(strcompress(mag.instrument,/remove_all)) EQ 'STOKESPOLARIMETER' then $
           title = 'University of Hawaii Stokes Polarimeter' $
        else title = 'University of Hawaii ' + mag.instrument 
    endif $
    else title = 'University of Hawaii Stokes Polarimeter'
  endif

  if keyword_set(nowords) then begin
     title=' '
     info_string=' '
  endif

  if (n_elements(lPs) EQ 0) then lPs=0  ; Default contour line styles
  if (n_elements(lMs) EQ 0) then begin
    if (out EQ 'print' OR out EQ 'printing') then lMs = 1 $
    else if (keyword_set(incolor)) then lMs = 0 $
    else lMs = 2
  endif

  if (n_elements(levPjz) EQ 0) then begin
     if keyword_set(tcurrent) then levPjz = [5.,8.,10.,15.,20.,30.] $
     else levPjz = [5.,8.,10.,15.,20.,30.]  ; Positive contours
     if keyword_set(current) OR keyword_set(tcurrent) then begin
        if tag_index(mag,'BX') GE 0 AND $
           tag_index(mag,'BY') GE 0 AND $
           tag_index(mag,'JZ') GE 0 then begin
           if n_elements(bcutoff) LE 0 then bcutoff = 300.
           sigmajz = jzerr(mag,bcutoff=bcutoff)
           if n_elements(lowsigma) LE 0 then lowsigma = 3.
           levPjz = sigmajz*(findgen(10)+lowsigma)
           print,lowsigma,lowsigma+1,lowsigma+2,$
                format="('INFO: J contours at ',f3.1,' sigma, ',f3.1,' sigma, ',f3.1,' sigma ...')"
           good = where( levPjz LE max(abs(mag.jz)), gcount)
           if gcount LT 2 then  begin
              good = [0,1] & gcount = 2
           endif
           if gcount GT 0 then levPjz = levPjz(good) 
        endif $
        else print,"WARNING: Can't compute Jz errors: contours at fixed intervals"
     endif
  endif

  if (n_elements(levP) EQ 0) then begin
    if keyword_set(unproject) then $
      levP = [100.0,200.0,400.0,800.0,1600.0,3200.0] $ ; Positive contours
    else $
      levP = [50.0,100.0,200.0,400.0,800.0,1600.0,3200.0] ; Positive contours
;   levP = [25.0,50.0,100.0,200.0,400.0,800.0,1600.0,3200.0] ; Positive contours
;   levP = [100.0,250.0,500.0,1000.0,1500.0,2000.0,4000.0]   ; Positive contours
  endif
  lev_labP = strcompress(string(levP,format='(f10.1)'),remove_all=1)
  lev_labPjz = strcompress(string(levPjz,format='(f10.1)'),remove_all=1)

  nP = n_elements(levP)
  flipper = intarr(nP)   ; Vector of element numbers, reversed
  for loop=0,nP-1 do flipper(loop) = nP-1-loop

  levM = levP*(-1.0) & levM = levM(flipper)     ; Negative Contours
  lev_labM = lev_labP & lev_labM = lev_labM(flipper)

  nPjz = n_elements(levPjz)
  flipper = intarr(nPjz)   ; Vector of element numbers, reversed
  for loop=0,nPjz-1 do flipper(loop) = nPjz-1-loop

  levMjz = levPjz*(-1.0) & levMjz = levMjz(flipper)     ; Negative Contours
  lev_labMjz = lev_labPjz & lev_labMjz = lev_labMjz(flipper)

  sclab = intarr(n_elements(levP))
  if (NOT keyword_set(noclab)) then sclab(*)=1 $
  else sclab(*)=0
    
  sclabjz = intarr(n_elements(levPjz))
  if (NOT keyword_set(noclab)) then sclabjz(*)=1 $
  else sclabjz(*)=0

  xlabel = 'Contours (G): !M+!3 '
  Clevs = strcompress(string(levP,format='(f10.1)'))  ; Get string of contours
  for i=0,n_elements(levP)-1 do xlabel = xlabel + Clevs(i)

  if keyword_set(current) and (NOT keyword_set(noblong)) then $
     xlabeljz = '!C!CContours (mAmp/m**2): !M+!3 '  $
  else $
     xlabeljz = 'Contours (mAmp/m**2): !M+!3 ' 
  Clevsjz = strcompress(string(levPjz,format='(f10.1)')) ;Get string of contours
  for i=0,n_elements(levPjz)-1 do xlabeljz = xlabeljz + Clevsjz(i)

  if tag_index(mag,'POINT') GE 0 then begin
     if tag_index(mag.point,'PIX_SIZE') GE 0 then begin
        pix_size = mag.point.pix_size
        if n_elements(pix_size) EQ 1 then pix_size=[pix_size(0),pix_size(0)]
        if NOT keyword_set(current) then begin
           xlabel = xlabel + '!C!CPixel Size = ' $
                           + strcompress(string(pix_size(0), $
                                                format="(f12.1)"),/remove_all) $
                           + '"'
        endif $
        else begin
           xlabeljz = xlabeljz + '!C!CPixel Size = ' $
                           + strcompress(string(pix_size(0), $
                                                format="(f12.1)"),/remove_all) $
                           + '"'
        endelse
     endif
  endif

  if keyword_set(nowords) then begin 
    xlabel = ' ' 
    xlabeljz = ' '
  endif

  if (out EQ 'print' OR out EQ 'printing') then begin
    if (out EQ 'print') then begin
      save_device = !D.NAME
      set_plot,'ps'
      device,/landscape
    endif
    default_color = 0
  endif $
  else begin
     if !d.name EQ 'PS' then default_color = 0 $
     else default_color = !d.table_size-1
  endelse

  x0=min(x)   ; Get plotting range
  x1=max(x)
  y0=min(y)
  y1=max(y)

  ; Set up the plot with
  ; an appropriate aspect ratio (need square pixels!).

  save_window = !P.POSITION   ; Save old plotting parameters
  save_windowX = !X
  save_windowY = !Y

  ; Since there is a one pixel border around the plot, nx and ny are
  ;              ; effectively increased by 2

  edge = 2

  !P.POSITION = pixsquare(nx+edge,ny+edge,position=position, $
                          xstep=xstep,xyscale=scale,edge=cmedge)

  ; Do the plot:
  ; trmvect uses a non-default x and y range to give a one pixel border
  ; around the plot and hence the xrange and yrange
  ; values in the calls to contour have seemingly bizarre ranges.
  ; trmvect uses color index 12 while Bvector uses 10,11

  ; Set up colors if requested

  if NOT keyword_set(noct0) then loadct,0

  use_colorP = 1  ; Default in case not on a color monitor
  use_colorM = 1
  use_colorV = [1,0]
  use_colorPjz = 1
  use_colorMjz = 1

  if (out EQ 'print' OR out EQ 'printing') then begin
    use_colorP = default_color  ; Default in case printing
    use_colorM = default_color
    use_colorV = [default_color,default_color]
    use_colorPjz = default_color  ; Default in case printing
    use_colorMjz = default_color
  endif

  if (!d.table_size GT 2) then begin
    use_colorP = 10
    use_colorM = 11
    use_colorV = [12,13]
    use_colorPjz = 8
    use_colorMjz = 9
    if (keyword_set(incolor)) then begin
      if keyword_set(print_colors) then begin
         ; Use only values of 255 here so that the printed line is solid
         specialct,use_colorP,255,0,0           ; Red contours (positive field)
         specialct,use_colorM,0,0,255           ; Blue contours (negative field)
         ;specialct,use_colorM,255,255,0        ; Yellow contours (negative field)
         specialct,use_colorV(0),50,50,50          ; Vectors
         specialct,use_colorV(1),50,50,50    ; Vectors
         specialct,use_colorPjz,255,255,0       ; Yellow contours (positive Jz)
         ;specialct,use_colorMjz,0,255,0       ; Green contours (negative Jz)
         specialct,use_colorMjz,255,0,255       ; Purple contours (negative Jz)
         if !d.name NE 'PS' then begin
            specialct,!d.table_size-1,0,0,0  ; Set the foreground to black
            specialct,0,255,255,255        ; Set the background to white
         endif
      endif $
      else begin
         specialct,use_colorP,255,0,0           ; Red contours (positive field)
         specialct,use_colorM,0,125,255         ; Blue contours (negative field)
         specialct,use_colorV(0),50,50,50          ; Vectors
         specialct,use_colorV(1),50,50,50    ; Vectors
         specialct,use_colorPjz,255,255,0       ; Yellow contours (positive Jz)
         specialct,use_colorMjz,0,255,0         ; Green contours (negative Jz)
         ;specialct,use_colorMjz,255,0,255       ; Purple contours (negative Jz)
         if !d.name NE 'PS' then begin
            specialct,!d.table_size-1,0,0,0  ; Set the foreground to black
            specialct,0,255,255,255        ; Set the background to white
         endif
      endelse
    endif $
    else begin
      specialct,use_colorP,255,255,255
      specialct,use_colorM,20,20,20
      specialct,use_colorV(0),50,50,50
      specialct,use_colorV(1),50,50,50
      specialct,use_colorPjz,255,255,255
      specialct,use_colorMjz,255,255,255
    endelse
  endif

  ; Use thick dots for dotted lines
  if (lMs EQ 1) then cMthick=[2,2] else cMthick=[1,1]
  if (lPs EQ 1) then cPthick=[2,2] else cPthick=[1,1]

  if n_elements(thick) EQ 1 then cMthick=[thick,thick]
  if n_elements(thick) EQ 1 then cPthick=[thick,thick]

  erase

  stepx = abs(x(1)-x(0))
  stepy = abs(y(1)-y(0))

  ; Plot an image under the contours
  if n_elements(dimage) GT 0 and $
     (NOT (out EQ 'print' OR out EQ 'printing')) then begin  
     ; Set up the data coordinates using a "fake" call to contour which
     ; matches the "real" call below.
     contour, Blong, x, y, levels=levP, c_line=[lPs,lPs], c_thick=cPthick, $
              title=title, /xst, /yst,c_labels=sclab, $
              xrange=[x0-stepx,x1+stepx],yrange=[y0-stepy,y1+stepy], $
              c_colors=[use_colorP],color=default_color, $
              /noerase,/nodata,/noclip
     ; Get parameters for resizing the image
     px = !x.window*(float(nx)/(nx+edge-1)) * !d.x_vsize
     py = !y.window*(float(ny)/(ny+edge-1)) * !d.y_vsize
     sx = px(1)-px(0)+1
     sy = py(1)-py(0)+1
     sz = SIZE(dimage)
     if !d.name EQ 'PS' then begin
        factor=float(sx/nx) 
        sx = sx/factor
        sy = sy/factor
     endif else factor=1.
     ; Resize the image.  The bytscl call ensures that the image will not use
     ; the first 15 colors which will be reserved for contours.
     if keyword_set(no_scale_image) then $ 
        sclimage = dimage $
     else $
        sclimage = (15b+bytscl(dimage,top=(!d.table_size-3-15)))
     if keyword_set(negative) then begin
        sclimage = max(sclimage)+min(sclimage)-sclimage
     endif

     newimage = poly_2d( sclimage, $
                         [[0,0],[sz(1)/sx,0]], $
                         [[0,sz(2)/sy],[0,0]], $
                         0,sx,sy )
     dev = convert_coord(x(0)-0.5*stepx,y(0)-0.5*stepy,/data,/to_device)
     tv,newimage,dev(0),dev(1),/device,xsize=sx*factor,ysize=sy*factor
  endif

  ; Plot contours

  if keyword_set(neutral_line) then neutral=get_neutral(Blong)

  if (keyword_set(noclab)) then begin    ; No contour labels included
    if NOT keyword_set(noblong) then begin
       contour, Blong, x, y, levels=levP, c_line=[lPs,lPs], c_thick=cPthick, $
                title=title, /xst, /yst,c_labels=sclab, $
                xrange=[x0-stepx,x1+stepx],yrange=[y0-stepy,y1+stepy], $
                xtitle=xlabel,c_colors=[use_colorP],color=default_color, $
                /noerase,ytitle=info_string,/noclip
       contour, Blong, x, y, levels=levM, c_line=[lMs,lMs], c_thick=cMthick, $
                /noerase,/xst,/yst,xrange=[x0-stepx,x1+stepx], $
                yrange=[y0-stepy,y1+stepy],c_labels=sclab, $
                c_colors=[use_colorM],color=default_color,/noclip
    endif 
    if keyword_set(neutral_line) then begin
         contour, neutral,x,y,levels=[0.0], c_line=[0],$
                  c_thick=5,/noerase,/xst,/yst,xrange=[x0-stepx,x1+stepx],$
                  yrange=[y0-stepy,y1+stepy], c_labels=[0], $
                  c_colors=[150], color=default_color,spline=1./nx,/noclip
    endif
    if keyword_set(current) then begin
       contour, jz, x, y, levels=levPjz, c_line=[lPs,lPs],c_thick=cPthick, $
                title=title, /xst, /yst,c_labels=sclabjz, $
                xrange=[x0-stepx,x1+stepx],yrange=[y0-stepy,y1+stepy], $
                xtitle=xlabeljz,c_colors=[use_colorPjz],color=default_color, $
                /noerase,ytitle=info_string,/noclip
       contour, jz, x, y, levels=levMjz, c_line=[lMs,lMs],c_thick=cMthick, $
                /noerase,/xst,/yst,xrange=[x0-stepx,x1+stepx], $
                yrange=[y0-stepy,y1+stepy],c_labels=sclabjz, $
                c_colors=[use_colorMjz],color=default_color,/noclip
    endif
  endif else begin    ;  Contour labels included 
    if NOT keyword_set(noblong) then begin
       contour, Blong,x,y,levels=levP, c_annotation=lev_labP,c_line=[lPs,lPs], $
                c_thick=cPthick,title=title, /xst, /yst,c_labels=sclab, $
                xrange=[x0-stepx,x1+stepx],yrange=[y0-stepy,y1+stepy],$
                xtitle=xlabel,c_colors=[use_colorP],color=default_color, $
                /noerase,ytitle=info_string,/noclip
       contour, Blong,x,y,levels=levM, c_annotation=lev_labM, c_line=[lMs,lMs],$
                c_thick=cMthick,/noerase,/xst,/yst,xrange=[x0-stepx,x1+stepx],$
                yrange=[y0-stepy,y1+stepy],c_labels=sclab, $
                c_colors=[use_colorM], color=default_color,/noclip
    endif
    if keyword_set(neutral_line) then begin
         contour, neutral,x,y,levels=[0.0], c_line=[0],$
                  c_thick=5,/noerase,/xst,/yst,xrange=[x0-stepx,x1+stepx],$
                  yrange=[y0-stepy,y1+stepy], c_labels=[0],  $
                  c_colors=[150], color=default_color,spline=1./nx,/noclip
    endif
    if keyword_set(current) then begin
       contour,jz,x,y,levels=levPjz, c_annotation=lev_labPjz,c_line=[lPs,lPs], $
                c_thick=cPthick,title=title, /xst, /yst,c_labels=sclabjz, $
                xrange=[x0-stepx,x1+stepx],yrange=[y0-stepy,y1+stepy], $
                xtitle=xlabeljz,c_colors=[use_colorPjz],color=default_color, $
                /noerase,ytitle=info_string,/noclip
       contour, jz,x,y,levels=levMjz, c_annotation=lev_labMjz,c_line=[lMs,lMs],$
                c_thick=cMthick,/noerase,/xst,/yst,xrange=[x0-stepx,x1+stepx],$
                yrange=[y0-stepy,y1+stepy],c_labels=sclabjz, $
                c_colors=[use_colorMjz], color=default_color,/noclip
    endif
  endelse

  ; Plot a legend

  if NOT keyword_set(nowords) then begin
    ;xyouts,0.03,0.5,info_string,color=default_color,/normal,alignment=0.5, $
    ;       orientation=90.,charsize=charsize,/noclip

    if NOT keyword_set(noblong) then begin
       xyouts,0.06,0.05,'B positive',color=default_color,/normal, $
          charsize=charsize,/noclip
       xyouts,0.06,0.08,'B negative',color=default_color,/normal, $
          charsize=charsize,/noclip
       plots,/normal,[0.01,0.05],[0.055,0.055],linestyle=lPs, $
          color=use_colorP,thick=thick
       plots,/normal,[0.01,0.05],[0.085,0.085],linestyle=lMs, $
          color=use_colorM,thick=thick
    endif
    if NOT keyword_set(novectors) then begin
       btkeylen = 1000.0  ; Must be a power of 10
       btdivisor = 2.0     
       REPEAT begin
          btkeylen = btkeylen/btdivisor
          if btdivisor EQ 2.0 then btdivisor=5.0 $  ; This gives 10000,5000,1000,
          else btdivisor=2.0                        ; 500,100,50,10
          if keyword_set(unproject) then begin
             trlabel = "horizontal" 
          endif $
          else begin
             trlabel = "transverse" 
          endelse
          maxBT = max(abs(Btrans))
          if keyword_set(maxT) then begin
             if maxT LT maxBT then data1000 = btkeylen/maxT $
             else data1000 = btkeylen/maxBT
          endif $
          else data1000 = btkeylen/maxBT
          if n_elements(length) GT 0 then data1000 = data1000 * length
          tline = fltarr(2,2)
          tline(0,1) = data1000(0)
          tline = convert_coord(tline,/data,/to_normal)
          tline(0,1) = tline(0,1) - tline(0,0) + 0.01
          tline(0,0) = 0.01
          tline(1,*) = [0.025,0.025]
       endrep UNTIL ( abs(tline(0,1)-tline(0,0)) LE 0.04 OR btkeylen LE 10.0 )
       xyouts,0.06,0.02,'B '+trlabel+" ("+strcompress(string(btkeylen,format="(f12.1)"),/remove_all)+" G)",color=default_color,/normal, $
              charsize=charsize,/noclip
       plots,/normal,tline,color=default_color,thick=thick
    endif

    if keyword_set(current) then begin
       if keyword_set(tcurrent) then begin
          xyouts,0.78,0.03,'J!Dtotal!N positive',color=default_color,/normal, $
                 charsize=charsize,/noclip
          xyouts,0.78,0.08,'J!Dtotal!N negative',color=default_color,/normal, $
                 charsize=charsize,/noclip
       endif $
       else begin
          xyouts,0.8,0.03,'J!Dz!N positive',color=default_color,/normal, $
                 charsize=charsize,/noclip
          xyouts,0.8,0.08,'J!Dz!N negative',color=default_color,/normal, $
                 charsize=charsize,/noclip
       endelse
       plots,/normal,[0.90,0.95],[0.035,0.035],linestyle=lPs,color=use_colorPjz,thick=thick
       plots,/normal,[0.90,0.95],[0.085,0.085],linestyle=lMs,color=use_colorMjz,thick=thick
    endif
  endif

  ; Plot transverse vectors on top of contours
 
  if NOT keyword_set(novectors) then $
     trmvect,u,v,x,y,length=length, missing=missing,maxT=maxT, $
             incolor=use_colorV,color_on=incolor,image=sclimage,/noerase, $
             arrows=arrows


  if ((keyword_set(reduce) AND NOT keyword_set(no_magnify)) OR $
     keyword_set(magnify)) AND  $
     (out EQ 'display') then begin
     ; Invoke the magnifying glass
     bmagnify,mag,pxd,pyd,b_window=!D.WINDOW,set_scale=set_scale,length=length,$
             missing=missing,maxT=maxT,levels=levels,LineP_style=LineP_style, $
             LineM_style=LineM_style,scale=scale,xstep=inxstep, $
             no_c_lab=noclab,incolor=incolor,current=current,noblong=noblong,$
             novectors=novectors,image=simage,arrows=arrows, $
             unproject=unproject,tcurrent=tcurrent,print_colors=print_colors, $
             negative=negative,JLevels=levPjz
  endif

  if NOT keyword_set(splot) then begin
     !P.POSITION = save_window  ; Restore old plotting parameters
     !X = save_windowX
     !Y = save_windowY
  endif

  !P.charsize = pcharsave

  if (out EQ 'print') then begin
    device, /close_file
    file = 'idl.' + strlowcase(!D.NAME)
    cmd = 'lpr -r ' + file     ; UNIX specific command to print/remove file
    spawn,cmd
    print, 'File ',file,' printed'
    set_plot,save_device    ; Restore old device
  endif

end
