;*******************************************************************************
;
; Procedure bdiv
; Calculate and plot the X/Y divergence of the magnetic field
; The calculation is done by using the image coordinates instead of doing
; interpolation.  It is much faster than doing interpolation and keep the
; resolution of the original measured data.
;
; Input:
;	1)Bdxarraym: X component of the photospherice magnetic field
;	2)Bdyarraym: Y component of the photospheric magnetic field
;	3)Bdzarraym: Normal component of the photospheric magnetic field
;       4)pxd: X coordinate on the photosphere for each of the pixel
;		in the above array 
;	5)pyd: Y coordinate on the photosphere for each of the pixel
;		in the above array
;	6)B0: The heliocentric lattitude of disk center
;	7)P: The angle between the disk projection of the solar spin axis
;		and the north point of the disk. (eastward is positive)
;		(both B0 and P are in radiants)
;	8)LC: The heliocentric longitude of the center point of the scan
;		(in radiants)
;	9)BC: The heliocentric lattitude of the center point of the scan
;		(in radiants)
;
; Output:
;	1)jzarrayh: X/Y divergence of the field
;	2)jpxd: The X coordinate on the photosphere for each pixel in
;		jzarrayh array
;	3)jpyd: The Y coordinate on the photosphere for each pixel in
;		jzarrayh array
;	4)Photospheric vector magnetogram superimposed by divergence
;
;*******************************************************************************
;
PRO bdiv, bdxarraym,bdyarraym,bdzarraym,pxd,pyd,b0,p,lc,bc,$
jzarrayh,jpxd,jpyd,noplot=noplot,dBdz=dBdz,pix_size=pix_size
ss=size(bdxarraym)
nx=ss(1)
ny=ss(2)
if n_elements(pix_size) EQ 0 then begin
   pix_size = [5.656,5.656]
   print,'BDIV: Using default pixel size of 5.656 arcseconds'
endif
cxx=cos(p)*cos(lc)-sin(p)*sin(b0)*sin(lc)
cxy=(-1)*cos(p)*sin(bc)*sin(lc)-sin(p)*(cos(b0)*cos(bc)+sin(b0)* $
sin(bc)*cos(lc))
cyx=sin(p)*cos(lc)+cos(p)*sin(b0)*sin(lc)
cyy=(-1)*sin(p)*sin(bc)*sin(lc)+cos(p)*(cos(b0)*cos(bc)+ $
sin(b0)*sin(bc)*cos(lc))
x0=nx*(pix_size(0)/min(pix_size))/2.
y0=ny*(pix_size(1)/min(pix_size))/2.
delt=cxx*cyy-cxy*cyx
   lylxbx=shift(bdxarraym,-1,-1)
   lylxby=shift(bdyarraym,-1,-1)
   lxbx=shift(bdxarraym,-1,0)
   lxby=shift(bdyarraym,-1,0)
   lybx=shift(bdxarraym,0,-1)
   lyby=shift(bdyarraym,0,-1)
   bx=bdxarraym
   by=bdyarraym
   dev1=0.5*(lylxby+lxby-lyby-by)   ; dBy/dx   cxx = dx/dX
   dev2=0.5*(lylxby+lyby-lxby-by)   ; dBy/dy   cyx = dy/dX
   dev3=0.5*(lylxbx+lxbx-lybx-bx)   ; dBx/dx   cxy = dx/dY
   dev4=0.5*(lylxbx+lybx-lxbx-bx)   ; dBx/dy   cyy = dy/dY
; The constant 0.0193 is derived as follows:
; 10^3 * 10^(-4) / (mu * pxlength)
; where mu = 4*pi*10^(-7)
; pxlength = length of each pixel in meters (5.7 arcsec corresponds to 4.13e6 m)
; 10^3 is the conversion between mAmp and Amp
; 10^(-4) is the conversion between Gauss and Tesla
; jz is thus in units of mAmp/m^2
   ;jzarrayh=0.0193*(cxx*dev1+cyx*dev2-cxy*dev3-cyy*dev4)
   ; Divergence of B
   jzarrayh=0.0193*(cxx*dev3+cyx*dev4+cxy*dev1+cyy*dev2)*(5.656/min(pix_size))
   if n_elements(dBdz) GT 0 then jzarrayh = jzarrayh + dBdz
   jzarrayh(nx-1,*)=0.
   jzarrayh(*,ny-1)=0.
   xcoor=(indgen(nx)#replicate(1,ny)+0.5)*pix_size(0)/min(pix_size)
   ycoor=(replicate(1,nx)#indgen(ny)+0.5)*pix_size(1)/min(pix_size)
   dircos=cos(bc)*cos(lc)*cos(b0)+sin(bc)*sin(b0)
   nmax=fix(sqrt(float(nx*nx+ny*ny))/dircos/2.)
   jpxd=(cyy*(xcoor-x0)-cxy*(ycoor-y0))/delt+nmax
   jpyd=(cxx*(ycoor-y0)-cyx*(xcoor-x0))/delt+nmax
   jzarrayh(0,*)=0.
   jzarrayh(*,0)=0.
   jpxd(0,*)=pxd(0,*)
   jpxd(*,0)=pxd(*,0)
   jpxd(nx-1,*)=pxd(nx-1,*)
   jpxd(*,ny-1)=pxd(*,ny-1)
   jpyd(0,*)=pyd(0,*)
   jpyd(*,0)=pyd(*,0)
   jpyd(nx-1,*)=pyd(nx-1,*)
   jpyd(*,ny-1)=pyd(*,ny-1)
xsize=max(pxd)-min(pxd)
ysize=max(pyd)-min(pyd)
if (NOT keyword_set(noplot)) then begin
   print,'Tell me the window number to plot the result'
   read,nw
   window,nw,xs=xsize*20,ys=ysize*20
   vfield2,bdxarraym,bdyarraym,pxd,pyd,thick=1,color=1,background=0,length=1.0
   lv=[25,50,100,200,400,800,1600,3200]
   lv=[-rotate(lv,2),lv]
   contour,bdzarraym,pxd,pyd,levels=lv,$
   c_color=(lv lt 0)*100+100,/noerase, xstyle=5,ystyle=5
      jlv=[-20,-15,-10,-8,-5,5,8,10,15,20]
      contour,jzarrayh,jpxd,jpyd,levels=jlv,c_color=(jlv gt 0)*20+35,$
      xstyle=5,ystyle=5,/noerase
   whitebg
   specialct,100,0,0,255
   specialct,200,0,255,0
   specialct,55,255,0,0
   specialct,35,255,0,255
   print,'Do you want to draw alpha histogram? Yes: y , No: n'
   answer='ansewer'
   read,answer
   if answer eq 'n' then goto,finish
   print,'You want to choose pixels with the magnitude of vertical current and'
   print,'normal magnetic field strength greater than j0 and Bz0 respectively'
   print,'Tell me the value of j0 and Bz0 please. '
   print,'j0='
   read,j0
   print,'Bz0='
   read,bz0
   ;smoothing the vertical magnetic field
      bdzarr=0.25*(shift(bdzarraym,-1,-1)+shift(bdzarraym,-1,0)+ $
             shift(bdzarraym,0,-1)+bdzarraym)
      bdzarr(*,ny-1)=0.
      bdzarr(nx-1,*)=0.
   ncrit=where(abs(bdzarr) lt bz0 or abs(jzarrayh) lt j0)
   crit=where(abs(bdzarr) ge bz0 and abs(jzarrayh) ge j0)
   ;nonzero the bdzarr array
   bdzarr(ncrit)=fix(bdzarr(ncrit))+0.1
   alpha=jzarrayh/bdzarr
   alpha(ncrit)=0.
   print,'Do you want to take a look at the position of those pixels chosen?'
   print,'Yes: y , No: n'
   read,answer
   if answer eq 'n' then goto, skip
   print,'Tell me the window number you want to plot'
   read,nw
   window,nw,xs=xsize*10,ys=ysize*10
   whitebg
      contour,bdzarr,jpxd,jpyd,levels=lv,c_color=(lv lt 0)*100+100,$
      xstyle=1,ystyle=1
      oplot,jpxd(crit),jpyd(crit),color=1,psym=1
   skip: critplus=where(bdzarr ge bz0 and abs(jzarrayh) ge j0)
         critminus=where(bdzarr le -(bz0) and abs(jzarrayh) ge j0)
      alfaplus=alpha(critplus)
      alfaminus=alpha(critminus)
      sss=[min(alfaminus),min(alfaplus)]
      ttt=[max(alfaminus),max(alfaplus)]
      dimlow=min(sss)
      dimhigh=max(ttt)
      bdzarrplus=bdzarr(critplus)
      bdzarrminus=bdzarr(critminus)
      alphaerr=4./bz0
      dim=fix((dimhigh-dimlow)/alphaerr)+1
      fluxplus=fltarr(dim)
      fluxminus=fltarr(dim)
      scr=size(bdzarrplus)
      imaxplus=scr(1)
      for i=0, imaxplus-1 do begin
      binalpha=fix((alfaplus(i)-dimlow)/alphaerr)
      fluxplus(binalpha)=fluxplus(binalpha)+bdzarrplus(i)
      endfor
      scr = size(bdzarrminus)
      imaxminus=scr(1)
      for i= 0, imaxminus-1 do begin
      binalpha=fix((alfaminus(i)-dimlow)/alphaerr)
      fluxminus(binalpha)=fluxminus(binalpha)+bdzarrminus(i)
      endfor
   print,'**fluxplus**'
   print,fluxplus
   print,'**fluxminus**'
   print,fluxminus
   print,'Tell me the number for the window in which you want to plot the '
   print,'alpha histogram'
   read,nw
   window,nw
   pxlength=5.7*(1.496e+11)*!pi/3600./180.
   pxarea=pxlength^2/delt
   fluxplus=fluxplus*pxarea*(1.e+4)/(1.e+20)
   fluxminus=fluxminus*pxarea*(1.e+4)/(1.e+20)
   xline=(indgen(dim)*alphaerr+dimlow+0.5*alphaerr)
   plot,xline,-fluxminus,psym=-7,xrange=[-0.1,0.1],yrange=[0,50.],$
   xtitle='alpha (Gauss*sqM/mA)',ytitle='Bflux (10^20 Maxwell)'
   xline=(indgen(dim)*alphaerr+dimlow+0.5*alphaerr)
   oplot,xline,fluxplus,psym=-6,$
   xtitle='alpha (Gauss*sqM/mA)',ytitle='Bflux (10^20Maxwell)'
   jzbala=total(jzarrayh(crit))*pxarea*(1.e-3)
   jzabs=total(abs(jzarrayh(crit)))*pxarea*(1.e-3)
   bzbala=total(bdzarr(crit))*pxarea*(1.e+4)
   bzabs=total(abs(bdzarr(crit)))*pxarea*(1.e+4)
   jzper=abs(jzbala)/jzabs
   bzper=abs(bzbala)/bzabs
   jzsumerr=sqrt(n_elements(crit))*j0*pxarea*(1.e-3)
   bzsumerr=sqrt(n_elements(crit))*bz0*pxarea*(1.e+4)
   print,'**************************Table 1 *******************************'
   print,' total jz	total |jz|	total bz	total |bz| '
   print,' (A)		(A)		(Maxwell)	(Maxwell)'
   print, jzbala,jzabs,bzbala,bzabs
   print,'total jz err	total bz err 	pixel number'
   print,jzsumerr,bzsumerr,n_elements(crit)
   print,'******************************************************************'
endif
finish: ;print,'finished'
end
