
;+
;NAME:
;	current
;PURPOSE:
;	Resolve the 180 degree ambiguity and compute vertical current.
;CATEGORY:
;
;CALLING SEQUENCE:
;	mag = current(fits_file)
;INPUTS:
;	fits_file = Name of file to get field data from OR a magnetic field
;                   structure (see bfits.pro for structure format).
;OPTIONAL INPUT PARAMETERS:
;	none
;KEYWORD PARAMETERS
;	scale = scale of plots.  1.0 is the default 1.2 is 20% larger etc.
;       integral_structure = Magnetic field from integral method.  Merged
;                            with the field from the fits file at low field
;                            strength.
;       no_minimize = turn minimization off
;       no_acute = turn acute angle resolution off
;       do_atest = turn on Acute PANG test
;       header = saves the FITS header from the bfits routine
;       title = title for plot of current
;       alpha = linear force-free alpha for initial ambiguity resolution.
;               alpha=0 is a potential field.
;       /normal = Use the vertical field to compute the linear force free field
;                 instead of the LOS field.
;       finite_energy = if set, restrict the force free solution to
;                          the finite energy solutions.
;       truea = an array of azimuths giving the "correct" field azimuth.  If
;               included, this array is used to resolve the 180 degree
;               ambiguity, and all the other techniques are turned off.
;       no_los = skip the resolution by the computed line-of-sight potential 
;                field.  Don't use this unless the input field is nearly 
;                correct.
;       no_deproject = Do not rotate to disk center.  Caution: Don't do this
;                      unless the data is already deprojected!
;       /verbose = print out lots of diagnostics
;       /noask = Don't ask any questions ... just use the defaults.  CAREFUL!
;OUTPUTS:
;	mag = combined magnetic/current structure
;COMMON BLOCKS:
;	none
;SIDE EFFECTS:
;	Modifies color table
;RESTRICTIONS:
;	none
;PROCEDURE:
; Throughout, +x is assumed West and +y is assumed North, and the azimuth is
; assumed to be measured CCW from north
;MODIFICATION HISTORY:
;	TRM 8/91  Version 1.0
;       TRM 11/4/93  If a structure and a FITS header are passed, the 
;                    POINT structure overrides the position info in 
;                    the FITS header.
;       AAP 4/8/98   Added call jxyz.pro if keyword_set(noask)
;-

function current,fits_file,scale=scale,integral_structure=imag,no_minimize=nomin,no_acute=no_acute,do_atest=do_atest,header=head,title=title,alpha=alpha,truea=truea,no_los=nolos,no_deproject=no_deproject,finite_energy=finite_energy,verbose=VERBOSE,normal=normal,quiet=quiet,radial=radial,noask=noask

;  Analysis flags

   MINIMIZE = 1   ; If true, do the current or divergence minimization
   ACUTE = 1      ; If true, do the acute angle resolution after potential 
                  ; field resolution and before minimizing current/divergence
   RANDM = 0      ; If true randomize the order of the current tests
   INTER = 3      ; If 1 do pixels in sequential order
                  ; If 2 do even pixels first then odd pixels
                  ; If 3 do in radial order from specified point
                  ; If 4 do in descending order of transverse field strength
                  ; If 5 do in increasing order of transverse field strength
                  ; INTER is not relevant if RANDM is set
   NOISE = 75.    ; All pixels with total field below NOISE are analyzed even
                  ; when APTEST and/or PTEST are set.  Pixels with field 
                  ; strength less than 1.0 are never analyzed since these are
                  ; probably points for which the field was not computed.

;  Acute angle flags

   NAITER = 20    ; Specifies the maximum number of iterations
   APTEST = 0     ; If set, only do points at least APANG deg from potential
   APANG = 25.0   ; The test angle for APTEST = true (should be less than 90)

;  Current/Divergence Minimization Flags

   NITER = 10     ; specifies the maximum number of iterations
   INDEP = 0      ; If true do each point independently from the rest
                  ; RANDM and INTER are not relevant if INDEP is set
   DIVB = 1       ; If true minimize the divergence of B when Btrans > DIVBCUT
                  ; else minimize the current at every point
   DIVBCUT = 1100.; Total field strength cutoff for current/divB calculations:
                  ; set to 0 to do divergence calculation at every point
   SUB_SIZE = 2   ; Number of points on either side of selected point used
                  ; in the computation of the current/divergence
   PTEST = 1      ; If set, only do points at least PANG deg from potential
   PANG = 85.0    ; The test angle for PTEST = true (should be less than 90)

;   DIVB = 0
;   DIVBCUT = 0

; Diagnostic flags

   if n_elements(VERBOSE) NE 1 then VERBOSE = 0    ; If true print diagnostic information

;   ----------------  BEGIN CODE -------------------


   if keyword_set(nomin) then MINIMIZE=0
   if keyword_set(no_acute) then ACUTE=0
   if keyword_set(do_atest) then APTEST=1

   alpha_on = 1
   if n_elements(alpha) NE 1 then begin 
      alpha = 0.
      alpha_on = 0
   endif

   ; Check the file 

   sfits = size(fits_file)
   if sfits(n_elements(sfits)-2) NE 7 then begin
      if sfits(n_elements(sfits)-2) EQ 8 then begin
         mag = fits_file
      endif $
      else begin
         print
         fits_file=' '
         print, 'FITS file?'     ; Ask for data file name
         read, fits_file
         mag = bfits(fits_file,head=head)
      endelse
   endif $ 
   else mag = bfits(fits_file,head=head)   ; Read in data

; Check that there are no blank lines in the header

   nhead = n_elements(head)
   if nhead GT 0 then begin
      i = nhead-1
      REPEAT begin
         bhead = byte(head(i))
         blank = where(bhead EQ 32B,count)    ; 32B is a space
         if count EQ n_elements(bhead) then begin
            temp = strarr(nhead-1)
            if i NE 0 then temp(0:i-1) = head(0:i-1)
            if i NE nhead-1 then temp(i:nhead-2) = head(i+1,nhead-1)
            head = temp
            nhead = nhead - 1
            i = nhead
         endif
         i = i - 1
      endrep UNTIL i LT 0
   endif

   if keyword_set(no_deproject) then begin
      mag.point.b0 = 0.
      mag.point.p = 0.
      mag.point.lat = 0.
      mag.point.lon = 0.
   end

   if n_elements(alpha) LE 0 then a=0 else a=alpha
   hcomment = "Ambiguity resolved "+systime()+"; Alpha="+ $
                 strcompress(string(a,format="(f6.2)"),/remove_all)

   if n_elements(imag) NE 0 then begin
      imgsize = size(imag)
      if imgsize(n_elements(imgsize)-2) NE 8 then $
         message,'ERROR: integral field must be a structure' $
      else $
         mag = bmerge(mag,imag,/quiet)
   endif
   
   mag.b_trans = abs(mag.b_trans)

   ; check whether the latitude and CMD are in the structure:

   pos_exists = (tag_index(mag,'LATITUDE') GE 0) AND $
                (tag_index(mag,'CMD') GE 0)

   ; Get method of selecting order points are processed in

   print
   if RANDM then print,'RANDOMIZING point selection' 
   if INDEP then print,'Points treated independently' 
   if INTER EQ 1 then print,'SEQUNTIAL point selection' $
   else if INTER EQ 2 then print,'EVEN/ODD point selection' $
   else if INTER EQ 3 then print,'Radial point selection' $
   else if INTER EQ 4 then print,'Descending transverse field point selection' $
   else if INTER EQ 5 then print,'Ascending transverse field point selection' $
   else begin
      print,'ERROR, INTER = ',INTER,' is undefined'
      return,mag
   endelse
   print

   if (nhead GT 0) AND (tag_index(mag,'POINT',/quiet) LT 0) then begin
      instrument = strupcase( $
                      strcompress( $
                         rkey(head,'INSTRUME','CHAR'),/remove_all))
      print,'Instrument: ',instrument
      b0 = rkey(head,'B0','REAL')*!dtor         ; Get info from FITS header
      p = rkey(head,'P','REAL')*!dtor
      pos_cmd = rkey(head,'POS-CMD','REAL')*!dtor
      pos_lat = rkey(head,'POS-LAT','REAL')*!dtor
      if strpos(instrument,'STOKESPOLARIMETER') GE 0 then begin
         raster = rkey(head,'RASTTYPE','CHAR')
         if strpos('RECTANGULAR',raster) LT 0 then begin
            print,'ERROR: This routine cannot handle raster type ',raster
            return,mag
         endif
      endif
      pix_size1 = abs(rkey(head,'CDELT1','REAL'))
      pix_size2 = abs(rkey(head,'CDELT2','REAL'))
      if VERBOSE then print,'Pixel size: ',pix_size1,pix_size2
      pix_size = double([pix_size1,pix_size2])
   endif $
   else begin
      b0 = mag.point.b0
      p = mag.point.p
      pos_cmd = mag.point.cmd
      pos_lat = mag.point.lat
      if n_elements(mag.point.pix_size) EQ 1 then $
         pix_size = double([mag.point.pix_size(0),mag.point.pix_size(0)]) $
      else pix_size = double(mag.point.pix_size)
   endelse
   mpix_size = min(pix_size)
   
   btotal = sqrt( mag.b_trans^2 + mag.b_long^2 )  ; Total field strength

   nx = n_elements(mag.b_long(*,0))
   ny = n_elements(mag.b_long(0,*))

;  The first step is to resolve the ambiguity by selecting the azimuth which
;  is closest to the potential field.

   if NOT keyword_set(truea) then begin

; Compute the potential field to get a crude resolution of the azimuth

  ;; potential,b0,p,pos_cmd,pos_lat,mag.b_long,px,py,/noplot
   pot = fff(mag.b_long,/image,b0=b0,pang=p,lat=pos_lat,cmd=pos_cmd, $
              alpha=0.,pixel=pix_size)
   pz = mag.b_long
   px = pot(*,*,0)   ; these are image coordinates
   py = pot(*,*,1)
   check = check_math(0,1)
      ;pa_azimuth = 360.0 - atan(px,py)/!dtor  ; Az is CCW from North
      pa_azimuth = atan(-px,py)/!dtor  ; Az is CCW from North
   check = check_math(1,0)

;  Fix the azimuthal angle from the potential field

   if NOT keyword_set(nolos) then begin

      fixa = where(abs(arange(mag.B_AZIM(*),pa_azimuth(*))-pa_azimuth(*)) GT $
                   90.,count)
      if (count gt 0) then mag.B_AZIM(fixa) = mag.B_AZIM(fixa) - 180.
      mag.B_AZIM = arange(mag.B_AZIM,0.)
      if VERBOSE then begin
         print
         print,'Resolved azimuth with LOS potential field:',count,' flips'
      endif
   endif

; Now compute the force free field and resolve the azimuth again.  It is
; necessary to transform to heliographic coordinates, so the field needed
; to be resolved with the potential field first.

   if alpha_on then begin

      ; Transform to heliographic coords

      disktran, -mag.b_trans*sin(mag.b_azim*!dtor), $
                 mag.b_trans*cos(mag.b_azim*!dtor), $
                 mag.b_long,b0,p,pos_cmd,pos_lat,bdx,bdy,bdz,pxd,pyd,/noplot,pix_size=pix_size
   
      if tag_index(mag,'BX',/quiet) LT 0 then mag=add2str(mag,['bx'],bdx) $
      else mag.bx=bdx
      if tag_index(mag,'BY',/quiet) LT 0 then mag=add2str(mag,['by'],bdy) $
      else mag.by=bdy
      if tag_index(mag,'BZ',/quiet) LT 0 then mag=add2str(mag,['bz'],bdz) $
      else mag.bz=bdz
      if tag_index(mag,'B_PXD',/quiet) LT 0 then mag=add2str(mag,['b_pxd'],pxd) $
      else mag.b_pxd=pxd
      if tag_index(mag,'B_PYD',/quiet) LT 0 then mag=add2str(mag,['b_pyd'],pyd) $
      else mag.b_pyd=pyd
   
      ;;pot = bff(mag,nz=1,alpha=alpha,/quiet,/no_heliographic, $
      ;;          finite_energy=finite_energy)  ; Compute the force-free field
      ;;px = reform(pot.bxp(*,*,0))  ; These are heliographic
      ;;py = reform(pot.byp(*,*,0))
      ;;pz = reform(pot.bzp(*,*,0))
   
      ;;xf = set_xform( b0, p, pos_lat, pos_cmd)
      ;;xform, px, py, pz, transform=xf.helio2image  ; Put back into image plane
   
      ;;Bvector,pot,/unproject,/arrows,/inc,/no_c,len=2, $
      ;;       title='Linear Force Free Field (alpha=' + $
      ;;          strcompress(string(alpha,format="(f5.1)")+")",/remove_all), $
      ;;       image=mag.i_cont
   
      if keyword_set(normal) then $
         pot = fff(mag.bz,/normal,/image,b0=b0,pang=p,lat=pos_lat, $
                   cmd=pos_cmd,alpha=alpha,pixel=pix_size, $
                   bx=mag.bx,by=mag.by) $
      else $
         pot = fff(mag.b_long,/image,b0=b0,pang=p,lat=pos_lat,cmd=pos_cmd, $
                   alpha=alpha,pixel=pix_size,bx=mag.bx,by=mag.by)
      px = pot(*,*,0)   ; these are image coordinates
      py = pot(*,*,1)
      pz = pot(*,*,2)
      
      if NOT keyword_set(quiet) then begin
         Bvector,{b_trans:sqrt(px^2+py^2),b_long:pz,b_azim:atan(-px,py)/!dtor, $
                  point:mag.point},/arrows,/inc,/no_c,len=2, $
                title='Linear Force Free Field (alpha=' + $
                   strcompress(string(alpha,format="(f5.1)")+")",/remove_all), $
               image=mag.i_cont

         print
         print,'PRESS RETURN TO CONTINUE ...'
         print
         junk = 'string '
         read,junk,format="(a80)"
      endif

      check = check_math(0,1)
         ;pa_azimuth = 360.0 - atan(px,py)/!dtor  ; Az is CCW from North
         pa_azimuth = atan(-px,py)/!dtor  ; Az is CCW from North
      check = check_math(1,0)
   
   ; Fix the azimuthal angle from the force-free field
   
     fixa = where(abs(arange(mag.B_AZIM(*),pa_azimuth(*))-pa_azimuth(*)) GT $
                  90.,count)
   
     if (count gt 0) then mag.B_AZIM(fixa) = mag.B_AZIM(fixa) - 180.
     mag.B_AZIM = arange(mag.B_AZIM,0.)
     if VERBOSE then begin
        print
        print,'Resolved azimuth with force-free field:',count,' flips'
     endif

  endif  ; End of force free computation
   
; Compute the derivative of the force-free\potential field wrt z, using the 
; 2D divergence computed in bdiv.pro.  -dBz/dz = dBx/dx + dBy/dy, since DivB=0.

   disktran, px,py,pz,b0,p,pos_cmd,pos_lat,pdx,pdy,pdz,pxd,pyd,/noplot,pix_size=pix_size

   bdiv,pdx,pdy,pdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd,jpyd, $
        /noplot,pix_size=pix_size
   dBdz = -jz

; Get x,y,z componets of magnetic field for current computation

  bx = -mag.b_trans * sin(mag.b_azim*!dtor)
  by = mag.b_trans * cos(mag.b_azim*!dtor)
  bz = mag.b_long

; Display the current computed after resolving the 180 degree ambiguity with
; the potential field

  disktran, bx,by,bz,b0,p,pos_cmd,pos_lat,bdx,bdy,bdz,pxd,pyd,/noplot,pix_size=pix_size

  if INTER EQ 3 OR (VERBOSE and (NOT ACUTE)) then begin
     jzhigh,bdx,bdy,bdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd,jpyd,pix_size, $
            jz_window=0,scale=scale,title=title,/noask,noplot=keyword_set(quiet)
  endif

 ; Get the radial point from the user if INTER is 3

  if INTER EQ 3 then begin
     if n_elements(radial) NE 2 then begin
        REPEAT begin
          minI = min(mag.i_cont,minIpix)
          print
          print,'Minimum intensity is at',minIpix MOD nx, long(minIpix/nx)
          print,'Starting from the lower left corner of a daily-plot,
          print,'"row" is the horizontal axis, and "col" is the vertical'
          print
          print,'Input the radiant point: row col'
          radial = [-1,-1]
          if keyword_set(noask) then radial = [minIpix MOD nx,long(minIpix/nx)] $
          else read, radial
          print
          radial = long(radial)
          if radial(0) LT 0 OR radial(0) GE nx OR $
             radial(1) LT 0 OR radial(1) GE ny then $
             print, "Huh?  Try that again.'
        endrep UNTIL radial(0) GE 0 and radial(0) LT nx and $
                     radial(1) GE 0 and radial(1) LT ny
     endif
     hcomment = hcomment + "; RP=("  $
                         + strcompress(string(radial(0)),/remove_all) $
                         + ","+ strcompress(string(radial(1)),/remove_all) + ")"
  endif

  sremain = lindgen(nx*ny)  ; Default is sequential order

  if INTER EQ 2 then begin         ; Even/odd pixels
     sremain = [sremain(where(sremain MOD 2 EQ 0)), $
               sremain(where(sremain MOD 2 NE 0))]
  endif $
  else if INTER EQ 3 then begin  ; Points by distance from radial point
      if n_elements(radial) NE 2 OR radial(0) LT 0 OR radial(1) LT 0 $
      then radial = [nx/2,ny/2]
      print, 'RADIANT point is ',radial & print
      ; Note that row = sremain mod nx
      ; and       col = long(sremain/nx)
      distance = sqrt( ((sremain mod nx)-radial(0))^2 + $
                       ((long(sremain/nx))-radial(1))^2 )
      sremain = sremain(sort(distance))
  endif $
  else if INTER EQ 4 then $    ; Descending order of transverse field
     sremain = sremain(sort(-ABS(mag.B_TRANS(sremain)))) $ 
  else if INTER EQ 5 then $    ; Ascending order of transverse field
     sremain = sremain(sort(ABS(mag.B_TRANS(sremain))))

; ***********************
; ACUTE angle resolution
; ***********************

  if ACUTE then begin

  remain = sremain
  if APTEST then begin  ; Use only points far from potential
     diff = abs( arange(mag.B_AZIM(remain),pa_azimuth(remain))  $
                 - pa_azimuth(remain) )
     diff = where(diff GT APANG OR (btotal LE NOISE AND btotal GT 1.), count)
     if count GT 0 then remain = remain(diff) else remain=[0]
  endif

  flip = -1
  dot1 = fltarr(nx,ny)
  dot2 = fltarr(nx,ny)
  flip_count = bytarr(nx,ny)

  for iteration = 1,NAITER do begin

     old_flip = flip
     flip = -1

     if INDEP or iteration EQ 1 then begin
        bx = -mag.b_trans * sin(mag.b_azim*!dtor)
        by = mag.b_trans * cos(mag.b_azim*!dtor)
        bz = mag.b_long
        disktran, bx,by,bz,b0,p,pos_cmd,pos_lat,bdx,bdy,bdz,pxd,pyd,/noplot,pix_size=pix_size
     endif

     bxs = -mag.b_trans * sin(mag.b_azim*!dtor+!pi)
     bys = mag.b_trans * cos(mag.b_azim*!dtor+!pi)
     bzs = mag.b_long
     disktran,bxs,bys,bzs,b0,p,pos_cmd,pos_lat,bdsx,bdsy,bdsz,pxds,pyds,/noplot,pix_size=pix_size

     ; Compute the sum of the dot product of the each field vector with all
     ; its neighbors.  Do this for both azimuth possibilities - the maximum
     ; dot product gives the orientation closest to the neighbors.  Note
     ; that using the dot product weights the neighbors by their respective
     ; field strengths.

     for loop=0L,n_elements(remain)-1 do begin
        row = remain(loop) mod nx
        col = long(remain(loop)/nx)
        if row GE 1 then  rl=1 else rl=0
        if row LE nx-2 then rh=1 else rh=0
        if col GE 1 then cl=1 else cl=0
        if col LE ny-2 then ch=1 else ch=0
        dot1(row,col) =  $
           total(bdx(row,col)*bdx(row-rl:row+rh,col-cl:col+ch)) $
               - bdx(row,col)^2 $
         + total(bdy(row,col)*bdy(row-rl:row+rh,col-cl:col+ch)) $
               - bdy(row,col)^2 $
         + total(bdz(row,col)*bdz(row-rl:row+rh,col-cl:col+ch)) $
               - bdz(row,col)^2
        dot2(row,col) = $
           total(bdsx(row,col)*bdx(row-rl:row+rh,col-cl:col+ch)) $
               - bdsx(row,col)*bdx(row,col) $
         + total(bdsy(row,col)*bdy(row-rl:row+rh,col-cl:col+ch)) $
               - bdsy(row,col)*bdy(row,col) $
         + total(bdsz(row,col)*bdz(row-rl:row+rh,col-cl:col+ch)) $
               - bdsz(row,col)*bdz(row,col)

        if NOT INDEP then begin   ; Change field immmediately
           if dot2(row,col) GT dot1(row,col) then  begin
              mag.B_AZIM(row,col) = arange(mag.B_AZIM(row,col) + 180.,0.)
              bx(row,col)=-mag.b_trans(row,col) * sin(mag.b_azim(row,col)*!dtor)
              by(row,col)=mag.b_trans(row,col) * cos(mag.b_azim(row,col)*!dtor)
              bz(row,col)=mag.b_long(row,col)
              disktran,bx(row,col), $
                       by(row,col), $
                       bz(row,col), $
                       b0,p,pos_cmd,pos_lat, $
                       tbdx,tbdy,tbdz,tpxd,tpyd,/noplot,pix_size=pix_size
              bdx(row,col) = tbdx
              bdy(row,col) = tbdy
              bdz(row,col) = tbdz
              sflip = size(flip)
              if sflip(0) GT 0 then flip = [flip,remain(loop)] $
              else flip=[remain(loop)]
           endif
        endif

     endfor

     if INDEP then begin
        flip = where(dot2 GT dot1,count) 
        if count GT 0 then mag.B_AZIM(flip) = mag.B_AZIM(flip) + 180.
        mag.B_AZIM = arange(mag.B_AZIM,0.)
     endif $
     else if flip(0) LT 0 then count=0 $
     else count = n_elements(flip)

     print,'iteration',iteration,' had',count, $
           ' flips from acute angle computation'
  
     if count EQ 0 then begin
        if (VERBOSE) then print,'Converged at iteration',iteration
        goto,DONE_ACUTE
     endif 

     flip_count(flip) = flip_count(flip) + 1
     flip_test = where(flip EQ old_flip,flip_test_count)
     if (flip_test_count EQ count AND n_elements(old_flip) EQ count) $
          OR (min(flip_count(flip)) GT 3) then begin
        ; If the same points are repeatedly flipped, or if all the flipped
        ; points have been flipped more than 3 times, terminate the iteration
        print,'REGION of CONFLICT:  Iteration terminated'
        conflict_row = flip mod nx ;The first element of places is alway 0
        conflict_col = long(flip/nx)
        rsort = sort(conflict_row)
        output = intarr(2*count)
        output(lindgen(count)*2)=conflict_row(rsort)
        output(lindgen(count)*2+1)=conflict_col(rsort)
        conflict = intarr(n_elements(conflict_row),2)
        conflict(*,0) = conflict_row(*)
        conflict(*,1) = conflict_col(*)
        ;print,output,format='(2i5)'
        acute_conflict = flip ;These points will be included in the minimization
        goto,DONE_ACUTE
     endif

  endfor  ; end of iteration loop

DONE_ACUTE:

  bx = -mag.b_trans * sin(mag.b_azim*!dtor)
  by = mag.b_trans * cos(mag.b_azim*!dtor)
  bz = mag.b_long

  disktran, bx,by,bz,b0,p,pos_cmd,pos_lat,bdx,bdy,bdz,pxd,pyd,/noplot,pix_size=pix_size
  if VERBOSE AND MINIMIZE then begin
     jzhigh,bdx,bdy,bdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd,jpyd,pix_size, $
            jz_window=0,scale=scale,title=title,/noask
  endif

  endif  ; End of ACUTE computation

; **********************************
; MINIMIZATION of current/divergence
; **********************************

  if MINIMIZE then begin

    changes = intarr(NITER)   ; Tracks number of changes in each iteration
    places = 0  ; Tracks which pixels flip in each iteration
    flip_count = bytarr(nx,ny)

    if PTEST then begin  ; Use only points far from potential
       diff = abs( arange(mag.B_AZIM(sremain),pa_azimuth(sremain))  $
                   - pa_azimuth(sremain) )
       if n_elements(acute_conflict) GT 0 then begin
          ; Include any points which the ACUTE method couldn't handle.
          ; The order of points (from INTER) is maintained by the code below.
          match, sremain, acute_conflict, suba
          ; diff contains indices of sremain elements to use in the minimization
          diff=[where((diff GT PANG OR (btotal LE NOISE AND btotal GT 1.)), count),suba] 
          diff=diff(rem_dup(diff)) ; remove duplicates
          diff=diff(sort(diff))    ; Put indices in the right order
          count = n_elements(diff)
       endif $
       else $
          diff=where(diff GT PANG OR (btotal LE NOISE AND btotal GT 1.), count)
       if count GT 0 then sremain = sremain(diff) else sremain=[0]
    endif

; Begin the iteration to the "minimum" current/divergence solution

    print
    print,'Minimize Current\Divergence'
    print

    for iteration = 1,NITER do begin

       old_places = places       ; Save old pixels flipped to compare with new
       places = 0
       changes(iteration-1) = 0

       remain = sremain   ; remain must be reset since RANDM alter it

       if INDEP then begin
          save_bx = bx
          save_by = by
          save_bz = bz
       endif 
 
       nremain = n_elements(remain)
       if VERBOSE then print,'  count',' changes',' pixel', '    Type'

       for loop = 0L,n_elements(remain)-1 do begin

          if (RANDM) then begin   ; Randomize the order of the examination
             rentry = long(randomu(seed)*(nremain-1)) 
             pixel = remain(rentry)
             if nremain GT 1 then begin
                if rentry EQ 0 then remain = remain(1:nremain-1) $
                else if rentry EQ nremain-1 then  remain = remain(0:nremain-2) $
                else remain = [remain(0:rentry-1),remain(rentry+1:nremain-1)]
                nremain = nremain - 1
             endif 
          endif $
          else pixel = remain(loop)
          row = pixel mod nx
          col = long(pixel/nx)

; Get a small subarray since the full data array need not be computed for
; every iteration

          rrange1 = max([0,row-SUB_SIZE])
          rrange2 = min([nx-1,row+SUB_SIZE])
          crange1 = max([0,col-SUB_SIZE])
          crange2 = min([ny-1,col+SUB_SIZE])

          bxs = bx(rrange1:rrange2,crange1:crange2)
          bys = by(rrange1:rrange2,crange1:crange2)
          bzs = bz(rrange1:rrange2,crange1:crange2)

          nxs = n_elements(bxs(*,0))
          nys = n_elements(bxs(0,*))

; The lat/cmd of the center of the sub-array is not quite right below when the
; FITS file did not have the CMD and latitude stored in it (this is the case
; with FITS file generated from a PM*A* file).  It should always be 
; mag.lat(row,col) and mag.lat(row,cmd) instead of pos_lat and
; pos_cmd.  The approximate position below is probably good enough, however, 
; since the final current is computed correctly.  The calls below merely 
; compare the currents with the azimuth pointed both directions so an 
; approximation is good enough.

          if pos_exists then begin
             disktran,bxs,bys,bzs,b0,p,mag.cmd(row,col)*!dtor, $
                      mag.latitude(row,col)*!dtor, $
                      bdx,bdy,bdz,pxd,pyd,/noplot,pix_size=pix_size
          endif $
          else begin
             disktran,bxs,bys,bzs,b0,p,pos_cmd,pos_lat,bdx,bdy,bdz, $
                      pxd,pyd,/noplot,pix_size=pix_size
          endelse

          if DIVB then begin
            if (total(sqrt(bxs^2+bys^2+bzs^2))/float(nxs*nys) GT DIVBCUT) then $
               DoDivB = 1 $
            else DoDivB = 0
          endif $
          else DoDivB = 0

          if DoDivB then begin
            if pos_exists then $
               bdiv,bdx,bdy,bdz,pxd,pyd,b0,p,mag.cmd(row,col)*!dtor, $
                        mag.latitude(row,col)*!dtor,jz,jpxd,jpyd, $
                        /noplot,dBdz=dBdz,pix_size=pix_size $
            else $
               bdiv,bdx,bdy,bdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd,jpyd, $
                    /noplot,dBdz=dBdz ,pix_size=pix_size
          endif $
          else begin
            if pos_exists then $
               jzhigh,bdx,bdy,bdz,pxd,pyd,b0,p,mag.cmd(row,col)*!dtor, $
                      mag.latitude(row,col)*!dtor,jz,jpxd,jpyd,pix_size, $
                      /noplot $
            else $
               jzhigh,bdx,bdy,bdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd, $
                      jpyd,pix_size,/noplot
          endelse
          ;old_total_current = total(abs(jz))
          old_total_current = total(jz^2)

          mag.b_azim(row,col) = arange(mag.b_azim(row,col) + 180.,0.)

          bx(row,col) = -mag.b_trans(row,col) * sin(mag.b_azim(row,col)*!dtor)
          by(row,col) = mag.b_trans(row,col) * cos(mag.b_azim(row,col)*!dtor)
 
          if INDEP then begin
             temp_bx = save_bx(row,col)
             temp_by = save_by(row,col)
             save_bx(row,col) = bx(row,col)
             save_by(row,col) = by(row,col)
             bxs = save_bx(rrange1:rrange2,crange1:crange2)
             bys = save_by(rrange1:rrange2,crange1:crange2)
             bzs = save_bz(rrange1:rrange2,crange1:crange2)
             save_bx(row,col) = temp_bx
             save_by(row,col) = temp_by
          endif $
          else begin
             bxs = bx(rrange1:rrange2,crange1:crange2)
             bys = by(rrange1:rrange2,crange1:crange2)
             bzs = bz(rrange1:rrange2,crange1:crange2)
          endelse

          disktran,bxs,bys,bzs,b0,p,pos_cmd,pos_lat,bdx,bdy,bdz,pxd,pyd,/noplot,pix_size=pix_size
          if DoDivB then begin
            if pos_exists then $
               	bdiv,bdx,bdy,bdz,pxd,pyd,b0,p,mag.cmd(row,col)*!dtor, $
                     mag.latitude(row,col)*!dtor,jz,jpxd,jpyd, $
                     /noplot,dBdz=dBdz,pix_size=pix_size $
            else $
               	bdiv,bdx,bdy,bdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd,jpyd, $
                     /noplot,dBdz=dBdz ,pix_size=pix_size
          endif $
          else begin
            if pos_exists then $
               jzhigh,bdx,bdy,bdz,pxd,pyd,b0,p,mag.cmd(row,col)*!dtor, $
                      mag.latitude(row,col)*!dtor,jz,jpxd,jpyd,pix_size, $
                      /noplot $
            else $
               jzhigh,bdx,bdy,bdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd,jpyd, $
                      pix_size,/noplot
          endelse
          ;new_total_current = total(abs(jz))
          new_total_current = total(jz^2)

          if abs(new_total_current) LT abs(old_total_current) then begin
             changes(iteration-1) = changes(iteration-1) + 1
             flip_count(row,col) = flip_count(row,col) + 1  ; Track flips
             places = [places,pixel]
          endif $
          else begin
             mag.b_azim(row,col) = arange(mag.b_azim(row,col) - 180.,0.)
             bx(row,col) = -mag.b_trans(row,col)*sin(mag.b_azim(row,col)*!dtor)
             by(row,col) = mag.b_trans(row,col)*cos(mag.b_azim(row,col)*!dtor)
          endelse

          if VERBOSE and (loop MOD 10 EQ 0) then begin
             if DoDivB then ctype = "  Divergence" $
             else ctype = "  Current"
             print,loop,changes(iteration-1),pixel,ctype,format="(i5,i8,i7,a12)"
          endif

       endfor   ; end of pixels loop

       print,'Iteration',iteration,' had ',changes(iteration-1),' flips'

       flip_test = where(places EQ old_places,flip_test_count)
       nplaces = n_elements(places)
       nold_places = n_elements(old_places)
       if VERBOSE and nplaces GT 1 then begin  ; Print which pixels are flipped
          ;print
          ;print,'Pixels flipped:'
          row = places(1:nplaces-1) mod nx 
          col = long(places(1:nplaces-1)/nx)
          rsort = sort(row)
          output = intarr(2*(nplaces-1))
          output(lindgen(nplaces-1)*2)=row(rsort)
          output(lindgen(nplaces-1)*2+1)=col(rsort)
          ;print,output,format='(2i5)'
          ;print
       endif
       if changes(iteration-1) EQ 0 then begin
          print,'Convergence at iteration',iteration
          NITER = iteration
          goto,DONE_MIN
       endif $
       else if ((flip_test_count EQ nplaces) AND (nplaces EQ nold_places)) OR $
            (min(flip_count(places(1:nplaces-1))) GT 2) then begin
          ; If the same points are repeatedly flipped, or if all the flipped
          ; points have been flipped more than twice, terminate the iteration
          ; and print the points which give the conflict.
          ; The first element of places is always 0.
          print,'REGION of CONFLICT:  Iteration terminated'
          conflict_row = places(1:nplaces-1) mod nx 
          conflict_col = long(places(1:nplaces-1)/nx)
          rsort = sort(conflict_row)
          output = intarr(2*(nplaces-1))
          output(lindgen(nplaces-1)*2)=conflict_row(rsort)
          output(lindgen(nplaces-1)*2+1)=conflict_col(rsort)
          ;print,output,format='(2i5)'
          conflict = intarr(n_elements(conflict_row),2)
          conflict(*,0) = conflict_row(*)
          conflict(*,1) = conflict_col(*)
          NITER = iteration
          goto,DONE_MIN
       endif

    endfor   ; end of iteration loop


DONE_MIN:   print & print,'done' & print

    for iteration = 1,NITER do begin
      print,'Iteration',iteration,' had',changes(iteration-1),' flips'
    endfor

  endif    ; End of minimization code

  endif $ ; End of truea test
  else begin  ; Use truea to resolve the ambiguity
     if n_elements(truea(*,0)) NE nx OR $
        n_elements(truea(0,*)) NE ny  then begin
        print,'ERROR: Dimensions of truea are incompatable'
        stop
     endif
  ;  Fix the azimuthal angle
     fixa= where(abs(arange(mag.B_AZIM(*),truea(*))-truea(*)) GT $
                 90.,count)
     if (count gt 0) then mag.B_AZIM(fixa) = mag.B_AZIM(fixa) - 180.
     mag.B_AZIM = arange(mag.B_AZIM,0.)
     if VERBOSE then begin
        print
        print,'Resolved azimuth with truea field:',count,' flips'
     endif

  endelse

  bx = -mag.b_trans * sin(mag.b_azim*!dtor)
  by = mag.b_trans * cos(mag.b_azim*!dtor)
  bz = mag.b_long
  disktran, bx,by,bz,b0,p,pos_cmd,pos_lat,bdx,bdy,bdz,pxd,pyd,/noplot,pix_size=pix_size
  jzhigh,bdx,bdy,bdz,pxd,pyd,b0,p,pos_cmd,pos_lat,jz,jpxd,jpyd,pix_size, $
         /noask,/noplot,jz_window=0,scale=scale

  point = {jzpoint,b0:b0,p:p,cmd:pos_cmd,lat:pos_lat,pix_size:pix_size}

;  if tag_index(mag,'POINT',/quiet) LT 0 then $
;     mag = add2str(mag, $
;                ['BX','BY','BZ','B_PXD','B_PYD','JZ','J_PXD','J_PYD','POINT'], $
;                  bdx, bdy, bdz, pxd,    pyd,    jz,  jpxd,   jpyd,  point) $
;  else $
;     if tag_index(mag,'B_PXD',/quiet) LT 0 then $
;       mag = add2str(mag, $
;                   ['BX','BY','BZ','B_PXD','B_PYD','JZ','J_PXD','J_PYD'], $
;                     bdx, bdy, bdz, pxd,    pyd,    jz,  jpxd,   jpyd)

  ; Store the current etc in the structure.  Check whether the tag already
  ; exists in the structure.  If so, replace the variable in the structure
  ; with the variable computed above.  If not, add the variable to the 
  ; structure.

  add_names = [ 'BX', 'BY', 'BZ','B_PXD','B_PYD','JZ','J_PXD','J_PYD','POINT'] 
  add_vars =  ['bdx','bdy','bdz',  'pxd',  'pyd','jz', 'jpxd', 'jpyd','point'] 
  add_index = tag_index(mag,add_names,/quiet)
  for add=0,n_elements(add_names)-1 do begin
     tag = add_names(add)
     var = add_vars(add)
     check = execute(strcompress('value='+var,/remove_all))
     if check then begin
        if add_index(add) GE 0 then begin  ; Tag already exists
           vsize = size(value)
           ; Structure must be treated seperately
           if vsize(n_elements(vsize)-2) NE 8 then begin
              check=execute(strcompress('mag.'+tag+'=value',/remove_all))
              if NOT check then $
                 message,'Could not fill '+tag+' from '+var,/info
           endif $
           else begin
              for i=0,n_tags(value)-1 do begin
                 check=execute(strcompress('mag.'+tag+'.('+string(i)+ $
                               ') = value.('+string(i)+')',/remove_all))
                 if NOT check then $
                    message,'Could not fill '+tag+' from '+var,/info
              endfor 
           endelse
        endif $
        else begin  ; tag needs to be created 
           check = execute(strcompress('value='+var,/remove_all))
           if check then mag = add2str(mag,tag,value)
        endelse
     endif $
     else message,'Variable "'+var+'" not found',/info
  endfor

  ; Run the magnifier and allow the user to flip vectors by hand

  if (NOT keyword_set(quiet)) and (NOT keyword_set(noask)) then $
     jzmagnify,mag,jz_window=0,conflict=conflict,title=title

;  print,'Do you want to see the force-free vector current? (y/n)'
;  answer='n'
;  read,answer
;  answer = byte(strcompress(answer,/remove_all))
;  if answer(0) EQ max(byte('y')) OR answer(0) EQ max(byte('Y')) then begin
;     alpha = (mag.jz/float(mag.bz) + mag.jz/shift(float(mag.bz),-1,-1) )/2.0
;     bad = where(mag.bz LT 50.,count)  ; alpha uncertain at these pixels
;     if count GT 0 then alpha(bad)=0.
;     bhelio_current,((alpha*mag.bx)<20.)>(-20.),((alpha*mag.by)<20.)>(-20.),mag.jz, $
;            mag.b_pxd,mag.b_pyd,b_window=0,scale=scale,/current, $
;            length=max(sqrt((((alpha*mag.bx)<20.)>(-20.))^2+ $
;                            (((alpha*mag.by)<20.)>(-20.))^2))/10.0
;  endif

  if NOT keyword_set(quiet) then begin
     print,'Do you want to compute the magnitude of the force-free current? (y/n)'
     answer='n'
     if keyword_set(noask) then answer='y' else read,answer
     answer = byte(strcompress(answer,/remove_all))
  endif else answer=byte('y')
;
	if keyword_set(noask) then mag=jxyz(mag,/quiet)
;AAP
  if answer(0) EQ max(byte('y')) OR answer(0) EQ max(byte('Y')) then $
     mag = jtotal(mag)

  if nhead GT 0 then begin
     junk = wkey(head,'HISTORY','COMMENT',hcomment,1,"","END")
     junk = wkey(head,'HISTORY','COMMENT','Force-free solution unrestricted (no finite energy)',1,"","END")
  endif

  return,mag

end
