
function jxyz,j,quiet=quiet,verbose=verbose,integral=integral

;+
;NAME:
;     JXYZ
;PURPOSE:
;     Computes the vector electric current using an observed magnetogram
;     for Jz and the force free current for Jx and Jy
;CATEGORY:
;CALLING SEQUENCE:
;     j = jxyz(b)
;INPUTS:
;     b = A magnetic field structure which has the 180 degree ambiguity
;         resolved and includes Bx, By and Bz.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;    /integral = Use the integral form of Ampere's law to derive the
;                vertical current. Very slow, but possibly less sensitive to
;                noise. 
;    /quiet = Keep quiet
;    /verbose = Do lots of plots
;OUTPUTS:
;    j = A new magnetic field structure which contains tags for jx, jy and
;        the value of alpha used in the force-free equation.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;    The input magnetic field structure mus have the 180 degree ambiguity
;    resolved, and it must contain the vertical and horizontal magnetic
;    field.
;PROCEDURE:
;    Computes the vector electric current from the curl of the magnetic
;    field.  The force-free equations are used to determine the horizontal
;    current density.
;
;      Jz = dBy/dx - dBx/dy
;      alpha = Jz/Bz
;      Jx = alpha*Bx
;      Jy = alpha*By
;
;      alpha is noisy and is taken as Jz/Bz only in regions where
;      both Jz and Bz are significant.  In other areas, the computed alpha
;      is smoothed with a median filter before currents are computed.
;MODIFICATION HISTORY:
;    Written by T. Metcalf 31-Mar-94
;    2004-Dec-03  Added /integral keyword.
;    2005-Jan-13  Use get_berrs instead of bterror
;-

   nx = n_elements(j.bx(*,0))
   ny = n_elements(j.bx(0,*))

   ; Compute relations between image and helio derivatives

   ; These c's convert helio to image

   b0 = double(j.point.b0)
   p = double(j.point.p)
   if tag_index(j,'CMD',/quiet) GE 0 then lc = double(j.cmd)*!dtor $
   else lc = double(j.point.cmd)
   if tag_index(j,'LATITUDE',/quiet) GE 0 then bc = double(j.latitude)*!dtor $
   else bc = 	double(j.point.lat)

   cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
   cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
         - COS(bc)*COS(b0)*SIN(p) 
   cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
   cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
         + COS(bc)*COS(b0)*COS(p)

   dxidX = cxx   ; At pixel corners, not centers
   dyidX = cyx
   dxidY = cxy
   dyidY = cyy

   determ = cxx*cyy-cyx*cxy  ; Get image to helio transformation
   icxx = cyy/determ
   icxy = -cxy/determ
   icyx = -cyx/determ
   icyy = cxx/determ

   dXdxi = icxx
   dYdxi = icyx
   dXdyi = icxy
   dYdyi = icyy

   pix_size = double(j.point.pix_size)
   if n_elements(pix_size) EQ 1 then $
      pix_size = [pix_size(0),pix_size(0)]
   mpix_size = min(pix_size)

   x = (lindgen(nx,ny) MOD nx)*(pix_size(0)/mpix_size); image pixel coordinates
   y = (lindgen(nx,ny) / nx)*(pix_size(1)/mpix_size) 
   bpxd = dXdxi*x + dXdyi*y           ; Heliographic pixel coordinates
   bpyd = dYdxi*x + dYdyi*y
   jpxd = dXdxi*(x+0.5) + dXdyi*(y+0.5)
   jpyd = dYdxi*(x+0.5) + dYdyi*(y+0.5)


   if keyword_set(integral) then begin
      jz = fltarr(nx,ny)
      jzerror = fltarr(nx,ny)
      if NOT keyword_set(quiet) then progress,0.0,/reset,label='%Complete'
      nintegrals = long(nx-1)*long(ny-1)
      for ii=0L,nx-2L do begin
         for jj=0L,ny-2L do begin
            ss = [ii,ii,ii+1,ii+1] + [jj,jj+1,jj+1,jj]*nx
            except = !except
            !except=0
            jz[ii,jj] = ampere(j,ss=ss,area=area,error=jzerr) ; compute current for this pixel
            check = check_math(mask=32) ; Ignore underflows 
            !except = except
            jz[ii,jj] = jz[ii,jj]*1000/area ; Amps -> mA/m^2
            jzerror[ii,jj] = jzerr*1000/area
            if NOT keyword_set(quiet) then $
               progress,(100.0*(ii*(ny-1)+jj+1))/nintegrals,frequency=5.0
         endfor
      endfor
      if NOT keyword_set(quiet) then progress,100.0,/last
      if NOT keyword_set(quiet) then begin
         print
         print,'Jz error: ',mean(jzerror),median(jzerror)
      endif
   endif else begin 
      ; Compute the derivatives required for the curl

      lylxbx=shift(j.bx,-1,-1)
      lylxby=shift(j.by,-1,-1)
      lxbx=shift(j.bx,-1,0)
      lxby=shift(j.by,-1,0)
      lybx=shift(j.bx,0,-1)
      lyby=shift(j.by,0,-1)

      dBxdx = 0.5*(lylxbx+lxbx-lybx-j.bx)
      dBxdy = 0.5*(lylxbx+lybx-lxbx-j.bx)

      dBydx = 0.5*(lylxby+lxby-lyby-j.by)
      dBydy = 0.5*(lylxby+lyby-lxby-j.by)
 
      ; The constant 0.0193 is derived as follows:
      ; 10^3 * 10^(-4) / (mu * pxlength)
      ; where mu = 4*pi*10^(-7)
      ; pxlength = length of each pixel in meters 
      ;            (5.7 arcsec corresponds to 4.13e6 m)
      ; 10^3 is the conversion between mAmp and Amp
      ; 10^(-4) is the conversion between Gauss and Tesla
      ; j is thus in units of mAmp/m^2

      units = 1.0/(4.*!pi*1.0e-3*721.9*mpix_size)  ; ~ 0.0193*5.656/pixsize

      jz = (dxidX*dBydx + dyidX*dBydy - dyidY*dBxdy - dxidY*dBxdx)*units

      ; Set the edges to zero since the above shifts will give garbage at edges

      jz(nx-1,*)=0 & jz(*,ny-1)=0
   endelse

   ; Compute error in the magnetic field

   errs = get_berrs(j)
   bsigma = errs.sig_btrans
   ;bsigma = bterror(j)
   if keyword_set(verbose) then $
      message,/info,strcompress('B sigma = '+string(long(bsigma)))+' G'

   ; Force-Free alpha

   jzalpha = fltarr(nx,ny)
   bzcenter = bilinxy(j.bz,jpxd,jpyd,xarr=bpxd,yarr=bpyd)  ; at current posit.

   goodbz=where(abs(bzcenter) NE 0.,cgoodbz)
   if cgoodbz GT 0 then jzalpha(goodbz) = jz(goodbz)/bzcenter(goodbz)
   jzalpha = median(jzalpha,3)   ; Apply median filter for noise reduction

   badbt = where(j.b_trans LE bsigma)
   if n_elements(badbt) LE 1 then jzerror = 0. $
   else jzerror = stdev(abs(jz(badbt)))
   goodbz=where(abs(bzcenter) GT bsigma and abs(jz) GT jzerror,cgoodbz)
   if cgoodbz GT 0 then jzalpha(goodbz) = jz(goodbz)/bzcenter(goodbz)

   jx = jzalpha*j.bx
   jy = jzalpha*j.by
   jtotal = sqrt(jx^2 + jy^2 + jz^2)*sign(jz)

   jout = j

   if tag_index(j,'JZ',/quiet) GE 0 then jout.jz = jz $
   else jout = add2str(jout,'JZ',jz,/quiet)
   if tag_index(j,'J_TOTAL',/quiet) GE 0 then jout.j_total = jtotal $
   else jout = add2str(jout,'J_TOTAL',jtotal,/quiet)
   if tag_index(j,'JX',/quiet) GE 0 then jout.jx = jx $
   else jout = add2str(jout,'JX',jx,/quiet)
   if tag_index(j,'JY',/quiet) GE 0 then jout.jy = jy $
   else jout = add2str(jout,'JY',jy,/quiet)
   if tag_index(j,'JXY_ALPHA',/quiet) GE 0 then begin
      if n_elements(j.JXY_ALPHA) EQ n_elements(jzalpha) then $
         jout.jxy_alpha = jzalpha $
      else $
         jout = add2str(jout,'JXY_ALPHA',jzalpha,/quiet,delete=['JXY_ALPHA'])
   endif else jout = add2str(jout,'JXY_ALPHA',jzalpha,/quiet)
   if tag_index(j,'B_PXD',/quiet) GE 0 then jout.b_pxd=bpxd $
   else jout = add2str(jout,'B_PXD',bpxd,/quiet)
   if tag_index(j,'B_PYD',/quiet) GE 0 then jout.b_pyd=bpyd $
   else jout = add2str(jout,'B_PYD',bpyd,/quiet)
   if tag_index(j,'J_PXD',/quiet) GE 0 then jout.j_pxd=jpxd $
   else jout = add2str(jout,'J_PXD',jpxd,/quiet)
   if tag_index(j,'J_PYD',/quiet) GE 0 then jout.j_pyd=jpyd $
   else jout = add2str(jout,'J_PYD',jpyd,/quiet)

   if NOT keyword_set(quiet) then begin
      display = add2str(dummy,['BX','BY','BZ','POINT'],jx,jy,jz,j.point,/quiet)
      error = jzerr(jout)
      vmgram,display,/incolor,image=j.i_cont,/xyz, $
              levels=(findgen(10)+2.0)*error,/arrow,length=2., $
              title='Vector Electric Current',/quiet
   endif

   return,jout

end 
