pro jzmagnify,mag,jz_window=jz_window,set_scale=set_scale,conflict=conflict, $
              title=title


;+
;NAME:
;	jzmagnify
;PURPOSE:
;	Plots a Jz map and a zoomed map.  Allows flipping vectors "by hand".
;CATEGORY:
;CALLING SEQUENCE:
;	jzmagnify,mag_structure
;INPUTS:
;	mag_structure = a structure containing B and Jz info.  Written by
;	                the current function.
;OPTIONAL INPUT PARAMETERS:
;	conflict = specifies row,col of regions of conflict.  These are
;                  drawn ar red blobs until corrected.
;KEYWORD PARAMETERS
;	jz_window = specifies the IDL window in which to plot Jz (unzoomed).
;	set_scale = Set the scale of the zommed image: gives the number of
;	            pixels along half of each dimension (2*set_scale+1) pixels
;	            in each dimension).
;       title = title for plot
;OUTPUTS:
;	The mag_structure is updated to reflect any changes made in the 
;	orientation of the vectors.
;COMMON BLOCKS:
;	None
;SIDE EFFECTS:
;	Messes with the color table
;RESTRICTIONS:
;	None
;PROCEDURE:
;MODIFICATION HISTORY:
;	T. Metcalf  June 1991
;-

   SCALE = 5   ; Number of pixels in zoomed image is 2*SCALE+1
               ; Hence, smaller SCALE means more zoom

   if n_elements(set_scale) GT 0 then SCALE = set_scale

   if n_elements(jz_window) EQ 0 then jz_window = 0

   if tag_index(mag,'BX') LT 0  OR $
      tag_index(mag,'BY') LT 0  OR $
      tag_index(mag,'BZ') LT 0  OR $
      tag_index(mag,'B_PXD') LT 0  OR $
      tag_index(mag,'B_PYD') LT 0  OR $
      tag_index(mag,'JZ') LT 0  OR $
      tag_index(mag,'J_PXD') LT 0  OR $
      tag_index(mag,'J_PYD') LT 0  OR $
      tag_index(mag,'POINT') LT 0 then begin
      print,'ERROR: jzmagnify: structure does not contain enough information'
      return
   endif $
   else begin
      if tag_index(mag.point,'B0') LT 0  OR $
         tag_index(mag.point,'P') LT 0  OR $
         tag_index(mag.point,'CMD') LT 0  OR $
         tag_index(mag.point,'LAT') LT 0  OR $
         tag_index(mag.point,'PIX_SIZE') LT 0  then begin
         print,'ERROR: jzmagnify: structure does not contain enough information'
         return
      endif
   endelse

   bdx = mag.BX
   bdy = mag.BY
   bdz = mag.BZ
   pxd = mag.B_PXD-min(mag.B_PXD)
   pyd = mag.B_PYD-min(mag.B_PYD)
   jz = mag.JZ
   jpxd = mag.J_PXD-min(mag.B_PXD)
   jpyd = mag.J_PYD-min(mag.B_PYD)

   nx = n_elements(bdx(*,0))
   ny = n_elements(bdx(0,*))

   if n_elements(mag.point.pix_size) EQ 1 then $
      pix_size = [double(mag.point.pix_size(0)),double(mag.point.pix_size(0))] $
   else $
      pix_size = double(mag.point.pix_size)
   mpix_size = min(pix_size)

   wset,jz_window 
   jzhigh,bdx,bdy,bdz,pxd,pyd, mag.point.b0,mag.point.p,mag.point.cmd, $
          mag.point.lat,jz,jpxd,jpyd,pix_size,jz_window=jz_window, $
          title=title,/noask
   plot_conflicts,conflict,pxd,pyd

   window,/free,title='Zoomed Image: Window Active When Filled'
   zoom_window = !D.WINDOW

   wset,jz_window
   save_plot = !P
   save_x = !X
   save_y = !Y
  
   print,'Click left button on spot to be magnified'
   print,'Click middle button to flip a vector'
   print,'Click right button to exit magnifier'

   bxi = -mag.b_trans * sin(mag.b_azim*!dtor)
   byi = mag.b_trans * cos(mag.b_azim*!dtor)
   bzi = mag.b_long

   tvcrs,1            ; enable cursor
   done = 0           ; a flag to end the loop when finished

   REPEAT begin
      cursor,x,y,/down,/device             ; Get place mouse clicked at
      case !D.WINDOW of 
         jz_window: begin
                       !P = save_plot    ; Reset the plotting parameters
                       !X = save_x
                       !Y = save_y
                    end
         zoom_window: begin
                        !P = save_zplot    ; Reset the plotting parameters
                        !X = save_zx
                        !Y = save_zy
                      end
      endcase
      spot = convert_coord(x,y,/device,/to_data)   ; Put in data coords
      button_pushed = !ERR

      distance = (pxd-spot(0))^2 + (pyd-spot(1))^2
      closest_pixel = sort(distance)
      closest_pixel = closest_pixel(0)
      px = closest_pixel mod nx          ; px,py are the row,col of the pixel
      py = fix(closest_pixel/nx)         ; nearest to the place clicked

      case button_pushed of
         1: begin    ; Left button = magnify
               wset,zoom_window
               xlow = max([0,px-SCALE])
               xhi = min([nx-1,px+SCALE])
               ylow = max([0,py-SCALE])
               yhi = min([ny-1,py+SCALE])
               jzhigh,bdx(xlow:xhi,ylow:yhi), $  ; Compute and plot zoomed Jz
                      bdy(xlow:xhi,ylow:yhi), $
                      bdz(xlow:xhi,ylow:yhi), $
                      pxd(xlow:xhi,ylow:yhi), $
                      pyd(xlow:xhi,ylow:yhi), $
                      mag.point.b0,mag.point.p,mag.point.cmd,mag.point.lat, $
                      tjz,jpxd(xlow:xhi,ylow:yhi),jpyd(xlow:xhi,ylow:yhi), $
                      pix_size,jz_window=zoom_window,$
                      /noask, $
                      nx_big = nx, $
                      ny_big = ny, $
                      xoffset = xlow + (xhi-xlow+1)/2. - nx/2., $
                      yoffset = ylow + (yhi-ylow+1)/2. - ny/2.
               save_zplot = !P
               save_zx = !X
               save_zy = !Y
            end
         2: begin    ; Middle button = flip vector
               mag.b_azim(px,py) = arange(mag.b_azim(px,py) + 180.,0.)
               bxi(px,py) = -mag.b_trans(px,py) * sin(mag.b_azim(px,py)*!dtor)
               byi(px,py) = mag.b_trans(px,py) * cos(mag.b_azim(px,py)*!dtor)
               ;;disktran, bx,by,bz,mag.point.b0,mag.point.p,mag.point.cmd, $
               ;;          mag.point.lat,bdx,bdy,bdz, $
               ;;          junkpxd,junkpyd,/noplot
               bxyz,bxi,byi,bzi,bdx,bdy,bdz,lat=mag.point.lat,cmd=mag.point.cmd, $
                    p=mag.point.p,b0=mag.point.b0
               if !D.WINDOW EQ jz_window then begin
                  jzhigh,bdx,bdy,bdz,pxd,pyd, $
                         mag.point.b0,mag.point.p,mag.point.cmd,mag.point.lat, $
                         jz,jpxd,jpyd,pix_size,jz_window=jz_window, $
                         title=title,/noask
               endif else begin
                  jzhigh,bdx,bdy,bdz,pxd,pyd, $
                         mag.point.b0,mag.point.p,mag.point.cmd,mag.point.lat, $
                         jz,jpxd,jpyd,pix_size,jz_window=jz_window, $
                         /noask,/noplot
                  jzhigh,bdx(xlow:xhi,ylow:yhi), $
                         bdy(xlow:xhi,ylow:yhi), $
                         bdz(xlow:xhi,ylow:yhi), $
                         pxd(xlow:xhi,ylow:yhi), $
                         pyd(xlow:xhi,ylow:yhi), $
                         mag.point.b0,mag.point.p,mag.point.cmd,mag.point.lat, $
                         tjz,jpxd(xlow:xhi,ylow:yhi),jpyd(xlow:xhi,ylow:yhi), $
                         pix_size, $
                         jz_window=zoom_window,/noask,scale=2, $
                         nx_big = nx, $
                         ny_big = ny, $
                         xoffset = xlow + (xhi-xlow+1)/2. - nx/2., $
                         yoffset = ylow + (yhi-ylow+1)/2. - ny/2.
               endelse
            end
         4: begin    ; Right button = exit window
               if !D.WINDOW EQ zoom_window then begin
                  erase
                  wset,jz_window 
                  jzhigh,bdx,bdy,bdz,pxd,pyd, $
                         mag.point.b0,mag.point.p,mag.point.cmd,mag.point.lat, $
                         jz,jpxd,jpyd,pix_size,jz_window=jz_window, $
                         title=title,/noask
               endif $
               else begin
                  done = 1
                  print,'Exiting ...'
               endelse
            end
         else: print,'Undefined result from cursor',button_pushed
      endcase
   endrep UNTIL done

   wdelete,zoom_window    ; Clean up 

; Save any changes which have been made

   mag.BX = bdx
   mag.BY = bdy
   mag.BZ = bdz
   mag.JZ = jz
   mag = jxyz(mag,/quiet)

   return

end
