
pro bxyz,Bxi,Byi,Bzi,Bxh,Byh,Bzh,latitude=bc,cmd=lc,p=p,b0=b0, $
         helio2image=helio2image,image2helio=image2helio, $
         last_transform=last_transform

;+
;NAME:
;     BXYZ
;PURPOSE:
;     Transforms vectors between image and heliographic coordinates.
;CATEGORY:
;CALLING SEQUENCE:
;     bxyz,Bxi,Byi,Bzi,Bxh,Byh,Bzh,latitude=latitude,cmd=cmd,p=p,b0=b0
;INPUT/OUTPUT:
;     Bxi = x component of the vector in image coordinates (scalar or array)
;     Byi = y component of the vector in image coordinates (scalar or array)
;     Bzi = y component of the vector in image coordinates (scalar or array) 
;     Bxh = x component of the vector in heliographic coordinates 
;           (scalar or array)
;     Byh = y component of the vector in heliographic coordinates  
;           (scalar or array)
;     Bzh = y component of the vector in heliographic coordinates  
;           (scalar or array) 
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     latitude = latitude.  If scalar, the latitude of the center of the FOV.
;                If array, must be the same size as the B arrays. In RADIANS.
;     cmd = cmd.  If scalar, the latitude of the center of the FOV.
;           If array, must be the same size as the B arrays.  In RADIANS.
;     p  = solar p-angle.  In RADIANS.
;     b0 = solar b-angle  in RADIANS.
;     /helio2image = if set convert heliographic vector to image vector.  In
;                    this case, B*i is output and B*h is input.
;     /last_transform = Use the last transformation ... do not compute a
;                       new transformation.  latitude,cmd,p,b0 are not
;                       required in this case.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The /last_transform keyword uses a common block to store the last
;     transformation.  This has the potential to introduce subtle bugs into
;     programs since it may not always be obvious what the last transform
;     was.
;PROCEDURE:
;     If latitude and cmd are passed as arrays, the transformation is done
;     independently for each point so that solar curvature is accounted for.
;MODIFICATION HISTORY:
;     T. Metcalf  April, 1994
;-

   common bxyz_private,axx,axy,axz,ayx,ayy,ayz,azx,azy,azz

   if NOT keyword_set(last_transform) then begin

      ; These are the transformation matrix elements given by eq. (1)
      ; of Gary and Hagyard, Solar Phys. 126, 21 (1990).
      ; The components of the magnetic field in heliographic coordinates (H)
      ; are given in terms of the image-plane components (I) by BH = A BI,
      ; where the B's are column vectors and A = [axx axy axz]
      ;                                          [ayx ayy ayz]
      ;                                          [azz azy azz]
      ; Note that A is orthogonal, so Inverse(A) = Transpose(A).
      ; Also note last column of A gives direction cosines of LOS in 
      ; helio-coords.

      axx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
      axy = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
      axz = - COS(b0)*SIN(lc)
      ayx = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
            - COS(bc)*COS(b0)*SIN(p)
      ayy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
            + COS(bc)*COS(b0)*COS(p)
      ayz = - COS(b0)*SIN(bc)*COS(lc) + SIN(b0)*COS(bc)
      azx = + COS(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
            - SIN(bc)*COS(b0)*SIN(p)
      azy = - COS(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
            + SIN(bc)*COS(b0)*COS(p)
      azz = + COS(bc)*COS(b0)*COS(lc) + SIN(bc)*SIN(b0)

   endif

   if NOT keyword_set(helio2image) then begin
      Bxh = axx*Bxi + axy*Byi + axz*Bzi   ; Image --> Heliographic
      Byh = ayx*Bxi + ayy*Byi + ayz*Bzi
      Bzh = azx*Bxi + azy*Byi + azz*Bzi
   endif else begin
      Bxi = axx*Bxh + ayx*Byh + azx*Bzh   ; Heliographic --> Image
      Byi = axy*Bxh + ayy*Byh + azy*Bzh
      Bzi = axz*Bxh + ayz*Byh + azz*Bzh
   endelse

end
