
function current2,bin,tfactor=tfactor,converge=converge,kb=kb,start=start, $
                  done=done,alpha=alpha, $
                  trecord=trecord,erecord=erecord,crecord=crecord, $
                  bterror=bterror,quiet=quiet,verbose=verbose,plot=plot, $
                  get_alpha=get_alphain,nopotential=nopotential, $
                  novectors=novectors,jweight=jweight,nogenetic=nogenetic, $
                  jtotal=jtotal,notty=notty,fast=fast,slow=slow, $
                  skippotential=skippotential,guard=guard,dweight=dweight, $
                  eweight=eweight,btlength=btlength,binfactor=binfactor, $
                  mask=maskin,shear=shear,azshearlimit=shearlimit, $
                  nofinalplot=nofinalplot,sweight=sweight,hweight=hweight, $
                  use_pikaia=use_pikaia,ngenerations=ngenerations, $
                  pweight=pweight,gweight=gweight,jpower=jpower,enclose=enclose, $
                  dBzdz=dBzdzin,divbin=divbin

;+
;NAME:
;     CURRENT2
;PURPOSE:
;     Resolves the 180 degree ambiguity in vector magneotgrams and computes
;     the vertical and force-free horizontal currents.
;     Metcalf, T. R., Solar Physics, 155, 235, 1994.
;CATEGORY:
;CALLING SEQUENCE:
;     j = current2(b)
;
;     Type "qq" to stop the iteration gracefully.
;INPUTS:
;     b = magnetic field structure from bfits.pro
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     alpha = linear force-free alpha for the d/dz calculations (def = 0.0).
;     /get_alpha = Compute a "best" value for alpha from a least squares fit 
;                  of the force-free transverse field to the observed 
;                  transverse field.  You should almost always set
;                  this.
;     /shear = Use the field weighted shear angle in get_alpha.  Not
;              recommended. 
;     dBzdz = the vertical derivative of the vertical field.  If you
;             know this somewhow, passing it in avoids the approximate
;             computation of dBzdz from the force-free extrapolation.
;             the units should be Gauss per pixel.  This must be an
;             array with the same dimension as all the other field
;             arrays. 
;     /nopotential = Do not use the force-free field for a crude initial
;                    resolution of the ambiguity.  Don't use this unless
;                    the input is already close to the correct resolution of
;                    the 180 degree ambiguity.  Does not do much as
;                    the initial state of the ambiguity is not very
;                    important. 
;     skippotential = Pixel with azimuth this close to potential are not
;                     included in the iteration to save time.  Degrees.
;                     default = 0.0.   Only applies to pixels with B_trans
;                     larger than 3*bterror.
;              CAUTION:  Don't set this too high:  skipping points can confuse
;                        the algorithm.
;     binfactor = Gram is reduced in size by this amount for speed and
;                 to provide some smoothing of the result.  The result
;                 is binned back up before returning.  Set this to
;                 2 or 3 to speed things up.  Float GE 1.
;                 binfactor = fix(5.7/mean(b.point.pix_size)) works
;                 well and smooths the gram to 5 or 6 arcsec.
;     azshearlimit = if binfactor is GT 1, then the final gram is
;                    produced by comparing the azimuth of the gram
;                    binned back to full resolution with the input
;                    gram.  If the shear angle bewteen the two grams is
;                    greater than azshearlimit (in degrees), then current2
;                    is called recursively with binfactor ->
;                    binfactor/divbin to fix those vectors that 
;                    have a shear angle larger than this value.  This
;                    should be pretty quick if there are not too many
;                    suspect vectors.  Set azshearlimit to 90.0 (or larger)
;                    to turn this feature off.  Default=90 (off).
;                    10.0 would be a good value if you want to turn
;                    this on.
;     divbin = divisor for the binfactor when current2 is called
;              recursively when the azshearlimit condition is met.
;              Default = 2.0.
;     bterror = Pixels with transverse field (absolute value) less than bterror
;              will be ignored in the iteration, saving time (def =
;              0.0). If /nopotential is set, the ignored pixels will
;              be left alone, otherwise their azimuth is set to the FFF. 
;              CAUTION:  Don't set this too high:  skipping points can confuse
;                        the algorithm.  An excellent compromise is to
;                        set this to bterror(b,bins=1.).
;     mask = Array with a 1 to indicate that the azimuth for that
;            pixel is to be tested or a 0 to indicate that it is
;            already set and should not be tested.  If /nopotential is
;            set, then untested points will use the input azimuth.  If
;            /nopotential is not set, then the azimuth of the untested
;            points will be set to the FFF azimuth.
;     jweight = weighting of the current minimization relative to the 
;               divergence minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def
;               = 1.0.
;     jpower = the 'energy' in the simulated annealing is raised to
;              this power.  For example, jpower=2 means use E^2.
;              Larger jpower tends to reduce large, isolated currents.
;              Default = 2.0. 
;     sweight = weighting of the sheath current minimization defined
;               by Georgoulis, LaBonte & Metcalf.  May be scalar or an
;               array with the same dimension as the magnetic field
;               arrays.  Def = 0.0.
;     dweight = weighting of the divergence minimization relative to the 
;               current minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def
;               = 1.0.
;     eweight = weighting of the entropy maximization relative to the current
;               and divergence minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def = 0.0.
;               +1 = maximum entropy, 0.0 = turns off, -1 = minimum
;               entropy
;     hweight = weighting of the current helicity (j*B) minimization
;               relative to the current and divergence
;               minimization. May be scalar or an array with the same
;               dimension as the magnetic field arrays.  Def = 0.0. 
;     pweight = Weight of small term to tip the solution towards the
;               potential or LFFF when it is hard to decide.  Def =
;               0.0
;     gweight = Weight of |J|/grad(|B|) term. |B| is unabmiguous, so this
;               tries to put J where the gradient of |B| is large. Def=0.0
;     /jtotal = Minimize Jtotal rather than the default Jz.  Use
;               this when the magnetogram is far from disk center.
;               Jx and Jy are only partially computed since the
;               vertical deivatives are not known.
;     /nogenetic = Do not use genetic annealing.  Faster, but less robust.
;     converge = convergence criteria in percent (def = 5.0).
;     kb = One over Boltzmann constant for the annealing iteration.  Larger
;          numbers start the iteration at a higher "temperature" (def = 5.0). 
;     tfactor = maximum factor to decrease the "temperature" by at each 
;               iteration.  The actual factor at each step is derived from 
;               tfactor and from the rate of convergence (def = 0.95).
;     crecord = vector of "specific heat" at each "temperature" (output)
;     erecord = vector of "energy" at each "temperature" (output)
;     trecord = vector of "temperatures" (output)
;     /fast = Run the routine as fast as possible; possibly less robust.
;             Uses kb = 3.0,
;                  tfactor = 0.90,
;                  /nogenetic
;                  binfactor=3
;     /slow = Run slowly to do the best possible resolution of the ambiguity.
;             Uses kb = 10.0
;                  tfactor = 0.98
;                  converge = 2.
;                  guard = 2
;     /quiet = work quietly.
;     /verbose = work loudly.
;     /plot = plot the field after every temperature.
;     /novectors = do not plot the horizontal vectors.
;     btlength = length of horizontal field vectors (def = 1)
;     /notty = run non-interactively
;     guard = size of fff.pro guard band (def = 1)
;OUTPUTS:
;     j = magnetic field structure with 180 degree ambiguity resolved and
;         currents and vertical/horizontal magnetic field computed.
;COMMON BLOCKS:
;SIDE EFFECTS:
;EXAMPLES:
;
;     b = bfits('B911005.1702')
;     j = current2(b,bterror=bterror(b,bins=1.),/plot)
;     wbfits,j,'J911005.1702',head=head
;
;     The following works well.  The bin factor provides spatial
;     smoothing to 5 or 6 arcsec or so.   Use less smoothing for grams
;     with smaller spots, etc.
;
;     j = current2(b,bterror=bterror(b,bins=1.),/plot,/get_alpha, $
;                  bin=fix(5.7/mean(b.point.pix_size)), $
;                  azshearlimit=10.,/slow)
;
;RESTRICTIONS:
;     Requires a large FOV encompassing all the magnetic flux in the region
;     since the derivative of Bz wrt to z is computed from a linear force
;     free or potential extrapolation.
;
;PROCEDURE:
;     Uses simulated annealing to simultaneously minimize the divergence of 
;     the magnetic field and the current density.  The vertical derivatives 
;     are computed with a linear force free extrapolation when alpha is 
;     nonzero, or from a potential extrapolation when alpha is undefined or 
;     zero.  To get the initial extrapolation correct, a potential field 
;     computed from the LOS field is first used for a crude resoltion of
;     the 180 degree ambiguity.  In some cases this may not be good enough
;     and the final solution will suffer.  See examples above.
;
;     For the simulated annealing, see Numerical Recipes (2nd ed.) and 
;     Kirkpatrick, Gelatt, and Vecchi, Science, 220, 671, 1983.
;
;     The input magnetic field structure is assumed to have at least the
;     following fields:
;
;     B_LONG:  Line of sight magnetic field (2-D Array)
;     B_TRANS: Transverse magnetic field (2-D Array)
;     B_AZIM:  Azimuth of the magnetic field in degrees CCW from terrestrial 
;              north (2-D Array)
;     POINT:   Sub-structure with the following pointing information:
;              B0:  Solar B angle in radians
;              P :  Solar P angle in radians
;              CMD: CMD in radians of center of field arrays, east negative
;              LAT: Latitude in radians of center of field arrays, 
;                   south negative
;              PIX_SIZE: The size of the pixels in the field arrays in arcsec
;                        If it is a single number, the x and y sizes are
;                        both set to this size.  Otherwise, pix_size should
;                        be a 2-element vector:  [xsize,ysize]
;     I_CONT:  Continuum image (2-D array).  Required only if /plot is set.
;
;    In addition, the following fields will be used if present, but are not
;    required:
;
;    CMD:      The CMD in degrees of each point in the field arrays, 
;              east negative (2-D Array).
;    LATITUDE: The latitude in degrees of each point in the field arrays,
;              south negative (2-D Array).
;
;    The output structure will have the following additional fields:
;
;    BX,BY,BZ: The vertical and horizontal magnetic fields.  BX is in the
;              solar east-west direction (west positive) and BY is in the
;              solar north-south direction (north positive).
;    B_PXD,B_PYD: The coordinates in pixels of the field in heliographic 
;                 coordinates (after rotation to disk center).
;    JX,JY,JZ: The vertical and horizontal electric current. JZ is computed
;              directly from the data, but JX and JY use the linear force
;              free extrapolation for the vertical derivatives of the field.
;    J_PXD,J_PYD: The coordinates in pixels of the currents in heliographic 
;                 coordinates (after rotation to disk center).
;    J_TOTAL:  The total current ... sqrt(JX^2+JY^2+JZ^2)
;
;MODIFICATION HISTORY:
;     T. Metcalf March 1994
;     1994-09-20 Added Genetic annealing algorithm.  See K. Price, Dr. 
;                Dobb's Journal, Vol. 19, Issue 11, October 1994, p. 127.
;     1994-10-21 Made genetic annealing the default.  Changed default kb to 5,
;                and the default tfactor to 0.95.
;     1995-11-30 Added skippotential keyword.
;     2003-04-01 Added binfactor keyword
;     2003-04-02 Replaced jzonly keyword with jtotal keyword.
;     2003-04-25 Added azshearlimit keyword.
;     2004-04-01 Fixed bug in which azshearlimit was not passed on in
;                the recursive call to current2.
;     2004-07-22 Added /sum flag in calls to brebin.  Added btskip
;                keyword in calls to vmgram.
;     2004-05-25 Added jpower.  default pweight is now 0.0.
;     2005-09-07 Pixels that do not fully converge when /genetic is
;                set, are now considered suspect when going to a
;                higher resolution.
;     2005-09-20 Added dBzdz keyword.
;     2005-10-06 Added divbin keyword.
;-

   sj = size(bin)
   if sj(n_elements(sj)-2) NE 8 then message,'Structure required as input'

   j = bin   ; protect the input

   ; Check the input structure for consistency

   if keyword_set(get_alphain) then get_alpha = 1 else get_alpha = 0
   if tag_index(j,'B_LONG') LT 0 OR tag_index(j,'B_TRANS') LT 0 OR $
      tag_index(j,'B_AZIM') LT 0 OR tag_index(j,'POINT') LT 0 OR $
      (keyword_set(plot) AND tag_index(j,'I_CONT') LT 0) then $
      message,'Input structure is incomplete'
   sp = size(j.point)
   if sp(n_elements(sp)-2) NE 8 then $
      message,'Pointing Structure required in input'
   if tag_index(j.point,'P') LT 0 OR tag_index(j.point,'B0') LT 0 OR $
      tag_index(j.point,'LAT') LT 0 OR tag_index(j.point,'CMD') LT 0 OR $
      tag_index(j.point,'PIX_SIZE') LT 0 then $
      message,'Pointing data is incomplete'
   sb = size(j.b_long)
   st = size(j.b_trans)
   sa = size(j.b_azim)
   if sb(0) NE 2 then message,'B_LONG must be a 2-D array'
   if st(0) NE 2 then message,'B_TRANS must be a 2-D array'
   if sa(0) NE 2 then message,'B_AZIM must be a 2-D array'
   if sb(1) NE st(1) OR sb(1) NE sa(1) OR sb(2) NE st(2) OR $
      sb(2) NE sa(2) then message,'Magnetic field dimensions are incompatible'
   if tag_index(j,'LATITUDE') GE 0 then begin
      sl = size(j.latitude)
      if sl(0) NE 2 then message,'LATITUDE must be a 2-D array'
      if sb(1) NE sl(1) OR sb(2) NE sl(2) then $
         message,'Latitude array is incompatible with field arrays'
   endif
   if tag_index(j,'CMD') GE 0 then begin
      sc = size(j.cmd)
      if sc(0) NE 2 then message,'CMD must be a 2-D array'
      if sb(1) NE sc(1) OR sb(2) NE sc(2) then $
         message,'CMD array is incompatible with field arrays'
   endif
   nx = sb[1]
   ny = sb[2]

   if n_elements(alpha) NE 1 then alpha = 0.0 else alpha = float(alpha)
   if n_elements(guard) LE 0 then begin
      if keyword_set(slow) then guard = 2 else guard = 1
   endif
   if n_elements(skippotential) NE 1 then $
      skippotential = 0.0 else skippotential = float(abs(skippotential))
   if n_elements(bterror) NE 1 then bterror = 0.0 else bterror = float(bterror)
   if n_elements(jweight) LE 0 then jweight = 1.0d0
   if n_elements(sweight) LE 0 then sweight = 0.0d0
   if n_elements(hweight) LE 0 then hweight = 0.0d0
   if n_elements(dweight) LE 0 then dweight = 1.0d0
   if n_elements(eweight) LE 0 then eweight = 0.0d0
   if NOT keyword_set(btlength) then btlength = 2.
   if keyword_set(nogenetic) OR keyword_set(fast) then genetic = 0 else genetic = 1
   if n_elements(binfactor) LE 0 and keyword_set(fast) then binfactor=3.
   if keyword_set(jtotal) then jzonly = 0 else jzonly = 1
   if n_elements(shear) LE 0 then shear = 0
   if n_elements(shearlimit) LE 0 then shearlimit = 90.  ; 90. turns this feature off
   if n_elements(maskin) EQ nx*ny then begin
      mask = fix(maskin)   ; protect the input
      ss = where(mask)
      ; make sure the mask is only zeros and ones.
      if ss[0] GE 0 then mask[ss] = 1 else message,'All points are masked out.'
   endif else if keyword_set(maskin) then message,/info,'Mask has the wrong dimension: ignoring mask'
   if keyword_set(dbzdzin) then dbzdz = dbzdzin
   if NOT keyword_set(divbin) then divbin=2.0
   if divbin LE 1.0 then message,'ERROR: divbin must be larger than 1.0.'

   ; Degrade resolution for speed and smoothing
   if keyword_set(binfactor) then begin
      if binfactor GT 1. then begin
         ; Need alpha for call to brebin
         if n_elements(alpha) LE 0 or keyword_set(get_alpha) then begin
            if keyword_set(plot) then begin
               noplot=0 
               loadct,0
            endif else noplot = 1
            ; Run get_alpha here, but make sure it is not run twice.
            ; The call should look like the one in divb.pro.
            alpha = get_alpha(j,btcut=bterror,/quiet,/resolve, $
                        noplot=noplot, window=30, $
                        shear=shear, $
                        alimit=0.10*(mean(j.point.pix_size)/0.55), $
                        astep=0.001*(mean(j.point.pix_size)/0.55), $
                        guard=guard)
            get_alpha = 0   ; Don't need to run this twice!
            if NOT keyword_set(quiet) then begin
               if NOT !quiet then print
               message,/info,strcompress('Best Alpha is '+string(alpha))
               if NOT !quiet then print
            endif
         endif
         ; Resolve with FF field since brebin requires the 180 degree
         ; ambiguity to be already resolved.
         if NOT keyword_set(nopotential) then $
            j = fffresolve(j,alpha,guard=guard)
         j = brebin(j,fix(sb[1]/binfactor),fix(sb[2]/binfactor),alpha,/sum)
         if keyword_set(dBzdz) then begin
            ; use congrid to make the array an integer multiple
            ; so that rebin's averaging can be used.
            rnx = fix(sb[1]/binfactor)
            rny = fix(sb[2]/binfactor)
            dBzdz = rebin(congrid(dbzdz,rnx*ceil(binfactor),rny*ceil(binfactor),/center,cubic=-0.5), $
                          rnx,rny)*binfactor
         endif
         if keyword_set(mask) then begin
            mask = congrid(float(mask),n_elements(j.b_long[*,0]),n_elements(j.b_long[0,*]), $
                           /center,/interp)
            mask = byte(mask+0.99) ; Only zero is left as a zero
         endif
      endif else binfactor = 1.0
  endif else binfactor = 1.0

   ; do the minimization

   noconvergence = -1
   if keyword_set(use_pikaia) then $
      j = jgenetic(j,state=state,quiet=quiet,plot=plot, $
                 alpha=alpha*binfactor,verbose=verbose, $
                 get_alpha=get_alpha,nopotential=nopotential,novectors=novectors,$
                 jweight=jweight,jzonly=jzonly,notty=notty, $
                 skippotential=skippotential,guard=guard, $
                 dweight=dweight,eweight=eweight,sweight=sweight, $
                 btlength=btlength,mask=mask,shear=shear,hweight=hweight, $
                 ngenerations=ngenerations,pweight=pweight,gweight=gweight,jpower=jpower, $
                 enclose=enclose,dBzdz=dBzdz) $
   else $
      j = janneal(j,state=state,quiet=quiet,plot=plot, $
                 tfactor=tfactor,kb=kb,converge=converge,start=start,done=done, $
                 trecord=trecord,erecord=erecord,bterror=bterror, $
                 crecord=crecord,alpha=alpha*binfactor,verbose=verbose,fast=fast, $
                 get_alpha=get_alpha,nopotential=nopotential,novectors=novectors,$
                 jweight=jweight,genetic=genetic,jzonly=jzonly,notty=notty, $
                 skippotential=skippotential,guard=guard,slow=slow, $
                 dweight=dweight,eweight=eweight,sweight=sweight, $
                 btlength=btlength,mask=mask,shear=shear,hweight=hweight, $
                 pweight=pweight,gweight=gweight,jpower=jpower, $
                 noconvergence=noconvergence,enclose=enclose,dBzdz=dBzdz)

   ; Go back to full resolution and fix azimuth
   if keyword_set(binfactor) then begin
      if binfactor GT 1. then begin
         js = j
         jl = brebin(js,nx,ny,alpha*binfactor,/sum)
         if keyword_set(dbzdz) then dbzdz = dbzdzin

         ; Reset mask to full res version.  If /nopotential is NOT set,
         ; must still deal with masked pixels since they are to be set
         ; to the FFF and may be flipped (and /nopotential is always set
         ; in the recusrive call, so have to be careful with mask here
         ; since it changes meaning if /nopotential is NOT set).
         mask = 1b+bytarr(nx,ny)
         if keyword_set(nopotential) AND n_elements(maskin) EQ nx*ny then $
            mask = maskin 
         j = fix_azimuth(bin,jl,shear=azshear,mask=mask)
         
         ; If /genetic was set in janneal then include the non-convergent
         ; pixels in the suspect array.

         if noconvergence[0] GE 0 then begin
           sjs = size(js.b_long)
           convtest = fltarr(sjs[1],sjs[2])
           convtest[noconvergence] = 1.0
           convtest = congrid(convtest,nx,ny)
         endif else convtest = bytarr(nx,ny) 

         ;suspect = where((abs(azshear) GT shearlimit OR convtest NE 0) AND $
         ;                j.b_trans GT bterror AND mask,nsuspect)
         ; without mask in the where statement, we can lose
         ; touch with maskin during the recursive calls, but 
         ; suspect points need to be fixed.
         suspect = where((abs(azshear) GE shearlimit OR convtest NE 0) AND $
                         j.b_trans GT bterror,nsuspect)

         delvarx,js,jl,mask,convtest  ; save memory since there may be recursive calls

         if nsuspect GT 0 then begin
            ; Try to fix up the suspect points with a hopefully
            ; quick recursive call to current2 with half the 
            ; binfactor.  This will be quick if there are not
            ; too many suspect points. binfactor must be reduced
            ; here to avoid an infinite loop.
            azmask = bytarr(nx,ny)
            azmask[*] = 0b
            azmask[suspect] = 1b
            ;if (((2*nx)/binfactor) > ((2*ny)/binfactor)) GT 150 or keyword_set(novectors) then $
            ;   nov=1 else nov=0
            if keyword_set(plot) then $
               vmgram,j,/inc,/xyz,title='Mask',/quiet,/novectors, $
                      image=azmask,btskip=fix(nx/80.)>0
            j = current2(j,tfactor=tfactor,converge=converge,kb=kb,alpha=alpha, $
                         bterror=bterror,quiet=quiet,verbose=verbose,plot=plot, $
                         /nopotential,get_alpha=0, $
                         novectors=novectors,jweight=jweight,nogenetic=nogenetic, $
                         jtotal=jtotal,notty=notty,fast=fast,slow=slow, $
                         skippotential=skippotential,guard=guard,dweight=dweight, $
                         eweight=eweight,sweight=sweight,btlength=btlength, $
                         hweight=hweight,azshearlimit=shearlimit,jpower=jpower, $
                         binfactor=float(binfactor)/divbin,pweight=pweight,gweight=gweight, $
                         mask=azmask,shear=shear,/nofinalplot,enclose=enclose,dBzdz=dBzdz)
         endif
      endif
   endif

   ; plot the final result
   if keyword_set(plot) and NOT keyword_set(nofinalplot) then $
      vmgram,j,/inc,/xyz,/arr,len=2,image=j.i_cont,title='final result', $
               /quiet,novector=novectors,/current,btskip=fix(nx/80.)>0

   ; Add a comment to the structure
   if NOT keyword_set(nofinalplot) then begin
      if n_elements(alpha) LE 0 then salpha = 'potential' else salpha=alpha
      salpha=strcompress(string(salpha),/remove_all)
      icstr = 0L
      REPEAT begin
         cstr = strtrim(strcompress('COMMENT'+string(icstr),/remove_all),2)
         icstr = icstr + 1L
      endrep UNTIL tag_index(j,cstr,/quiet) LT 0 OR icstr GT 10000L
      if icstr LE 10000L then begin
         j = add2str(j,cstr,'CURRENT2: 180 Ambig. Resolved: '+ $
                       strmid(!stime,0,17)+ $
                       ' alpha = '+salpha,delete='TOTALDIV',/quiet)
      endif
   endif

   ; compute the current
   j = jxyz(j,/quiet,verbose=verbose)

   return,j

end
