
function dbzdz_fff, b, alpha, nopotential=nopotential, get_alpha=get_alpha, $
                    enclose=enclose, guard=guard, plot=plot,quiet=quiet,jresolved=j

;+
;NAME:
;     dbzdz_fff
;PURPOSE:
;     Computes dBzdz from a linear force free field extrapolation
;CATEGORY:
;CALLING SEQUENCE:
;INPUTS:
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;OUTPUTS:
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
;MODIFICATION HISTORY:
;-

   nx = n_elements(b.b_long(*,0))
   ny = n_elements(b.b_long(0,*))
   nx1 = nx-1L
   ny1 = ny-1L

   ; These are the transformation matrix elements given by eq. (2)
   ; of Gary and Hagyard, Solar Phys. 126, 21 (1990).
   ; The image-plane coordinates xI are related to the heliographic coordinates
   ; xH by xI = C xH, where C = [cxx cxy]
   ;                            [cyx cyy]
   ; THIS ASSUMES zH=0 I.E. ON PHOTOSPHERE, so 2x2 matrix instead of 3x3.
   ; Note that zI = czx*xH + czy*yH, which they do not define, but 
   ; it exists and is nonzero.

   ; Get geometric factors:
   ; cxx = dx/dX     X is heliographic and x is image
   ; cyx = dy/dX
   ; cxy = dx/dY
   ; cyy = dy/dY

   p = double(b.point.p)
   b0 = double(b.point.b0)
   if tag_index(b,'CMD',/quiet) GE 0 then lc = double(b.cmd)*!dtor $
   else lc = double(b.point.cmd)
   if tag_index(b,'LATITUDE',/quiet) GE 0 then bc = double(b.latitude)*!dtor $
   else bc = double(b.point.lat)
   pix_size = double(b.point.pix_size)
   if n_elements(pix_size) EQ 1 then pix_size = [pix_size(0),pix_size(0)]
   mpix_size = min(pix_size)

   ; These c's convert helio to image coordinates

   cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
   cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
         - COS(bc)*COS(b0)*SIN(p) 
   cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
   cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
         + COS(bc)*COS(b0)*COS(p)

   ; These ic's (inverse c's) convert image to helio

   determ = cxx*cyy-cyx*cxy
   icxx = cyy/determ
   icxy = -cxy/determ
   icyx = -cyx/determ
   icyy = cxx/determ

   ; Heliographic pixel coordinates from image coordinates

   xi = (lindgen(nx,ny) MOD nx)*(pix_size(0)/mpix_size); Image pixel coordinates
   yi = (lindgen(nx,ny) / nx)*(pix_size(1)/mpix_size)
   pxd = icxx*xi + icxy*yi           ; Heliographic pixel coordinates
   pyd = icyx*xi + icyy*yi
   jpxd = icxx*(xi+0.5) + icxy*(yi+0.5) ; Where the derivates are computed
   jpyd = icyx*(xi+0.5) + icyy*(yi+0.5)

   j = b

   if tag_index(j,'B_PXD',/quiet) LT 0 then $
      j=add2str(j,'B_PXD',pxd,/quiet) $
   else j.b_pxd = pxd
   if tag_index(j,'B_PYD',/quiet) LT 0 then $
      j=add2str(j,'B_PYD',pyd,/quiet) $
   else j.b_pyd = pyd
   degree = (sqrt(n_elements(pxd))-1.0) < 3
   polywarp,jpxd,jpyd,pxd,pyd,long(degree),jkx,jky  ; Map to center
   dxidX =  poly_2d(cxx,jkx,jky,2)
   dxidY =  poly_2d(cxy,jkx,jky,2)
   dyidX =  poly_2d(cyx,jkx,jky,2)
   dyidY =  poly_2d(cyy,jkx,jky,2)

   ; Compute the potential field to get a crude resolution of the azimuth

   if keyword_set(get_alpha) or n_elements(alpha) NE 1 then begin
      loadct,0
      if keyword_set(plot) then noplot=0 else noplot = 1
      ;alpha = get_alpha(j,btcut=(bterror>(bterror(j))),/quiet,/resolve, $
      ;                  noplot=noplot,guard=guard)
      alpha = get_alpha(j,btcut=bterror,/quiet,/resolve, $
                        ;noplot=noplot, $
                        noplot=noplot, win=30, $
                        shear=shear, $
                        alimit=0.10*(mean(j.point.pix_size)/0.55), $
                        astep=0.0015*(mean(j.point.pix_size)/0.55), $
                        guard=guard)
      if NOT !quiet then print
      message,/info,strcompress('Best Alpha is '+string(alpha))
      if NOT !quiet then print
   endif

   if NOT keyword_set(nopotential) then begin

      jblong = j.b_long
      pot = fff(b0=b0,pang=p,cmd=lc,lat=bc,jblong,/image,pixel=pix_size, $
                alpha=alpha,/quiet,guard=guard)
      pxi = pot(*,*,0)  ; These are image coordinates
      pyi = pot(*,*,1)
      pzi = pot(*,*,2)

      check = check_math(0,1)
         pa_azimuth = atan(-pxi,pyi)/!dtor  ; Az is CCW from North
      check = check_math(1,0)

      ; Fix the azimuthal angle from the potential field.  This is necessary
      ; for the computation of the initial temperature in janneal.pro.
      ; If we don't make an initial guess, the temperature starts too high
      ; and wastes time.

      fixa = where(abs(arange(j.B_AZIM(*),pa_azimuth(*))-pa_azimuth(*)) GT $
                   90.,count)
      if (count gt 0) then j.B_AZIM(fixa) = j.B_AZIM(fixa) + 180.
      j.B_AZIM = j.B_AZIM MOD 360.

      print
      if alpha EQ 0.0 then $
         print,'Resolved azimuth with LOS potential field:',count,' flips' $
      else $
         print,'Resolved azimuth with LOS force-free field:',count,' flips'
   endif

   azim = double(j.b_azim) * !dtor
   bxyz,-j.b_trans*sin(azim), $
         j.b_trans*cos(azim), $
         j.b_long, $
         bx,by,bz,p=p,b0=b0,lat=bc,cmd=lc
   if tag_index(j,'BX',/quiet) GE 0 then j.bx=bx $
   else j=add2str(j,'BX',bx,/quiet)
   if tag_index(j,'BY',/quiet) GE 0 then j.by=by $
   else j=add2str(j,'BY',by,/quiet)
   if tag_index(j,'BZ',/quiet) GE 0 then j.bz=bz $
   else j=add2str(j,'BZ',bz,/quiet)

   j = jxyz(j,/quiet)

   ; Get the potential\force-free field in heliographic coordinates.  It is
   ; used in the calculation of the divergence.

   jbz = j.bz
   if keyword_set(enclose) then begin
      pot = fff(jbz,/normal,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pix_size, $
                alpha=0.0,bx=j.bx,by=j.by,/enclose)
   endif else begin
      pot = fff(jbz,/normal,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pix_size, $
                alpha=alpha,bx=j.bx,by=j.by,guard=guard)
   endelse
   px = pot(*,*,0)  ; These are heliographic coordinates
   py = pot(*,*,1)
   pz = pot(*,*,2)

   ;if keyword_set(plot) and NOT keyword_set(quiet) then $
   ;   vmgram,j,/inc,/xyz,/arr,len=2,image=j.i_cont, $
   ;   title='Initial Field',novector=novectors,/quiet,current=havejz, $
   ;   btskip=fix(nx/80.)>0

   lylxpx=shift(px,-1,-1)    ; Shifts for potential field derivatives
   lylxpy=shift(py,-1,-1)
   lylxpz=shift(pz,-1,-1)
   lxpx=shift(px,-1,0)
   lxpy=shift(py,-1,0)
   lxpz=shift(pz,-1,0)
   lypx=shift(px,0,-1)
   lypy=shift(py,0,-1)
   lypz=shift(pz,0,-1)

   ; Derivatives in the image coordinates of the potential field

   dPxdxi = 0.5*(lylxpx+lxpx-lypx-px)
   dPxdyi = 0.5*(lylxpx+lypx-lxpx-px)
   dPydxi = 0.5*(lylxpy+lxpy-lypy-py)
   dPydyi = 0.5*(lylxpy+lypy-lxpy-py)
   dPzdxi = 0.5*(lylxpz+lxpz-lypz-pz)
   dPzdyi = 0.5*(lylxpz+lypz-lxpz-pz)


   ; Derivatives in the heliographic coordinates of the potential field.
   ; Use div P = 0 to derive z derivative

   dPzdZ = -(dPxdxi * dxidX + dPxdyi * dyidX + $     ; dPxdX
             dPydyi * dyidY + dPydxi * dxidY)        ; dPydY
   dPzdX = dPzdxi*dxidX + dPzdyi*dyidX
   dPzdY = dPzdxi*dxidY + dPzdyi*dyidY

   return,dPzdZ

end
