
;+
;   Computes all possible field orientations and divergences\currents for a 
;   magnetic field.  Called by janneal.pro, not intended to be called
;   otherwise.  All possible "energies" are returned in divarray.
;
;   T. Metcalf March 1994
;   March 27, 2003 TRM In calculation of Jx and Jy, use alpha = Jz/Bz
;   April 2, 2003  TRM Added azimuth difference from potential field
;                      to the minimized funtion.
;   Jan 14, 2005 TRM Use get_berrs instead of bterror
;-

pro divb,j,divarray,state=state,reset=reset,alpha=alpha,quiet=quiet, $
         plot=plot,nopotential=nopotential,get_alpha=get_alpha, $
         bterror=bterror,novectors=novectors,jweight=jweight,jzonly=jzonly, $
         potazim=pa_azimuth,guard=guard,dweight=dweight,eweight=eweight, $
         sweight=sweight,shear=shear,hweight=hweight,pweight=pweightin, $
         gweight=gweight,jpower=jpower,mask=mask,enclose=enclose,dBzdz=dBzdzin

   ; Set up

   nx = n_elements(j.b_long(*,0))
   ny = n_elements(j.b_long(0,*))
   nx1 = nx-1L
   ny1 = ny-1L

   if n_elements(jweight) LE 0 then jweight = 1.0d0 $
   else jweight=double(abs(jweight))
   if n_elements(jpower) LE 0 then jpower = 2.0d0 $
   else jpower=double(jpower)
   if n_elements(sweight) LE 0 then sweight = 0.0d0 $
   else sweight=double(abs(sweight))
   if n_elements(dweight) LE 0 then dweight = 1.0d0 $
   else dweight=double(abs(dweight))
   if n_elements(eweight) LE 0 then eweight = 0.0d0 $
   else eweight=double((eweight))
   if n_elements(hweight) LE 0 then hweight = 0.0d0 $
   else hweight=double((hweight))
   if n_elements(pweightin) LE 0 then pweightin = 0.0d0 $
   else pweightin=double((pweightin))
   if n_elements(gweight) LE 0 then gweight = 0.0d0 $
   else gweight=double((gweight))
   if n_elements(guard) LE 0 then guard = 1
   if keyword_set(jzonly) then jhweight=0.0 else jhweight=1.0
   if n_elements(jweight) NE 1 and n_elements(jweight) NE nx*ny then begin
      help,jweight
      message,'Dimension of jweight is incompatible with field arrays'
   endif
   if n_elements(dweight) NE 1 and n_elements(dweight) NE nx*ny then begin
      help,dweight
      message,'Dimension of dweight is incompatible with field arrays'
   endif
   if n_elements(eweight) NE 1 and n_elements(eweight) NE nx*ny then begin
      help,eweight
      message,'Dimension of eweight is incompatible with field arrays'
   endif
   if n_elements(shear) LE 0 then shear = 0

   
   ; These are the transformation matrix elements given by eq. (2)
   ; of Gary and Hagyard, Solar Phys. 126, 21 (1990).
   ; The image-plane coordinates xI are related to the heliographic coordinates
   ; xH by xI = C xH, where C = [cxx cxy]
   ;                            [cyx cyy]
   ; THIS ASSUMES zH=0 I.E. ON PHOTOSPHERE, so 2x2 matrix instead of 3x3.
   ; Note that zI = czx*xH + czy*yH, which they do not define, but 
   ; it exists and is nonzero.

   ; Get geometric factors:
   ; cxx = dx/dX     X is heliographic and x is image
   ; cyx = dy/dX
   ; cxy = dx/dY
   ; cyy = dy/dY

   p = double(j.point.p)
   b0 = double(j.point.b0)
   if tag_index(j,'CMD',/quiet) GE 0 then lc = double(j.cmd)*!dtor $
   else lc = double(j.point.cmd)
   if tag_index(j,'LATITUDE',/quiet) GE 0 then bc = double(j.latitude)*!dtor $
   else bc = double(j.point.lat)
   pix_size = double(j.point.pix_size)
   if n_elements(pix_size) EQ 1 then pix_size = [pix_size(0),pix_size(0)]
   mpix_size = min(pix_size)

   ; These c's convert helio to image coordinates

   cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
   cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
         - COS(bc)*COS(b0)*SIN(p) 
   cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
   cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
         + COS(bc)*COS(b0)*COS(p)

   ; These ic's (inverse c's) convert image to helio

   determ = cxx*cyy-cyx*cxy
   icxx = cyy/determ
   icxy = -cxy/determ
   icyx = -cyx/determ
   icyy = cxx/determ

   ; Heliographic pixel coordinates from image coordinates

   xi = (lindgen(nx,ny) MOD nx)*(pix_size(0)/mpix_size); Image pixel coordinates
   yi = (lindgen(nx,ny) / nx)*(pix_size(1)/mpix_size)
   pxd = icxx*xi + icxy*yi           ; Heliographic pixel coordinates
   pyd = icyx*xi + icyy*yi
   jpxd = icxx*(xi+0.5) + icxy*(yi+0.5) ; Where the derivates are computed
   jpyd = icyx*(xi+0.5) + icyy*(yi+0.5)

   if tag_index(j,'B_PXD',/quiet) LT 0 then $
      j=add2str(j,'B_PXD',pxd,/quiet) $
   else j.b_pxd = pxd
   if tag_index(j,'B_PYD',/quiet) LT 0 then $
      j=add2str(j,'B_PYD',pyd,/quiet) $
   else j.b_pyd = pyd

   if (tag_index(j,'TOTALDIV',/quiet)) LT 0 then $
      j = add2str(j,'TOTALDIV',0.0,/quiet)

   ; Derivatives between coordinate systems at pixel centers.  This is
   ; necessary since the divergence is computed at the centers later.

   ;dXdxi = icxx   ; At pixel corners, not centers
   ;dYdxi = icyx
   ;dXdyi = icxy
   ;dYdyi = icyy

   ;dxidX = cxx   ; At pixel corners, not centers
   ;dyidX = cyx
   ;dxidY = cxy
   ;dyidY = cyy

  ; lylxCx=shift(pxd,-1,-1)
  ; lylxCy=shift(pyd,-1,-1)
  ; lxCx=shift(pxd,-1,0)
  ; lxCy=shift(pyd,-1,0)
  ; lyCx=shift(pxd,0,-1)
  ; lyCy=shift(pyd,0,-1)

  ; dXdxi = 0.5*(lylxCx+lxCx-lyCx-pxd)
  ; dYdxi = 0.5*(lylxCy+lxCy-lyCy-pyd)
  ; dXdyi = 0.5*(lylxCx+lyCx-lxCx-pxd)
  ; dYdyi = 0.5*(lylxCy+lyCy-lxCy-pyd)
  ; determ = dXdxi*dYdyi - dYdxi*dXdyi   ; invert the 2 variable system
  ; dxidX = dYdyi/determ
  ; dxidY = -dXdyi/determ
  ; dyidX = -dYdxi/determ
  ; dyidY = dXdxi/determ

   ; Patch up the edges where the shifts get messed up

  ; if n_elements(cxx) GT 1 then begin
  ;    dxidX(nx-1,*)=cxx(nx-1,*) & dxidX(*,ny-1)=cxx(*,ny-1)
  ;    dxidY(nx-1,*)=cxy(nx-1,*) & dxidY(*,ny-1)=cxy(*,ny-1)
  ;    dyidX(nx-1,*)=cyx(nx-1,*) & dyidX(*,ny-1)=cyx(*,ny-1)
  ;    dyidY(nx-1,*)=cyy(nx-1,*) & dyidY(*,ny-1)=cyy(*,ny-1)
  ; endif else begin
  ;    dxidX(nx-1,*)=cxx & dxidX(*,ny-1)=cxx
  ;    dxidY(nx-1,*)=cxy & dxidY(*,ny-1)=cxy
  ;    dyidX(nx-1,*)=cyx & dyidX(*,ny-1)=cyx
  ;    dyidY(nx-1,*)=cyy & dyidY(*,ny-1)=cyy
  ; endelse

   degree = (sqrt(n_elements(pxd))-1.0) < 3
   polywarp,jpxd,jpyd,pxd,pyd,long(degree),jkx,jky  ; Map to center
   dxidX =  poly_2d(cxx,jkx,jky,2)
   dxidY =  poly_2d(cxy,jkx,jky,2)
   dyidX =  poly_2d(cyx,jkx,jky,2)
   dyidY =  poly_2d(cyy,jkx,jky,2)

   ; Arrays that define the state of the field

   field = dblarr(2,3,nx,ny)       ; Bx, By, Bz
   state = bytarr(nx,ny)           ; flipped?
   divarray = dblarr(16,nx,ny)     ; Div 

   ; Compute the potential field to get a crude resolution of the azimuth

   if NOT keyword_set(nopotential) then begin

      jblong = j.b_long
      pot = fff(b0=b0,pang=p,cmd=lc,lat=bc,jblong,/image,pixel=pix_size, $
                alpha=alpha,/quiet,guard=guard)
      pxi = pot(*,*,0)  ; These are image coordinates
      pyi = pot(*,*,1)
      pzi = pot(*,*,2)

      check = check_math(0,1)
         pa_azimuth = atan(-pxi,pyi)/!dtor  ; Az is CCW from North
      check = check_math(1,0)

      ; Fix the azimuthal angle from the potential field.  This is necessary
      ; for the computation of the initial temperature in janneal.pro.
      ; If we don't make an initial guess, the temperature starts too high
      ; and wastes time.

      fixa = where(abs(arange(j.B_AZIM(*),pa_azimuth(*))-pa_azimuth(*)) GT $
                   90.,count)
      if (count gt 0) then j.B_AZIM(fixa) = j.B_AZIM(fixa) + 180.
      j.B_AZIM = j.B_AZIM MOD 360.

      print
      if alpha EQ 0.0 then $
         print,'Resolved azimuth with LOS potential field:',count,' flips' $
      else $
         print,'Resolved azimuth with LOS force-free field:',count,' flips'
   endif else pa_azimuth=j.b_azim

   azim = double(j.b_azim) * !dtor
   bxyz,-j.b_trans*sin(azim), $
         j.b_trans*cos(azim), $
         j.b_long, $
         bx,by,bz,p=p,b0=b0,lat=bc,cmd=lc
   if tag_index(j,'BX',/quiet) GE 0 then j.bx=bx $
   else j=add2str(j,'BX',bx,/quiet)
   if tag_index(j,'BY',/quiet) GE 0 then j.by=by $
   else j=add2str(j,'BY',by,/quiet)
   if tag_index(j,'BZ',/quiet) GE 0 then j.bz=bz $
   else j=add2str(j,'BZ',bz,/quiet)

   j = jxyz(j,/quiet)
   jzerror = jzerr(j)

   ;if tag_index(j,'JZ',/quiet) LT 0 then havejz=0 else havejz=1
   ;if havejz then jzerror = jzerr(j) else jzerror = jzerr(jxyz(j,/quiet))

   ; Get the potential\force-free field in heliographic coordinates.  It is
   ; used in the calculation of the divergence.

   if keyword_set(get_alpha) then begin
      loadct,0
      if keyword_set(plot) then noplot=0 else noplot = 1
      ;alpha = get_alpha(j,btcut=(bterror>(bterror(j))),/quiet,/resolve, $
      ;                  noplot=noplot,guard=guard)
      alpha = get_alpha(j,btcut=bterror,/quiet,/resolve, $
                        ;noplot=noplot, $
                        noplot=noplot, win=30, $
                        shear=shear, $
                        alimit=0.10*(mean(j.point.pix_size)/0.55), $
                        astep=0.001*(mean(j.point.pix_size)/0.55), $
                        guard=guard)
      if NOT !quiet then print
      message,/info,strcompress('Best Alpha is '+string(alpha))
      if NOT !quiet then print
   endif

   ; Get the FFF to use in computing div B and Jtotal terms.
   ; The FF field is independent of the ambiguity resolution
   ; unless /nopotential is set.

   ;if NOT keyword_set(nopotential) then begin
   ;   jblong = j.b_long
   ;   pot = fff(jblong,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pix_size, $
   ;             alpha=alpha,bx=j.bx,by=j.by,guard=guard) 
   ;endif else begin
   ;   jbz = j.bz
   ;   pot = fff(jbz,/normal,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pix_size, $
   ;             alpha=alpha,bx=j.bx,by=j.by,guard=guard)
   ;endelse
   jbz = j.bz
   if keyword_set(enclose) then begin
      pot = fff(jbz,/normal,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pix_size, $
                alpha=0.0,bx=j.bx,by=j.by,/enclose)
   endif else begin
      pot = fff(jbz,/normal,b0=b0,pang=p,cmd=lc,lat=bc,pixel=pix_size, $
                alpha=alpha,bx=j.bx,by=j.by,guard=guard)
   endelse
   px = pot(*,*,0)  ; These are heliographic coordinates
   py = pot(*,*,1)
   pz = pot(*,*,2)

   if keyword_set(plot) and NOT keyword_set(quiet) then $
      vmgram,j,/inc,/xyz,/arr,len=2,image=j.i_cont, $
      title='Initial Field',novector=novectors,/quiet,current=havejz, $
      btskip=fix(nx/80.)>0

   ; Compute and store the two possible orientations of the field

   ;print
   ;print,'Computing field array ... ',!stime
   ;print

   for i=0,1 do begin                      ; Two possible orientations
      azim = double(j.b_azim) * !dtor
      bxyz,-j.b_trans*sin(azim), $
            j.b_trans*cos(azim), $
            j.b_long, $
            bx,by,bz,/last_transform
      field(i,0,*,*) = bx
      field(i,1,*,*) = by
      field(i,2,*,*) = bz

      j.b_azim = (j.b_azim + 180.0) MOD 360.  ; called twice so not permanent

   endfor

   ; Compute and store all possible divergence\currents 

   ; The divergence is computed between four data points.  Since each data
   ; point has two possible states, there are 16 possible states for the
   ; divergence\current at each point in the magnetogram.

   ; divergence state = 0..15 (4 bits)

   ; bit2   +-----+ bit3       y ^
   ;        |     |              |
   ;        |     |              |
   ;        |     |              |
   ; bit0   +-----+ bit1         +-----> x

   ; The derivative arrays are defined so that the derivative is placed
   ; in the array element represented by bit0, although it is actually
   ; the derivative at the center of the 4 pixels.  jpxd and jpyd give the
   ; coordinates of the center of the 4 pixels.

   lylxpx=shift(px,-1,-1)    ; Shifts for potential field derivatives
   lylxpy=shift(py,-1,-1)
   lylxpz=shift(pz,-1,-1)
   lxpx=shift(px,-1,0)
   lxpy=shift(py,-1,0)
   lxpz=shift(pz,-1,0)
   lypx=shift(px,0,-1)
   lypy=shift(py,0,-1)
   lypz=shift(pz,0,-1)

   ; Derivatives in the image coordinates of the potential field

   dPxdxi = 0.5*(lylxpx+lxpx-lypx-px)
   dPxdyi = 0.5*(lylxpx+lypx-lxpx-px)
   dPydxi = 0.5*(lylxpy+lxpy-lypy-py)
   dPydyi = 0.5*(lylxpy+lypy-lxpy-py)
   dPzdxi = 0.5*(lylxpz+lxpz-lypz-pz)
   dPzdyi = 0.5*(lylxpz+lypz-lxpz-pz)

   ; Compute the deriviatives of the total magnetic field
   
   btot = sqrt(j.bx^2 + j.by^2 + j.bz^2)

   lylxbtot=shift(btot,-1,-1)
   lxbtot=shift(btot,-1,0)
   lybtot=shift(btot,0,-1)

   dBtotdxi = 0.5*(lylxbtot+lxbtot-lybtot-btot)
   dBtotdyi = 0.5*(lylxbtot+lybtot-lxbtot-btot)

   ; Derivatives in the heliographic coordinates of the potential field.
   ; Use div P = 0 to derive z derivative

   dPzdZ = -(dPxdxi * dxidX + dPxdyi * dyidX + $     ; dPxdX
             dPydyi * dyidY + dPydxi * dxidY)        ; dPydY
   dPzdX = dPzdxi*dxidX + dPzdyi*dyidX
   dPzdY = dPzdxi*dxidY + dPzdyi*dyidY

   ; Get force-free field at the positions where the current and divergence
   ; are computed.  Used to compute the horizontal current.

   pxcenter = bilinxy(px,jpxd,jpyd,xarray=pxd,yarray=pyd)
   pycenter = bilinxy(py,jpxd,jpyd,xarray=pxd,yarray=pyd)
   pzcenter = bilinxy(pz,jpxd,jpyd,xarray=pxd,yarray=pyd)

   for i=0,15 do begin   ; compute divergence

      s0 = keyword_set(i AND 1)       ; state of pixel 0   (0,0)
      s1 = keyword_set(i AND 2)       ; state of pixel 1   (1,0)
      s2 = keyword_set(i AND 4)       ; state of pixel 2   (0,1)
      s3 = keyword_set(i AND 8)       ; state of pixel 3   (1,1)

      bxcenter = bilinxy(reform(field(s0,0,*,*)),jpxd,jpyd,xarray=pxd,yarray=pyd)
      bycenter = bilinxy(reform(field(s0,1,*,*)),jpxd,jpyd,xarray=pxd,yarray=pyd)
      bzcenter = bilinxy(reform(field(s0,2,*,*)),jpxd,jpyd,xarray=pxd,yarray=pyd)
      ; We'll be dividing by btotal, so set a minimum of 1.0
      btotal = sqrt(bxcenter^2 + bycenter^2 + bzcenter^2) > 1.0

      lylxbx=shift(reform(field(s3,0,*,*)),-1,-1)  ; Shifts for derivatives
      lylxby=shift(reform(field(s3,1,*,*)),-1,-1)
      lylxbz=shift(reform(field(s3,2,*,*)),-1,-1)
      lxbx=shift(reform(field(s1,0,*,*)),-1,0)
      lxby=shift(reform(field(s1,1,*,*)),-1,0)
      lxbz=shift(reform(field(s1,2,*,*)),-1,0)
      lybx=shift(reform(field(s2,0,*,*)),0,-1)
      lyby=shift(reform(field(s2,1,*,*)),0,-1)
      lybz=shift(reform(field(s2,2,*,*)),0,-1)

      ; Derivatives in the image coordinates

      dBxdxi = 0.5*(lylxbx+lxbx-lybx-reform(field(s0,0,*,*)))
      dBxdyi = 0.5*(lylxbx+lybx-lxbx-reform(field(s0,0,*,*)))
      dBydxi = 0.5*(lylxby+lxby-lyby-reform(field(s0,1,*,*)))
      dBydyi = 0.5*(lylxby+lyby-lxby-reform(field(s0,1,*,*)))
      dBzdxi = 0.5*(lylxbz+lxbz-lybz-reform(field(s0,2,*,*)))
      dBzdyi = 0.5*(lylxbz+lybz-lxbz-reform(field(s0,2,*,*)))

      ; Computing the derivatives as above can lead to a problem.  The 
      ; derivative will be artificially small if the field has a 
      ; "checkerboard" pattern, e.g.,
      ;
      ;      large +---+ small    positive +---+ negative
      ;            |   |                   |   |
      ;            |   |                   |   |
      ;      small +---+ large    negative +---+ positive
      ;
      ; In this case,  the derivatives along each row/col will be large, but 
      ; of opposite sign, and averaging two row/cols will give a small 
      ; derivative.  In this case, the field is far from smooth, but the 
      ; derivative is small and very uncertain.  I test for this using the 
      ; second derivative.

      ; The second derivatives give the uncertainty in the first derivatives.
      ; If the error is large, don't try to average to the center of the pixel.
      ; This effectively gives an energy penalty to these pixels so the 
      ; uncertain state will not usually be selected.
  
      ; In the "where" statements below the units look strange, but the second
      ; derivative is multiplied by the length of one half pixel.

      if 1 then begin

         ; TRM 2005-05-25 changed to 1.5 from 0.5
         ddfactor = 1.0  ; Fractional error in first derivative 
                         ; considered to be too high 

         d2Bxdxidyi = lylxbx-lxbx-lybx+reform(field(s0,0,*,*))
         bad = where(abs(0.5*d2Bxdxidyi) GT ddfactor*abs(dBxdxi))
         if bad(0) GE 0 then dBxdxi(bad) = lylxbx(bad)-lybx(bad)
         bad = where(abs(0.5*d2Bxdxidyi) GT ddfactor*abs(dBxdyi))
         if bad(0) GE 0 then dBxdyi(bad) = lylxbx(bad)-lxbx(bad)

         d2Bydxidyi = lylxby-lxby-lyby+reform(field(s0,1,*,*))
         bad = where(abs(0.5*d2Bydxidyi) GT ddfactor*abs(dBydxi))
         if bad(0) GE 0 then dBydxi(bad) = lylxby(bad)-lyby(bad)
         bad = where(abs(0.5*d2Bydxidyi) GT ddfactor*abs(dBydyi))
         if bad(0) GE 0 then dBydyi(bad) = lylxby(bad)-lxby(bad)

         d2Bzdxidyi = lylxbz-lxbz-lybz+reform(field(s0,2,*,*))
         bad = where(abs(0.5*d2Bzdxidyi) GT ddfactor*abs(dBzdxi))
         if bad(0) GE 0 then dBzdxi(bad) = lylxbz(bad)-lybz(bad)
         bad = where(abs(0.5*d2Bzdxidyi) GT ddfactor*abs(dBzdyi))
         if bad(0) GE 0 then dBzdyi(bad) = lylxbz(bad)-lxbz(bad)
      endif

      ; Derivatives in the heliographic coordinates:

      ; dBxdX = dBxdxi * dxidX + dBxdyi * dyidX
      ; dBydY = dBydyi * dyidY + dBydxi * dxidY

      ; divarray = |Divergence| + |Current| - Entropy + |dAzimuth|

      ; If alpha=0 then J is just Jz (Jx=0, Jy=0), so for alpha=0
      ; we just minimize Jz

      ; Jz is  dBydX-dBxdY
      Jz =   dBydxi*dxidX + dBydyi*dyidX $    ; dBydX
           - dBxdyi*dyidY - dBxdxi*dxidY      ; dBxdY
      if 0 then begin
         if (0) then begin  ; Compute alpha from Jz/Bz in strong field regions
            ; For b_trans LT bsigma, alpha is the linear FF alpha
            ; For b_trans GT 1.*bsigma and b_trans LT 3.*bsigma, 
            ;    alpha = median smoohted alpha
            ; For b_trans GT 3.*bsigma, alpha = Jz/Bz
            errs = get_berrs(j)
            bsigma = errs.sig_btrans
            ;bsigma = bterror(j)
            alphanlfff = median(f_div(Jz,Pzcenter),3)
            goodbz=where(abs(Pzcenter) GT 3.*bsigma  $
                         ;and abs(jz) GT jzerror $
                         ,cgoodbz)
            if cgoodbz GT 0 then alphanlfff[goodbz] = jz[goodbz]/Pzcenter[goodbz]
            bad = where(abs(pzcenter) LT bsigma,nbad)
            if nbad GT 0. then alphanlfff[bad] = alpha
         endif else alphanlfff = alpha  ; Use linear force free alpha
         ; The following is not so good since dPzdX should be approximately
         ; equal to (dBzdxi*dxidX+dBzdyi*dyidX), so all we are minimizing
         ; is pycenter.  Better to set dBxdZ and dBydZ to 0.0 so that we
         ; are minimizing the horizontal gradient of Bz.
         ; Compute dBxdZ and dBydZ from the FFF
         ; Jy is dBxdZ-dBzdX
         Jy =   dPzdX + alphanlfff*pycenter   $  ; dBxdZ
              - dBzdxi*dxidX - dBzdyi*dyidX      ; dBzdX
         ; Jx is dBzdY-dBydZ
         Jx =   dBzdxi*dxidY + dBzdyi*dyidY   $  ; dBzdY
              - dPzdY + alphanlfff*pxcenter      ; dBydZ
       endif else begin
         ; TRM 2005-05-25 make this the default
         ; Just set dBxdZ and dBydZ to 0 since we have no idea what 
         ; they are.  This puts a condition on the smoothness of
         ; Bz.
         ; Jy is dBxdZ-dBzdX
         Jy =   0.0   $                          ; dBxdZ
              - dBzdxi*dxidX - dBzdyi*dyidX      ; dBzdX
         ; Jx is dBzdY-dBydZ
         Jx =   dBzdxi*dxidY + dBzdyi*dyidY   $  ; dBzdY 
              - 0.0                              ; dBydZ
       endelse



      Jtot =  sqrt( $
                   ( Jz )^2  + $                            ; Jz^2
                   (( Jx )^2)*jhweight  + $                 ; Jx^2
                   (( Jy )^2)*jhweight    $                 ; Jy^2
                  )                                         ; |J| 

      ;if n_elements(Jtotmax) LE 0 then Jtotmax = max(Jtot)  ; don't let the scale vary
      ;Jtot = Jtotmax*(Jtot/Jtotmax)^jpower

      ; The mean current is rougly at the noise level since currents
      ; are not typically widespread.  Hold fixed at 100 sigma.
      ;if n_elements(Jtotscale) LE 0 then Jtotscale = 100.0*(total(Jtot)/n_elements(Jtot))  ; don't let the scale vary
      ;Jtot = Jtotscale*(Jtot/Jtotscale)^jpower

      gradB = sqrt((dBtotdxi*dxidX + dBtotdyi*dyidX)^2 + (dBtotdxi*dxidY + dBtotdyi*dyidY)^2) > 1.0
      mingradB = min(gradB)
      Jgrad = Jtot/(gradB/mingradB)

      ; Sheath current from Georgoulis, LaBonte, and Metcalf.

      bxrel = bxcenter/btotal
      byrel = bycenter/btotal
      ;bhrel = sqrt((bxrel)^2+(byrel)^2)
      ;hrelbad = where(bhrel EQ 0.0,nhrelbad)
      ;if nhrelbad GE 1 then bhrel[hrelbad] = 1.0
      Js = abs((dBtotdxi*dxidX + dBtotdyi*dyidX)*(byrel) - $
               (dBtotdxi*dxidY + dBtotdyi*dyidY)*(bxrel)) ;/ bhrel

           ;+ $   ; (dBtot/dz)'
           ;((dBtotdxi*dxidX + dBtotdyi*dyidX)*(bxcenter/btotal) + $
           ; (dBtotdxi*dxidY + dBtotdyi*dyidY)*(bycenter/btotal)) * $
           ;(bzcenter/(bxcenter^2+bycenter^2))

      ; The 100 should be about exp(1)*jzerr(j)
      ;junits = 100.0*(4.*!pi*(1.0e-3)*721.9*mpix_size) ; 100 mA/m^2 scale
      ;lambda = 10.
      ;jscale = -f_div(jweight,lambda*eweight)
      ;;junits = exp(-jscale)*jzerror*(4.*!pi*(1.0e-3)*721.9*mpix_size)
      ;;Jlog = (alog((Jtot > (max(Jtot)/1.e30))/junits)) > jscale
      ;lambda=1.
      ;junits = jzerror*(4.*!pi*(1.0e-3)*721.9*mpix_size)
      ;;Jlog = alog(1.+Jtot/junits)

      ; Adding the entropy is like a spatially varying jweight:
      ; jweight -> jweight - eweight*lambda*Jlog/junits
      ; The j weight is reduced where the current is strong and increased
      ; where the current is weak, i.e. the divergence condition is
      ; favored where the current is strong and the current condition
      ; is favored where the current is weak.  If eweight is negative
      ; this is reversed.

      lambda = 1.0/alog(mean(Jtot))
      Jlog = alog(Jtot > (max(Jtot)/1.e30))
      entropy = lambda * Jtot * Jlog

      ; The current helicity is essentially the field weighted current
      ; so the current is weighted more where the field is strong.  Kind
      ; of the reverse of the entropy term when it has positive weight.
      jhelicity = sqrt(Jz^2+Jx^2+Jy^2) * (btotal/max(btotal)) ; Current helicity

      ; A weak term to make the angle with the potential field acute.
      ; This should only come into play when the algorithm is having
      ; a hard time deciding on orientations based on the field 
      ; derivatives.  In that case, use the solution with azimuth 
      ; closest to the LFF/potential field.

      pweight = 3.*pweightin*jzerror/180.0  ; Never larger than 3*jzerror
      bxyz,bxi,byi,bzi,bxcenter,bycenter,bzcenter, $
         p=p,b0=b0,lat=bc,cmd=lc,/helio2image
      bxyz,pxi,pyi,pzi,pxcenter,pycenter,pzcenter, $
         p=p,b0=b0,lat=bc,cmd=lc,/helio2image
      bazim = atan(-bxi,byi)*!radeg
      pazim = atan(-pxi,pyi)*!radeg
      pdiff = abs(arange(bazim,pazim)-pazim)  ; in degrees

      if NOT keyword_set(dBzdzin) then begin
         if i eq 0 then message,/info,'Using lfff dBzdz'
         dBzdz = dPzdz
      endif else begin
         if n_elements(dBzdzin) NE n_elements(dPzdz) then $
            message,'ERROR: dBzdz has the wrong size.'
         dBzdz = dBzdzin
      endelse

      divarray(i,*,*) = abs( $
                              dBxdxi * dxidX + dBxdyi * dyidX + $ ; dBxdX
                              dBydyi * dyidY + dBydxi * dxidY + $ ; dBydY
                              dBzdz $                             ; dBzdZ
                           )*dweight $                            ; |DivB|
                        + Jtot*jweight $                          ; |J|
                        + Jgrad*gweight $                         ; |J|/grad(B)
                        + Js*sweight $                            ; |Js|
                        + entropy*eweight $                       ; neg J entropy
                        + pdiff*pweight $                         ; |dAzimuth|
                        + jhelicity*hweight                       ; |Current Helicity|

      divarray(i,*,*) = divarray(i,*,*)^jpower 
                         
   endfor

   ; Set the edges to zero since the above shifts will give garbage at edges

   divarray(*,nx-1,*)=0. 
   divarray(*,*,ny-1)=0.

   j.totaldiv = total(divarray(0,*,*))    ; Initial "energy"

end
