;+
;NAME:
;     JANNEAL
;PURPOSE:
;     Resolves 180 degree ambiguity in vector magnetic field data.
;     Implements simulated annealing for the simultaneous minimization of the
;     divergence of B and the vertical electric current.
;CATEGORY:
;CALLING SEQUENCE:
;     jnew = janneal(j)
;INPUTS:
;     j = Magnetic field structure whose divergence\current is minimized.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     skippotential = Pixel with azimuth this close to potential are not
;                     included in the iteration to save time.  Degrees.
;                     default = 0.0.  Only applies to pixels with B_trans
;                     larger than 3*bterror.
;              CAUTION:  Don't set this too high:  skipping points can confuse
;                        the algorithm.
;     bterror = smallest transverse magnetic field (absolute value) to consider.
;              CAUTION:  Don't set this too high: skipping points can confuse
;              the algorithm. (default = 0.0)
;     mask = Array with a 1 to indicate that the azimuth for that
;            pixel is to be tested or a 0 to indicate that it is
;            already set and should not be tested.
;     jweight = weighting of the current minimization relative to the 
;               divergence minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def
;               = 1.0.
;     sweight = weighting of the sheath current minimization relative to the 
;               divergence minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def
;               = 1.0.
;     dweight = weighting of the divergence minimization relative to the 
;               current minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def = 1.0.
;     eweight = weighting of the entropy maximization relative to the 
;               current minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def = 0.0.
;     hweight = weighting of the current helicity minimization
;               relative to the current and divergence
;               minimization. May be scalar or an array with the same
;               dimension as the magnetic field arrays.  Def = 0.0. 
;     /jzonly = Minimize only Jz, do not use the force-free Jx and Jy.
;     kb = The variation in energy is multiplied by this value to derive the
;          initial temperature (default = 5.0).
;     converge = convergence criterion in percent.  Smaller means iterate
;                longer (default = 5.0).
;     tfactor = The temperature is reduced by factor tfactor at each iteration.
;               (default = 0.95).
;     /genetic = Use genetic annealing.  Slower, but potentially more
;                robust.
;     noconvergence = on output the index of the pixels that did not
;                     fully converge under the genetic algorithm.  It
;                     the genetic algorithm is not used, or if all
;                     pixels converged, then noconvergence will be -1.
;     /nopotential = do not use the force-free field for a crude initial 
;                    resolution of the ambiguity.  Does not do much as
;                    the initial state of the ambiguity is not very
;                    important. 
;     /get_alpha = Compute a "best" value of alpha for the force-free
;                  field
;     /shear = Use the field-weighted shear angle in the call to
;              get_alpha.  Not reccommended. 
;     trecord = returns with the teperatures used
;     erecord = returns with the energy at each temperature
;     crecord = returns with the specific heat at each temperature after the
;               first (one less element than trecord).
;     /fast = Run as fast as possible; possibly less robust
;             Uses kb = 3.0,
;                  tfactor = 0.90,
;                  /nogenetic
;                  binfactor = 3
;     /slow = Run slowly to get a more robust ambiguity resolution
;             Uses kb = 10.0
;                  tfactor = 0.98
;                  converge = 2.
;                  guard = 2
;     /quiet = Work quietly.
;     /verbose = Work loudly.
;     /plot = Plot the field after every temperature.
;     /novector = do not plot horizontal field vectors.
;     btlength = lenght of horizontal field vectors (def = 1)
;     alpha = force free alpha to use for the computation of dBz/dz.
;     /notty = run non-interactively
;OUTPUTS:
;     jnew = answer (same format as j) with the minimum energy.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Requires a large FOV encompassing all the magnetic flux in the region
;     since the derivative of Bz wrt to z is computed from a linear force
;     free or potential extrapolation.
;
;     The pixels are assumed to be square.
;
;PROCEDURE:
;     See Numerical Recipes 2nd edition for detail of simulated annealing.
;     Type "qq" to quit after the current temperature is complete.
;MODIFICATION HISTORY:
;     T. Metcalf  March 1994
;     1994-09-20  Added Genetic annealing algorithm.  See K. Price, Dr. 
;                 Dobb's Journal, Vol. 19, Issue 11, October 1994, p. 127.
;     1995-11-30  Added skippotential keyword
;     2003-04-11  TRM Added mask keyword.
;     2004-07-22  TRM Added btskip keyword in calls to vmgram.
;     2005-09-07  TRM Added noconvergence keyword.
;-

function janneal,jin,state=state,kb=k,quiet=quiet,plot=plot, $
                converge=converge,tfactor=tfactor,start=start,done=done, $
                TRecord=TRecord,ERecord=ERecord,CRecord=CRecord, $
                bterror=bterror,alpha=alpha,verbose=verbose, $
                get_alpha=get_alpha,nopotential=nopotential, $
                novectors=novectors,jweight=jweight,genetic=genetic, $
                jzonly=jzonly,notty=notty,fast=fast,slow=slow, $
                skippotential=skippotential,guard=guard,dweight=dweight, $
                eweight=eweight,sweight=sweight,btlength=btlength, $
                mask=mask,shear=shear,hweight=hweight, $
                pweight=pweight,gweight=gweight,jpower=jpower, $
                noconvergence=noconvergence,enclose=enclose,dBzdz=dBzdz

   ; Define some constants

   MINRAN = 6.0e-7         ; The minimum random number we need to check
   MINEXP = alog(MINRAN)   ; The smallest exponent which must be computed

   if keyword_set(genetic) then genetic = 1 else genetic = 0

   if genetic then nstates = 10L else nstates = 1L

   start = !stime
   j = jin   ; protect the input

   if n_elements(bterror) NE 1 then bterror = 0.0 else bterror = float(bterror)
   if NOT keyword_set(btlength) then btlength = 2.
   if n_elements(skippotential) NE 1 then $
      skippotential = 0.0 else skippotential = float(abs(skippotential))
   if n_elements(jweight) LE 0 then jweight = 1.0d0
   if n_elements(dweight) LE 0 then dweight = 1.0d0
   if n_elements(eweight) LE 0 then eweight = 0.0d0
   if n_elements(sweight) LE 0 then sweight = 0.0d0
   if n_elements(hweight) LE 0 then hweight = 0.0d0
   if n_elements(shear) LE 0 then shear = 0

   divb,j,divarray,state=state,/reset,alpha=alpha,quiet=quiet,plot=plot, $
        get_alpha=get_alpha,nopotential=nopotential,bterror=bterror, $
        novectors=novectors,jweight=jweight,jzonly=jzonly,potazim=potazim, $
        guard=guard,dweight=dweight,eweight=eweight,sweight=sweight,hweight=hweight, $
        shear=shear,pweight=pweight,gweight=gweight,jpower=jpower,mask=mask, $
        enclose=enclose,dBzdz=dBzdz

   nx = n_elements(j.b_long(*,0))
   ny = n_elements(j.b_long(0,*))
   nx1 = nx-1L
   ny1 = ny-1L
   nxy = nx*ny

   xarr = lindgen(nx,ny) MOD nx
   yarr = lindgen(nx,ny) / nx

   tx0 = lindgen(nxy) MOD nx    ; Used to compute the total energy
   ty0 = lindgen(nxy) / nx
   tx = [[tx0],[tx0+1L],[tx0],[tx0+1L]]
   ty = [[ty0],[ty0],[ty0+1L],[ty0+1L]]
   tlbits = [[replicate(1L,nxy)],[replicate(2L,nxy)], $
           [replicate(4L,nxy)],[replicate(8L,nxy)]]

   ;;btotal = sqrt(j.b_long^2 + j.b_trans^2)
   good = where(abs(j.b_trans) GE bterror,nmag)
   if nmag LE 0 then begin
      message,/info,strcompress('No transverse field above '+strcompress(string(bterror))+' G')
      return,fix_azimuth(j,j.b_azim)
   endif
   if bterror GT 0. then begin
      print
      print,'Neglecting all fields below',strcompress(string(bterror)), $
            ' G (transverse),', $
            strcompress(string(nmag)),' of '+strcompress(string(nxy))+ $
            ' remaining'
      print
   endif
   if n_elements(mask) EQ nxy then begin
      good2 = where(mask[good] NE 0,nmag)
      if nmag LE 0 then begin
         if NOT keyword_set(quiet) then begin
            print
            print,'All points are masked out.'
            print
         endif
         return,fix_azimuth(j,j.b_azim)
      endif else begin
         good = good[good2]
         print
         print,strcompress('After MASK, there are '+string(nmag)+ $
                           ' points of '+string(nxy)+' remaining.')
         print
      endelse
   endif
   if n_elements(potazim) EQ nxy and skippotential GT 0. then begin
      ;good1 = where(abs(arange(j.B_AZIM(good),potazim(good))-potazim(good)) GT $
      ;             skippotential,nmag)
      good1 = where(abs(arange(j.B_AZIM(good),potazim(good))-potazim(good)) GT $
                   skippotential OR abs(j.b_trans(good)) LT bterror*3.0,nmag)
      if nmag LE 0 then begin
         if NOT keyword_set(quiet) then begin
            print
            print,'All points are within '+strcompress(string(skippotential))+ $
                  ' degrees of force-free field'
            print
         endif
         return,fix_azimuth(j,potazim)
      endif
      print
      print,'Neglecting field where azimuth is within ', $
            strcompress(string(skippotential)),' degrees of potential and '
      print,'b_trans is above '+string(bterror*3.0)+ $
            ' ('+strcompress(string(nmag))+' points of '+ $
            strcompress(string(nxy))+' remain)'
      print
      good = good(good1)
   endif
   xmag = tx0(good)
   ymag = ty0(good)
   xmagm1 = xmag-1L
   ymagm1 = ymag-1L
   nmag1 = nmag - 1L

   ; Number of flips to do "at once" with vector arithmetic
   ; nflips << nmag (no order to the flips) but >> 1 (speed)

   ;nflips = (nmag/10L)>3L
   nflips = (nmag/32L)>3L  ; 8 neighbors ==> Only 25% will have adjacent flips
   nf1 = nflips-1L
   nf10 = nflips*10L

   bits = [ [[replicate(1L,nflips,4L)]],[[replicate(2L,nflips,4L)]], $
            [[replicate(4L,nflips,4L)]],[[replicate(8L,nflips,4L)]] ]
   lbits = [[replicate(1L,nflips)],[replicate(2L,nflips)], $
           [replicate(4L,nflips)],[replicate(8L,nflips)]]

   ;print
   ;print,'Done with setup ... ',!stime
   ;print

   ; Measure the variation of the energy from some perturbations to set
   ; the temperature scale.

   ; There are only about 10^7 (2^24) possible numbers from randomu so it
   ; won't work if nmag>10^6 or so.  In this case the ranlong function could
   ; be used to generate long integer random numbers.  ranlong is about 6 
   ; times slower than randomu though.  Alternatively, x0 and y0 could
   ; be addressed by 2 random numbers, ie x0(xpix,ypix)=x0(xpix+nx*ypix).

   ;if nmag GT 16777215L then $
   ;   message,'WARNING: randomu may not work for this many pixels'
   ;if nmag GE 16777215L then $
   if nmag GT 8388607L then $   ; 2^23 - 1
      pix = (long(nmag*randomu(seed,nflips,/double))<nmag1)>0L $
   else $
      pix = (long(nmag*randomu(seed,nflips))<nmag1)>0L
   x0 = xmag(pix)
   y0 = ymag(pix)
   x0m1 = xmagm1(pix)
   y0m1 = ymagm1(pix)
   x = [[x0],[x0m1],[x0],[x0m1]]
   y = [[y0],[y0],[y0m1],[y0m1]]
   dx = [[[x]],[[x+1L]],[[x]],[[x+1L]]]
   dy = [[[y]],[[y]],[[y+1L]],[[y+1L]]]
   dstate = long(total(state(dx,dy)*bits,3))
   energy = total(divarray(dstate XOR lbits,x,y),2) - $
            total(divarray(dstate,x,y),2)

   b1 = lonarr(nflips*nstates,4,4)
   lb1 = lonarr(nflips*nstates,4)
   sselect1 = lonarr(nflips*nstates,4,4)
   sselect2 = lonarr(nflips*nstates)
   for i=0L,nstates-1L do begin
      b1(i*nflips:(i+1L)*nflips-1L,*,*) = bits
      lb1(i*nflips:(i+1L)*nflips-1L,*) = lbits 
      sselect1(i*nflips:(i+1L)*nflips-1L,*,*) = i
      sselect2(i*nflips:(i+1L)*nflips-1L) = i
   endfor
   bits = temporary(b1)
   lbits = temporary(lb1)

   ; Initializations

   qverbose = keyword_set(verbose)
   qnquiet = NOT keyword_set(quiet)
   command = ''
   sxc = size(xcommand) & if sxc(sxc(0)+1) NE 7 then xcommand=''  ; string?
   if n_elements(k) ne 1 then begin
      if keyword_set(fast) then k=3.0 else if keyword_set(slow) then k=10.0 else k=5.0
   endif else k = abs(k)
   if n_elements(converge) NE 1 then begin
      if keyword_set(slow) then converge=2. else converge=5.
   endif else converge=float(abs(converge))
   if n_elements(tfactor) NE 1 then begin
      if keyword_set(fast) then tfactor=0.90 else if keyword_set(slow) then tfactor = 0.98 else tfactor=0.95 
   endif else tfactor=abs(float(tfactor)<0.999)
   if tfactor LE 0.0 then tfactor=0.9
   T = K*float(stdev(energy(1:*)))   ; initial temperature
   ; Stop if T goes below Tstop.  If tfactor=0.9, T/1.e5 gives 110 iterations
   Tstop = double(T)/1.d6
   if keyword_set(slow) then Tstop = double(T)/1.d7
   ;Tstop = 0.0
   energy = j.totaldiv

   states = bytarr(nx,ny,nstates)
   for i=0,nstates-1 do states(*,*,i) = state
   state = temporary(states)

   ; Proceed to the next temperature when naccept flips have been accepted
   ; or when nreconf flips have been attempted.

   naccept = long(10L*nmag*((nstates/3.0)>1.0))
   nreconf = 10L*naccept
   MINCF = (tfactor/0.9)*(nreconf/float(naccept/10.0))/100.0

   nrecord = 0L          ; n_elements(TRecord)-1L
   TRecord = T
   ERecord = energy
   C0 = K*energy/T         ; Sets a scale for the specific heat

   cfactor = 1.0     ; variable to control T changes based on the specific heat
   done = 0

   splicesz = long(sqrt(nxy)/5.0+0.5)>1L  ; max size of a splice

   ; Now do the annealing minimization

   REPEAT begin
      if qnquiet then begin
         print
         print,'Temperature = ',strcompress(string(T),/remove_all),', ', $
               'Energy = ',strcompress(string(ERecord(nrecord)),/remove_all), $
               ', ', $
               'Steps = ',strcompress(string(Cfactor),/remove_all)
         print
      endif
      reconfig = 0L    ; Count number of reconfigurations at this temperature
      accepted = 0L    ; Count accepted reconfigurations
      tenergy = 0.0d0  ; Used to compute mean energy at present temperature
      time0 = systime(1)
      REPEAT begin

         if genetic then begin   ; Attempt to splice from one state to another
            pixsplice = (long(nmag*randomu(seed,nstates))<nmag1)>0L
            other = long((randomu(seed,nstates)*NSTATES)) MOD NSTATES
            for i=0L,nstates-1L do begin  ; Splice
               if other(i) NE i then begin
                  x1 = xmag(pixsplice(i))
                  y1 = ymag(pixsplice(i))
                  xsize = (long((randomu(seed,nstates)*splicesz)+1L))(0) < (nx-1-x1)
                  ysize = (long((randomu(seed,nstates)*splicesz)+1L))(0) < (ny-1-y1)
                  x0 = xarr(x1:x1+xsize,y1:y1+ysize)
                  y0 = yarr(x1:x1+xsize,y1:y1+ysize)
                  nsplice = n_elements(x0)
                  x0 = reform(x0,nsplice)
                  y0 = reform(y0,nsplice)
                  darr = make_array(nsplice,4,4,value=i,/long)
                  oarr = make_array(nsplice,4,4,value=other(i),/long)
                  differ = where(state(x0,y0,darr(*,0,0)) NE state(x0,y0,oarr(*,0,0)),ndiffer)
                  if ndiffer GT 0 then begin
                     x = [[x0],[x0-1],[x0],[x0-1]]
                     y = [[y0],[y0],[y0-1],[y0-1]]
                     dx = [[[x]],[[x+1L]],[[x]],[[x+1L]]]
                     dy = [[[y]],[[y]],[[y+1L]],[[y+1L]]]
                     sbits = [ [[replicate(1L,nsplice,4L)]],[[replicate(2L,nsplice,4L)]], $
                          [[replicate(4L,nsplice,4L)]],[[replicate(8L,nsplice,4L)]] ]
                     dstate = long(total(state(dx,dy,darr)*sbits,3))
                     ostate = long(total(state(dx,dy,oarr)*sbits,3))
                     newenergy = total(divarray(ostate,x,y))
                     oldenergy = total(divarray(dstate,x,y))

                     if (newenergy LE oldenergy) OR $
                        (randomu(seed,1))(0) LT exp((oldenergy-newenergy)/(T*ndiffer)) then begin
                        darr = make_array(nsplice,value=i,/long)
                        oarr = make_array(nsplice,value=other(i),/long)
                        state(x0,y0,darr) = state(x0,y0,oarr)
                     endif
                  endif
               endif
            endfor
         endif

         pix = (long(nmag*randomu(seed,nflips*nstates))<nmag1)>0L

         x0 = xmag(pix)
         y0 = ymag(pix)
         x0m1 = xmagm1(pix)
         y0m1 = ymagm1(pix)

         x = [[x0],[x0m1],[x0],[x0m1]]
         y = [[y0],[y0],[y0m1],[y0m1]]
         dx = [[[x]],[[x+1L]],[[x]],[[x+1L]]]
         dy = [[[y]],[[y]],[[y+1L]],[[y+1L]]]
         dstate = long(total(state(dx,dy,sselect1)*bits,3))
         newenergy = total(divarray(dstate XOR lbits,x,y),2)
         oldenergy = total(divarray(dstate,x,y),2)

         oracle = bytarr(nflips*nstates)   ; starts out all zeroes (no flips)
         good = where(newenergy LT oldenergy,ngood)
         if ngood GT 0 then oracle(good) = 1b
         maybe = where(newenergy GE oldenergy,nmaybe)
         if nmaybe GT 0 then begin
             e_T = (oldenergy(maybe)-newenergy(maybe))/T
             needexp = where(e_T GE MINEXP,nneedexp)   ; MINEXP is negative
             if nneedexp GT 0 then $
                oracle(maybe(needexp)) = $
                   (randomu(seed,nneedexp) LT exp(e_T(needexp)))
         endif

         accept = where(oracle,naccepted)
         if naccepted GT 0 then begin  ; accept
            state(x0(accept),y0(accept),sselect2(accept)) = state(x0(accept),y0(accept),sselect2(accept)) XOR 1b
            accepted = accepted + naccepted
         endif
         reconfig = reconfig + nflips*nstates

         if qverbose then if (reconfig MOD nf10) EQ 0L then begin
            print,'#' + $
                  strcompress(string(reconfig),/remove_all) + $
                  ',  accepted: ' + $
                  strcompress(string(accepted),/remove_all) + $
                  ' ',string(systime(1)-time0),'s'
            time0 = systime(1)
         endif
      endrep UNTIL (reconfig GE nreconf) OR (accepted GE naccept)
      if (NOT keyword_set(notty)) then $
         command = command+strupcase(get_kbrd(0)+get_kbrd(0))

      ; Compute the new energy

      if genetic then begin
         ; state0 is the consensus between all the states
         state0 = byte(0.5+total(state,3)/float(nstates))  ; Average the states
      endif else $
         state0 = reform(state)
      dstate = long(total(state0(tx,ty)*tlbits,2))
      j.totaldiv = total(divarray(dstate,tx0,ty0))

      ; How well are we doing?

      TRecord = [TRecord,T]
      ERecord = [ERecord,j.totaldiv]
      nrecord = nrecord + 1L
      if nrecord GT 2L then begin   ; Compute the Specific Heat
         ; Cfactor is the number of T steps to advance or retard.
         ; When the specific heat is large, the data is at the freezing point
         ; and the temperature steps should be small.
         ; Specific Heat = dE/dT
         ;CRecord = ((ERecord(1:*)-shift(ERecord(1:*),1)) / $
         ;           (TRecord(1:*)-shift(TRecord(1:*),1)))
         CRecord = deriv(TRecord(1:*),Erecord(1:*))/C0
         if CRecord(nrecord-1L) LE 0.0 then Cfactor = 1.0 $
         else Cfactor = (abs(C0/CRecord(nrecord-1L))<1.0)>MINCF
         if qnquiet then begin
            print
            print,'Specific Heat: '
            print,CRecord
            print
         endif
      endif
      if qnquiet then begin
         print
         print,'Accepted ',strcompress(string(accepted),/remove_all), $
               ' of ',strcompress(string(reconfig),/remove_all), $
               ' reconfigurations'
         print
      endif

      ; Plot the current guess

      if qnquiet AND keyword_set(plot) then begin
         flip = where(state0)
         azim = j.b_azim
         if flip(0) GE 0 then azim(flip) = ((azim(flip) + 180.) MOD 360.)
         azim = azim * !dtor
         bxyz,-j.b_trans*sin(azim), $
               j.b_trans*cos(azim), $
               j.b_long, $
               bx1,by1,bz1,/last_transform
         vmgram,{point:j.point,bx:bx1,by:by1,bz:bz1}, $
                 /inc,/xyz,/arr,image=j.i_cont,/quiet, $
                 title=strcompress('Temperature: ' + $
                                   string(T,format="(f18.1)") + ' Energy: ' + $
                                   string(ERecord(nrecord), $
                                          format="(f18.1)")), $
                 novector=novectors,length=btlength,btskip=fix(nx/80.)>0
      endif

      ; Set up for the next iteration

      T = ((tfactor^Cfactor)*T)>0.0     ; Set the new temperature

      if genetic then begin             ; Check if we're done (all flips in noise?)
         totalstate = total(state,3)
         bad = where(totalstate GT 1 and totalstate LT nstates-1,nbad)
         if n_elements(bt_error) LE 0 then $
            if bterror GT 0 then bt_error=bterror else bt_error = bterror(j.b_trans)
         if nbad GT 0 then begin
            bbad1 = where(j.b_trans(bad) GT 1.0*bt_error,nbbad1)
            bbad2 = where(j.b_trans(bad) GT 2.0*bt_error,nbbad2)
            ;done = (float(nbbad2)/float(nbad+nxy-nmag) LE 0.00d0*converge) and $
            ;       (float(nbbad1)/float(nbad+nxy-nmag) LE 0.01d0*converge) and $
            ;       (reconfig GE nreconf)
         endif ;else done=1
         state(*,*,0) = state0  ; Iterate the consensus
      endif

      if (T LE 0.0) then done = 1
      ;if (T LE Tstop) then done=1
      if (T LE Tstop/5.d0) then done=1
      if (accepted EQ 0) OR (strpos(command,'QQ') GE 0) then done = 1
      ;if (accepted/double(reconfig) LE 0.01d0*converge) then done=1
      if ((accepted/double(reconfig) LE 0.01d0*converge) OR (T LE Tstop)) AND (NOT done) then begin
         T = Tstop/10.d0 ;One last iteration at T=0 
         message,/info,'Beginning final cold iteration.'
      endif

   endrep UNTIL done

   ; Update the azimuth to the final values

   if genetic then begin 
      print
      print,strcompress('There are '+string(nbad)+' pixels out of '+string(nmag)+' which did not fully converge')
      if nbad GT 0 then begin
         print,strcompress('Of these, '+string(nbbad1)+' have B_trans > '+string(1.0*bt_error))
         print,'     And, '+strtrim(strcompress(string(nbbad2)+' have B_trans > '+string(2*bt_error)))
      endif else begin
         print,strcompress('Of these, none have B_trans > '+string(bt_error))
      endelse
      print
      noconvergence=bad
  endif else noconvergence=-1

   flip = where(state0)
   if flip(0) GE 0 then j.b_azim(flip) = ((j.b_azim(flip) + 180.) MOD 360.)

   done = !stime
   return,fix_azimuth(j,j.b_azim)

end
