   

function jgenetic_func,n,x
   common jgenetic_private1, divarray, state, tx, ty, tx0, ty0, tlbits, txy, good
   state[good] = nint(x)
   dstate = long(total(state[tx,ty]*tlbits,2))
   energy = total(divarray[dstate,tx0,ty0])
   return,-energy  ; negative since pikaia is a maximizer not a minimizer
end

;+
;NAME:
;     JGENETIC
;PURPOSE:
;     Resolves 180 degree ambiguity in vector magnetic field data.
;     Implements ipikaia genetic algorithm for the simultaneous minimization of the
;     divergence of B and the vertical electric current.
;CATEGORY:
;CALLING SEQUENCE:
;     jnew = jgenetic(j)
;INPUTS:
;     j = Magnetic field structure whose divergence\current is minimized.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     skippotential = Pixel with azimuth this close to potential are not
;                     included in the iteration to save time.  Degrees.
;                     default = 0.0.  Only applies to pixels with B_trans
;                     larger than 3*bterror.
;              CAUTION:  Don't set this too high:  skipping points can confuse
;                        the algorithm.
;     bterror = smallest transverse magnetic field (absolute value) to consider.
;              CAUTION:  Don't set this too high: skipping points can confuse
;              the algorithm. (default = 0.0)
;     mask = Array with a 1 to indicate that the azimuth for that
;            pixel is to be tested or a 0 to indicate that it is
;            already set and should not be tested.
;     jweight = weighting of the current minimization relative to the 
;               divergence minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def
;               = 1.0.
;     sweight = weighting of the sheath current minimization relative to the 
;               divergence minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def
;               = 1.0.
;     dweight = weighting of the divergence minimization relative to the 
;               current minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def = 1.0.
;     eweight = weighting of the entropy maximization relative to the 
;               current minimization.  May be scalar or an array with
;               the same dimension as the magnetic field arrays.  Def = 0.0.
;     hweight = weighting of the current helicity minimization
;               relative to the current and divergence
;               minimization. May be scalar or an array with the same
;               dimension as the magnetic field arrays.  Def = 0.0. 
;     /jzonly = Minimize only Jz, do not use the force-free Jx and Jy.
;     /nopotential = do not use the force-free field for a crude initial 
;                    resolution of the ambiguity.  Does not do much as
;                    the initial state of the ambiguity is not very
;                    important. 
;     /get_alpha = Compute a "best" value of alpha for the force-free
;                  field
;     /shear = Use the field-weighted shear angle in the call to
;              get_alpha.  Not reccommended. 
;     /quiet = Work quietly.
;     /verbose = Work loudly.
;     /plot = Plot the energy vs. generation.
;     /novector = do not plot horizontal field vectors.
;     btlength = lenght of horizontal field vectors (def = 1)
;     alpha = force free alpha to use for the computation of dBz/dz.
;     /notty = run non-interactively
;OUTPUTS:
;     jnew = answer (same format as j) with the minimum energy.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Requires a large FOV encompassing all the magnetic flux in the region
;     since the derivative of Bz wrt to z is computed from a linear force
;     free or potential extrapolation.
;
;     The pixels are assumed to be square.
;
;PROCEDURE:
;MODIFICATION HISTORY:
;     T. Metcalf  janneal.pro --> jgenetic.pro
;-

function jgenetic,jin,state=state,quiet=quiet,plot=plot, $
                bterror=bterror,alpha=alpha,verbose=verbose, $
                get_alpha=get_alpha,nopotential=nopotential, $
                novectors=novectors,jweight=jweight, $
                jzonly=jzonly,notty=notty, $
                skippotential=skippotential,guard=guard,dweight=dweight, $
                eweight=eweight,sweight=sweight,btlength=btlength, $
                mask=mask,shear=shear,hweight=hweight,ngenerations=ngenerations, $
                pweight=pweight,gweight=gweight,jpower=jpower,enclose=enclose, $
                dBzdz=dBzdz

   common jgenetic_private1, divarray, sstate, tx, ty, tx0, ty0, tlbits, txy, good

   ; Define some constants

   start = !stime
   j = jin   ; protect the input

   if n_elements(bterror) NE 1 then bterror = 0.0 else bterror = float(bterror)
   if NOT keyword_set(btlength) then btlength = 2.
   if n_elements(skippotential) NE 1 then $
      skippotential = 0.0 else skippotential = float(abs(skippotential))
   if n_elements(jweight) LE 0 then jweight = 1.0d0
   if n_elements(dweight) LE 0 then dweight = 1.0d0
   if n_elements(eweight) LE 0 then eweight = 0.0d0
   if n_elements(sweight) LE 0 then sweight = 0.0d0
   if n_elements(hweight) LE 0 then hweight = 0.0d0
   if n_elements(shear) LE 0 then shear = 0

   divb,j,divarray,state=state,/reset,alpha=alpha,quiet=quiet,plot=plot, $
        get_alpha=get_alpha,nopotential=nopotential,bterror=bterror, $
        novectors=novectors,jweight=jweight,jzonly=jzonly,potazim=potazim, $
        guard=guard,dweight=dweight,eweight=eweight,sweight=sweight,hweight=hweight, $
        shear=shear,pweight=pweight,gweight=gweight,jpower=jpower,mask=mask, $
        enclose=enclose,dBzdz=dBzdz

   nx = n_elements(j.b_long(*,0))
   ny = n_elements(j.b_long(0,*))
   nxy = nx*ny

   xarr = lindgen(nx,ny) MOD nx
   yarr = lindgen(nx,ny) / nx

   tx0 = lindgen(nxy) MOD nx    ; Used to compute the total energy
   ty0 = lindgen(nxy) / nx
   tx = [[tx0],[tx0+1L],[tx0],[tx0+1L]]
   ty = [[ty0],[ty0],[ty0+1L],[ty0+1L]]
   tlbits = [[replicate(1L,nxy)],[replicate(2L,nxy)], $
           [replicate(4L,nxy)],[replicate(8L,nxy)]]

   good = where(abs(j.b_trans) GE bterror,nmag)
   if nmag LE 0 then begin
      message,/info,strcompress('No transverse field above '+strcompress(string(bterror))+' G')
      return,fix_azimuth(j,j.b_azim)
   endif
   if bterror GT 0. then begin
      print
      print,'Neglecting all fields below',strcompress(string(bterror)), $
            ' G (transverse),', $
            strcompress(string(nmag)),' of '+strcompress(string(nxy))+ $
            ' remaining'
      print
   endif
   if n_elements(mask) EQ nxy then begin
      good2 = where(mask[good] NE 0,nmag)
      if nmag LE 0 then begin
         if NOT keyword_set(quiet) then begin
            print
            print,'All points are masked out.'
            print
         endif
         return,fix_azimuth(j,j.b_azim)
      endif else begin
         good = good[good2]
         print
         print,strcompress('After MASK, there are '+string(nmag)+ $
                           ' points of '+string(nxy)+' remaining.')
         print
      endelse
   endif
   if n_elements(potazim) EQ nxy and skippotential GT 0. then begin
      ;good1 = where(abs(arange(j.B_AZIM(good),potazim(good))-potazim(good)) GT $
      ;             skippotential,nmag)
      good1 = where(abs(arange(j.B_AZIM(good),potazim(good))-potazim(good)) GT $
                   skippotential OR abs(j.b_trans(good)) LT bterror*3.0,nmag)
      if nmag LE 0 then begin
         if NOT keyword_set(quiet) then begin
            print
            print,'All points are within '+strcompress(string(skippotential))+ $
                  ' degrees of force-free field'
            print
         endif
         return,fix_azimuth(j,potazim)
      endif
      print
      print,'Neglecting field where azimuth is within ', $
            strcompress(string(skippotential)),' degrees of potential and '
      print,'b_trans is above '+string(bterror*3.0)+ $
            ' ('+strcompress(string(nmag))+' points of '+ $
            strcompress(string(nxy))+' remain)'
      print
      good = good(good1)
   endif

   sstate = state    ; sent to jgenetic_func via the common block
   state0 = state[good]
   print,'Intitial energy: ',-(jgenetic_func(nmag,state0)),total(divarray(0,*,*))

   if NOT keyword_set(ngenerations) then ngenerations = 1000
   npopulation = 80
   pikaia,'jgenetic_func',nmag,state0,energy,pop=npopulation,generations=ngenerations, $
          replacement=2,variable=0,verbose=verbose,decimalplaces=1, $
          fitwatch=plot,creepmutation=1,twopoint=1,testinput=state0, $
          muteprob=0.005

   print,'Final energy: ',-(jgenetic_func(nmag,state0))
   state[good] = nint(state0)

   ; Update the azimuth to the final values

   flip = where(state)
   if flip(0) GE 0 then j.b_azim(flip) = ((j.b_azim(flip) + 180.) MOD 360.)

   done = !stime
   return,fix_azimuth(j,j.b_azim)

end
