;+

function current3,mag,_extra=_extra,nviews=nviews,scale=scale, $
                  binfactor=binfactor,novector=novector,mask=maskin, $
                  nopotential=nopotential,alpha_best=alpha_best, $
                  get_alpha=get_alpha,enclose=enclose, $   ; don't allow these in _extra
                  sfactor=sfactor,azshearlimit=azshearlimit, $
                  subazshearlimit=subazshearlimit,quiet=quiet,noplot=noplot, $
                  guard=guardin,resolvesubviews=resolvesubviews

;NAME:
;     CURRENT3
;PURPOSE:
;     Resolve the 180 degree ambiguity in vector magnetograms by
;     breaking up the gram into subgrams and computing dBzdz in each
;     subgram.  The idea is to model the variation of alpha over the
;     gram by doing many subgrams and finding the best alpha subgram
;     by subgram and then patching the results back together.
;CATEGORY:
;CALLING SEQUENCE:
;     j = current3(mag)
;INPUTS:
;     mag = vector magnetic field structure to be resolved
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS:
;     NViews = Number of views to divide gram into in each dimension.
;              If nviews is a 2 element array then it is
;              [xnviews,ynviews].  If it is a scalar, then xnviews and
;              ynviews are both set to this scalar.
;              Setting nviews implies an xnviews X ynviews grid and
;              ((xnviews-1)*sfactor+1)*((ynviews-1)*sfactor+1) calls
;              to current2.  The larger this number is, the smaller
;              the subgrams.  The default is to divide into roughly 60
;              arcsec pieces. But, the default is expanded if the
;              sub-grams are not at least 8x8 pixels. 
;     scale = scale in arcsec used to compute the default value of
;             xnviews and ynviews.  Def = 60.0.  Ignored if nviews is
;             passed in directly.  The idea is to set this to the
;             spatial scale of variations in alpha.  60" is the
;             default since this is a typical sunspot size.
;     sfactor = number of steps required to shift by a full subgram.
;               The default is 2 so that each subgram is shifted by
;               1/2 its size at each step.  With sfactor=3, it is
;               shifted by 1/3 its size, etc.  The routine slows down
;               rapidly as this number is increased, but may be more
;               robust.  Must be an integer GE 2.
;     subazshearlimit = azshearlimit for the subviews.  Default = 180.
;     azshearlimit = azshearlimit for the final, full gram.  Default =
;                    10.
;     /resolvesubviews = call current2 on each subview and "vote" to
;                        get the best azimuth.  Much slower!  This is
;                        the old default behavior.
;     alpha_best = best alpha for gram as a whole.  Computed with
;                  get_alpha if not set.
;
;     Accepts all the keywords that can be passed to current2.pro.
;OUTPUTS:
;     j = vector magnetic field structure with the 180 degree
;         ambiguity resolved.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     And you thought current2 was slow!  current3 is 2 to 5 times
;     slower. 
;PROCEDURE:
;     Breaks the magnetogram into smaller chunks and computes dBzdz
;     repeatedly.  Optionally calls current2 on each subview as well.
;     The "window" is shifted by 1/sfactor of it's size
;     at each step.  
;
;     If current2 is called for each subview, then. at the end, most 
;     pixels will have been considered
;     from several fields-of-view and the majority consensus on the
;     orientation of each vector is accepted.  If there is no
;     consensus, then the azimuth closest to the FFF is used.  The
;     /get_alpha keyword is always set in the calls to current2 since
;     this routine would not make much sense unless trying to allow a
;     variable alpha over the FOV.  The final step is to glue the
;     pieces back together by running current2 on the full FOV with
;     but only allow flips for those pixels that did not have a
;     unanimous vote. 
;EXAMPLE:
;     Here's a typical call that seems to work quite well.  The bin=3
;     is applied to the subgrams, but not to the final full FOV call
;     to current2 and is used to speed things up (may be less robust
;     though)..  
;     j = current3(mag,bin=3,/plot,btlen=4,sfact=3,bterror=bterror(mag,bins=1.), $
;                  guard=4)
;        ... or ...
;     j =
;     current3(mag,bin=8,azshear=2.,/plot,sfact=3,bterror=bterror(mag,bins=1.)/3., $
;              guard=4,/slow)
;     ; ... go home and get some sleep ...
;     vmgram,j,/inc,/curr,/novect,len=2,/xyz
;MODIFICATION HISTORY:
;     T. Metcalf 2003-Apr-17
;     2005-Sep-16 TRM Set weight to zero for 2*binsize pixels along
;                     the edges instead of 1*binsize.  Added
;                     subazshearlimit keyword.  Make sure that xnviews
;                     and ynviews are at least 2 in the default
;                     calculation. 
;     2005-Sep-21 TRM Added dBzdz calculation and /resolvesubviews
;                     keyword. The new default is to compute dBzdz in
;                     the subviews, but not to call current2 for each
;                     subview. 
;-

if n_elements(binfactor) LE 0 then binfactor = 1
if n_elements(azshearlimit) LE 0 then azshearlimit = 10.0
if n_elements(subazshearlimit) LE 0 then subazshearlimit = 180.0
if NOT keyword_set(guardin) then guard=1 else guard=guardin

; Get pointing information

p = double(mag.point.p)
b0 = double(mag.point.b0)
if tag_index(mag,'CMD',/quiet) GE 0 then lc = double(mag.cmd)*!dtor $
else lc = double(mag.point.cmd)
if tag_index(mag,'LATITUDE',/quiet) GE 0 then bc = double(mag.latitude)*!dtor $
else bc = double(mag.point.lat)
pix_size = double(mag.point.pix_size)
if n_elements(pix_size) EQ 1 then pix_size = [pix_size[0],pix_size[0]]

smag = size(mag.b_long)
nx = long(smag[1])
ny = long(smag[2])

if n_elements(maskin) EQ nx*ny then mask = maskin $
else begin
   if keyword_set(maskin) then message,/info,'Mask has the wrong dimension: ignoring.'
   mask = 1b+bytarr(nx,ny)   ; test all points
endelse

if keyword_set(nviews) then begin
   if n_elements(nviews) EQ 2 then begin
      xnviews = nviews[0]
      ynviews = nviews[1]
   endif else begin
      xnviews = nviews[0]   ; Only one element
      ynviews = nviews[0]
   endelse
endif else begin
   multiplier = 1.0
;   REPEAT begin
      ; 60" subgrams are the default, but get at least 8x8 
      if NOT keyword_set(scale) then scale = 60.0
      message,/info,strcompress('Scale set to '+string(scale*multiplier))
      xnviews = long(floor((nx*pix_size[0])/(scale*multiplier)))>1L
      xnviews = (xnviews < long((nx/8.)+0.5)) > 1L  ; Subgram at least 8x8
      ynviews = long(floor((ny*pix_size[1])/(scale*multiplier)))>1L
      ynviews = (ynviews < long((ny/8.)+0.5)) > 1L  ; Subgram at least 8x8
;      if xnviews GT 1 or ynviews GT 1 or  $
;         xnviews GE long((nx/8.)+0.5) or  $
;         ynviews GE long((ny/8.)+0.5) then done=1 $
;      else begin
;         done=0
;         multiplier=multiplier*0.95
;       endelse                  ;     endelse
;   endrep UNTIL done
endelse
message,/info,strcompress('Using '+string(xnviews)+' X '+string(ynviews)+' grid.')
if xnviews EQ 1 and ynviews eq 1 then $
   message,/info,'WARNING: xnviews and ynviews are both 1.  '+ $
                 'In this case, current3 adds no value to current2.  ' + $
                 'You should reduce the scale parameter or use current2.'

; Divide into nviews X nviews grid of views

xbinsize = nx/xnviews
ybinsize = ny/ynviews

; But the view is only shift 1/sfactor step each time.
; For sfactor=2, shift 1/2 the subgram at each step
; so there are (2*nviews-1)x(2*nviews-1)(nviews-1) calls to current2.

if NOT keyword_set(sfactor) then sfactor = 2L
sfactor = long(sfactor) > 1L  ; must be integer type GE 1
sfactor = (sfactor < xbinsize) < ybinsize

xshift = xbinsize/sfactor
yshift = ybinsize/sfactor

nxbin = nx / xshift
nybin = ny / yshift

azimuth = mag.b_azim

; Get orientation of the FF field

if n_elements(alpha_best) LE 0 then $
   alpha_best = get_alpha(mag,btcut=bterror,quiet=quiet,/resolve, $
                          noplot=noplot, window=30, $
                          shear=shear, $
                          alimit=0.10*(mean(mag.point.pix_size)/0.55), $
                          astep=0.001*(mean(mag.point.pix_size)/0.55), $
                          guard=guard<4)

pot = fff(b0=b0,pang=p,cmd=lc,lat=bc,mag.b_long,/image,pixel=pix_size, $
          alpha=alpha_best,/quiet,guard=guard) ; can't set /enclose for LOS field
pxi = pot(*,*,0)  ; These are image coordinates
pyi = pot(*,*,1)
pzi = pot(*,*,2)
check = check_math(0,1)
   pa_azimuth = atan(-pxi,pyi)/!dtor  ; Az is CCW from North
check = check_math(1,0)
;fixa = where(abs(arange(mag.B_AZIM,pa_azimuth) - pa_azimuth) GT 90.,count)

dBzdz_best = dbzdz_fff(mag,alpha_best,_extra=_extra,guard=guard,enclose=enclose, $
                       nopotential=0,quiet=quiet)

; Set up the array to hold the vote.  Start with the FF field.
azflip = intarr(nx,ny)  ; Stores the vote on whether or not to flip
azflip[*] = 0

unanimous = intarr(nx,ny)  ; keeps track of the value of a unanimous vote
unanimous[*] = 0
dBzdz_full = fltarr(nx,ny) ; must be zeroed
dBzdz_nval = lonarr(nx,ny) ; must be zeroed

; Loop through the various views and run current2 on each
for i=0L,nxbin-sfactor do begin
   for j=0L,nybin-sfactor do begin
      message,/info,strcompress('Starting ('+string(i)+','+string(j)+') of ('+ $
                                  string(nxbin-sfactor)+','+string(nybin-sfactor)+')')
      xrange = [i*xshift,(i*xshift+xbinsize-1L)<(nx-1L)]
      yrange = [j*yshift,(j*yshift+ybinsize-1L)<(ny-1L)]
      subnx = xrange[1]-xrange[0]+1
      subny = yrange[1]-yrange[0]+1

      weight = intarr(subnx,subny) ; The spatial weighting (must be integer type)
      weight[*] = 1
      weight[0:((binfactor*2-1)<(subnx-1)),*] = 0  ; Ignore the edges
      weight[*,0:((binfactor*2-1)<(subny-1))] = 0
      weight[((subnx-binfactor*2)>0):(subnx-1),*] = 0
      weight[*,((subny-binfactor*2)>0):(subny-1)] = 0

      ; get alpha for this subview
      loadct,0
      alpha = get_alpha(mag,btcut=bterror,quiet=quiet,/resolve, $
                        noplot=noplot, window=30, $
                        shear=shear, $
                        alimit=0.10*(mean(mag.point.pix_size)/0.55), $
                        astep=0.001*(mean(mag.point.pix_size)/0.55), $
                        guard=guard<2, $
                        xrange=xrange,yrange=yrange)  ; xrange and yrange specify the subview

      ; Use the whole FOV to get dBzdz at this alpha.  do not 
      ; set /enclose here.
      dBzdz = dbzdz_fff(mag, alpha, nopotential=nopotential,_extra=_extra, $
                        jresolved=jresolved,guard=guard,quiet=quiet)
      dBzdz_full[xrange[0]:xrange[1],yrange[0]:yrange[1]] = $
         dBzdz_full[xrange[0]:xrange[1],yrange[0]:yrange[1]]+dBzdz[xrange[0]:xrange[1],yrange[0]:yrange[1]]
      dBzdz_nval[xrange[0]:xrange[1],yrange[0]:yrange[1]] = $
         dBzdz_nval[xrange[0]:xrange[1],yrange[0]:yrange[1]] + 1L

      jsmall =  substruc(jresolved,xrange[0],xrange[1],yrange[0],yrange[1])

      if keyword_set(resolvesubviews) then begin

         ; Resolve ambiguity for this alpha in this subview
         ; do not set /enclose here.
         jj = current2(jsmall,dBzdz=dBzdz[xrange[0]:xrange[1],yrange[0]:yrange[1]], $
                       _extra=_extra,alpha=alpha,/nopotential,quiet=quiet, $
                       binfactor=binfactor,novector=novector,azshearlimit=subazshearlimit, $
                       mask=mask[xrange[0]:xrange[1],yrange[0]:yrange[1]],guard=guard)
         azim = azimuth[xrange[0]:xrange[1],yrange[0]:yrange[1]]
         dAzim = abs(arange(jj.b_azim,azim) - azim)
         ; flip is -1 for no flip and +1 for flip
         flip = (fix(0.5 + (dAzim / 180.0)) < 1)*2 - 1
         ; Now cast a vote on a flip or a no flip
         azflip[xrange[0]:xrange[1],yrange[0]:yrange[1]] = $
            azflip[xrange[0]:xrange[1],yrange[0]:yrange[1]] + flip*weight
         ; Keep a record of what a unanimous vote would be
         unanimous[xrange[0]:xrange[1],yrange[0]:yrange[1]] = $
            unanimous[xrange[0]:xrange[1],yrange[0]:yrange[1]] + weight
     endif else vmgram,jsmall,/inc,btskip=fix(max([subnx,subny])/80.)>0,/arr,len=3,/xyz,/quiet
   endfor
endfor

; If the consensus is to flip, then azflip will be positive
; If the consensus is to not flip, then azflip will be negative
; If there was no consensus, then azflip will be zero

flip = where(azflip GT 0,nflip)
unsure = where(azflip EQ 0,nunsure)

; If the vote is to flip, then do so
if nflip GT 0 then mag.b_azim[flip] = (mag.b_azim[flip] + 180.) MOD 360.

; If the vote was a tie, use the azimuth closest to the FFF
if nunsure GT 0 then begin
   print,strcompress('*** There are '+string(nunsure)+ $
                     ' pixels that are uncertain')
   ; Fix the azimuthal angle from the potential field.
   fixa = where(abs(arange(mag.B_AZIM[unsure],pa_azimuth[unsure]) - $
                    pa_azimuth[unsure]) GT $
                90.,count)
   if (count gt 0) then begin
      mag.B_AZIM[unsure[fixa]] = mag.B_AZIM[unsure[fixa]] + 180.
      mag.B_AZIM[unsure[fixa]] = mag.B_AZIM[unsure[fixa]] MOD 360.
   endif
endif

dBzdz_full = f_div(dBzdz_full,dBzdz_nval)  ; mean dBzdz

bad = where(dBzdz_nval LE 0,nbad)
if nbad GT 0 then dBzdz_full[bad]=dBzdz_best[bad]

; Finally, run current2 one last time over the whole FOV to get the
; uncertain areas right. Hopefully there are not too many uncertain
; pixels and this will go fast.  The following tests all areas which
; are not unanimous. 

; if /resolvesubviews is not set, then unanimous is 0 everywhere so all
; pixels will be resolved.

bad = where(((abs(azflip) NE unanimous) OR (unanimous LE 1)),nbad)
if nbad GT 0 then begin
   azmask = bytarr(nx,ny)
   azmask[*] = 0b
   azmask[bad] = 1b   ; Only these points are tested
   if keyword_set(resolvesubviews) then np = 1 else np = 0
   ;if keyword_set(enclose) then ga=0 else ga=1
   jj = current2(fix_azimuth(mag,mag.b_azim),mask=azmask, $
                 alpha = alpha_best, $
                 binfactor=binfactor,_extra=_extra, $
                 nopotential=np,get_alpha=0,enclose=enclose, $
                 azshearlimit=azshearlimit,guard=guard, $
                 dBzdz=dBzdz_full)
   mag.b_azim = jj.b_azim
endif

; Add a comment
icstr = 0L
REPEAT begin
   cstr = strtrim(strcompress('COMMENT'+string(icstr),/remove_all),2)
   icstr = icstr + 1L
endrep UNTIL tag_index(mag,cstr,/quiet) LT 0 OR icstr GT 10000L
if icstr LE 10000L then begin
   mag = add2str(mag,cstr,strcompress('CURRENT3: 180 Ambig. Resolved: '+ $
                 strmid(!stime,0,17)+ $
                 ' sfac,nxv,nyv: '+ $
                 string(sfactor)+' '+string(xnviews)+' '+string(ynviews)),/quiet)
endif


; Compute Bx, By, and Bz and then return after calculating J

return,fix_azimuth(mag,mag.b_azim)

end
