
;+

function ffenergy, j, error, alpha=alphain, guard=guardin, $
                   blong=blong, extrapolate=extrapolate,step=stepin, $
                   mask=mask,ubound=ubound,fx=fx,fy=fy,fz=fz,fp=fp, $
                   randomize=randomize,noprogress=noprogress, $
                   tx=tx,ty=ty,tz=tz, penergy=penergy, label=label, $
                   errfx=errfx, errfy=errfy, errfz=errfz, errfp=errfp, $
                   wheatland=wheatland,wenergy=wenergy,montecarlo=montecarlo, $
                   x0=x0in,y0=y0in,quiet=quiet

;NAME:
;     FFENERGY
;PURPOSE:
;     Computes the energy of a force-free field extrapolation 
;CATEGORY:
;CALLING SEQUENCE:
;     energy = ffenergy(j,error)
;INPUTS:
;     j = magnetic field structure from bfits.pro.  j must have the 180 degree
;         ambiguity reloved, unless both the /blong and /extrapolate keywords
;         are set.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     /extrapolate = Do a force free extrapolation and calculate the energy of
;                    the extrapolated field.  If /extrapolate is not set, then
;                    calculate the energy of the measured field.  If the field
;                    is not force-free, this energy will not mean much.
;     alpha = the force-free parameter to use.  Default = 0.0 (potential
;             field).  Only used if /extrapolate or /wheatland is set.
;     guard = the size of the guard band in the extrapolation.  Default = 2.
;             Only used of /extrapolate is set.  Passed through to fff.pro.
;     /blong = Use the longitudinal field as the boundary condition.  The
;              default is to use the vertical field.  Only used if
;              /extrapolate is set.
;     /randomize = randomize the field using the magnitude of the
;                  field errors.  Useful for a monte carlo error
;                  estimation. 
;     /montecarlo = Repeat the energy calculation for many origins of
;                   the coordinate system.  This is not very useful
;                   anymore since Wheatland derived an analytic
;                   expression for the energy error. 
;     step = integer giving the step size when calculating the energy for
;            different coordinate systems.  Default = 1 which uses every
;            possible coordinate system.  step=2 would skip every other point,
;            etc.   Higher values are less accurate, but faster.  Set step GE
;            nx,ny to avoid the multiple coordinate systems.  Ignored
;            unless /montecarlo is set.
;     mask = Index array giving the pixels to include in the energy 
;            calculation.  don't use this as it does not make physical 
;            sense to exclude some points from the energy calculation.
;     /ubound = Calculate an upper bound to the energy ala Aly 1984.  If this
;               keyword is set, the actual energy is not computed and not
;               returned.  The returned value is only the upper bound.
;     /wheatland = apply the Wheatland alpha correction to the energy
;                  estimate.  Ignored if /extrapolate or /ubound is
;                  set.  The value of the correction depends on alpha
;                  so alpha also needs to be set to compute the
;                  Whealand correction to the energy.
;     wenergy = returns with the value of the wheatland energy corection.
;     fx,fy,fz,fp = outputs with Lorentz force integrals and the
;                   pressure integral 
;     errfx,errfy,errfz,errfp = uncertainty in Lorentz force integrals
;     tx,ty,tz = outputs with Lorentz torque integrals
;     penergy = outputs with potential field energy from Aly
;               (1989). DOES NOT WORK.  DO NOT USE.
;     /noprogress = Do not print progress status
;     /quiet = work quietly
;OUTPUTS:
;     energy = total energy of the magnetic field in ergs
;     error = estimated error on the energy
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     - The calculation assumes that all the flux is with the field-of-view
;     - mag.point.pix_size assumed to be in arc seconds
;     - mag.bx, mag.by, mag.bz must exist with 180 ambiguity resolved
;PROCEDURE:
;     Uses the magnetic virial theorem. If the field is not force-free, 
;     the computed energy will not mean much.
;
;     To calculate the energy of the observed field (must be force-free!) use:
;
;          ffenergy(j)
;
;     To calculate the potential field energy from the virial theorem use
;
;          ffenergy(j,/extrapolate).
;
;     If the 180 deg ambiguity has not been resolved, then the only energy
;     calculation which will work is an extrapolation from b_long:
;
;          ffenergy(j,/blong,/extrapolate,alpha=whatever)
;
;MODIFICATION HISTORY:
;     T. Metcalf  1998-Mar-18
;                 1999-Apr-19 - Added coordinate transformations to account for
;                               projection effects.
;                             - The default is now to compute the energy of the
;                               observed field.  To compute the energy of the
;                               extrapolation use the /extraploate keyword.
;                 2001-Jul-03 - Changed the area of integration from dx*dy 
;                               to dA which is more accurate.  The dx*dy did
;                               not include a "parallelogram" factor.
;                 2004-Nov-30 - Fixed indexing bug in x0 and y0.
;     G. Barnes   2004-Nov-23 - Added option to also calculate the net magnetic
;                               torque on the system, the potential energy,
;                               using expression from Aly (1989) and uncertainties
;                               in the net force.
;     T. Metcalf  2005-Jan-11 - Merged Metcalf's and Barnes' changes.
;     T. Metcalf  2005-May-26 - Added Wheatland correction and
;                               corrected the sign error in the
;                               forces.
;     T. Metcalf  2005-Jul-06 - Fixed typo in fxden equation and remap
;                               jz onto the field grid (0.5 pixel
;                               shift) in the Wheatland calculation. 
;-

   common ffenergy_private,seed  ; make seed persistent

   if n_elements(alphain) NE 1 then alpha = 0.0 else alpha = float(alphain)
   if n_elements(guardin) NE 1 then guard = 2 else guard = float(guardin)
   if n_elements(stepin) NE 1 then step = 1L else step = long(stepin)

   if NOT keyword_set(label) then label = 'ffenergy'

   p = double(j.point.p)
   b0 = double(j.point.b0)
   if tag_index(j,'CMD') LT 0 OR tag_index(j,'LATITUDE') LT 0 then begin
      latlon = get_latlon(mag,/add)
   endif

   lc = double(j.cmd)*!dtor           ; cmd, latitude assumed to be in degrees
   bc = double(j.latitude)*!dtor      ; Convert to radians

   cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
   cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
         - COS(bc)*COS(b0)*SIN(p) 
   cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
   cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
         + COS(bc)*COS(b0)*COS(p)

   ;dxidX = cxx   ; At pixel corners, not centers
   ;dyidX = cyx
   ;dxidY = cxy
   ;dyidY = cyy

   ; the determinant is equal to cosa (below)
   determ = cxx*cyy-cyx*cxy  ; Get image to helio transformation
   icxx = cyy/determ
   icxy = -cxy/determ
   icyx = -cyx/determ
   icyy = cxx/determ

   dXdxi = icxx
   dYdxi = icyx
   dXdyi = icxy
   dYdyi = icyy

   pix_size = double(j.point.pix_size)
   if n_elements(pix_size) EQ 1 then pix_size = [pix_size(0),pix_size(0)]

   bsize = size(j.b_long)
   nx = bsize(1)
   ny = bsize(2)

   berr = get_berrs(j)

   ; Compute the force-free field.

   if keyword_set(extrapolate) then begin
      if keyword_set(blong) then begin
         if keyword_set(randomize) then $
            err=randomn(seed,nx,ny)*berr.sig_blong $
         else err = 0.0
         pot = fff(b0=b0,pang=p,cmd=lc,lat=bc,j.b_long+err,/image, $
                   pixel=pix_size, $
                   alpha=alpha,/quiet,guard=guard)
         pxi = pot(*,*,0)  ; These are image coordinates
         pyi = pot(*,*,1)
         pzi = pot(*,*,2)
         bxyz,pxi,pyi,pzi,px,py,pz,latitude=bc,cmd=lc,p=p,b0=b0
      endif else begin
         if keyword_set(randomize) then $
            err=randomn(seed,nx,ny)*berr.sig_bzh $
         else err = 0.0
         pot = fff(j.bz+err,/normal,b0=b0,pang=p,cmd=lc,lat=bc, $
                   pixel=pix_size, $
                   alpha=alpha,bx=j.bx,by=j.by,guard=guard,/quiet)
         px = pot(*,*,0)  ; These are heliographic coordinates
         py = pot(*,*,1)
         pz = pot(*,*,2)
      endelse
   endif else begin
      px = j.bx  ; If these don't exist, then the 180 deg ambiguity was not
      py = j.by  ; resolved and the energy calculation won't work anyway.
      pz = j.bz
      if keyword_set(randomize) then begin
         px = px + randomn(seed,nx,ny)*berr.sig_bxh ;berr.err_bxh
         py = py + randomn(seed,nx,ny)*berr.sig_byh ;berr.err_byh
         pz = pz + randomn(seed,nx,ny)*berr.sig_bzh ;berr.err_bzh
      endif
   endelse

   psize = size(pz)
   nx = psize(1)
   ny = psize(2)

   if NOT keyword_set(mask) then begin
      mask = lindgen(nx,ny)
   endif
   umask = bytarr(nx,ny)
   umask[mask] = 1b
   umask = where(umask NE 1b,numask)

   x = (lindgen(nx,ny) MOD nx)*pix_size(0)          ; image pixel coordinates in arcsec
   y = (lindgen(nx,ny) / nx)*pix_size(1) 
   bpxd = dXdxi*x + dXdyi*y                         ; Heliographic pixel coordinates
   bpyd = dYdxi*x + dYdyi*y
   ;jpxd = dXdxi*(x+0.5) + dXdyi*(y+0.5)
   ;jpyd = dYdxi*(x+0.5) + dYdyi*(y+0.5)

   ; Redefine x,y to be heliographic coordinates in cm.

   x = bpxd*7.27e7  ; pix_size assumed to be in arc seconds
   y = bpyd*7.27e7

   cosa = SIN(b0)*SIN(bc) + COS(b0)*COS(bc)*COS(lc) ;ang. dist. from sun centre
   offlimb = where(cosa LE 0.0,nofflimb)
   if nofflimb GT 0 then begin
      message,/info,'Warning: Ignoring points off the limb'
      mask = kill_index(mask,offlimb)
      cosa[offlimb] = 1.0 ; will be ignored anyway, can set to anything
   endif

   ;dx = x - shift(x,1,0) & dx[0,*] = dx[1,*]
   ;dy = y - shift(y,0,1) & dy[*,0] = dy[*,1]
   ;dA = dx*dy

   dA = (pix_size[0]*pix_size[1]*7.27e7*7.27e7)/cosa  ; Pixel areas

   ; energy integral in ergs

   ; Do the energy integral for all possible coordinate systems to get a good
   ; mean and standard deviation.  If the field were really force free, every
   ; coordinate system would give the same answer.  Also, the different weighting
   ; gives a good estimate of the error

   if numask GT 0 then dA[umask] = 0.0  ; easy way to turn off pixels

   energy = [-1]
   if keyword_set(ubound) then begin
      pz2 = pz^2
      pz2dA = pz2*dA
      integral1 = total(pz2dA)
   endif
   pzdA = pz*dA

   nxy = long(nx)*long(ny)
   nxy1 = nxy-1L
;   penergy=0.d0
   for ipix = 0L,nxy-1L,long(step) do begin
      ii = ipix MOD nx
      ij = ipix / nx
      x0 = x[ii,ij]
      y0 = y[ii,ij]
      if NOT keyword_set(montecarlo) then begin
         if n_elements(x0in) GT 0 then x0=x0in else x0 = x[(nx-1)/2,(ny-1)/2]
         if n_elements(y0in) GT 0 then y0=y0in else y0 = y[(nx-1)/2,(ny-1)/2]
      endif else begin
         ii = ipix MOD nx
         ij = ipix / nx
         x0 = x[ii,ij]
         y0 = y[ii,ij]
      endelse
      xx0 = x-x0
      yy0 = y-y0
      if keyword_set(ubound) then begin
         d2 = xx0^2+yy0^2
         ; calculate an upper bound to the energy from Aly (1984)
         integral2 = total(d2*pz2dA)
         energy = [energy,sqrt(integral1*integral2)/(4.0*!pi)]
      endif else begin
         ; Calculate the energy from the virial theorem
         energy = [energy,total( (xx0*px+yy0*py)*pzdA )/(4.0*!pi)]
      endelse
      ; calculate the potential energy
;      non0=where(d2 NE 0.0)
;      penergy=penergy+pzdA[ii,ij]*total(pzdA[non0]/sqrt(d2[non0])) 
      if NOT keyword_set(noprogress) AND NOT keyword_set(quiet) AND step LT nxy and keyword_set(montecarlo) then $
         progress,(100.*float(ipix+1))/nxy,frequency=5,label=label, $
                  reset=keyword_set(ipix EQ 0L), $
                  last=keyword_set(ipix GE nxy-step)
      if NOT keyword_set(montecarlo) then break
   endfor
   energy = energy[1:*]
;   penergy=penergy/(16.*!PI^2)

   if keyword_set(ubound) then begin
      energyvalue = min(energy)
   endif else begin
      energyvalue = total(energy)/(n_elements(energy))  ; Mean energy
   endelse

   ; Do the force integrals as well

   fx = -total(px*pzdA)/(4.0*!pi)
   fy = -total(py*pzdA)/(4.0*!pi)
   fz = -total((pz^2-px^2-py^2)*dA)/(8.0*!pi)
   fp = total((pz^2+px^2+py^2)*dA)/(8.0*!pi)

   ; and the torque integrals

   tz = total((y*px-x*py)*pzdA)/(4.0*!pi)
   tx = total(y*(pz^2-px^2-py^2)*dA)/(8.0*!pi)
   ty = total(x*(pz^2-px^2-py^2)*dA)/(8.0*!pi)

   ; estimate uncertainty in net force

   pxe=berr.err_bxh
   pye=berr.err_byh
   pze=berr.err_bzh

   dA2 = dA^2
   errfx=sqrt(total(((pxe*pz)^2+(px*pze)^2)*dA2))/(4.0*!PI)
   errfy=sqrt(total(((pye*pz)^2+(py*pze)^2)*dA2))/(4.0*!PI)
   errfz=sqrt(total(((pze*pz)^2+(px*pxe)^2+(py*pye)^2)*dA2))/(4.0*!PI)
   errfp=errfz

   ; estimate error in energy from Wheatland's expression

   Lx = nx*pix_size[0]*7.27e7  ; mgram x dimension
   Ly = ny*pix_size[0]*7.27e7  ; mgram y dimension
   terror = sqrt(abs(((cyy*Lx)^2+(cxy*Ly)^2)*fx^2-2.0*fx*fy*Lx*Ly(cyy*cyx+cxy*cxx)+((cyx*Lx)^2+(cxx*Ly)^2)*fy^2)/(cosa*2.0*sqrt(3.)))
   ss = where(finite(terror),nss)
   if nss GT 0 then error = total(terror[ss])/nss else terror=0.0 ; mean over the FOV
   ;if n_elements(energy) GT 1 then print,stdev(energy),error

   ; Do the Wheatland energy correction
   wenergy = 0.0
   if keyword_set(wheatland) and NOT keyword_set(ubound) and $
      NOT keyword_set(extrapolate) and tag_index(j,'JZ') GE 0 then begin
      ; Compute error from magnetic field errors, under the assumption
      ; that the Wheatland correction fixes any probolems with forces.
      if n_elements(x0in) GT 0 then x0=x0in else x0 = x[(nx-1)/2,(ny-1)/2]
      if n_elements(y0in) GT 0 then y0=y0in else y0 = y[(nx-1)/2,(ny-1)/2]
      xx0 = x-x0
      yy0 = y-y0
      dEdBx = total(xx0*pzdA)/(4.0*!pi)
      dEdBy = total(yy0*pzdA)/(4.0*!pi)
      dEdBz = total( (xx0*px+yy0*py)*dA )/(4.0*!pi)
      ;print, 'Error incl. force: ',error
      error = sqrt((dEdBx*berr.sig_bxh)^2+(dEdBy*berr.sig_byh)^2+(dEdBz*berr.sig_bzh)^2)
     ; print, 'Error excl. force: ',error
      ; now compute the Wheatland correction
      jz = remapjz(j)  ; put Jz onto the same grid at the magnetic field
      fxden = py*(alpha*pz-jz)/(4.0*!pi)
      fyden = px*(jz-alpha*pz)/(4.0*!pi)
      if n_elements(x0in) GT 0 then x0=x0in else x0 = x[(nx-1)/2,(ny-1)/2]
      if n_elements(y0in) GT 0 then y0=y0in else y0 = y[(nx-1)/2,(ny-1)/2]
      xavg = f_div(total(fxden*(x-x0)*dA),total(fxden*dA))
      yavg = f_div(total(fyden*(y-y0)*dA),total(fyden*dA))
      wenergy = xavg*fx + yavg*fy
      werror = sqrt((xavg*errfx)^2 + (yavg*errfy)^2) ; neglecting error on xavg and yavg
      error = sqrt(error^2+werror^2)  ; combine the energy error with werror
      if NOT keyword_set(quiet) then print,'Applied Wheatland correction of ',wenergy,' +\- ',werror,' ergs'
   endif else if tag_index(j,'JZ') LT 0 then $
      message,'j.jz in not defined so the wheatland correction cannot be computed.'

   return,energyvalue+wenergy

end
