

;+

function fix_azimuth,b1in,b2,shear=shear,mask=maskin

;NAME:
;     FIX_AZIMUTH
;PURPOSE:
;     Fix the azimuth of a vector magnetogram by selecting the 
;     closest orientation to an input azimuth within 180 degrees.
;CATEGORY:
;CALLING SEQUENCE:
;     gram_out = fix_azimuth(gram_in,azimuth)
;INPUTS:
;     gram_in = magnetogram structure to be fixed.  gram_in.b_azim
;               (degrees) is modified by adding 180 degrees if
;               necessary to get a good match with the reference
;               azimuth. 
;     azimuth = "correct" azimuth to comapre with the azimuth in
;               gram_in.  If azimuth is a structure, then the azimuth
;               is taken to be azimuth.b_azim.  Degrees.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     shear = returns with the shear angle between the azimuth from
;             gram_in and azimuth after fixing the azimuth.  Degress between
;             -90 and +90.
;OUTPUTS:
;     gram_out = magnetogram structure with the azimuth fixed by 180
;                degree increments to best match the azimuth input
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The azimuth array and gram_in.b_azim must have the same dimension.
;PROCEDURE:
;     Adds 180 degrees to gram_in if necessary to get a good match
;     with the reference azimuth.  Calculates Bx, BY, and BZ on this
;     field and then runs jxyz to get currents, etc.
;MODIFICATION HISTORY:
;     April 1, 2003  T. Metcalf
;-

b1 = b1in

; Fix azimuth of b1 by finding the closest orientation to b2

if strupcase(size(b2,/tname)) EQ 'STRUCT' then azimuth = b2.b_azim $
else azimuth = b2

saz = size(azimuth)

if n_elements(maskin) EQ n_elements(azimuth) then mask=maskin $
else mask=1b+bytarr(saz[1],saz[2])

fixa = where(abs(arange(b1.B_AZIM,azimuth)-azimuth) GT 90. AND mask,count)
if (count gt 0) then b1.B_AZIM[fixa] = b1.B_AZIM[fixa] + 180.
b1.B_AZIM = b1.B_AZIM MOD 360.
shear = arange(b1.B_AZIM,azimuth)-azimuth

p = double(b1.point.p)
b0 = double(b1.point.b0)
if tag_index(b1,'CMD',/quiet) GE 0 then lc = double(b1.cmd)*!dtor $
else lc = double(b1.point.cmd)
if tag_index(b1,'LATITUDE',/quiet) GE 0 then bc = double(b1.latitude)*!dtor $
else bc = double(b1.point.lat)

azim = double(b1.b_azim) * !dtor
bxyz,-b1.b_trans*sin(azim), $
      b1.b_trans*cos(azim), $
      b1.b_long, $
      bx,by,bz,p=p,b0=b0,lat=bc,cmd=lc
if tag_index(b1,'BX',/quiet) GE 0 then b1.bx=bx $
else b1=add2str(b1,'BX',bx,/quiet)
if tag_index(b1,'BY',/quiet) GE 0 then b1.by=by $
else b1=add2str(b1,'BY',by,/quiet)
if tag_index(b1,'BZ',/quiet) GE 0 then b1.bz=bz $
else b1=add2str(b1,'BZ',bz,/quiet)

return,jxyz(b1,/quiet)

end
