
;+

function fix_b_glitch,bin,quiet=quiet,ncall=ncall,maxcall=maxcall, $
                      ssglitch=cssglitch,gfactor=gfactorin

;NAME:
;     fix_b_glitch
;PURPOSE:
;     Patch places where the magnetic field is impossibly large or
;     where it is much higher than all its neighbors.
;CATEGORY:
;CALLING SEQUENCE:
;     bnew = fix_b_glitch(b)
;INPUTS:
;     b = magnetic field structure from bfits.  Will be modified in place.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     maxcall = maximum number of recursive calls.  Set to 0 for just
;               one run through.  Default = 10.
;     ssglitch = returns with the glitch indices.
;     gfactor = glitch factor.  Default = 1.8.  Must be larger than 1.0.
;     /quiet = work quietly
;OUTPUTS:
;     bnew = new structure with corrected field.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Only fixes b_long, b_trans, and b_zaim.  So this should be done
;     before ambiguity resolution.
;     May need to ".run xform" before running.
;PROCEDURE:
;     Looks for pixels where Btotal is much larger than all its
;     neighbors or is above 4000 G and interpolates.  Calls itself
;     recursively until there are no more glitches detected.
;MODIFICATION HISTORY:
;     T. Metcalf 2004-May-28
;     TRM 2004-Nov-30 Changed glitch test to allow a big field
;                     difference on 3 sides rather than 4.  Also look
;                     for holes as well as peaks.  Added ssglitch
;                     keyword and gfactor keyword.
;-

   if NOT keyword_set(ncall) then ncall = 1L
   ncall = long(ncall)
   if ncall LT 1L then ncall = 1L
   if n_elements(maxcall) LE 0 then maxcall = 10L 
   maxcall = long(maxcall)>0L

   b = bin

   szb = size(b.b_long)
   if szb[0] NE 2 then message,'ERROR: odd size on b.b_long'
   nx = szb[1]
   ny = szb[2]

   if n_elements(alpha) LE 0 then alpha = 0.0
   if NOT keyword_set(guard) then guard = 2

   berr = get_berrs(b)

   if keyword_set(gfactorin) then gfactor = float(gfactorin) $
   else gfactor = 1.8   ; glitch factor
   if gfactor LE 1.0 then message,'gfactor must be greater than 1.0'
   bthresh = 4000. ; max allowed total field

   btot = sqrt(b.b_long^2+b.b_trans^2)
   btot2 = btot/gfactor
   btot3 = btot*gfactor
   bti = b.b_trans
   bti2 = bti/gfactor
   bti3 = bti*gfactor
   bzi = b.b_long
   bzi2 = (abs(bzi)*(gfactor-1.0)) > (berr.sig_blong*3.0)

;   ssglitch = where((shift(btot,1,0) LT btot2 AND $
;                     shift(btot,0,1) LT btot2 AND $
;                     shift(btot,-1,0) LT btot2 AND $
;                     shift(btot,0,-1) LT btot2 AND $
;                     btot GT berr.sig_btrans*3.0) OR $
;                    (shift(bti,1,0) LT bti2 AND $
;                     shift(bti,0,1) LT bti2 AND $
;                     shift(bti,-1,0) LT bti2 AND $
;                     shift(bti,0,-1) LT bti2 AND $
;                     bti GT berr.sig_btrans*3.0) OR $
;                    (shift(bzi,1,0) LT bzi2 AND $
;                     shift(bzi,0,1) LT bzi2 AND $
;                     shift(bzi,-1,0) LT bzi2 AND $
;                     shift(bzi,0,-1) LT bzi2 AND $
;                     bti GT berr.sig_blong*3.0) OR $
;                     btot GT bthresh)

   btottestlo = (shift(btot,1,0) LT btot2) + $    ; look for a peak
                (shift(btot,0,1) LT btot2) + $
                (shift(btot,-1,0) LT btot2) + $
                (shift(btot,0,-1) LT btot2)

   btottesthi = (shift(btot,1,0) GT btot3) + $  ; look for a hole
                (shift(btot,0,1) GT btot3) + $
                (shift(btot,-1,0) GT btot3) + $
                (shift(btot,0,-1) GT btot3)

   bttestlo = (shift(bti,1,0) LT bti2) + $   ; look for a peak
              (shift(bti,0,1) LT bti2) + $
              (shift(bti,-1,0) LT bti2) + $
              (shift(bti,0,-1) LT bti2)

   bttesthi = (shift(bti,1,0) GT bti3) + $  ; look for a hole
              (shift(bti,0,1) GT bti3) + $
              (shift(bti,-1,0) GT bti3) + $
              (shift(bti,0,-1) GT bti3)

   bztest = (abs(shift(bzi,1,0)-bzi) GT bzi2) + $  ; look for large difference
            (abs(shift(bzi,0,1)-bzi) GT bzi2) + $
            (abs(shift(bzi,-1,0)-bzi) GT bzi2) + $
            (abs(shift(bzi,0,-1)-bzi) GT bzi2)

   ssglitch = where((btottestlo GE 3 AND (btot GT berr.sig_btrans*3.0)) OR $
                    (btottesthi GE 3 AND (btot GT berr.sig_btrans*3.0)) OR $
                    (bttestlo GE 3 AND (bti GT berr.sig_btrans*3.0)) OR $
                    (bttesthi GE 3 AND (bti GT berr.sig_btrans*3.0)) OR $
                    (bztest GE 3 AND (abs(bzi) GT berr.sig_blong*3.0)) OR $
                    (btot GT bthresh))

   if ssglitch[0] GE 0 then begin
      if NOT keyword_set(quiet) then begin
         ;junk = fltarr(nx,ny)
         ;junk[ssglitch] = 1.
         ;tvscl,junk
         message,/info,strcompress(string(n_elements(ssglitch))+' glitches corrected.')
      endif

      bxi = -b.b_trans*sin(b.b_azim*!dtor)
      byi = b.b_trans*cos(b.b_azim*!dtor)
      bzi = b.b_long

      bxi = (shift(bxi,0,1)+shift(bxi,1,0)+shift(bxi,-1,0)+shift(bxi,0,-1))/4.
      byi = (shift(byi,0,1)+shift(byi,1,0)+shift(byi,-1,0)+shift(byi,0,-1))/4.
      bzi = (shift(bzi,0,1)+shift(bzi,1,0)+shift(bzi,-1,0)+shift(bzi,0,-1))/4.

      b.b_long[ssglitch] = bzi[ssglitch]
      b.b_trans[ssglitch] = sqrt((bxi[ssglitch])^2+(byi[ssglitch])^2)
      b.b_azim[ssglitch] = atan(-bxi[ssglitch],byi[ssglitch])/!dtor
   endif

   if ncall GT maxcall OR ssglitch[0] LT 0 then answer=b $
   else answer=fix_b_glitch(b,ncall=ncall+1,maxcall=maxcall, $
                            ssglitch=ssg,gfactor=gfactorin)

   if n_elements(ssg) GT 0 then $
      cssglitch = squeeze([ssglitch,ssg]) else cssglitch=ssglitch

   return,answer

end
