function get_berrs, magin, errout, cont_lev=cont_lev, roi=roi,$
	sigbl=sigbl,sigbt=sigbt,sigazim=sigazim,nozero=nozero, $
        plot=doplot,integral=integral

;+
;NAME: get_berrs
;PURPOSE: calculate errors in a magnetic-field structure
;CATEGORY:
;CALLING SEQUENCE:
;INPUTS:
;OPTIONAL INPUT PARAMETERS:
;	cont_level: used to restrict area to use for getting B_long,B_trans
;			starting errors.
;	roi: same, feed in array found by defroi or the like
;	sigbl,sigbt: override that found by the histogram method
;			below, helpful if distributions are strange
;			and shenanigans are required to determine them by hand.
;       nozero: if set, then precludes points whose values are identically zero from
;		being considered.
;       plot: if set, plot the LOS and transverse histograms
; 	integral: if set, computes the 1-sigma errors for blong, btrans
; 		by integrating the histogram of field strengths to 68% of the 
;		area under the curve.  IT IS STRONGLY RECOMMENDED that you
;		set a continuum level or quiet-sun roi when using this.
;OUTPUTS:
;	errout, structure with **1-sigma** errors:
;		   SIG_BLONG       FLOAT 	1-sigma b_long error, gauss
;		   SIG_BTRANS      FLOAT        1-sigma b_trans error, gauss
;	 	   SIG_AZIM       FLOAT        1-sigma b_azim error, radians
;	 	   SIG_BXH	   FLOAT	1-sigma Bx error, gauss
;	 	   SIG_BYH	   FLOAT	1-sigma By error, gauss
;	 	   SIG_BZH	   FLOAT	1-sigma Bz error, gauss
;	 	   ERR_BXI         FLOAT     Array[256, 256] image-plane bx uncertainties, each point
;		   ERR_BYI         FLOAT     Array[256, 256] image-plane by ""    ""
;	           ERR_BXH         FLOAT     Array[256, 256] heliogr.plane bx uncertainties
;                  ERR_BYH         FLOAT     Array[256, 256] heliogr.plane by uncertainties
;		   ERR_BZH         FLOAT     Array[256, 256] heliogr.plane bz uncertainties
;		   ERR_JZ          FLOAT     Array[256, 256] jz errors 

;PROCEDURE:  generic b_long, b_trans uncertainties determined by
;		histograms, then fed into propagation of errors
;		using coordinate transforms, etc for heliographic
;		bx,by,bz error 'images'.  does not rely on ambig-resolved
;		bx, etc...rather just uses coordinate transforms.
;		err_bzi will be err_blong, since what goes into the
;		image-plane bx,by uncertainties is only the azimuth and the b_trans
; 	     a rudimentary guess is used for a continuum level to determine quiet-sun,
;		but the 'cont_level' or 'roi' keyword should be used for best results.
;MODIFICATION HISTORY:
;	written KDL 9/01 from get_errs.pro and get_berr.pro written
;		for flares, alpha projects.
;       TRM 2004-Apr-23 Use built in histogram instead of histo.pro.
;                       Use guassfit.pro rather than
;                       ivm_gaussfit.pro. The one sigma error in the
;                       transverse field is at the peak of the
;                       histogram. 
;       KDL 2004-Jul-13 Added structure name to final structure.
;       TRM 2004-Nov-17 Added /plot and took locations keyword out of
;                       histogram calls.
;	KDL 2004-Nov-18 Added integral option, made a default continuum level
;			so that by default it at least tries to use
;			quiet sun only.
;       TRM 2005-Sep-13 Added some digits to the structure name so
;                       that it will not repeat even if called within
;                       the same second. 
;-

mag=magin
xf=set_xform(mag)
;stname=strcompress(string(long(systime(1))),/remove_all) ; make unique name
stname=strcompress(string(long64(systime(/seconds,/utc)*1000.)),/remove_all)
; if not already fed in, find 1-sigma uncertainties:
if (not keyword_set(sigbl) or not keyword_set(sigbt)) then begin
; restrict where to look by continuum-level (default, but level is over-ride-able)
; or other means
;   dd=findgen(n_elements(mag.b_long))
   icmoment=moment(mag.i_cont,mdev=icmdev)
   cl=icmoment(0)-icmdev
   if (keyword_set(cont_lev)) then cl=cont_lev
   dd=where(mag.i_cont gt cl)
   if dd[0] LT 0 then dd=where(finite(mag.i_cont))
; the following is provided so can do blah=defroi(image), then pass the roi=blah array in.
   if (keyword_set(roi)) then dd=roi
; nozero is set so that the distributions don't get weirded out by a spike at zero.
   if (keyword_set(nozero)) then dd=where(mag.b_long ne 0.0 AND mag.i_cont gt cl)
   if (keyword_set(nozero) AND keyword_set(cont_lev)) then $
   	dd=where((mag.i_cont gt cont_lev) AND (mag.b_long ne 0.0))
   
; get basic uncertainty level from quiet-sun or wherever width of Blong, Btrans
; distributions...
   IF (KEYWORD_SET(integral)) THEN BEGIN
      binsize = 1.
      yy = histogram(abs(mag.b_long(dd)),min=0,max=199,binsize=binsize)
      xx = findgen(n_elements(yy))*binsize ; locations of bins
      xx = xx + binsize/2.0
      totbl=fltarr(n_elements(yy))
      for i=0,n_elements(yy)-1 do $
         totbl(i)=total(yy(0:i))
      sigbl=xx(min(where(totbl ge 0.68*totbl(n_elements(yy)-1))))
   ENDIF ELSE BEGIN 
      binsize = 10.
      minv = -500.
      yy = histogram(mag.b_long(dd),min=minv,max=499,binsize=binsize)
      xx = minv+findgen(n_elements(yy))*binsize ; locations of bins
      xx = xx + binsize/2.0
      aa=[max(yy),(xx(where(yy eq max(yy))))(0),50,0.,0.,0.] ; 50 is width initial-guess 
      fit=gaussfit(xx,yy,aa)
      if (not keyword_set(sigbl)) then sigbl=aa(2)*sqrt(-2.*alog(0.5)) ; 
                                ; width is the error: min(good) = yfit(2)*wc+yfit(1)
                                ; sqrt(-2.*alog(0.5)) gives half-width
   ENDELSE
   if keyword_set(doplot) then begin
      smulti = !p.multi
      !p.multi = [0,1,2,0]
      plot,xx,yy,title='LOS Field',psym=10
      oplot,[sigbl,sigbl],[0,max(yy)*2],line=1
      oplot,[-sigbl,-sigbl],[0,max(yy)*2],line=1
      xyouts,/norm,0.6,0.8,strcompress('LOS Error = '+string(sigbl))
   endif
   IF (KEYWORD_SET(integral)) THEN BEGIN
      binsize = 1.
      minv = sigbl
      yy = histogram(abs(mag.b_trans(dd)),min=minv,max=699,binsize=binsize)
      xx = findgen(n_elements(yy))*binsize ; locations of bins
      xx = xx + binsize/2.0
      totbt=fltarr(n_elements(yy))
      for i=0,n_elements(yy)-1 do $
      	totbt(i)=total(yy(0:i))
      ;sigbt=min(where(totbt ge 0.68*totbt(n_elements(yy)-1)))
      sigbt=xx(min(where(totbt ge 0.68*totbt(n_elements(yy)-1))))
   ENDIF ELSE BEGIN
      minv = sigbl
      binsize = 10.
      yy = histogram(mag.b_trans(dd),min=minv,max=699,binsize=binsize)
      xx = minv+findgen(n_elements(yy))*binsize
      xx = xx + binsize/2.0
      aa=[max(yy),(xx(where(yy eq max(yy))))(0),100,0.,0.,0.]
      fit = gaussfit(xx,yy,aa,nterms=6)
      hs = rotate(sort(fit),2)
      if hs(0) NE 0 then hplace = hs(0) else hplace=hs(1)
      ; 1 sigma error in the transverse magnetic field at histogram peak
      if (not keyword_set(sigbt)) then sigbt = xx[hplace]  
   ENDELSE
   if keyword_set(doplot) then begin
      plot,xx,yy,title='Transverse Field',psym=10
      oplot,[sigbt,sigbt],[0,max(yy)*2],line=1
      !p.multi = smulti
      xyouts,/norm,0.6,0.2,strcompress('Transverse Error = '+string(sigbt))
   endif
endif
sigazim=5*!dtor  ; 5 deg -> radians.

sigbxh=SQRT(xf.dircos(0)^2*sigbl^2 + (1-xf.dircos(0)^2)*sigbt^2)
sigbyh=SQRT(xf.dircos(1)^2*sigbl^2 + (1-xf.dircos(1)^2)*sigbt^2)
sigbzh=SQRT(xf.dircos(2)^2*sigbl^2 + (1-xf.dircos(2)^2)*sigbt^2) ; dircos(2) = mu

; from propagation of errors.  Sku has convinced me that the
; variables are independent, hence covariance is zero.
; even though the mean of mag.b_azim is not well defined...

cosazerrxy=sin(mag.b_azim*!dtor)*sigazim
sinazerrxy=cos(mag.b_azim*!dtor)*sigazim
bxierrarr=sqrt((sin(mag.b_azim*!dtor)*sigbt)^2 +$
	    (mag.b_trans*sinazerrxy)^2 +$
			sigbt^2*sinazerrxy^2)
byierrarr=sqrt((cos(mag.b_azim*!dtor)*sigbt)^2 +$
	    (mag.b_trans*cosazerrxy)^2 +$
			 sigbt^2*cosazerrxy^2)
; use matrix, get errors w/ covariance between terms
xfm=xf.image2helio

bxherr=sqrt((xfm(0,0)*bxierrarr)^2 + (xfm(0,1)*byierrarr)^2 + (xfm(0,2)*sigbl)^2)
byherr=sqrt((xfm(1,0)*bxierrarr)^2 + (xfm(1,1)*byierrarr)^2 + (xfm(1,2)*sigbl)^2)
bzherr=sqrt((xfm(2,0)*bxierrarr)^2 + (xfm(2,1)*byierrarr)^2 + (xfm(2,2)*sigbl)^2)

pixsize=mag.point.pix_size
if (n_elements(pixsize) gt 1) then pixsize=mean(pixsize)
C = (1e3)/(4.*!pi*1e-7*pixsize*1e4*xf.ARCSEC2MM*1e6/xf.mu)
; 10^3 mA/A, mu=4 pi x10^-7,1.1arcsec,10^6m/Mm,10^4G/tesla
jzerr=sqrt(4)*C*sqrt(bxherr^2+byherr^2)

errs=create_struct(name='bs'+stname,'sig_blong',sigbl,'sig_btrans',sigbt,'sig_azim',sigazim,'sig_bxh',sigbxh,$
	'sig_byh',sigbyh,'sig_bzh',sigbzh,'err_bxi',bxierrarr,'err_byi',byierrarr,$
	'err_bxh',bxherr,'err_byh',byherr,'err_bzh',bzherr,$
	'err_jz',jzerr)

return,errs
end

