
;+

function ivm_mdi_align,mag,mindex,mdata, $
                         ivm_mask=rmaskin,mdi_mask=tmaskin, $
                         tmdi2ivmguess=tmdi2ivmguess, $
                         xvr=xvr,yvr=yvr,xvi=xvi,yvi=yvi, $
                         quiet=quiet,mdi2ivm=mdi2ivm, $
                         blindguess=blindguess, xyivm=xyivmsub, $
                         noparallax=noparallax,_EXTRA=extra

;NAME:
;     IVM_MDI_ALIGN
;PURPOSE:
;     Use an MDI white light image to fix the pointing in an IVM
;     white light image by assuming that the MDI pointing is correct.
;CATEGORY:
;CALLING SEQUENCE:
;     new_ivm_gram = ivm_mdi_align(mag,mindex,mdata)
;INPUTS:
;     mag  = IVM magnetogram (a structure)
;     mindex  = MDI index (a structure)
;     mdata   = MDI image
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS:
;     ivm_mask = Index giving points in the IVM image to neglect in the
;                  alignment.
;     mdi_mask = Index giving points in the MDI image to neglect in the
;                alignment.
;     mdi2ivm = the final transformation from the MDI image to the IVM
;                 image (output).  You probably don't really want this.
;     /quiet = work quietly
;     /noparallax = Do not include the parallax correction.  This is
;                   only for debugging purposes, do not use this keyword.
;     _extra = Any extra kewords are passed to auto_align_image.  E.g. scale,
;             /amoeba, etc.
;
;     INPUT PARAMETERS FOR THE INITIAL GUESS (NOT REQUIRED):
;
;     These keywords allow you to avoid clicking on points in the two images
;     which may be useful for automated alignments etc. 
;
;     /blindguess = If set, use the alignment in the MDI and IVM headers
;                   to set the initial guess: do not ask the user to click
;                   on points in the two images.  If this works, fine, but if
;                   the guess is too far off, it is possible that
;                   auto_align_images will fail.  If /blindguess is set, the
;                   tmdi2ivm and xvr, xvi etc. keywords are ignored (but
;                   will be set on output).
;     xyivm = 4 element array giving the LL and UR corners of a subregion
;               to use in the alignment.  If this is not set, you will be
;               asked to define the subregion with the mouse. [x0,x1,y0,y1]
;               Ignored if /blindguess is not set.
;
;     If the following are passed as input, they must all be passed in and they must
;     all be consistent.  I.e., the transformation must apply to the subimages
;     defined by xvr,yvr,xvi,yvi.  If they are not consistent, the answer you 
;     get will be incorrect.  Garbage in, garbage out.  Be careful with these,
;     getting them right can be a bit tricky!  If these are not set on input,
;     they will be set on output so that they can be used for a subsequent call.
;
;     tmdi2ivmguess = Tranform vector (from caltrans.pro) giving the initial
;                       guess for the alignment.  If this is not specified,
;                       the user will be prompted to click on corresponding
;                       points in the IVM and MDI images in order to get
;                       the initial guess.  If this is passed, xvr,yvr,xvi,yvi
;                       must also be passed in.  If undefined, it will be
;                       set on output to the guess from setpts_roi.
;     xvr,yvr = 2 element arrays giving LL a UR corners of the IVM subimage
;               to use.  Ignored unless tmdi2ivmguess is also passed.
;               If undefined, it will be set on output to the guess from 
;               setpts_roi.
;     xvi,yvi = 2 element arrays giving LL a UR corners of the MDI subimage
;               to use.  Ignored unless tmdi2ivmguess is also passed.
;               If undefined, it will be set on output to the guess from 
;               setpts_roi.
;OUTPUTS:
;     new_ivm_gram = ivm gram with the pointing corrected.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     VERY SLOW
;
;     Obviously the IVM and MDI images should be close in time to avoid
;     evolution in the AR (but solar rotation and foreshortening are 
;     accounted for).
;
;     MDI pointing is assumed to be correct.  Not well tested when the
;     MDI p anlge is other than 0.
;PROCEDURE:
;     The user first gives an initial guess for the transformation between
;     IVM and MDI using setpts_roi and then the guess is refined using 
;     auto_align_image.   The transformation is then used to set the pointing
;     of IVM assuming that the MDI pointing is correct.
;EXAMPLE:
;     ; First get the MDI data
;     mdata  = readfits('Some_MDI_FITS_File',smdihead)
;     mindex = fitshead2struct(smdihead)
;     ; Second get the IVM data
;     j = ivm_bfits('Some_IVM_FITS_file')
;     ; Third align the MDI and IVM data
;     jnew = ivm_mdi_align(j,mindex,mdata)
;MODIFICATION HISTORY:
;     T. Metcalf 2001-Oct-17
;     T. Metcalf 2001-Nov-14 Added a warning for a large coordinate shift.
;     T. Metcalf 2001-Nov-15 Also update crpix1,crpix2,crval1,crval2 so that
;                            the new index with the new pointing will work
;                            with index2map and plot_map which gets pointing
;                            from crpix and crval rather than xcen and ycen.
;     T. Metcalf 2001-Nov-19 Added the /blindguess keyword.  Also mask
;                            off-limb points.
;     T. Metcalf 2002-Apr-04 Added correction for parallax in longitude
;                            between the earth and SOHO points of view
;                            (a small correction).  Parallax in latitude
;                            was already accounted for.
;     T. Metcalf 2002-Apr-23 Cleaned up code a bit.
;     T. Metcalf 2002-Oct-22 Added /eit keyword and call to get_orbit
;     T. Metcalf 2005-Mar-10 Converted trace_mdi_align to ivm_mdi_align.
;-


nxmdi = n_elements(mdata[*,0])
nymdi = n_elements(mdata[0,*])

if NOT keyword_set(eit) then begin
  bad = where(mdata LT mindex.d_min0 OR mdata GT mindex.d_max0,nbad)
  if nbad GT 0 then mdata[bad] = 0
endif

; mdidate should be equivalent time at earth but the time at L1 and the
; time observed at earth only differ by 5 seconds or so
; and I'm not sure if date_obs already has the 5 sec correction, so ignore it
; here since differential rotation is small over 5 seconds anyway. 

mdidate    = anytim(mindex.date_obs,/yoh)  
ivmdate0 = anytim(mag.date_obs+' '+mag.time_obs,/yoh)

; Get the pixel size of MDI image

mdipix = mindex.im_scale
; This corrects the known problem with the MDI plate scale.  It might be
; necessary to update the value periodically???
if abs(mdipix-1.9778400) LT 0.00001 then begin
   mdipix = 1.982
   message,/info,'Correcting MDI plate scale of 1.97784 to '+string(mdipix)
endif
mdipix_earth = mdipix*mindex.EARTH_R0/mindex.OBS_R0

;nxivm = long(n_elements(mag.b_long(*,0)))
;nyivm = long(n_elements(mag.b_long(0,*)))
tdata = mag.i_cont
xyivm = lonlat2xy([mag.point.cmd*!radeg,mag.point.lat*!radeg], $
                  b0=mag.point.b0,radius=mag.point.radius)
sbl = size(mag.b_long)
nxivm = sbl[1]
nyivm = sbl[2]

; The first task is to get an initial guess for the MDI->IVM
; transformation.  If tmdi2ivmguess is passed in, we are all set.  If
; /blindguess is set, then use the IVM and MDI pointing.  
; Otherwise prompt the user to select ROI's and click on corresponding 
; points in the subimages.

if keyword_set(blindguess) then begin
   if n_elements(xyivmsub) NE 4 then begin
      nxr = n_elements(tdata[*,0])
      nyr = n_elements(tdata[0,*])
      binr = 1L+max([nxr,nyr])/640L
      REPEAT nxr = nxr - 1L UNTIL nxr MOD binr EQ 0
      REPEAT nyr = nyr - 1L UNTIL nyr MOD binr EQ 0
      ; Select a smaller FOV which encompasses only the features to be
      ; aligned.  This is for speed; it'd be OK to take the whole FOV.
      message,/info,'Select a sub region for the alignment.  The smaller the '+$
                    'region, the faster the calculation.'
      junk = wdefroi(rebin(tdata[0:nxr-1,0:nyr-1],nxr/binr,nyr/binr,/sample), $
                     xvroi,yvroi)
      xyivmsub = [min(xvroi),max(xvroi),min(yvroi),max(yvroi)]*binr
   endif
   xvr = [xyivmsub[0]>0,xyivmsub[1]<(nxivm-1L)]
   yvr = [xyivmsub[2]>0,xyivmsub[3]<(nyivm-1L)]
   ; Now get the equivalent region in the MDI data
   ;xcen,ycen are the coordinates of the center of the FOV (arcsec)
   ; account for shift in center when using restricted FOV
   xpshift = (xvr[0]/2.0 + (xvr[1]-(nxivm-1L))/2.0)*mag.point.pix_size[0]
   ypshift = (yvr[0]/2.0 + (yvr[1]-(nyivm-1L))/2.0)*mag.point.pix_size[1]
   ; convert the shift from image coords to solar coords
   xshift = xpshift*cos(-mag.point.p)-ypshift*sin(-mag.point.p)
   yshift = xpshift*sin(-mag.point.p)+ypshift*cos(-mag.point.p)
   xcen = xyivm[0] + xshift
   ycen = xyivm[1] + yshift
   ; Correct these for solar rotation
   ll = xy2lonlat([xcen,ycen],b0=mag.point.b0,radius=mag.point.radius)
   ddays = int2secarr(mdidate,ivmdate0)/86400.
   ll[0] = ll[0] + diff_rot(ddays,ll[1],/allen)
   xy = lonlat2xy(ll,mdidate)
   xcen = xy[0]
   ycen = xy[1]
   ; Now convert to MDI pixel coordinates
   xvi = fltarr(2) & yvi = fltarr(2)
   xvi[0] = mindex.crpix1+xcen/mdipix_earth - $
           ((xvr[1]-xvr[0])/2)*mag.point.pix_size[0]/mdipix_earth
   xvi[1] = mindex.crpix1+xcen/mdipix_earth + $
           ((xvr[1]-xvr[0])/2)*mag.point.pix_size[0]/mdipix_earth
   yvi[0] = mindex.crpix2+ycen/mdipix_earth - $
           ((yvr[1]-yvr[0])/2)*mag.point.pix_size[1]/mdipix_earth
   yvi[1] = mindex.crpix2+ycen/mdipix_earth + $
           ((yvr[1]-yvr[0])/2)*mag.point.pix_size[1]/mdipix_earth
   ; Shift by the fractional part of xvi,yvi
   xshift = xvi[0]-long(xvi[0]) & yshift = yvi[0]-long(yvi[0]) 
   xvi = long(xvi)
   yvi = long(yvi)
   ; Make sure the boundaries are not out of range
   if xvi[0] LT 0 then begin
      xshift = xshift - xvi[0]
      xvi[0] = 0
   endif
   if yvi[0] LT 0 then begin
      yshift = yshift - yvi[0]
      yvi[0] = 0
   endif
   if xvi[1] GT nxmdi-1 then begin
      xvi[1] = nxmdi-1
   endif
   if yvi[1] GT nymdi-1 then begin
      yvi[1] = nymdi-1
   endif
   ; Get the transformation
   tmdi2ivmguess = rss2pq(xshift=xshift, $
                          yshift=yshift, $
                          rot12=mag.point.p*!radeg-mindex.solar_p, $
                          xscale=mdipix_earth/mag.point.pix_size[0], $
                          yscale=mdipix_earth/mag.point.pix_size[1], $
                          xvi[1]-xvi[0]+1,yvi[1]-yvi[0]+1,/center)
   if NOT keyword_set(quiet) then begin
      tvscl,mag.i_cont[xvr[0]:xvr[1],yvr[0]:yvr[1]]
      ocontour2,poly_2d(mdata[xvi[0]:xvi[1],yvi[0]:yvi[1]], $
                        tmdi2ivmguess[*,0],tmdi2ivmguess[*,1],1, $
                        xvr[1]-xvr[0]+1,yvr[1]-yvr[0]+1)
   endif
endif

if n_elements(tmdi2ivmguess) LE 0 then begin

   ; Get the initial guess from the user if one
   ; is not passed in and /blindguess is not set

   message,/info,'Select the regions of the IVM and MDI images to ' + $
                 'analyze.  Then click on corresponding points in the two ' + $
                 'images to set the initial guess for the transformation ' + $
                 'between them.'

   transtype = {type:'g', $
                ref_x:mag.point.pix_size[0], $
                ref_y:mag.point.pix_size[1], $
                pix_x:mdipix_earth, $
                pix_y:mdipix_earth, $
                transpose:0,phi:mag.point.p*!radeg+mindex.solar_p} ; angle signs???

   pp = setpts_roi(mdata,tdata,key=transtype, $
                   xvi=xvi,yvi=yvi,xvr=xvr,yvr=yvr)

   ; Get back to the restricted FOV since we'll use that in the 
   ; alignment

   pp(0,0,*) = pp(0,0,*) -  xvr(0)  ; x-coord of ref image
   pp(1,0,*) = pp(1,0,*) -  yvr(0)  ; y-coord of ref image
   pp(0,1,*) = pp(0,1,*) -  xvi(0)  ; x-coord of transformed image
   pp(1,1,*) = pp(1,1,*) -  yvi(0)  ; y-coord of transformed image

   tmdi2ivmguess = caltrans(pp,transtype=transtype)
   if NOT keyword_set(quiet) then begin
      print,'Restricted initial guess for transformation:'
      pq2rss,tmdi2ivmguess(*,0),tmdi2ivmguess(*,1), $
         erot,exscl,eyscl,exshft,eyshft,enrss
   endif
endif

rimage = tdata[xvr[0]:xvr[1],yvr[0]:yvr[1]]
timage = mdata[xvi[0]:xvi[1],yvi[0]:yvi[1]]

; Convert IVM FOV center from solar coorinages to image coordinates
; which are required as input to get_latlon 
ixivm = xyivm[0]*cos(mag.point.p) + xyivm[1]*cos(mag.point.p)
iyivm = xyivm[0]*sin(mag.point.p) + xyivm[1]*sin(mag.point.p)
; add shift from restricted FOV to full FOV
; x0,y0 are in IVM coords, not solar coords, so do not need to
; correct the shifts for the p-angle
xshift = ((xvr[1]+xvr[0])/2.0 - (nxivm-1.)/2.)
yshift = ((yvr[1]+yvr[0])/2.0 - (nyivm-1.)/2.)
x0 = ixivm/mag.point.pix_size[0] + xshift
y0 = iyivm/mag.point.pix_size[1] + yshift
ll = get_latlon(nx=xvr[1]-xvr[0]+1L, $
                ny=yvr[1]-yvr[0]+1L, $
                x0=x0, $
                y0=y0, $
                b0=mag.point.b0, $
                pang=mag.point.p, $
                pixs=mag.point.pix_size, $
                rad=mag.point.radius)

rho = acos(cos(ll.latitude*!dtor)*cos(mag.point.b0)*cos(ll.cmd*!dtor) + $
           sin(ll.latitude*!dtor)*sin(mag.point.b0))
offlimb = where(rho GT 85.*!dtor,nofflimb)
delvarx,ll,rho

rmask = tdata & rmask[*] = 0.0
if n_elements(rmaskin) GT 0 then rmask[rmaskin] = 1.0
rmask = rmask[xvr[0]:xvr[1],yvr[0]:yvr[1]]
if nofflimb GT 0 then begin
   rmask[offlimb] = 1.0
   if NOT keyword_set(quiet) then $
      message,/info,strcompress('Masking out '+string(nofflimb)+' off-limb points.')
endif
rmask = where(rmask,nrmask)
if nrmask LE 0 then delvarx,rmask

if  n_elements(tmaskin) GT 0 then begin
   tmask = mdata & tmask[*] = 0.0
   tmask[tmaskin] = 1.0
   tmask = tmask[xvi[0]:xvi[1],yvi[0]:yvi[1]]
   tmask = where(tmask)
endif

; Use a 2nd degree transformation to allow for evolution and foreshortening
; due to solar rotation.

if sqrt(float(n_elements(tmdi2ivmguess[*,0]))) MOD 1.0 NE 0.0 then $
   message,'tmdi2ivmguess has bad dimension'
pin = tmdi2ivmguess[*,0]
qin = tmdi2ivmguess[*,1]
n = long(sqrt(n_elements(pin))+0.5)
n9 = (3>n)  ; at least order 2 transformation
p9 = fltarr(n9,n9) & p9[0,0] = reform(pin,n,n) & p9 = reform(p9,n9*n9)
q9 = fltarr(n9,n9) & q9[0,0] = reform(qin,n,n) & q9 = reform(q9,n9*n9)

if 0 then begin
   pout = p9  ; This just uses the initial guess for debugging purposes
   qout = q9
   inew = poly_2d(timage,pout,qout,2,xvr[1]-xvr[0]+1,yvr[1]-yvr[0]+1, $
                  cubic=-0.5)
endif else begin
   ; Use amoeba to improve the initial guess
   inew=auto_align_images(timage,rimage,p9,q9,pout,qout,missing=0.0, $
                          tmask=tmask,rmask=rmask,unrestricted=1, $
                          /quiet,/amoeba,itmax=250)
   p9 = pout & q9=qout
   inew=auto_align_images(timage,rimage,p9,q9,pout,qout,missing=0.0, $
                          tmask=tmask,rmask=rmask,unrestricted=1, $
                          quiet=quiet,_EXTRA=extra)
endelse

; Check the plate scales and rotation assumptions
p4 = (reform(pout,n9,n9))[0:1,0:1]
q4 = (reform(qout,n9,n9))[0:1,0:1]
pq2rss,p4,q4,erot,exscl,eyscl,exshft,eyshft,enrss,/quiet
if abs(exscl*mag.point.pix_size[0]-mdipix_earth) GT 0.1 OR $
   abs(eyscl*mag.point.pix_size[1]-mdipix_earth) GT 0.1 then begin
   message,/info,'Warning: IVM and MDI plate scales are inconsistent: ' + $
                 string(max([abs(exscl*mag.point.pix_size[0]-mdipix_earth), $
                             abs(eyscl*mag.point.pix_size[1]-mdipix_earth)])) + $
                 ' arcsec'
endif

solar_p = mindex.solar_p

mdi2ivm=[[pout],[qout]]

; Get Lat/Lon of MDI data

ivmdate = ivmdate0
ivmdate = anytim(ivmdate,/yoh)

ddays = int2secarr(anytim2ints(mdidate),anytim2ints(ivmdate))/86400.
if NOT keyword_set(quiet) then print,'ddays:',ddays

; get_latlon changes the pixs input to a 2-elements array so protect mdipix
jmdipix = mdipix 

; For MDI  ( (ny-1)/2.0 - y0 ) * im_scale = ycen 

orbit = get_orbit(mdidate)
earth_L0 = orbit.hel_lon_earth*!radeg
obs_L0 = orbit.hel_lon_soho*!radeg
earth_B0 = orbit.hel_lat_earth*!radeg
obs_B0 = orbit.hel_lat_soho*!radeg
earth_R0 =get_rb0p(mdidate,/rad)
obs_R0 = soho_fac(mdidate)*earth_R0
;x0 = mindex.crpix1 - mindex.crval1/mindex.cdelt1
;y0 = mindex.crpix2 - mindex.crval2/mindex.cdelt2

if keyword_set(noparallax) then begin
   message,/info,'WARNING: no parallax correction.'
   obs_L0 = earth_L0
   obs_B0 = earth_B0
endif

get_fits_par,mindex,xcen,ycen,dx,dy
x0 = (nxmdi+1)/2. - xcen/dx - 1.0  ; The location of disk center in pixels
y0 = (nymdi+1)/2. - ycen/dy - 1.0  ; with 0,0 defined as the LL corner pixel

latlon = get_latlon(nx=xvi[1]-xvi[0]+1L, $
                    ny=yvi[1]-yvi[0]+1L, $
                    x0=(xvi[0]+xvi[1])/2.0-x0, $
                    y0=(yvi[0]+yvi[1])/2.0-y0, $
                    b0=obs_b0*!dtor, $
                    pang=solar_p*!dtor, $  ; should be zero
                    pixs=jmdipix, $
                    rad=obs_r0)

; convert center of IVM restricted FOV to MDI coordinates
 
xivm = (xvr[1]-xvr[0])/2.0
yivm = (yvr[1]-yvr[0])/2.0
pq2xy,xivm,yivm,mdi2ivm[*,0],mdi2ivm[*,1],xivm_mdi,yivm_mdi

; What is the lat/cmd of this MDI point?  Remove solar rotation
; here since that piece of the transformation is not a pointing shift.
; Also remove parallax in the longitude.  The parallax in latitude is
; accounted for already since get_latlon has b0 = mindex.obs_b0 but below
; lonlat2xy uses the earth value.

; compute the parallax in longitude: mindex.obs_L0-mindex.earth_L0

value = (obs_L0-earth_L0+180.) MOD 360.
neg = where(value LT 0.,count)
if count GT 0 then value(neg) = value(neg) + 360.
plon = value - 180. ; this is mindex.obs_L0-mindex.earth_L0 in [-180,+180]

lat = bilinear(latlon.latitude,xivm_mdi,yivm_mdi)
cmd = bilinear(latlon.cmd,xivm_mdi,yivm_mdi)-diff_rot(ddays,lat,/allen) + plon

; x,y at IVM restricted FOV center in solar coordinates
xy = lonlat2xy([cmd,lat],b0=mag.point.b0,radius=mag.point.radius)
; add shift from restricted FOV to full FOV
xpshift = ((xvr[1]+xvr[0])/2.0 - (nxivm-1.)/2.)*mag.point.pix_size[0]
ypshift = ((yvr[1]+yvr[0])/2.0 - (nyivm-1.)/2.)*mag.point.pix_size[1]
; convert the shift from image coords to solar coords
xshift = xpshift*cos(-mag.point.p)-ypshift*sin(-mag.point.p)
yshift = xpshift*sin(-mag.point.p)+ypshift*cos(-mag.point.p)
xy[0] = xy[0] - xshift
xy[1] = xy[1] - yshift
ll = xy2lonlat(xy,b0=mag.point.b0,radius=mag.point.radius)

magout = mag
magout.point.cmd = ll[0]*!dtor
magout.point.lat = ll[1]*!dtor

if NOT keyword_set(quiet) then begin
   print,'Old center: ',xyivm[0],xyivm[1]
   print,'New center: ',xy[0],xy[1]
   print,'IVM off by ',$
         sqrt((xyivm[0]-xy[0])^2 + $
              (xyivm[1]-xy[1])^2),' arcsec'
   print
endif

llnew = get_latlon(magout,/add)

return,magout

end
