
;+

function ivm_mdi_embed,j,mindex,mdatain,binup=binup,quiet=quiet,jdata=jdata,bmask=bmask,alpha=alpha

;NAME:
;     IVM_MDI_EMBED
;PURPOSE:
;     Embed a magnetogram into a larger MDI FOV
;CATEGORY:
;CALLING SEQUENCE:
;     jnew = ivm_mdi_embed(j,mindex,mdata)
;INPUTS:
;     j = magnetic field structure
;     mindex = MDI index structure, e.g. mdata=mrdfits('file',0,head)
;              mindex = fitshead2struct(head)
;     mdata = MDI data line-of-sight magnetic field array
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     binup = the multiplier to determine the final magnetogram size.
;             Default = 2.0
;     bmask = array with 1's to indicate a good data point in the j
;             structure and 0's to indicate a bad point that should be
;             replaced with MDI data.
;     alpha = the force-free parameter to use in computing the MDI
;             transverse field.  drfault = 0.0.
;OUTPUTS:
;     jnew = new magnetic field structure with the IVM data embedded
;            in the MDI data.  The MDI transverse field is computed
;            from a potential or force-free extrapolation.
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     The MDI data is assumed to be "raw" and is converted to physical
;     units as mindex.bzero+mindex.bscale*mdata
;PROCEDURE:
;     The IVM data is embedded in a larger MDI field-of-view.  The
;     MDI data is supplemented with a potential or LFF extrapolation so
;     that the new magnetoram has bx, by, and bz at all points.
;MODIFICATION HISTORY:
;     T. Metcalf 2005 June 24
;-

; Must use pointing corrected IVM data

nxivm = n_elements(j.b_long[*,0])
nyivm = n_elements(j.b_long[0,*])

nxmdi = n_elements(mdatain[*,0])
nymdi = n_elements(mdatain[0,*])


; Make a map for the mgram for testing purposes
;xy = lonlat2xy([j.point.cmd,j.point.lat]*!radeg,b0=j.point.b0,radius=j.point.radius)
;imap = {data:j.i_cont,xc:xy[0],yc:xy[1],dx:j.point.pix_size[0],dy:j.point.pix_size[1], $
;        time:j.date_obs+' '+j.time_obs,id:j.instrument,roll_angle:-j.point.p*!radeg, $
;        roll_center:xy,dur:0.0,xunits:'arcsecs',yunits:'arcsecs',shoho:0b}

; Convert MDI to earth view

mdata = mindex.bzero+mindex.bscale*mdatain ; convert MDI data to physical units
mmap = mk_soho_map_earth(mindex,mdata,/use_warp_tri)

; Get SOHO orbital parameters
get_fits_time,mindex,sohotime
; orbit = get_orbit(sohotime)
; earth_L0 = orbit.hel_lon_earth*!radeg
; obs_L0 = orbit.hel_lon_soho*!radeg
; earth_B0 = orbit.hel_lat_earth*!radeg
; obs_B0 = orbit.hel_lat_soho*!radeg
; earth_R0 =get_rb0p(sohotime,/rad)
; obs_R0 = soho_fac(sohotime)*earth_R0

; Increase the size of the IVM arrays and use the MDI data to fill the
; blanks in the LOS field.

if NOT keyword_set(binup) then binup = 2.0
binup = float(binup) > 1.0

nxivmnew = long(nxivm*binup+0.5)
nyivmnew = long(nyivm*binup+0.5)
; make sure that the size difference is even to avoid a half pixel shift
if (nxivmnew-nxivm) MOD 2 NE 0 then nxivmnew = (nxivmnew - 1) > nxivm
if (nyivmnew-nyivm) MOD 2 NE 0 then nyivmnew = (nyivmnew - 1) > nyivm
dx = ((nxivmnew-nxivm)/2) > 0
dy = ((nyivmnew-nyivm)/2) > 0
mask = bytarr(nxivmnew,nyivmnew)

; Get coordiantes of new mgram map

x = lindgen(nxivmnew,nyivmnew) MOD nxivmnew
y = lindgen(nxivmnew,nyivmnew) / nxivmnew

; Convert the mgram coordinates to MDI coordinates

xy = lonlat2xy([j.point.cmd,j.point.lat]*!radeg,b0=j.point.b0,radius=j.point.radius)/j.point.pix_size
xy = xy # [[cos(j.point.p),-sin(j.point.p)], [sin(j.point.p),cos(j.point.p)]]
llivm = get_latlon(nx=nxivmnew,ny=nyivmnew, $
                   x0=xy[0],y0=xy[1], $
                   b0=j.point.b0, $
                   pangle = j.point.p, $
                   pixs=j.point.pix_size, $
                   radius=j.point.radius)
ll = llivm
; apply differential rotation
ddays = int2secarr(j.date_obs+' '+j.time_obs,sohotime)/86400.
llivm.cmd = llivm.cmd - diff_rot(ddays,llivm.latitude)

xyivm = lonlat2xy( $
                  reform( $
                          transpose([[[llivm.cmd]],[[llivm.latitude]]],[2,0,1]), $
                          2,nxivmnew*nyivmnew), $
                  b0=j.point.b0,radius=j.point.radius)
xivm = reform(xyivm[0,*],nxivmnew,nyivmnew)
yivm = reform(xyivm[1,*],nxivmnew,nyivmnew)

x0 = (nxmdi+1)/2.-mmap.xc/mmap.dx-1.0
y0 = (nymdi+1)/2.-mmap.yc/mmap.dx-1.0

xm = xivm/mmap.dx + x0
ym = yivm/mmap.dy + y0

;blong = warp_tri(x,y,xm,ym,float(mmap.data),output_size=[nxivmnew,nyivmnew])
degree=3
polywarp,xm,ym,x,y,degree,p,q
blong = poly_2d(float(mmap.data),p,q,2,nxivmnew,nyivmnew,cubic=-0.5,missing=0.0)
bls = blong

if NOT keyword_set(quiet) then begin
endif

if NOT keyword_set(jdata) then jdata = j.b_long
blong[dx:dx+nxivm-1,dy:dy+nyivm-1] = jdata

if n_elements(bmask) EQ n_elements(jdata) then begin
   mask[dx:dx+nxivm-1,dy:dy+nyivm-1] = bmask
   ssmask = where(mask EQ 0b,nssmask)
   if nssmask GT 0 then blong[ssmask] = bls[ssmask]
endif

; Compute the linear force free field to set the transverse field

f = fff(blong,b0=j.point.b0,pangle=j.point.p, $
              lat=ll.latitude*!dtor,cmd=ll.cmd*!dtor, $
              pixel=j.point.pix_size, $
              guard=2,/image,z=0.0,alpha=alpha, $
              /quiet)

btrans = sqrt(f[*,*,0]^2+f[*,*,1]^2)
bazim = atan(-f[*,*,0],f[*,*,1])
bxyz,-btrans*sin(bazim),btrans*cos(bazim),blong, $
     bx,by,bz, $
     p=j.point.p,b0=j.point.b0,lat=ll.latitude*!dtor,cmd=ll.cmd*!dtor
bazim = bazim * !radeg
b_trans = btrans
b_azim = bazim
b_x = bx
b_y = by
b_z = bz
i_cont = fltarr(nxivmnew,nyivmnew)
b_trans[dx:dx+nxivm-1,dy:dy+nyivm-1] = j.b_trans
b_azim[dx:dx+nxivm-1,dy:dy+nyivm-1] = j.b_azim
b_x[dx:dx+nxivm-1,dy:dy+nyivm-1] = j.bx
b_y[dx:dx+nxivm-1,dy:dy+nyivm-1] = j.by
b_z[dx:dx+nxivm-1,dy:dy+nyivm-1] = j.bz
i_cont[dx:dx+nxivm-1,dy:dy+nyivm-1] = j.i_cont

if nssmask GT 0 then begin
   b_trans[ssmask] = btrans[ssmask]
   b_azim[ssmask] = bazim[ssmask]
   b_x[ssmask] = bx[ssmask]
   b_y[ssmask] = by[ssmask]
   b_z[ssmask] = bz[ssmask]
   i_cont[ssmask] = 0.0
endif

jnew = jxyz({b_long: blong,b_trans:b_trans,b_azim:b_azim, $
             bx:b_x, by:b_y, bz:b_z,i_cont:i_cont, $
             point:j.point,latitude:ll.latitude,cmd:ll.cmd},/quiet)

if NOT keyword_set(quiet) then begin
   ;ssscale = where(bls NE blong AND abs(blong) GE 50. AND abs(bls) GE 50.,nssscale)
   ;if nssscale GT 0 then print,'Scaling: MDI/IVM: ', $
   ;                            total(f_div(bls[ssscale],blong[ssscale]))/nssscale,nssscale
   loadct,0
   vmgram,jnew,/inc,btskip=(nxivmnew/128)>0,/arr,len=5,/latlon,/xyz;,image=bytscl(jnew.b_long,-1200.,1200.)
   ;window,/free
   ;loadct,0
   ;plot,bls[ssscale],blong[ssscale],psym=3,xrange=[-2000.,2000.],yrange=[-2000.,2000.],/xst,/yst,xtitle='MDI',ytitle='IVM'
   ;oplot,[-2000.,2000.],[-2000.,2000.],color=128
endif

return,jnew

end
