;+
; NAME:
;   jzerr
; PURPOSE:
;   This function estimates the noise level in the vertical current.  The
;   noise is estimated by computing the standard deviation ("sigma") of the
;   vertical current for those pixels whose transverse field strength is
;   below a certain threshold.  (A current is considered to be significant
;   if its magnitude is at least 3*sigma.)
; CALLING SEQUENCE:
;   sigma = jzerr(mag, bcutoff=bcutoff, mean=mean, npix=npix, help=help)
; INPUT:
;   mag = magnetic field structure.  Must contain the elements BX, BY and JZ.
; OUTPUT:
;   sigma = standard deviation of current values in regions of low transverse
;      field.
; OPTIONAL INPUT PARAMETER:
;   bcutoff = cutoff value of transverse field for defining low-field regions
;      (default: 300 G).
; OPTIONAL OUTPUT KEYWORDS:
;   mean = mean value of current in low-field regions.
;   npix = 2-element vector: [{# of low-field pixels}, {total # of pixels}].
;   help = nonzero to print calling sequence summary and exit.
; MODIFICATION HISTORY:
;   1992 Feb 25:  Written by J-F de La Beaujardiere.
;   1992 Mar  2:  NPIX keyword added (JdLB).
;   2003 Apr 10:  TRM Changed default bcuttoff from 300 to 2*bterror(mag)
;   2005 Jan 13:  TRM Use get_berrs instead of bterror
;-

FUNCTION jzerr, mag, bcutoff=bcutoff, mean=mean, npix=npix, help=help

if (keyword_set(help)) then begin
   print, 'CALLING SEQUENCE:'
   print, '   sigma = jzerr(mag, bcutoff=bcutoff, mean=mean, help=help)'
   return, 'JZERR function called with /HELP keyword.'
endif
 
; Compute transverse field strength
bperp=sqrt(mag.bx^2+mag.by^2)

; Find low-field regions
errs = get_berrs(mag)
if (n_elements(bcutoff) EQ 0) then bcutoff = 2.*errs.sig_btrans
tmpstr1 = strtrim(bcutoff,2)
lowbperp=where(bperp LT bcutoff)
; Make sure low-field region exists
if (n_elements(lowbperp) EQ 1) then begin
   if (lowbperp EQ -1) then begin
     print, 'JZERR: Cannot compute noise level of currents in low-field regions'
     print, '--no pixels were found with transverse field < ' + tmpstr1 + '.'
     print, 'Specify a higher value of BCUTOFF.'
     return, 0.0
   endif
endif
; Determine size of low-field region
npix = [n_elements(lowbperp), n_elements(mag.jz)]

; Determine current in low-field regions
jznoise=mag.jz(lowbperp)

; Determine standard deviation of noise current
sigma = stdev(jznoise,mean)


; Exit procedure
return, sigma
end

