
function lowfunc,r,phi0,phi1,r0,a,alpha

   ; The generating function for the Low solutions

   alpha = ((phi1-phi0)/(r0-a))/(cosh((r-a)/(r0-a))^2)
   return,phi0+(phi1-phi0)*tanh((r-a)/(r0-a))

end

function parallel_area,x,y,x10,y10,x01,y01

; Compute the area of a parallelogram
;
;   x01,y01
;      +---------------+
;     /               /
;    /               /
;   /               /
;  +---------------+
; x,y           x10,y10

   lx2 = (x10-x+0.0)^2+(y10-y+0.0)^2
   ly2 = (x01-x+0.0)^2+(y01-y+0.0)^2
   lx = sqrt(lx2)
   ly = sqrt(ly2)
   dg2 = (x01-x10+0.)^2+(y01-y10+0.)^2
   cosA = (lx2+ly2-dg2)/(2.0*lx*ly)
   sinA = sqrt(1.0-cosA^2)
   return,abs(lx*ly*sinA)

end


;+

function lowff,nx,ny,nz,func=func,phi0=phi0,phi1=phi1,r0=r0,a=a,B0=B0, $
               bangle=bangle,pangle=p,lat=lat,cmd=cmd,xnum=xnum,ynum=ynum

;NAME:
;     LOWFF
;PURPOSE:
;     Computes a force-free field model
;CATEGORY:
;CALLING SEQUENCE:
;     b = lowff(nx,ny,nz)
;INPUTS:
;     nx,ny,nz = dimention of the field.  Default = 40,40,1
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     func = name of function to use as the generating function.
;            Called as func(r,phi0,phi1,r0,a,dfdr).  Default = 'lowfunc'.
;            r,phi0,phi1,r0,a are input; dfdr is output as the 
;            derivative of the function wrt r.
;     phi0,phi1,r0,a = Controls the model.
;                      If the default generating function ('lowfunc') is used,
;                      then the parameters have the following meaning:
;                      phi0: The twist of the central field.  phi0=phi1 implies
;                            a purely potential field.  Radians.
;                      phi1: The orientation of the background potential field.
;                            Radians.
;                         a: The depth of the line current generating the
;                            background potential field.  Image Pixels.
;                        r0: The distance over which the field goes to the 
;                            background potential field.  Image Pixels.
;     B0 = The maximum magnetic field strength.  Default = 1000G.
;     bangle = solar b-angle.  Default = 0.
;     pangle = solar p-angle.  Default = 0.
;     lat = latitude of center of FOV.  Default = 0.
;     cmd = CMD of center of FOV.  Default = 0.
;     xnum,ynum = Number of sub-pixels to use in the integration over each
;                 pixel.  Should be an odd integer GE 1. Default = 5,5
;OUTPUTS:
;     b = magnetic field structure
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
;     B. C. Low, 1982, Solar Physics, 77, 43
;
;     If the coordinates are not for the center of the disk, the image plane
;     has a uniform grid, but the heliographic field does not (just as if the
;     field were observed).
;
;     The field is integrated over each pixel in the heliographic grid to 
;     better simulate observations.
;MODIFICATION HISTORY:
;     T. Metcalf 1994-12-01
;-

   if n_elements(nx) LE 0 then nx = 40L
   if n_elements(ny) LE 0 then ny = 40L
   if n_elements(nz) LE 0 then nz = 1L
   if n_elements(func) LE 0 then func='lowfunc'
   if n_elements(phi0) LE 0 then phi0 = -!pi/2.0
   if n_elements(phi1) LE 0 then phi1 = +!pi/2.0
   if n_elements(r0) LE 0 then r0 =sqrt(nx^2+ny^2)/(2.0*!pi)
   if n_elements(a) LE 0 then a = r0/3.
   if n_elements(B0) LE 0 then B0 = 1000.
   if n_elements(bangle) LE 0 then bangle=0.
   if n_elements(p) LE 0 then p=0.
   if n_elements(lat) LE 0 then lat=0.
   if n_elements(cmd) LE 0 then cmd=0.
   if n_elements(xnum) LE 0 then xnum=5L & xnum=long(xnum)>1L
   if n_elements(ynum) LE 0 then ynum=5L & ynum=long(ynum)>1L

   if nz GT 1 then begin
      bx = fltarr(nx,ny,nz)
      by = fltarr(nx,ny,nz)
      bz = fltarr(nx,ny,nz)
      alpha = fltarr(nx,ny,nz)
   endif else begin
      bx = fltarr(nx,ny)
      by = fltarr(nx,ny)
      bz = fltarr(nx,ny)
      alpha = fltarr(nx,ny)
   endelse

   B0a = B0*a
   nx2 = nx/2.0
   ny2 = ny/2.0
   nx1 = nx - 1L
   ny1 = ny - 1L

   xf = set_xform(bangle*!dtor,p*!dtor,lat*!dtor,cmd*!dtor)

   xi = float(lindgen(nx,ny) MOD long(nx))
   yi = float(lindgen(nx,ny) / long(nx))
   xform,xi,yi,xh0,yh0,transform=xf.zimage2helio

   xh = xh0 - mean(xh0)  ; Use Heliographic coordinates.  The image
   yh = yh0 - mean(yh0)  ; coordinates are a regular grid, not the helio coords

   ; The field is integrated over each pixel and xnum and ynum specify the
   ; the increase in the resolution for the integration, i.e. the number of
   ; sub-pixels in the summation.

   xsize = 1.0/(2.0*xnum)  ; The size of 1/2 sub-pixel in image coordinates
   ysize = 1.0/(2.0*ynum)
   valx = -0.5+(lindgen(xnum,ynum) MOD long(xnum))/float(xnum)+xsize
   valy = -0.5+(lindgen(xnum,ynum)/long(xnum))/float(ynum)+ysize

   for i=0L,nx-1L do begin
      for j=0L,ny-1L do begin
         for k=0L,nz-1L do begin

            z = float(k)

            ; Compute the area of the entire pixel (assumed a parallelogram)

            xedge = bilin(xh,[i-0.5,i+0.5,i-0.5],[j-0.5,j-0.5,j+0.5])
            yedge = bilin(yh,[i-0.5,i+0.5,i-0.5],[j-0.5,j-0.5,j+0.5])
            area = parallel_area(xedge(0),yedge(0),xedge(1),yedge(1), $
                                 xedge(2),yedge(2))

            ; Now integrate the field over the pixel.  xval and yval are
            ; the x and y positions within the pixel to compute the field.

            xval = bilin(xh,i+valx,j+valy)
            yval = bilin(yh,i+valx,j+valy)

            for i1=0L,xnum-1L do begin
               for j1=0L,ynum-1L do begin
                  
                  x1 = xval(i1,j1)
                  y1 = yval(i1,j1)
                  r1 = sqrt(x1^2+y1^2+(z+a)^2)

                  ; compute the area of the sub-pixel

                  xedge = bilin(xh,[i+valx(i1,j1)-xsize,  $
                                       i+valx(i1,j1)+xsize,  $
                                       i+valx(i1,j1)-xsize], $
                                      [j+valy(i1,j1)-ysize,  $
                                       j+valy(i1,j1)-ysize,  $
                                       j+valy(i1,j1)+ysize])
                  yedge = bilin(yh,[i+valx(i1,j1)-xsize,  $
                                       i+valx(i1,j1)+xsize,  $
                                       i+valx(i1,j1)-xsize], $
                                      [j+valy(i1,j1)-ysize,  $
                                       j+valy(i1,j1)-ysize,  $
                                       j+valy(i1,j1)+ysize])
                  area1 = parallel_area(xedge(0),yedge(0),xedge(1),yedge(1), $
                                        xedge(2),yedge(2))

                  phi = call_function(func,r1,phi0,phi1,r0,a,dpdr)
                  cphi = cos(phi)
                  sphi = sin(phi)
                  za = z+a
                  yza2 = y1^2+za^2
                  ryza2 = r1*yza2
                  bx(i,j,k) = bx(i,j,k) - area1*(B0a/r1)*cphi
                  by(i,j,k) = by(i,j,k) + $
                              area1*((B0a*x1*y1/ryza2)*cphi - (B0a*za/yza2)*sphi)
                  bz(i,j,k) = bz(i,j,k) + $
                              area1*((B0a*x1*za/ryza2)*cphi + (B0a*y1/yza2)*sphi)
                  alpha(i,j,k) = alpha(i,j,k) - area1*dpdr  ; Why is it minus??
               endfor
            endfor

            bx(i,j,k) = bx(i,j,k)/area
            by(i,j,k) = by(i,j,k)/area
            bz(i,j,k) = bz(i,j,k)/area
            alpha(i,j,k) = alpha(i,j,k)/area

         endfor
      endfor
   endfor

   xform,bx,by,bz,hbx,hby,blong,transform=xf.helio2image
   btrans = sqrt(hbx^2+hby^2)
   azim = atan(-hbx,hby)*!radeg

   point = {pointlowff,b0:float(bangle*!dtor),p:float(p*!dtor), $
                       lat:float(lat*!dtor), $
                       cmd:float(cmd*!dtor),pix_size:1.}
   return,add2str(dummy,['B_LONG','B_TRANS','B_AZIM','BX','BY','BZ','POINT', $
                         'I_CONT','B_PXD','B_PYD','ALPHA'], $
                        blong,btrans,azim,bx,by,bz,point,bz,xh0,yh0,alpha)

end