
;+

function mdi2bfits,imdi,dmdi,iwl,dwl,noplot=noplot,subarray=subarray, $
                   fixhole=fixhole,cutlimb=cutlimb,rotangle=rotangle, $
                   crholimb=crholimb,alpha=alpha

;NAME:
;     MDI2BFITS
;PURPOSE:
;     Convert MDI data to the UH FITS format
;CATEGORY:
;CALLING SEQUENCE:
;     b = mdi2bfits(imdi,dmdi,subarry=subarry)
;INPUTS:
;     imdi,dmdi = MDI index and data
;OPTIONAL INPUT PARAMETERS:
;     iwl,dwl = MDI WL index and data (to fill i_cont field)
;KEYWORD PARAMETERS
;     /noplot = no plots
;     subarry = [x1,x2,y1,y2].  Required.
;     /fixhole = attempt to inerpolate over hole in spot centers
;                (not recommended).
;     /cutlimb = if there are offlimb points don't return them.
;                Otherwise they are set to zero and returned.
;     crholimb = cos(rho) which is off the limb (def = 0.2)
;     rotangle = rotate the data (degrees, CCW).  Useful if the region
;                is on the limb.
;     alpha = force-free alpha used in extrapolation to derive BX, BY
;             and BZ.
;OUTPUTS:
;     b = UH FITS structure
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;PROCEDURE:
;MODIFICATION HISTORY:
;     T. Metcalf 2003-Oct-22
;     2004-feb-16 Added alpha keyword
;-

if n_elements(alpha) LE 0 then alpha = 0.0

mdidate = anytim(imdi.date_obs,/yoh)

; Get the pixel size of MDI image

mdipix = imdi.im_scale
; This corrects the known problem with the MDI plate scale.  It might be
; necessary to update the value periodically???
if abs(mdipix-1.9778400) LT 0.00001 then begin
   mdipix = 1.982
   message,/info,'Correcting MDI plate scale of 1.97784 to '+string(mdipix)
endif
mdipix_earth = mdipix*imdi.EARTH_R0/imdi.OBS_R0

nxmdi = n_elements(dmdi[*,0,0])
nymdi = n_elements(dmdi[0,*,0])

ll = get_latlon(nx=nxmdi,ny=nymdi, $
                x0=imdi.xcen/mdipix, $
                y0=imdi.ycen/mdipix, $
                b0=get_rb0p(mdidate,/b0),pangle=0.0, $
                pixsize=[mdipix,mdipix], $
                radius=get_rb0p(mdidate,/radius))

if NOT keyword_set(noplot) then begin
   ; Display the lon/lat as a check
   bin=1
   device,get_screen_size=screensize
   if nxmdi GE screensize[0] OR nymdi GE screensize[1] then begin
      bin = (nxmdi/screensize[0]+1)>(nymdi/screensize[1]+1)
   endif
   if !d.x_size NE nxmdi/bin OR !d.y_size NE nymdi/bin then $
      window,/free,xs=nxmdi/bin,ys=nymdi/bin
   tvscl,congrid((dmdi>(-1000.))<(+1000.),nxmdi/bin,nymdi/bin)
   ocontour2,ll.latitude,levels=(findgen(37)-18.)*5.,bin=bin
   ocontour2,ll.cmd,levels=(findgen(37)-18.)*5.,bin=bin
   empty
endif

if keyword_set(rotangle) then begin
   ; subarray will now give the coords *after* rotation
   rmdi=shift_img(dmdi,[0.,0.],rot12=rotangle,anchor=[imdi.mdi_x0,imdi.mdi_y0])
   ll.latitude = shift_img(ll.latitude,[0.,0.],rot12=rotangle,anchor=[imdi.mdi_x0,imdi.mdi_y0])
   ll.cmd = shift_img(ll.cmd,[0.,0.],rot12=rotangle,anchor=[imdi.mdi_x0,imdi.mdi_y0])
   pangle = rotangle/!radeg
endif else begin
   rmdi = dmdi
   pangle = 0.0
endelse

lat_cen = ll.latitude[(subarray[1]+subarray[0])/2,(subarray[3]+subarray[2])/2]
cmd_cen = ll.cmd[(subarray[1]+subarray[0])/2,(subarray[3]+subarray[2])/2]

jmdi = {b_long:rmdi[subarray[0]:subarray[1],subarray[2]:subarray[3]], $
        latitude: ll.latitude[subarray[0]:subarray[1],subarray[2]:subarray[3]], $
        cmd: ll.cmd[subarray[0]:subarray[1],subarray[2]:subarray[3]], $
        date_obs: mdidate, $
        point:{mdipoint, $
               lat:lat_cen*!dtor, $
               cmd:cmd_cen*!dtor, $
               b0:get_rb0p(mdidate,/b0), $
               p:pangle, $
               pix_size:[mdipix,mdipix], $
               radius:get_rb0p(mdidate,/radius)} $
       }

;vmgram,jmdi,/inc,/novect,/latlon

nx = n_elements(jmdi.b_long[*,0])
ny = n_elements(jmdi.b_long[0,*])

; Find the limb for later use
crho = cos(jmdi.latitude*!dtor)*cos(jmdi.point.b0)*cos(jmdi.cmd*!dtor)+ $
       sin(jmdi.latitude*!dtor)*sin(jmdi.point.b0)

if n_params() GE 4 then begin  ; Add WL
   nxwl = n_elements(dwl[*,0,0])
   nywl = n_elements(dwl[0,*,0])
   wldate = anytim(iwl.date_obs,/yoh)
   ; Get the pixel size of MDI WL image
   wlpix = iwl.im_scale
   ; This corrects the known problem with the MDI plate scale.  It might be
   ; necessary to update the value periodically???
   if abs(wlpix-1.9778400) LT 0.00001 then begin
      wlpix = 1.982
   endif
   wlpix_earth = wlpix*iwl.EARTH_R0/iwl.OBS_R0
   llwl = get_latlon(nx=nxwl,ny=nywl, $
                   x0=iwl.xcen/wlpix, $
                   y0=iwl.ycen/wlpix, $
                   b0=get_rb0p(wldate,/b0),pangle=0.0, $
                   pixsize=[wlpix,wlpix], $
                   radius=get_rb0p(wldate,/radius))
   if NOT keyword_set(noplot) then begin
      ; Display the lon/lat as a check
      bin=1
      device,get_screen_size=screensize
      if nxwl GE screensize[0] OR nywl GE screensize[1] then begin
         bin = (nxwl/screensize[0]+1)>(nywl/screensize[1]+1)
      endif
      if !d.x_size NE nxwl/bin OR !d.y_size NE nywl/bin then $
         window,/free,xs=nxwl/bin,ys=nywl/bin
      tvscl,congrid((dwl>(-10000.))<(+10000.),nxwl/bin,nywl/bin)
      ocontour2,llwl.latitude,levels=(findgen(37)-18.)*5.,bin=bin
      ocontour2,llwl.cmd,levels=(findgen(37)-18.)*5.,bin=bin
      empty
   endif
   if keyword_set(rotangle) then begin
      ; subarray will now give the coords *after* rotation
      rwl=shift_img(dwl,[0.,0.],rot12=rotangle,anchor=[iwl.mdi_x0,imdi.mdi_y0])
      llwl.latitude = shift_img(llwl.latitude,[0.,0.],rot12=rotangle,anchor=[iwl.mdi_x0,iwl.mdi_y0])
      llwl.cmd = shift_img(llwl.cmd,[0.,0.],rot12=rotangle,anchor=[iwl.mdi_x0,iwl.mdi_y0])
   endif else begin
      rwl = dwl
   endelse
   ddays = int2secarr(mdidate,wldate)/86400.
   if abs(ddays) LE 0.25 then begin
      dr = diff_rot(ddays,jmdi.point.lat*!radeg,/allen)
      ; Figure out what WL subarray to use.  It will differ from the magnetic
      ; subarray due to solar rotation.
      xy0 = lonlat2xy([jmdi.point.cmd,jmdi.point.lat]*!radeg,mdidate)
      xy1 = lonlat2xy([jmdi.point.cmd-dr/!radeg,jmdi.point.lat]*!radeg,wldate)
      xyshift = long(xy1-xy0+0.5)/wlpix_earth
      wlsubarray = subarray
      ; pad it by 10 on each side to make sure we get enough
      wlsubarray[0] = (wlsubarray[0]+xyshift[0]-10L)>0L
      wlsubarray[1] = (wlsubarray[1]+xyshift[0]+10L)<(nxwl-1)
      wlsubarray[2] = (wlsubarray[2]+xyshift[1]-10L)>0L
      wlsubarray[3] = (wlsubarray[3]+xyshift[1]+10L)<(nywl-1)
      ; Interpolate the WL to the magnetic grid
      wl = bilinxy(rwl[wlsubarray[0]:wlsubarray[1],wlsubarray[2]:wlsubarray[3]], $
                   jmdi.cmd,jmdi.latitude, $
                   xarr=llwl.cmd[wlsubarray[0]:wlsubarray[1], $
                                 wlsubarray[2]:wlsubarray[3]]+dr, $
                   yarr=llwl.latitude[wlsubarray[0]:wlsubarray[1], $
                                      wlsubarray[2]:wlsubarray[3]], $
                   degree=3)
      jmdi = add2str(jmdi,['i_cont','date_obs_i_cont'],wl,wldate)
      if keyword_set(fixhole) then begin
         ; check for a hole in the sunspot field and fix it approximately
         ;crho=cos(jmdi.latitude*!dtor)*cos(jmdi.point.b0)*cos(jmdi.cmd*!dtor)+ $
         ;     sin(jmdi.latitude*!dtor)*sin(jmdi.point.b0)
         ;  Linear function: 1300 G at 2000, 800 G at 5000
         blguess = (1633.33-jmdi.i_cont*500.0/3000.0)
         bad = where(abs(jmdi.b_long) LT blguess and $
                     jmdi.b_long GT -3000.0 and $
                     jmdi.i_cont GT 1000.0 and $
                     jmdi.i_cont LT 5000. and $
                     crho GT 0.2,nbad)
         if nbad GT 0 then jmdi.b_long[bad] = blguess[bad]*((jmdi.b_long[bad] GE 0)*2-1)
      endif
   endif else message,/info,'Time difference between the magnetogram and white light images is too long.  Not adding WL to structure.'+string(ddays)
endif   

; Compute Bx, By, Bz from a potential extrapolation and add them in

if not keyword_set(crholimb) then crholimb = 0.2
ondisk = where(crho GE crholimb,nondisk)
offdisk = where(crho LT crholimb,noffdisk)
if nondisk GE n_elements(jmdi.b_long) then begin
   ; All points are on the disk so just proceed
   fphotxyz = fff(jmdi,normal=0,z=[0.0],alpha=alpha)
   fphotlos = fff(jmdi,normal=0,image=1,z=[0.0],alpha=alpha)
   jmdi = add2str(jmdi,['bx','by','bz'],fphotxyz[*,*,0,0], $
                                        fphotxyz[*,*,0,1], $
                                        fphotxyz[*,*,0,2])
   
   jmdi = add2str(jmdi,['b_trans','b_azim'], $
                        sqrt(fphotlos[*,*,0,0]^2+fphotlos[*,*,0,1]^2), $
                        atan(-fphotlos[*,*,0,0],fphotlos[*,*,0,1])/!dtor)
endif else begin
   ; fff can't deal with offlimb points
   ; Cut out a rectangular region on the disk and zero out the rest
   xondisk = ondisk MOD nx
   yondisk = ondisk / nx
   xmin = min(xondisk)  ; boundaries of region on disk
   xmax = max(xondisk)
   ymin = min(yondisk)
   ymax = max(yondisk)
   repeat begin
      xmin = xmin + 1 & xmax = xmax - 1
      ymin = ymin + 1 & ymax = ymax - 1
      if xmin GT xmax or ymin GT ymax then message,'All point are off the limb'
   endrep until min(crho[xmin:xmax,ymin:ymax] GE crholimb)
   repeat begin
      notdone = 0
      if xmin GT 0 then begin
         if min(crho[xmin-1:xmax,ymin:ymax]) GE crholimb then begin
            xmin = xmin -1
            notdone = 1
         endif
      endif
      if xmax LT nx-1 then begin 
         if min(crho[xmin:xmax+1,ymin:ymax]) GE crholimb then begin
            xmax = xmax +1
            notdone = 1
         endif
      endif
      if ymin GT 0 then begin
         if min(crho[xmin:xmax,ymin-1:ymax]) GE crholimb then begin
            ymin = ymin -1
            notdone = 1
         endif
      endif
      if ymax LT ny-1 then begin
         if min(crho[xmin:xmax,ymin:ymax+1]) GE crholimb then begin
            ymax = ymax +1
            notdone = 1
         endif
      endif
   endrep until notdone eq 0
   sjmdi = substruc(jmdi,xmin*1,xmax*1,ymin*1,ymax*1) ; *1 so they don't change
   sjmdi.point.lat = jmdi.latitude[(xmax+xmin)/2,(ymax+ymin)/2]*!dtor
   sjmdi.point.cmd = jmdi.cmd[(xmax+xmin)/2,(ymax+ymin)/2]*!dtor
   fphotxyz = fff(sjmdi,normal=0,z=[0.0],alpha=alpha)
   fphotlos = fff(sjmdi,normal=0,z=[0.0],alpha=alpha,image=1)
   bx = fltarr(nx,ny) & by = fltarr(nx,ny) & bz = fltarr(nx,ny) 
   b_trans = fltarr(nx,ny) & b_azim = fltarr(nx,ny)
   bt = sqrt(fphotlos[*,*,0,0]^2+fphotlos[*,*,0,1]^2)
   ba = atan(-fphotlos[*,*,0,0],fphotlos[*,*,0,1])/!dtor
   bx[xmin,ymin] = fphotxyz[*,*,0,0]
   by[xmin,ymin] = fphotxyz[*,*,0,1]
   bz[xmin,ymin] = fphotxyz[*,*,0,2]
   b_trans[xmin,ymin] = bt
   b_azim[xmin,ymin] = ba
   jmdi = add2str(jmdi,['bx','by','bz','b_trans','b_azim'],bx,by,bz,b_trans,b_azim)
   sjmdi = add2str(sjmdi,['bx','by','bz','b_trans','b_azim'],fphotxyz[*,*,0,0], $
                                        fphotxyz[*,*,0,1], $
                                        fphotxyz[*,*,0,2],bt,ba)
   if keyword_set(cutlimb) then jmdi = sjmdi
endelse

if NOT keyword_set(noplot) then begin
   vmgram,jmdi,/inc,/novect,/latlon,/xyz, $
          levels=[150.,300.,600.,1200.,2400.],pixsize=mdipix
endif

if keyword_set(go_fff) then begin
   zh = findgen(100)*2
   f = fff(jmdi,normal=0,z=zh,xh=xh,yh=yh,alpha=alpha)
   nf = fff2bff(f,b0=jmdi.point.b0,pa=jmdi.point.p, $
                cmd=jmdi.cmd*!dtor,lat=jmdi.latitude*!dtor,$
                pixel=jmdi.point.pix_size, $
                alpha=alpha,xh=xh,yh=yh,zh=zh,$
                i_cont=jmdi.i_cont)
   jmdi = add2str(jmdi,['b_pxd','b_pyd'],nf.b_pxd,nf.b_pyd)

   v=blines(nf,image=jmdi.i_cont)
   junk=blines(color_blines(v,color=[255,255]), $
               /vertices, $
               /rot2image, $
               image=jmdi.i_cont^3)
   sep = separatrix(jmdi,fff=nf,vertices=vertices)
   ;save,jmdi,sep,v,nf,vertices,file='/u1/metcalf/010825/mdifff.sav'
endif

if keyword_set(ps) then begin
   ;restore,'/u1/metcalf/010825/mdifff.sav'
   ps,'fig_mdi_trace_blines.eps',/encaps
   device,bits_per_pix=8,/color
   junk = blines(color_blines(v,color=[0,0]), $
                 /vertices, $
                 /rot2image, $
                 image=jmdi.i_cont^3)
   psclose
endif

return,jmdi

end
