;+
; NAME:
;   mgram_helio
; PURPOSE: 
;   This procedure makes an ambiguity-resolved, disk-center, vector
;   magnetogram with the correct aspect ratio.
; CALLING SEQUENCE:
;   mgram_helio, mag
; INPUT:
;   mag = structure containing magnetic field data.  Required elements:
;         BX, BY, BZ, B_PXD, B_PYD, POINT.
; OPTIONAL INPUTS:
; KEYWORDS:
;   image = array to display as an image under the field contours.
;           Default=mag.i_cont if present.
;           Scales image to avoid colors 0 through 9.
;   /noimage = Do not display an underlying image
;   pixsize = pixel size in then image interpolation.  Larger values make
;             the interpolation smoother (and the .ps file larger).  Also
;             used as the pixel size for axis labels. Def = 1.0.
;   xstep = the size of an x pixel (def = 1.0) relative to a y pixel
;   lv = contour levels for longitudinal field
;   length = relative length of arrows used to plot transverse field.
;      Default is 1.0, which makes the longest arrow the length of one pixel.
;   missing = smallest value of transverse field to plot.
;   /new_window = nonzero to open new window for magnetogram. Ignored if 
;      postscript.  By default, the current window is resized and reopened.
;   /nodash = Use a solid contour for both positive and negative field
;   /novectors = Do not plot field vectors
;   c_thick = Thickness of contours. If scalar, the thickness is used for both
;             positive and negative contours.  If a 2 element vector, the first
;             element is for the positive contours and the second element is 
;             for the negative contours.  c_thick is multiplyed by (!p.thick>1)
;             Def = [1,1]. 
;   c_color = color of contours.  If scalar, the color index is used for both
;             positive and negative contours.  If a 2 element vector, the first
;             color is for the positive contours and the second element is 
;             for the negative contours.
;   arr_color = the color index for the arrows.
;   /incolor = Make the contours etc in color.  Adjusts the color table.
;   title = title for plot
;   xtitle = title for x axis
;   ytitle = title for y axis
; SIDE EFFECTS:
;   Either a magnetogram is displayed on the monitor screen or
;      a PostScript file is created in the current working directory.
;   The !P.POSITION system variable is modified.
;   The color table is changed if /incolor is set.
; MODIFICATION HISTORY:
;   25 Jun 1992:  Written by J-F de La Beaujardiere.
;   xx Feb 1994:  T. Metcalf: Extensive modifications.
;                             Added ability to plot underlying image.
;                             Added color and linestyle selections.
;                             Made pixels square using pixsquare.pro
;   28 Apr 1994:  T. Metcalf: Added /incolor keyword.
;                             Vectors now plot in the center of the pixels.
;                             Added titles to plot.
;-


PRO mgram_helio, mag, lv=lv, pixsize=pixsize, length=length, $
    missing=inmissing,novectors=novectors, image=image,title=title, $
    new_window=new_window, c_color=c_color,arr_color=arr_color, $
    nodash=nodash,c_thick=c_thick,noimage=noimage,incolor=incolor, $
    xtitle=xtitle,ytitle=ytitle,xstep=xstep

ps = strupcase(!d.name) EQ 'PS'

if n_elements(pixsize) LE 0 then pixsize = 1
case n_elements(c_color) of
   0: begin
         if keyword_set(noimage) then begin
            pcolor = 256
            mcolor = 256
         endif else begin
            pcolor = 0
            mcolor = 0
         endelse
      end
   1: begin
         pcolor = c_color
         mcolor = c_color
      end
   2: begin
         pcolor = c_color(0)
         mcolor = c_color(1)
      end
   else: message,'Too many elements in c_color vector'
endcase
case n_elements(c_thick) of
   0: begin
         pthick = 1
         if NOT keyword_set(nodash) and ps then mthick=2 else mthick = 1
      end
   1: begin
         pthick = c_thick
         mthick = c_thick
      end
   2: begin
         pthick = c_thick(0)
         mthick = c_thick(1)
      end
   else: message,'Too many elements in c_thick vector'
endcase
if n_elements(arr_color) LE 0 then begin
   if NOT keyword_set(noimage) then arr_color=0 else arr_color=!d.table_size-1
endif
if keyword_set(incolor) then begin
   mcolor=2
   pcolor=3
   arr_color = 4
   tvlct,/get,r,g,b
   r(mcolor)   =  0b & g(mcolor)   =  0b & b(mcolor)   =255b
   r(pcolor)   =255b & g(pcolor)   =  0b & b(pcolor)   =  0b
   if keyword_set(noimage) then begin
      r(arr_color)=0b & g(arr_color)=0b & b(arr_color)=0b
   endif else begin
      r(arr_color)=255b & g(arr_color)=255b & b(arr_color)=255b
   endelse
   nodash=1
   tvlct,r,g,b
endif
if NOT keyword_set(nodash) then begin
   pltype = 0
   if ps then mltype=1 else mltype=2
endif else begin
   pltype = 0
   mltype = 0
endelse

; Define coordinates of the contour surface
nx = n_elements(mag.bz(*,0))
ny = n_elements(mag.bz(0,*))
if tag_index(mag,'B_PXD') GE 0 AND tag_index(mag,'B_PYD') GE 0 then begin
   xpos = (mag.b_pxd-min(mag.b_pxd))*pixsize
   ypos = (mag.b_pyd-min(mag.b_pyd))*pixsize
endif else begin
   xpos = float(lindgen(nx,ny) MOD nx)*pixsize
   ypos = float(lindgen(nx,ny) / nx)*pixsize
   message,/info,'WARNING: B_PXD and/or B_PYD missing, field ' + $
                 'vector direction off by p-angle'
endelse
nnx = max(xpos)-min(xpos)+1
nny = max(ypos)-min(ypos)+1

; Determine image size
xsize = max(xpos)-min(xpos)
ysize = max(ypos)-min(ypos)

if NOT ps then begin
   zoom=500./((ysize+xsize)/2.+1)
   ; Open window (factor of 1.25 leaves room for titles; ZOOM magnifies)
   if (keyword_set(new_window) OR (!d.window EQ -1)) AND (!d.name NE 'Z') then begin
      window, /free, xsize=(xsize+1)*1.25*zoom, ysize=(ysize+1)*1.25*zoom
   endif else erase
endif

; Set contour levels
if (n_elements(lv) EQ 0) then lv = [100,200,400,800,1600]

if NOT keyword_set(title) then $
   if tag_index(mag,'INSTRUMENT') GE 0 then title=string(mag.instrument) $
   else title=''
if NOT keyword_set(xtitle) then begin
   xtitle='Contours at +\- '
   for i=0,n_elements(lv)-1 do xtitle=xtitle+strcompress(' '+string(lv(i)))
endif
if NOT keyword_set(ytitle) then ytitle='Vertical\Horizontal Magnetic Field'

if ps then contour,mag.bz,xpos,ypos,levels=lv,/nodata,xst=5,yst=5  ; New plot

; Make square pixels

!p.position = pixsquare(nnx,nny,position=[.1,.1,.9,.9],xstep=xstep)

if tag_index(mag,'I_CONT') GE 0 and NOT keyword_set(image) then $
   image=mag.i_cont

;  Warp the image to the correct shape and draw it

if keyword_set(image) and NOT keyword_set(noimage) then begin
   x = float(lindgen(nx,ny) MOD (nx))
   y = float(lindgen(nx,ny) / (nx))
   degree=3
   polywarp,x,y,xpos-min(xpos),ypos-min(ypos),degree,p,q
   if ps then missing = max(image) else missing = min(image)
   i_cont = poly_2d(image,p,q,1,nnx,nny,missing=missing)
   if NOT ps then begin
      dc = convert_coord([!p.position(0),!p.position(2)], $
                         [!p.position(1),!p.position(3)],/normal,/to_device)
      i_cont = congrid(i_cont,dc(0,1)-dc(0,0),dc(1,1)-dc(1,0))
   endif
   tv,bytscl(i_cont,top=!d.table_size-12)+10b,!p.position(0),!p.position(1), $
      /normal,xsize=!p.position(2)-!p.position(0), $
              ysize=!p.position(3)-!p.position(1)
endif

; Plot the longitudinal fields using contours
; Positive (out of the plane) fields 
contour, mag.bz, xpos, ypos, levels=lv,c_thick=pthick*(!p.thick>1), $
   xstyle=1, ystyle=1,/noerase,c_color=pcolor,c_linestyle=pltype,title=title, $
   xtitle=xtitle,ytitle=ytitle
; Negative fields
contour, -mag.bz, xpos, ypos, levels=lv, $
   xstyle=1, ystyle=1, c_linestyle=mltype, c_thick=mthick*(!p.thick>1),$
    c_labels=0,/noerase, c_color=mcolor

; Plot transverse fields using arrows
if NOT keyword_set(novectors) then begin
   if (n_elements(length) EQ 0) then length = 1
   if (n_elements(inmissing) EQ 0) then missing = -1 else missing = inmissing
   ; plot the vectors in the middle of the pixels
   xpos1 = 0.5*pixsize+xpos*float(nx)/float(nx+1.0)
   ypos1 = 0.5*pixsize+ypos*float(ny)/float(ny+1.0)
   vectorfield, mag.bx, mag.by, xpos1, ypos1, missing=missing, length=length, $
      g1000=g1000, color=arr_color, /noerase, /arrows

   ; Draw and label 1000-Gauss reference line segment for transverse field
   yout = -5*pixsize  ;vertical location of reference seqment
   plots, [0,g1000], [yout,yout]
   xyouts, g1000+pixsize, yout, '(1000 G)'  ;label the reference segment
endif

; Exit procedure
return
end

