
;+

function ffenergy, j, error, energy, alpha=alphain, guard=guardin, $
                   blong=blong, extrapolate=extrapolate,step=stepin, $
                   mask=mask,ubound=ubound,fx=fx,fy=fy,fz=fz,fp=fp, $
                   randomize=randomize,noprogress=noprogress

;NAME:
;     FFENERGY
;PURPOSE:
;     Computes the energy of a force-free field extrapolation 
;CATEGORY:
;CALLING SEQUENCE:
;     energy = ffenergy(j,error,earr)
;INPUTS:
;     j = magnetic field structure from bfits.pro.  j must have the 180 degree
;         ambiguity reloved, unless both the /blong and /extrapolate keywords
;         are set.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     /extrapolate = Do a force free extrapolation and calculate the energy of
;                    the extrapolated field.  If /extrapolate is not set, then
;                    calculate the energy of the measured field.  If the field
;                    is not force-free, this energy will not mean much.
;     alpha = the force-free parameter to use.  Default = 0.0 (potential
;             field).  Only used if /extrapolate is set.
;     guard = the size of the guard band in the extrapolation.  Default = 2.
;             Only used of /extrapolate is set.  Passed through to fff.pro.
;     /blong = Use the longitudinal field as the boundary condition.  The
;              default is to use the vertical field.  Only used if
;              /extrapolate is set.
;     /randomize = randomize the input to the extrapolation using the
;                  magnitude of the field errors.  Only used if
;                  /extrapolate is set.
;     step = integer giving the step size when calculating the energy for
;            different coordinate systems.  Default = 1 which uses every
;            possible coordinate system.  step=2 would skip every other point,
;            etc.   Higher values are less accurate, but faster.  Set step GE
;            nx,ny to avoid the multiple coordinate systems.
;     mask = Index array giving the pixels to include in the energy 
;            calculation.  don't use this as it does not make physical 
;            sense to exclude some points from the energy calculation.
;     /ubound = Calculate an upper bound to the energy ala Aly 1984.  If this
;               keyword is set, the actual energy is not computed and not
;               returned.  The returned value is only the upper bound.
;     fx,fy,fz,fp = Lorentz force integrals and the pressure integral
;     /noprogress = Do not print progress status
;OUTPUTS:
;     energy = total energy of the magnetic field in ergs
;     error = estimated error on the energy
;     earr = array of energy calculations for all coordinates systems
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     - The calculation assumes that all the flux is with the field-of-view
;     - mag.point.pix_size assumed to be in arc seconds
;     - mag.bx, mag.by, mag.bz must exist with 180 ambiguity resolved
;PROCEDURE:
;     Uses the magnetic virial theorem. If the field is not force-free, 
;     the computed energy will not mean much.
;
;     To calculate the energy of the observed field (must be force-free!) use:
;
;          ffenergy(j)
;
;     To calculate the potential field energy use
;
;          ffenergy(j,/extrapolate).
;
;     If the 180 deg ambiguity has not been resolved, then the only energy
;     calculation which will work is an extrapolation from b_long:
;
;          ffenergy(j,/blong,/extrapolate,alpha=whatever)
;
;MODIFICATION HISTORY:
;     T. Metcalf  1998-Mar-18
;                 1999-Apr-19 - Added coordinate transformations to account for
;                               projection effects.
;                             - The default is now to compute the energy of the
;                               observed field.  To compute the energy of the
;                               extrapolation use the /extraploate keyword.
;                 2001-Jul-03 - Changed the area of integration from dx*dy 
;                               to dA which is more accurate.  The dx*dy did
;                               not include a "parallelogram" factor.
;                 2004-Nov-30 - Fixed indexing bug in x0 and y0.
;-

   common ffenergy_private,seed  ; make seed persistent

   if n_elements(alphain) NE 1 then alpha = 0.0 else alpha = float(alphain)
   if n_elements(guardin) NE 1 then guard = 2 else guard = float(guardin)
   if n_elements(stepin) NE 1 then step = 1L else step = long(stepin)

   p = double(j.point.p)
   b0 = double(j.point.b0)
   if tag_index(j,'CMD') LE 0 OR tag_index(j,'LATITUDE') LE 0 then begin
      latlon = get_latlon(mag,/add)
   endif

   lc = double(j.cmd)*!dtor           ; cmd, latitude assumed to be in degrees
   bc = double(j.latitude)*!dtor      ; Convert to radians

   cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
   cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
         - COS(bc)*COS(b0)*SIN(p) 
   cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
   cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
         + COS(bc)*COS(b0)*COS(p)

   ;dxidX = cxx   ; At pixel corners, not centers
   ;dyidX = cyx
   ;dxidY = cxy
   ;dyidY = cyy

   determ = cxx*cyy-cyx*cxy  ; Get image to helio transformation
   icxx = cyy/determ
   icxy = -cxy/determ
   icyx = -cyx/determ
   icyy = cxx/determ

   dXdxi = icxx
   dYdxi = icyx
   dXdyi = icxy
   dYdyi = icyy

   pix_size = double(j.point.pix_size)
   if n_elements(pix_size) EQ 1 then pix_size = [pix_size(0),pix_size(0)]

   bsize = size(j.b_long)
   nx = bsize(1)
   ny = bsize(2)

   ; Compute the force-free field.

   if keyword_set(extrapolate) then begin
      berr = get_berrs(j)
      if keyword_set(blong) then begin
         if keyword_set(randomize) then $
            err=randomn(seed,nx,ny)*berr.sig_blong $
         else err = 0.0
         pot = fff(b0=b0,pang=p,cmd=lc,lat=bc,j.b_long+err,/image, $
                   pixel=pix_size, $
                   alpha=alpha,/quiet,guard=guard)
         pxi = pot(*,*,0)  ; These are image coordinates
         pyi = pot(*,*,1)
         pzi = pot(*,*,2)
         bxyz,pxi,pyi,pzi,px,py,pz,latitude=bc,cmd=lc,p=p,b0=b0
      endif else begin
         if keyword_set(randomize) then $
            err=randomn(seed,nx,ny)*berr.sig_bzh $
         else err = 0.0
         pot = fff(j.bz+err,/normal,b0=b0,pang=p,cmd=lc,lat=bc, $
                   pixel=pix_size, $
                   alpha=alpha,bx=j.bx,by=j.by,guard=guard,/quiet)
         px = pot(*,*,0)  ; These are heliographic coordinates
         py = pot(*,*,1)
         pz = pot(*,*,2)
      endelse
   endif else begin
      px = j.bx  ; If these don't exist, then the 180 deg ambiguity was not
      py = j.by  ; resolved and the energy calculation won't work anyway.
      pz = j.bz
   endelse

   if NOT keyword_set(mask) then begin
      mask = lindgen(nx,ny)
   endif

   x = (lindgen(nx,ny) MOD nx)*pix_size(0)          ; image pixel coordinates in arcsec
   y = (lindgen(nx,ny) / nx)*pix_size(1) 
   bpxd = dXdxi*x + dXdyi*y                         ; Heliographic pixel coordinates
   bpyd = dYdxi*x + dYdyi*y
   ;jpxd = dXdxi*(x+0.5) + dXdyi*(y+0.5)
   ;jpyd = dYdxi*(x+0.5) + dYdyi*(y+0.5)

   ; Redefine x,y to be heliographic coordinates in cm.

   x = bpxd*7.27e7  ; pix_size assumed to be in arc seconds
   y = bpyd*7.27e7

   cosa = SIN(b0)*SIN(bc) + COS(b0)*COS(bc)*COS(lc) ;ang. dist. from sun centre
   offlimb = where(cosa LE 0.0,nofflimb)
   if nofflimb GT 0 then begin
      message,/info,'Warning: Ignoring points off the limb'
      mask = kill_index(mask,offlimb)
      cosa[offlimb] = 1.0 ; will be ignored anyway, can set to anything
   endif

   ;dx = x - shift(x,1,0) & dx[0,*] = dx[1,*]
   ;dy = y - shift(y,0,1) & dy[*,0] = dy[*,1]
   ;dA = dx*dy

   dA = (pix_size[0]*pix_size[1]*7.27e7*7.27e7)/cosa  ; Pixel areas

   ; energy integral in ergs

   ; Do the energy integral for all possible coordinate systems to get a good
   ; mean and standard deviation.  If the field were really force free, every
   ; coordinate system would give the same answer.  Also, the different weighting
   ; gives a good estimate of the error

   energy = [-1]
   if keyword_set(ubound) then begin
      pz2 = pz^2
      pz2dA = pz2*dA
      integral1 = total((pz2dA)[mask])
   endif else begin
      pzda = pz*dA
   endelse
   nxy = long(nx)*long(ny)
   if NOT keyword_set(noprogress) then progress,0.0,/reset,label='% Complete'
   for ipix = 0L,nxy-1L,long(step) do begin
      ii = ipix MOD nx
      ij = ipix / nx
      x0 = x[ii,ij]
      y0 = y[ii,ij]
      if keyword_set(ubound) then begin
         ; calculate an upper bound to the energy from Aly (1984)
         integral2 = total((((x-x0)^2+(y-y0)^2)*pz2dA)[mask])
         energy = [energy,sqrt(integral1*integral2)/(4.0*!pi)]
      endif else begin
         ; Calculate the energy from the virial theorem
         energy = [energy,total( (((x-x0)*px+(y-y0)*py)*pzdA)[mask] )/(4.0*!pi)]
      endelse
      if NOT keyword_set(noprogress) then progress,100.*float(ipix+1.0)/nxy,frequency=5
   endfor
   if NOT keyword_set(noprogress) then progress,100.,/last
   energy = energy[1:*]

   if keyword_set(ubound) then begin
      energyvalue = min(energy)
   endif else begin
      energyvalue = total(energy)/(n_elements(energy))  ; Mean energy
   endelse
   if n_elements(energy) GT 1 then error = stdev(energy) else error = 0.

   ; Do the force integrals as well

   fx = total((px*pz*dA)[mask])/(4.0*!pi)
   fy = total((py*pz*dA)[mask])/(4.0*!pi)
   fz = total(((pz^2-px^2-py^2)*dA)[mask])/(8.0*!pi)
   fp = total(((pz^2+px^2+py^2)*dA)[mask])/(8.0*!pi)

   return,energyvalue

end
