;+
function rd_ivm,file,h,shead=header
;NAME:
;     RD_IVM
;PURPOSE:
;     Read an IVM standard FITS file into a structure for
;     use with other software.  Acts like the old bfits.pro
;     when used on the standard FITS files (as opposed to
;     the old bfits files)
;CATEGORY:
;CALLING SEQUENCE:
;     b = rd_ivm(file,h)
;INPUTS:
;     file = FITS file name
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     shead = string header output
;OUTPUTS:
;     b = magnetic field structure
;     h = header structure
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     IMPORTANT:
;     Do not use this routine to read the old bfits files.
;     You must use bfits.pro for those files!!!
;PROCEDURE:
;MODIFICATION HISTORY:
;     T. Metcalf 2000 Nov 21
;-
   common rd_ivm_private,sname_save,sname_version

   message,'OBSOLETE:  Use ivm_bfits.pro'
   b = readfits(file,header)
   nx = n_elements(b[*,0,0])
   ny = n_elements(b[0,*,0])
   h = fitshead2struct(header)

   htags = tag_names(h)

   ; Get array names

   stags = ['']
   sindx = [-1]
   scale = [-1.]
   for i=0L,h.naxis3-1L do begin
      simage0 = strtrim(strcompress('image'+string(i),/remove_all),2)
      simage1 = strtrim(strcompress('image0'+string(i),/remove_all),2)
      simage = simage0
      if tag_index(h,simage0) then simage = simage0
      if tag_index(h,simage1) then simage = simage1
      check = execute(strtrim(strcompress('name = h.'+simage,/remove_all),2))
      if check then begin
         stags = [stags,name]
         sindx = [sindx,i]
         ; The Scale values are supposed to go from 1..N rather than 
         ; 0..(N-1).  However, this is a source of potential confusion!
         siscale = 'ISCALE'+strtrim(strcompress(string(i+1),/remove_all),2)
         if tag_index(h,siscale) GE 0 then begin
            check = execute(strtrim(strcompress('value = h.'+siscale,/remove_all),2))
            if check then begin
               scale = [scale,value]
            endif else begin
               scale = [scale,1.0]
            endelse 
         endif else begin
            scale = [scale,1.0]
         endelse 
      endif
   endfor
   if n_elements(stags) LE 1 then message,'No image definitions'
   stags = stags[1:*]
   sindx = sindx[1:*]
   scale = scale[1:*]
   if n_elements(stags) NE n_elements(b[0,0,*]) then begin
      indx_missing = kill_index(lindgen(n_elements(b[0,0,*])),sindx)
      errstring = ''
      for i=0L,n_elements(indx_missing)-1L do $
         errstring=errstring+string(indx_missing[i])
      message,/info,strcompress('WARNING: no image definition for arrays '+errstring)
   endif

   ; generate a structure name

   sname = systime(0)
   day = strcompress(strtrim(strmid(sname,8,2),2),/remove_all)
   if strlen(day) EQ 1 then day = '0'+day
   sname = strmid(sname,20,4)+strmid(sname,4,3)+ $
           day+strmid(sname,11,2)+strmid(sname,14,2)+ $
           strmid(sname,17,2)
   if n_elements(sname_save) GT 0 then begin
      if sname_save EQ sname then begin
         if n_elements(sname_version) LE 0 then sname_version = 0L
         sname_version = sname_version + 1L
         sname = strtrim(strcompress(sname + '_' + string(sname_version),/remove_all),2)
      endif else sname_version = 0L
   endif
   sname_save = sname

   ; Make the pointing structure

   if tag_index(h,'ctype1') GE 0 then begin
      ; Attempt a sanity check if possible
      if h.ctype1 NE 'arcsec' then $
         message,'Unknown units for pixel size in FITS header: '+h.ctype1
   endif
   
   ns = 0.
   ew = 0.
   if tag_index(h,'POS_CMD') GE 0 then ew = h.pos_cmd
   if tag_index(h,'EW') GE 0 then ew = h.ew
   if tag_index(h,'POS_LAT') GE 0 then ns = h.pos_lat
   if tag_index(h,'NS') GE 0 then ns = h.ns
   point = {rd_ivm_point,lat:ns*!dtor, $
                         cmd:ew*!dtor, $
                         ;lat:h.pos_lat*!dtor, $
                         ;cmd:h.pos_cmd*!dtor, $
                         radius:h.radius, $
                         pix_size:[h.cdelt1,h.cdelt2], $
                         ;wavelength:h.wavelnth, $
                         ;date:h.date_obs, $
                         ;t_start:h.t_start, $
                         ;t_stop:h.t_stop, $
                         ;noaa:h.noaa_no, $
                         ;instrument:'IVM', $
                         b0:h.b0*!dtor, $
                         p:h.p*!dtor}

   ; Make the structure

   s = 's = {IVM'+sname+','
   for i=0L,n_elements(sindx)-1L do begin
      tag = stags[i]
      if tag EQ 'ICONTIN' then begin
         tag = 'I_CONT'  ; Use the old bfits.pro standard
         if min(b[*,*,sindx[i]]) LT -10000. then begin
            message,/info,'Swapping byte order of input array'
            b = swap_endian(temporary(b))
         endif
      endif
      ;if tag EQ 'B_AZIM' then b[*,*,sindx[i]] = b[*,*,sindx[i]]*!radeg ; Convert to degrees
      if tag EQ 'B_AZIM' then b[*,*,sindx[i]] = b[*,*,sindx[i]]
      if tag EQ 'B_LONG' or tag EQ 'B_TRANS' then begin
         ; Check for bogus values of the magnetic field
         bthresh = 10000.
         bad = where(abs(b[*,*,sindx[i]]) GT bthresh,nbad)
         if nbad GT n_elements(b[*,*,sindx[i]])/3 then begin
            bb = swap_endian(b)
            bbad = where(abs(bb[*,*,sindx[i]]) GT bthresh,nbbad)
            if nbbad LT nbad*0.9 then begin
               b = temporary(bb)
               nbad = nbbad
               bad = bbad
               message,/info,'Swapping byte order of input array'
            endif
         endif
         if nbad GT 0 then begin
            message,/info,strcompress('Setting '+string(nbad)+' bogus '+tag+ $
                                      ' to +/- '+string(bthresh)+' G')
            btemp = b[*,*,sindx[i]]
            btemp[bad] = bthresh*(btemp[bad]/abs(btemp[bad]))
            b[*,*,sindx[i]] = temporary(btemp)
         endif
      endif
      indxstr = strtrim(strcompress(string(sindx[i]),/remove_all),2)
      s = s + tag + ':' + 'b[*,*,' + indxstr + ']*scale['+indxstr+'],'
   endfor
   s = s + 'instrument: "IVM",'
   check = execute('junk=h.date_obs')
   if check then begin
      s = s + 'date_obs: ANYTIM( h.date_obs, /STIME, /DATE_ONLY ),'
      s = s + 'time_obs: ANYTIM( h.date_obs, /STIME, /TIME_ONLY ),'
   endif
   s = s + 'point:point}'
   check = execute(s)

   ; Add the position arrays
   if tag_index(s,'LATITUDE') LE 0 and tag_index(s,'CMD') LE 0 then $
      latlon = get_latlon( s, /add, /quiet)
   s.point.lat = s.latitude[nx/2,ny/2]*!dtor
   s.point.cmd = s.cmd[nx/2,ny/2]*!dtor

   print, tag_names(s)

   return,s

end
