;+
function separatrix,mag,alpha=alpha,fff=ff,nz=nz,zscale=zscale, $
                    surround=surround,xyexpand=xyexpand,get_alpha=get_alpha, $
                    vertices=vertices,distance=distance,helio=helio
;+
;NAME:
;     SEPARATRIX
;PURPOSE:
;     Finds separatrices in magnetograms.
;CATEGORY:
;CALLING SEQUENCE:
;     s = separatrix(j)
;INPUTS:
;     j = magnetic field structure with 180 degree ambiguity resolved
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS:
;     alpha = force-free alpha (passed to bff.pro)
;     nz = number of z planes (passed to bff.pro)
;     zscale = scale of heights (passed to bff.pro)
;     /surround = surround field with zeroes (passed to bff.pro)
;     xyexpand = expand field with zeroes (passed to bff.pro)
;     /get_alpha = compute alpha (passed to bff.pro)
;     /helio = Leave output in helio coords.  The default is to transform
;              back from the bff grid to the image grid.
;KEYWORD INPUT/OUTPUT:
;     fff = force-free field structure from bff.pro (not computed if input)
;     vertices = vertices structure from blines.pro (not computed if input)
;KEYWORD OUTPUT:
;     distance = fltarr(nx,ny,3) containing the changes in x,y,z between
;                the starting and ending footpoints.
;OUTPUTS:
;     s = image of separatrices
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Very slow!
;     Uses a linear force-free extrapolation (bff.pro)
;EXAMPLE:
;     IDL> j = bfits('J891020.1741')
;     IDL> s = separatrix(j,/get,/surr)
;     IDL> vmgram,j,/inc,/xyz,image=s
;PROCEDURE:
;     Computes the x,y,z seperation of the footpoints of 
;     extrapolated field lines.  To account for the direction of the field
;     line, these separations are each run through sobel and then summed.
;     Uses the sobel function to find regions where the seperation changes 
;     rapidly.
;MODIFICATION HISTORY:
;     Thomas R. Metcalf  June 3, 1994
;-

   ; compute the force-free extrapolation 

   if n_elements(ff) LE 0 then begin
      ff = bff(mag,alpha=alpha,nz=nz,zscale=zscale,surround=surround, $
                   xyexpand=xyexpand,get_alpha=get_alpha)
   endif

   nx = n_elements(ff.bxp(*,0,0))
   ny = n_elements(ff.bxp(0,*,0))
   nz = n_elements(ff.bxp(0,0,*))

   ; compute field lines for all pixels

   if n_elements(vertices) LE 0 then begin
      vertices = blines(ff,bcut=0.0,/quiet,/noselect)
   endif
   if size(vertices,/tname) EQ 'STRUCT' then $
      v = reform(vertices.vertices,3,n_elements(vertices.vertices)/3) $
   else $
      v = vertices

   ; Find the start and end indices of the field lines

   lstart = where(v(0,*) EQ -1000., nlines)  ; -1000 indicates start of line
   lend = where(v(0,*) EQ -2000., nelines)   ; -2000 indicates end of line

   if (nelines NE nlines) then $
      message,/info,'WARNING: # of line ends NE # of line starts'
   nlines = nlines < nelines
   if nlines NE nx*ny then message,/info,'nlines NE npixels'

   distance = fltarr(nx,ny,3)   ; must be zeroed
   complete = bytarr(nx,ny)     ; must be zeroed
   for pix=0L,nlines-1L do begin
      lst = lstart(pix)
      led = lend(pix)
      npoints = led - lst - 1L   ; number of vertices in the field line

      if npoints GE 2 then begin   ; else distance remains 0.0
         xstart = v(0,lst+1L)
         ystart = v(1,lst+1L)
         zstart = v(2,lst+1L)
         xend = v(0,led-1L)
         yend = v(1,led-1L)
         zend = v(2,led-1L)
         x = long(xstart+0.5)
         y = long(ystart+0.5)
         if v(1,led) NE 0 then complete(x,y) = 1b  ; Field line is complete
         while xend LT 0 OR xend GT nx OR yend LT 0 or yend GT ny do begin
            led = led - 1L
            xend = v(0,led-1L)
            yend = v(1,led-1L)
            zend = v(2,led-1L)
            complete(x,y) = 0b   ; Field line out of bounds ... not complete
         endwhile

         distance(x,y,*) = [xend-xstart,yend-ystart,zend-zstart]

      endif
   endfor

   ; Scale the distances.  Ignore any field line which is not complete

   c = where(complete)
   nc = where(complete EQ 0)
   cx = c MOD nx
   cy = c / nx
   for i=0L,2L do begin
      d = distance(*,*,i)
      distance(*,*,i) = bytscl(d,min=min(d(c)),max=max(d(c)))
   endfor

   ; Transform the separatrix image back to image coordinates

   if NOT keyword_set(helio) then begin    
      nx = n_elements(mag.bz(*,0))
      ny = n_elements(mag.bz(0,*))
 
      x = lindgen(nx,ny) MOD nx
      y = long(lindgen(nx,ny) / nx)
 
      ; pinv,qinv are the inverse transform which puts the grid back to 
      ; the image coords
 
      polywarp,mag.b_pxd-min(mag.b_pxd),mag.b_pyd-min(mag.b_pyd),x,y,3,$
              pinv,qinv
      dis = fltarr(nx,ny,3)
      dis(*,*,0) = poly_2d(distance(*,*,0),pinv,qinv,1,nx,ny,missing=0.)
      dis(*,*,1) = poly_2d(distance(*,*,1),pinv,qinv,1,nx,ny,missing=0.)
      dis(*,*,2) = poly_2d(distance(*,*,2),pinv,qinv,1,nx,ny,missing=0.)
      ;com = poly_2d(complete,pinv,qinv,1,nx,ny,missing=0.)
   endif else dis=distance

   ; Use bitwise OR rather than + so that we don't get a spurious factor of 3

   sepx = long(sobel(dis(*,*,0))) OR long(sobel(dis(*,*,1))) OR  $
          long(sobel(dis(*,*,2)))

   return,sepx

end
