;+
; NAME:
;   substruc
; PURPOSE:
;   This function creates a copy of a structure, replacing fields which are
;   2-D arrays with subarrays of the specified dimensions.  Scalar and
;   1-D fields are passed unchanged.
; CALLING SEQUENCE:
;   result = substruc (struc, [name,] x0, x1, y0, y1, help=help)
; INPUTS:
;   struc = original structure from which to extract subarrays.
;  [Optional:
;   name = string specifying new name of structure type (must be different from
;      any name currently in use; this function will test for uniqueness).]
;   x0, x1, y0, y1 = integers specifying subarray, as in array(x0:x1,y0:y1).
; OUTPUT:
;   result = new structure with reduced arrays.
; KEYWORDS:
;   help = nonzero to print calling sequence summary and exit.
; OTHER ROUTINES INVOKED:
;   strstat (T. R. Metcalf), get_def (TRM), makestr (TRM).
; RESTRICTIONS:
;   The makestr procedure is Unix-specific.
; MODIFICATION HISTORY:
;   26 Feb 1992:  Written by J-F de La Beaujardiere, with assistance from
;                 T. R. Metcalf.
;   920917 ANM: Allow "name" to be omitted, using anonymous structures
;-

FUNCTION substruc, struc, name, x0, x1, y0, y1, help=help

; Print help message and exit, if requested
if (keyword_set(help)) then begin
   print, 'CALLING SEQUENCE:'
   print, '   result = substruc (struc, [name,] x0, x1, y0, y1, help=help)'
   return, 'SUBSTRUC called with /HELP keyword.'
endif

; Check if structure name has already been used
IF N_PARAMS() EQ 5 THEN BEGIN						; ANM
	y1=y0 & y0=x1 & x1=x0 & x0=name & name='' & END ELSE $		; ANM
if (strstat(name, /quiet)) then begin
   tmpstring = "SUBSTRUC:  Structure name already exists."
   print, tmpstring
   return, tmpstring
endif

; Define global variables
ntags = n_tags(struc)  ;number of structure tags in original structure
tagnames = tag_names(struc)  ;original structure tag names
fieldstring = strarr(ntags)  ;strings to define new structure
reduce = bytarr(ntags)  ;flag indicating whether a field becomes a subarray

; Create individual structure tag definitions
for tag = 0, ntags-1 do begin
   field = struc.(tag)
   fieldsize = size(field)
   if (fieldsize(0) EQ 2) then begin  ;field is a 2-D array--shrink it
      fieldstring(tag) = ","+tagnames(tag)+":"+get_def(field(x0:x1,y0:y1))
      reduce(tag) = 1
   endif else begin  ;keep the existing field
      fieldstring(tag) = ","+tagnames(tag)+":"+get_def(field)
      reduce(tag) = 0
   endelse
endfor
;;;TAG DELETED BECAUSE IT IS IGNORED BY BFITS AND WBFITS
;;;; Additional tag to hold subarray coordinates
;;;fieldstring(ntags) = ",subarray:intarr(4)"

; Create complete structure definition
struc_string = "{"+strtrim(name,2)
for tag = 0, ntags-1 do begin
   struc_string = struc_string+fieldstring(tag)
endfor
struc_string = struc_string+"}"
IF name EQ '' THEN struc_string = "{"+STRMID(struc_string,2,9999)	; ANM

; Create new structure
result = makestr(struc_string)  ;Unix-specific routine

; Fill structure fields
for tag = 0, ntags-1 do begin
   field = struc.(tag)
   if (reduce(tag)) then begin
      result.(tag) = field(x0:x1,y0:y1) 
   endif else begin
      result.(tag) = field
   endelse    
endfor
;;;result.(ntags) = [x0,x1,y0,y1]

; Exit function
return, result
end
