PRO vectorfield,U,V,X,Y, Missing = Missing, Length = length, Dots = dots, $
            maxT=maxT, color=color, noerase=noerase, $
            arrows=arrows, g1000=g1000,device=device

;+
; NAME:
;	VECTORFIELD
; PURPOSE:
;	Produce a two dimensional velocity field plot.
;		Draw a line segment at each point showing
;		the direction and magnitude of the field.
;		
; CATEGORY:
;	Plotting
; CALLING SEQUENCE:
;	VECTORFIELD,U,V  ...  or:
;	VECTORFIRLD,U,V,X,Y
; INPUTS:
;	U = The X component of the two dimensional field.  Both
;		U and V must be two dimensional and have the same
;		dimensions.
;	V = The Y component of the two dimensional field.  The
;		vector at point (i,j) has the magnitude of
;		(U(i,j)^2 + V(i,j)^2)^0.5 and a direction of
;		ATAN2(V(i,j),U(i,j)).
; OPTIONAL INPUT PARAMETERS:
;	X = Optional abcissae values, array, Size of U,V.  This is the 
;           location of the center of the vector, not the endpoint.
;	Y = Optional ordinate values, array, Size of U,V.  This is the 
;           location of the center of the vector, not the endpoint.
; KEYWORD INPUT PARAMETERS:
;       /device = x and y are device coordinates rather than the default
;                 data coordinates.
;	Missing = missing data value.  (U,V) Vectors with a LENGTH shorter
;		than missing are ignored.
;       maxT = vectors with magnitudes above maxT are not plotted
;	Length = length factor, default = 1.0 makes the longest (U,V)
;		 vector the length of one cell.
;	/Dots = places a dot at each missing point, default = 
;		draw nothing for missing points.  Has effect only if
;		Missing is specified.
;       color = plots the vector in the color specified by color index "color".
;	/noerase = if set, do not erase window before plotting
;       /arrows = if set, draw arrow heads on the vectors
; KEYWORD OUTPUT PARAMETER:
;	g1000 = length of 1000-unit line segment (in data coordinates)
; OUTPUTS:
;	None.
; COMMON BLOCKS:
;	None.
; SIDE EFFECTS:
;	Plotting on the selected device is performed.  System
;	variables concerning Plotting are changed.
; RESTRICTIONS:
;	None.
; PROCEDURE:
;	Straightforward.  The system variables !Xtitle, !Ytitle and
;	!Mtitle may be set to title the axes.
; MODIFICATION HISTORY:
;	Dms, Rsi, Oct, 1983.
;	For Sun, DMS, RSI, April, 1989.
;
;       Changed to TRMVECT from VELOVECT 3/13/90.  
;       No Arrowheads, center of line segment at gridpoint
;
;       10/11/90 TRM Added color
;	06/22/92 Added G1000 keyword; removed unwanted conversion to integer
;	         in computation of SINA and COSA. (JdLB)
;       Feb, 1994 T. Metcalf ... Modified trmvect to accept arbitrary
;                 x,y locations and renamed vectorfield.pro.
;       1995-02-21 Changed noclip=0 to noclip=1 in plots call.
;-

	s = size(u)
	t = size(v)

	if s(0) ne 2 then begin
baduv:	 print,'VECTORFIELD -- U and V parameters must be 2D and same size'
		return
		endif
	if total(abs(s(0:2)-t(0:2))) ne 0 then goto,baduv
;
	if n_params(0) lt 3 then x = long(lindgen(s(1),s(2)) MOD s(1)) else $
	   if n_elements(x) ne s(1)*s(2) then begin
badxy:        print,'VECTORFIELD - X and Y arrays have incorrect size.'
              return
           endif
	if n_params(1) lt 4 then y = long(lindgen(s(1),s(2)) / s(1)) else $
		if n_elements(y) ne s(1)*s(2) then goto,badxy
;
	if n_elements(missing) le 0 then missing = -1.0e30
	if n_elements(maxT) le 0 then maxT = 1.0e30
	if n_elements(length) le 0 then length = 1.0

	mag = sqrt(u^2+v^2)		;magnitude.
		;Subscripts of good elements
	nbad = 0			;# of missing points
	if n_elements(missing) gt 0 or n_elements(maxT) gt 0 then begin
		good = where((mag gt missing) and (mag lt maxT)) 
		if keyword_set(dots) then $
                  bad = where(mag ge missing or mag le maxT, nbad)
	endif else begin
		good = lindgen(n_elements(mag))
	endelse

	mag = mag(good)			;Discard missing values
	maxmag = max(mag)
	ugood = u(good)
	vgood = v(good)
	x0 = min(x)			;get scaling
	x1 = max(x)
	y0 = min(y)
	y1 = max(y)
	sina = length * (x1-x0)/maxmag/s(1)*ugood ;sin & cosine components.
	cosa = length * (y1-y0)/maxmag/s(2)*vgood
	; Compute length of 1000-Gauss segment
	g1000 = length * (x1-x0)/maxmag/s(1)*1000.0

        r = .3                          ;len of arrow head
        angle = 22.5 * !dtor            ;Angle of arrowhead
        st = r * sin(angle)             ;sin 22.5 degs * length of head
        ct = r * cos(angle)

	for i=0L,n_elements(good)-1L do begin	;Each point
		x0 = x(good(i))	;get coords of start & end
		dx = sina(i)
		x1 = x0 + dx/2.0
		x0 = x0 - dx/2.0
		y0 = y(good(i))
		dy = cosa(i)
		y1 = y0 + dy/2.0
		y0 = y0 - dy/2.0
                if keyword_set(arrows) then begin
                   plots,[x0,x1,x1-(ct*dx-st*dy),x1,x1-(ct*dx+st*dy)], $
                         [y0,y1,y1-(ct*dy+st*dx),y1,y1-(ct*dy-st*dx)], $
                          color=color,device=device,noclip=1
                endif $
		else plots,[x0,x1],[y0,y1],color=color,device=device,noclip=1
		endfor
	if nbad gt 0 then $		;Dots for missing?
		oplot, x(bad mod s(1)), y(bad / s(1)), psym = 3,color=color, $
                     device=device

end
