;+

function virial,mag,error,energy

;NAME:
;     virial
;PURPOSE:
;     Compute the energy of a magnetic field from the virial theorem
;CATEGORY:
;CALLING SEQUENCE:
;     energy = virial(mag,error,earr)
;INPUTS:
;     mag = vector magnetogram structure (must have the 180 degree
;         ambiguity resolved)
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;OUTPUTS:
;     energy = energy in ergs
;     error = error estimate on the energy
;     earr = array of energies computed with different coordinate systems
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     - mag.point.pix_size assumed to be in arc seconds
;     - mag.bx, mag.by, mag.bz must exist with 180 ambiguity resolved
;PROCEDURE:
;     Computes the energy from the virial theorem.  The error
;     estimate is calculated by assuming several different coordinate
;     systems and returning the stdev of the energy from each.
;MODIFICATION HISTORY:
;     T. Metcalf 1999-April-19
;-

   message,'Obsolete: use ffenergy.pro'

   nx = n_elements(mag.bz(*,0))
   ny = n_elements(mag.bz(0,*))

   p = double(mag.point.p)      ; assumed to be in radians
   b0 = double(mag.point.b0)

   if tag_index(mag,'CMD',/quiet) LE 0 OR tag_index(mag,'LATITUDE',/quiet) LE 0 then begin
      latlon = get_latlon(mag,/add)
   endif

   lc = double(mag.cmd)*!dtor           ; cmd, latitude assumed to be in radians
   bc = double(mag.latitude)*!dtor

   cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
   cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
         - COS(bc)*COS(b0)*SIN(p) 
   cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
   cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
         + COS(bc)*COS(b0)*COS(p)

   dxidX = cxx   ; At pixel corners, not centers
   dyidX = cyx
   dxidY = cxy
   dyidY = cyy

   determ = cxx*cyy-cyx*cxy  ; Get image to helio transformation
   icxx = cyy/determ
   icxy = -cxy/determ
   icyx = -cyx/determ
   icyy = cxx/determ

   dXdxi = icxx
   dYdxi = icyx
   dXdyi = icxy
   dYdyi = icyy

   pix_size = double(mag.point.pix_size)
   if n_elements(pix_size) EQ 1 then pix_size = [pix_size(0),pix_size(0)]

   x = (lindgen(nx,ny) MOD nx)*pix_size(0)          ; image pixel coordinates in arcsec
   y = (lindgen(nx,ny) / nx)*pix_size(1) 
   bpxd = dXdxi*x + dXdyi*y                         ; Heliographic pixel coordinates
   bpyd = dYdxi*x + dYdyi*y
   ;jpxd = dXdxi*(x+0.5) + dXdyi*(y+0.5)
   ;jpyd = dYdxi*(x+0.5) + dYdyi*(y+0.5)

   ; Redefine x,y to be heliographic coordinates in cm.

   x = bpxd*7.27e7  ; pix_size assumed to be in arc seconds
   y = bpyd*7.27e7

   dx = x - shift(x,1,0) & dx[0,*] = dx[1,*]
   dy = y - shift(y,0,1) & dy[*,0] = dy[*,1]

   ; energy integral in ergs

   energy = dblarr(nx,ny)
   for i=0L,nx-1L do begin
      for j = 0L,ny-1L do begin
         x0 = x[i]
         y0 = y[j]
         energy[i,j] = total(((x-x0)*mag.bx+(y-y0)*mag.by)*mag.bz*dx*dy)/4.0/!pi
      endfor
   endfor

   error = stdev(energy)

   return,total(energy)/(double(nx)*double(ny))  ; Mean energy

end
