;+
; NAME:
;   VMGRAM
; PURPOSE: 
;   This procedure makes an image-plane vector magnetogram overlaid on an
;   image.  Optionally an alignment between the image and the magnetogram
;   can be supplied.
; CALLING SEQUENCE:
;   vmgram, mag, neutraldata, pp=pp,levels=lv,pixsize=pixsize
;           length=length, missing=missing
; INPUT:
;   mag = structure containing magnetic field data.  Required elements:
;         B_LONG, B_TRANS, B_AZIM, POINT unless /xyz is set in which case
;         BX, BY, BZ, and POINT are required.
; KEYWORD INPUT:
;   pp = output from setpts ... specifies the co-registration between the 
;        magnetogram and the image (default = coregistered) with the
;        image as the reference and the magnetogram as the transformed.
;   key = transformation type.  Used in conjunction with pp.
;         See "transtype" keyword in CALTRANS for details.
;         Default is a general linear transform.
; OPTIONAL INPUTS:
;   neutraldata = data to be used for the neutral line (thick contour).
;                 by default this is the logitudinal or vertical field,
;                 but neutraldata will override this.  Neutraldata must
;                 be a 2-d array which is the same size as b_long or bz.
;                 The co-alignment is assumed to be specified with pp, but
;                 this can be overriden with the ppneutral keyword.
; KEYWORDS:
;   /helio = The mgram is in helio coordinates, and the rotation by the
;            p angle should NOT be done.  Only applies if /xyz is set.
;   position = !p.position style vector to specify the largest possible
;              plotting window.  The actual image will be reduced so
;              that the pixels have the correct aspect ratio.
;              Default = [0.1,0.1,0.9,0.9].
;   levels = contour levels for longitudinal or vertical field
;   jlevels = contour levels for current
;   length = relative length of line segments used to plot transverse field.
;      Default is 1.0, which makes the longest segment the length of one pixel.
;   /arrows = draw arrow heads on the vectors
;   missing = smallest value of transverse field to plot.
;   image = 2-D image to display under the magnetogram.  Def = mag.i_cont.
;           The image is byte scaled to avoid the bottom 15 colors of the
;           color table.  These 15 colors can be used for contour colors, etc.
;   /noimage = do not plot an image under the magnetogram.
;   pixsize = scale to label axes in rather than in pixels. (default = 1.0)
;   x0,y0 = coordinates of the start of the x and y axes in pixsize
;           units.  Default = 0.
;   xstep = the size of an x pixel relative to a y pixel (def = 1.0).
;           Specifies non-square pixels.  IF the input magnetic field
;           structure contains a point.pix_size field which is a 2-element
;           vector, the default xstep is taken from point.pix_size.
;   /fixcolors = do not modify the color table
;   c_color = color of contours.  If scalar, the color index is used for all
;             positive and negative contours.  If a 2 element vector, the first
;             color is for the positive contours and the second element is 
;             for the negative contours.  If a 4 element vector, the first two
;             elements are the colors of the magnetic contours and the second 
;             two are the colors of the current contours.  If a 5 element
;             vector, the first two elements are the colors of the magnetic 
;             contours, the second two are the colors of the current contours,
;             and the last element is the color of the lat/lon contours.
;   c_thick = Thickness of contours. If scalar, the thickness is used for all
;             positive and negative contours.  If a 2 element vector, the first
;             element is for the positive contours and the second element is 
;             for the negative contours.  If a 4 element vector, the first two
;             elements are the thickness of the magnetic contours and the 
;             second two are the thickness of the current contours.  If a 5 
;             element vector, the first two elements are the thickness of the 
;             magnetic contours, the second two are the thickness of the 
;             current contours and the last element is the thickness of the 
;             lat/lon contours.
;   ncolor = neutral line color index
;   vcolor = color index for field vectors
;   acolor = color index for axes, labels, etc.
;   llcolor = color index for lat/lon contours
;   /incolor = use default colors with solid contours.
;   /new_window = nonzero to open new window for magnetogram.    By default,
;                 the current window is erased and used.
;   /old_window = use the old window
;   /xyz = plot vertical/horizontal field
;   /nodash = use solid contours for both positive and negative fields
;   btskip = Number of vectors to skip before plotting the next one.
;            default = 0, plot all vectors. if set to 1, plot every
;            other vector, etc.
;   /novectors = do not draw transverse/horizontal magnetic field vectors
;   /nocontours = do not plot magnetic contours
;   /neutralline = draw smoothed magnetic neutral line
;   /current = plot jz contours (/current and /tcurrent are mutually 
;               exclusive)
;   /tcurrent = plot j total contours  (/current and /tcurrent are mutually 
;               exclusive)
;   /latlon = Plot latitude and longitude contours.  Requires latitude and 
;             CMD tags in the field structure.
;   llinterval = interval between lat/lon contours. Default = 5.0.
;   nthick = thickness of neutral line (default = 5)
;   nsmooth = smoothing of neutral line (default = 0.25, none=0.0)
;   vertices = vertices structure from blines.pro.  Plots field lines on 
;              top of the image.
;   /nosample = do not use /sample when rebinning the image to the window size.
;               This is not relevant with PostScript.
;   neutralpp = co-alignment for neutral data if it is specified.
;   edge = output variable which returns the border around the image in 
;          device coordinates
;   B_label_pos = Number of pixel sizes to shift B\J labels up\down.
;   title = title for plot
;   xtitle = title for x axis
;   ytitle = title for y axis
;   /nowords = do not annotate the plot.
;   /nolabels = do not label the axes
;   /noaxes = do not draw axes on the plot.
;   /noerase = do not erase plotting device.
;   /quiet = work quietly.
; SIDE EFFECTS:
;   o A magnetogram is displayed on the monitor screen.
;   o The !P.POSITION system variable is modified.
;   o The color table is modified unless /fixcolors is set
; EXAMPLE:
;   vmgram,j,/inc
; RESTRICTIONS:
;   B_AZIM is assumed to be measured CCW from terrestrial north.
;   BX,BY are assumed to point solar west and solar north, respectively.
; MODIFICATION HISTORY:
;   25 Jun 1992:  Written by J-F de La Beaujardiere.
;   30 Jul 1992:  mgram modified by T. Metcalf for SXT use
;   04 May 1994:  extensive mods to sxt_mgram to avoid oplot_b and to use
;                 vectorfield.pro.
;   09 Nov 1994:  Added /latlon keyword.
;   13 Feb 1996:  Added /helio keyword.
;   13 Jul 2004:  Added btskip keyword.
;-


PRO vmgram, magin, neutraldata, pp=pp, pixsize=pixsize, levels=lv, $
               length=lengthin,neutralpp=ppneutral, B_Label_pos=Btpos, $
               missing=missing,new_window=new_window,image=image, $
               xyz=xyz,novectors=novectors,neutralline=neutral,current=jz, $
               tcurrent=jt,jlevels=jlv,position=position,noaxes=noaxes, $
               nthick=nthick,nsmooth=nsmooth,nodash=nodash,c_color=c_color, $
               vertices=vertices,nocontours=nocontours,nosample=nosample, $
               edge=edge,c_thick=c_thick,incolor=incolor,ncolor=ncolor, $
               arrows=arrows,noimage=noimage,title=titlein,xtitle=xtitlein, $
               ytitle=ytitlein,xstep=xstep,nowords=nowords,quiet=quiet, $
               noerase=noerase,vcolor=vcolor,acolor=acolor,latlon=latlon, $
               llinterval=llinterval,key=ttyp,nolabels=nolabels,llcolor=llcolor, $
               old_window=old_window,fixcolors=fixcolors,helio=helio, $
               border=border,bthick=bthick,bcolor=bcolor,btskip=btskip, $
               x0=x0,y0=y0

mag = magin  ; protect the input variable from change

if !d.name EQ 'PS' then ps = 1 else ps = 0

if n_elements(missing) LE 0 then missing = 0.0
if n_elements(lengthin) LE 0 then length = 1.0 else length=lengthin

if n_elements(pixsize) LE 0 then pixsize = 1.0

if n_elements(x0) LE 0 then x0 = 0. else x0 = float(x0)
if n_elements(y0) LE 0 then y0 = 0. else y0 = float(y0)

if NOT keyword_set(xstep) then begin
   if tag_index(magin,'POINT') GE 0 then begin
      if tag_index(magin.point,'PIX_SIZE') GE 0 then begin
         if n_elements(magin.point.pix_size) EQ 2 then begin
            xstep = double(magin.point.pix_size(0))/double(magin.point.pix_size(1))
            message,/info,'xstep = '+strtrim(string(xstep),2)
         endif
      endif
   endif
endif

; Set up color table

if NOT keyword_set(fixcolors) then begin
   specialct, 0, 255, 255, 255  ; white for background
   specialct, 1, 0, 0, 0        ; black for foreground
endif
if n_elements(ncolor) LE 0 then begin
   if keyword_set(noimage) then ncolor=255 else ncolor=0
endif
if n_elements(bcolor) LE 0 then begin
   if keyword_set(noimage) then bcolor=128 else bcolor=128
endif
if n_elements(vcolor) LE 0 then begin
   if keyword_set(noimage) then vcolor=255 else vcolor=0
endif
if n_elements(acolor) LE 0 then acolor=1
case n_elements(c_color) of
   0: begin
         if keyword_set(noimage) then begin
            pcolor = 255
            mcolor = 255
            jpcolor = 255
            jmcolor = 255
            llcolor = 255
         endif else begin
            pcolor = 0
            mcolor = 0
            jpcolor = 0
            jmcolor = 0
            llcolor = 0
         endelse
      end
   1: begin
         pcolor = c_color
         mcolor = c_color
         jpcolor = c_color
         jmcolor = c_color
         llcolor = c_color
      end
   2: begin
         pcolor = c_color(0)
         mcolor = c_color(1)
         jpcolor = c_color(0)
         jmcolor = c_color(1)
         llcolor = c_color(0)
      end
   3: begin
         pcolor = c_color(0)
         mcolor = c_color(1)
         jpcolor = c_color(2)
         jmcolor = c_color(2)
         llcolor = c_color(0)
      end
   4: begin
         pcolor = c_color(0)
         mcolor = c_color(1)
         jpcolor = c_color(2)
         jmcolor = c_color(3)
         llcolor = c_color(0)
      end
   5: begin
         pcolor = c_color(0)
         mcolor = c_color(1)
         jpcolor = c_color(2)
         jmcolor = c_color(3)
         llcolor = c_color(4)
      end
   else: message,'Too many elements in c_color vector'
endcase
case n_elements(c_thick) of
   0: begin
         pthick = 1
         if NOT keyword_set(nodash) and ps then mthick=2 else mthick = 1
         jpthick = 1
         if NOT keyword_set(nodash) and ps then jmthick=2 else jmthick = 1
         llthick = 1
      end
   1: begin
         pthick = c_thick
         mthick = c_thick
         jpthick = c_thick
         jmthick = c_thick
         llthick = c_thick
      end
   2: begin
         pthick = c_thick(0)
         mthick = c_thick(1)
         jpthick = c_thick(0)
         jmthick = c_thick(1)
         llthick = c_thick(0)
      end
   3: begin
         pthick = c_thick(0)
         mthick = c_thick(1)
         jpthick = c_thick(2)
         jmthick = c_thick(2)
         llthick = c_thick(0)
      end
   4: begin
         pthick = c_thick(0)
         mthick = c_thick(1)
         jpthick = c_thick(2)
         jmthick = c_thick(3)
         llthick = c_thick(0)
      end
   5: begin
         pthick = c_thick(0)
         mthick = c_thick(1)
         jpthick = c_thick(2)
         jmthick = c_thick(3)
         llthick = c_thick(4)
      end
   else: message,'Too many elements in c_thick vector'
endcase
if n_elements(arr_color) LE 0 then arr_color=!d.table_size-1
if NOT keyword_set(nodash) then begin
   pltype = 0
   if ps then mltype=1 else mltype=2
endif else begin
   pltype = 0
   mltype = 0
endelse
llltype = 0
if keyword_set(incolor) then begin
   mcolor=2
   pcolor=3
   vcolor=4
   ncolor=5
   jmcolor=6
   jpcolor=7
   llcolor=8
   if NOT keyword_set(fixcolors) then begin
      tvlct,/get,r,g,b
      r(mcolor)   =  0b & g(mcolor)   =  0b & b(mcolor)    = 255b
      r(pcolor)   = 255b & g(pcolor)  =  0b & b(pcolor)    = 0b
      r(jmcolor)  =  0b & g(jmcolor)  =  255b & b(jmcolor) = 0b
      r(jpcolor)  = 255b & g(jpcolor) =  255b & b(jpcolor) = 0b
      r(llcolor)  = 255b & g(llcolor) =  255b & b(llcolor) = 51b
      if keyword_set(noimage) then begin
         r(vcolor)=0b & g(vcolor)=0b & b(vcolor)=0b
      endif else begin
         r(vcolor)=255b & g(vcolor)=255b & b(vcolor)=255b
      endelse
      r(ncolor)=0b & g(ncolor)=0b & b(ncolor)=0b
      tvlct,r,g,b
   endif
   pltype=0
   mltype=0
   lltype=0
endif

; Determine image/field size

if NOT keyword_set(image) then begin
   if tag_index(mag,'I_CONT') GE 0 AND NOT keyword_set(noimage) then $
      image = mag.i_cont $
   else if keyword_set(xyz) then image = mag.bz else image = mag.b_long
endif
imsize = size(image)
xsize = imsize(1)
ysize = imsize(2)
;;if keyword_set(xyz) then begin
;;   if n_elements(mag.bx(*,0)) NE xsize OR $
;;      n_elements(mag.bx(0,*)) NE ysize then $
;;      message,'Image size is incompatible with field arrays, use /noimage'
;;endif else begin
;;   if n_elements(mag.b_long(*,0)) NE xsize OR $
;;      n_elements(mag.b_long(0,*)) NE ysize then $
;;      message,'Image size is incompatible with field arrays, use /noimage'
;;endelse

; Set axis scaling factor for relevant variables.  The axes are labeled
; in units of pixsize 

if n_elements(pixsize) NE 1 then pixsize = 1. else pixsize = float(pixsize)

; Set up and open output window, if necessary

if NOT ps and !d.name NE 'TEK' then begin
   if ( (keyword_set(new_window) OR (!d.window EQ -1)) AND (NOT keyword_set(old_window)) ) then begin
      zoomw=long(1500./(xsize+ysize))>1L
      wxsize = xsize*zoomw
      wysize = ysize*zoomw
      case !d.name of
         'Z':  device,set_resolution=[wxsize,wysize]
         else: window, /free, xsize=wxsize, ysize=wysize
      endcase
   endif
endif

; Set plot position in device coordinates, leaving room for the titles and
; forcing square pixels 

if NOT keyword_set(position) then position=[0.1,0.1,0.9,0.9]
!p.position = pixsquare(xsize,ysize,position=position,xstep=xstep)
zoomx = (!p.position(2)-!p.position(0))*!d.x_size/double(xsize)
zoomy = (!p.position(3)-!p.position(1))*!d.y_size/double(ysize)
edge = {edge,  left:!p.position(0)*!d.x_size, $
             bottom:!p.position(1)*!d.y_size, $
              right:(1.0 - !p.position(2))*!d.x_size, $
                top:(1.0 - !p.position(3))*!d.y_size}

; Set default contour levels

if (n_elements(lv) LE 0) then begin
   lv = [100,200,400,800,1600,3200]
endif
if (n_elements(jlv) LE 0) and (keyword_set(jz) or keyword_set(jt)) then begin
   jlv = jzerr(mag)*(findgen(6)+1.5) 
endif

; Set default titles

if n_elements(titlein) LE 0 then begin
   if tag_index(mag,'INSTRUMENT') GE 0 AND NOT keyword_set(nowords) then $
      title=string(mag.instrument) $
   else title=''
endif else title=string(titlein)
if n_elements(xtitlein) LE 0 then begin
   if NOT keyword_set(nowords) then begin
      xtitle='Magnetic Contours at !9+!3 '
      for i=0,n_elements(lv)-1 do $
         xtitle=xtitle+strcompress(' '+string(lv(i),format="(f15.1)"))
      xtitle=xtitle + ' G'
      if (keyword_set(jz) or keyword_set(jt)) then begin
         xtitle=xtitle+'!CCurrent Contours at !9+!3'
         for i=0,n_elements(jlv)-1 do $
            xtitle=xtitle+strcompress(' '+string(jlv(i),format="(f15.1)"))
         xtitle=xtitle+' mA/m!U2!N'
      endif
   endif else xtitle=''
endif else xtitle=string(xtitlein)
if n_elements(ytitlein) LE 0 then begin
   if NOT keyword_set(nowords) then begin
      if NOT keyword_set(xyz) then ytitle='LOS\Transverse Magnetic Field' $
      else ytitle='Vertical\Horizontal Magnetic Field'
   endif else ytitle=''
endif else ytitle=string(ytitlein)


; Define size of plotted image

if ps then begin
   nx = xsize
   ny = ysize
endif else begin
   nx = long((xsize)*zoomx)
   ny = long((ysize)*zoomy)
endelse

; Display underlying image if specified

if NOT keyword_set(noerase) then erase
if NOT keyword_set(noimage) then begin
   tv,congrid(bytscl(image,top=!D.TABLE_SIZE-19)+16b,nx,ny, $
              interp=keyword_set(nosample)), $
      edge.left,edge.bottom, $
      xsize=(!p.position(2)-!p.position(0))*!d.x_size, $
      ysize=(!p.position(3)-!p.position(1))*!d.y_size,/device
endif

; Compute x,y coordinates of the magnetic field data

if keyword_set(xyz) then begin
   bnx = n_elements(mag.bz(*,0))   ; size of magnetogram
   bny = n_elements(mag.bz(0,*))
endif else begin
   bnx = n_elements(mag.b_long(*,0))   ; size of magnetogram
   bny = n_elements(mag.b_long(0,*))
endelse

; xt,yt are the position of the magnetogram pixels for contours and vectors
; after transformation to the coordinates of the image.
; The magnetic field data are drawn at the centers of the image pixels.

x = lindgen(bnx,bny) MOD bnx   ; magnetogram pixel coordinates
y = lindgen(bnx,bny) / bnx
xt=make_array(size=size([x]),/double)  ; zeroed initially
yt=make_array(size=size([y]),/double)
if NOT keyword_set(pp) then pp=unit_pp()
ppnew = pp

;ppnew(0,0,*) = ppnew(0,0,*)*zoomx+edge.left   ; Convert to device coords
;ppnew(1,0,*) = ppnew(1,0,*)*zoomy+edge.bottom
;ppnew(0,1,*) = ppnew(0,1,*)-0.5               ; pixel center
;ppnew(1,1,*) = ppnew(1,1,*)-0.5
;;;t = caltrans(ppinvert(ppnew))
;pp_reverse,ppnew,ppnewi,ttyp,ttypi
;t = caltrans(ppnewi,transtype=ttypi)

ppnew(0,0,*) = ppnew(0,0,*)+edge.left/zoomx   ; Add offset
ppnew(1,0,*) = ppnew(1,0,*)+edge.bottom/zoomy
ppnew(0,1,*) = ppnew(0,1,*)-0.5               ; pixel center
ppnew(1,1,*) = ppnew(1,1,*)-0.5
;;t = caltrans(ppinvert(ppnew))
pp_reverse,ppnew,ppnewi,ttyp,ttypi
t = caltrans(ppnewi,transtype=ttypi)
t(*,0) = t(*,0)*zoomx                         ; apply zoom factor
t(*,1) = t(*,1)*zoomy

p = reform(t(*,0),2,2)
q = reform(t(*,1),2,2)
for m=0,1 do begin      ; Convert x,y from magnetogram coordinates
   for n=0,1 do begin   ; to image coordinates.
      xm = x^m & yn = y^n
      xt = xt + p(n,m)*xm*yn   ; device coords of transformed data
      yt = yt + q(n,m)*xm*yn
   endfor
endfor
pq2rss,p,q,rotation,/quiet   ; Get the rotation which corresponds to p,q

; Compute the y position of the field and current labels

if !p.charsize EQ 0.0 then chsize = 1.0 else chsize = !p.charsize
ychsize = !d.y_ch_size * chsize
xchsize = !d.x_ch_size * chsize
if n_elements(btpos) NE 1 then btpos = 0
yout = btpos*zoomy+(edge.bottom-5.*ychsize)
xscl = 2

; Draw contours and contour key

if NOT keyword_set(nocontours) then begin

   if keyword_set(xyz) then begin
      contourdata = mag.bz
      ctype='Z'
   endif else begin
      contourdata = mag.b_long
      ctype='LOS'
   endelse

   contour,contourdata,xt,yt,xstyle=5,ystyle=5,$
           position=[edge.left,edge.bottom,$
                     edge.left+xsize*zoomx, $
                     edge.bottom+ysize*zoomy], $
           xrange=[edge.left,edge.left+(xsize)*zoomx], $
           yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
           levels=lv,c_color=pcolor,/device,/noerase,c_lab=0, $
           c_linestyle=[pltype],thick=pthick

   if NOT keyword_set(nowords) then begin
      xout = (!d.x_size-edge.right-edge.left)*0.2*chsize + edge.left
      plots,[0,xscl*xchsize]+xout,[yout,yout]+0.25*ychsize, $
            color=pcolor>1,linestyle=pltype,/device
      xyouts,xscl*xchsize+xout+xchsize,yout,'Pos. B!D'+ctype+'!N', $
            color=acolor,/device,charsize=chsize
   endif

   contour,-contourdata,xt,yt,xstyle=5,ystyle=5,$
           position=[edge.left,edge.bottom,$
                     edge.left+xsize*zoomx, $
                     edge.bottom+ysize*zoomy], $
           xrange=[edge.left,edge.left+(xsize)*zoomx], $
           yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
           levels=lv,c_color=mcolor,/device,/noerase,c_lab=0,$
           c_linestyle=[mltype],thick=mthick

   if NOT keyword_set(nowords) then begin
      xout = (!d.x_size-edge.right-edge.left)*0.40*chsize + edge.left
      plots,[0,xscl*xchsize]+xout,[yout,yout]+0.25*ychsize, $
            color=mcolor>1,linestyle=mltype,/device
      xyouts,xscl*xchsize+xout+xchsize,yout,'Neg. B!D'+ctype+'!N', $
             color=acolor,/device,charsize=chsize
   endif

   if keyword_set(latlon) then begin
      ; Contour interval for lat/lon
      if n_elements(llinterval) NE 1 then llinterval = 5.0
      REPEAT begin
         lllevels = findgen(long(180./llinterval+1L))*llinterval-90.
         latlevels = where(lllevels GE min(mag.latitude) AND $
                           lllevels LE max(mag.latitude))
         if latlevels(0) LT 0 then latlevels=[0.]
         lonlevels = where(lllevels GE min(mag.cmd) AND lllevels LE max(mag.cmd))
         if lonlevels(0) LT 0 then lonlevels=[0.]
         latclabel = intarr(n_elements(latlevels))+1
         lonclabel = intarr(n_elements(lonlevels))+1
         bad = n_elements(latclabel) GT 30 OR n_elements(lonclabel) GT 30
         if bad then begin
            llinterval = llinterval*2.
         endif
      endrep UNTIL NOT bad
      contour,mag.latitude,xt,yt,xstyle=5,ystyle=5,$
              position=[edge.left,edge.bottom,$
                        edge.left+xsize*zoomx, $
                        edge.bottom+ysize*zoomy], $
              xrange=[edge.left,edge.left+(xsize)*zoomx], $
              yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
              levels=lllevels(latlevels),c_color=llcolor,/device,/noerase, $
              c_lab=latclabel,c_linestyle=[llltype],thick=llthick
      contour,mag.cmd,xt,yt,xstyle=5,ystyle=5,$
              position=[edge.left,edge.bottom,$
                        edge.left+xsize*zoomx, $
                        edge.bottom+ysize*zoomy], $
              xrange=[edge.left,edge.left+(xsize)*zoomx], $
              yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
              levels=lllevels(lonlevels),c_color=llcolor,/device,/noerase, $
              c_lab=lonclabel,c_linestyle=[llltype],thick=llthick
   endif
endif

if keyword_set(jz) OR keyword_set(jt) then begin

   if keyword_set(jz) then begin
      contourdata=remapjz(mag)
      ctype = 'Z'
   endif else begin
      contourdata=remapjz(mag,/total)
      ctype='TOTAL'
   endelse

   contour,contourdata,xt,yt,xstyle=5,ystyle=5,$
           position=[edge.left,edge.bottom,$
                     edge.left+xsize*zoomx, $
                     edge.bottom+ysize*zoomy], $
           xrange=[edge.left,edge.left+(xsize)*zoomx], $
           yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
           levels=jlv,c_color=jpcolor,/device,/noerase,c_lab=0, $
           c_linestyle=[pltype],thick=jpthick

   if NOT keyword_set(nowords) then begin
      xout = (!d.x_size-edge.right-edge.left)*0.6*chsize + edge.left
      plots,[0,xscl*xchsize]+xout,[yout,yout]+0.25*ychsize, $
            color=jpcolor>1,linestyle=pltype,/device
      xyouts,xscl*xchsize+xout+xchsize,yout,'Pos. J!D'+ctype+'!N', $
             color=acolor,/device,charsize=chsize
   endif

   contour,-contourdata,xt,yt,xstyle=5,ystyle=5,$
           position=[edge.left,edge.bottom,$
                     edge.left+xsize*zoomx, $
                     edge.bottom+ysize*zoomy], $
           xrange=[edge.left,edge.left+(xsize)*zoomx], $
           yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
           levels=jlv,c_color=jmcolor,/device,/noerase,c_lab=0,$
           c_linestyle=[mltype],thick=jmthick

   if NOT keyword_set(nowords) then begin
      xout = (!d.x_size-edge.right-edge.left)*0.8*chsize + edge.left
      plots,[0,xscl*xchsize]+xout,[yout,yout]+0.25*ychsize, $
            color=jmcolor>1,linestyle=mltype,/device
      xyouts,xscl*xchsize+xout+xchsize,yout,'Neg. J!D'+ctype+'!N', $
            color=acolor,/device,charsize=chsize
   endif

endif

; Draw neutral line

if keyword_set(neutral) or n_elements(neutraldata) GT 0 then begin
   if n_elements(nthick) LE 0 then nthick=5
   if n_elements(nsmooth) LE 0 then nsmooth=0.5
   if n_elements(neutraldata) LE 0 then begin
      if keyword_set(xyz) then neutraldata=get_neutral(mag.bz,nsmooth) $   
      else neutraldata=get_neutral(mag.b_long,nsmooth)
   endif
   if NOT keyword_set(ppneutral) then ppneutral=pp
   contour,neutraldata,xt,yt,xstyle=5,ystyle=5,$
           position=[edge.left,edge.bottom,$
                     edge.left+xsize*zoomx, $
                     edge.bottom+ysize*zoomy], $
           xrange=[edge.left,edge.left+(xsize)*zoomx], $
           yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
           levels=[0.0],c_color=ncolor,/device,/noerase,c_lab=0,$
           c_linestyle=[0],thick=nthick
endif

; Draw border

if keyword_set(border) then begin
   borderdata = mag.b_long
   borderdata[*] = 1.
   borderdata[0,*] = 0.
   borderdata[n_elements(borderdata[*,0])-1L,*] = 0.
   borderdata[*,0] = 0.
   borderdata[*,n_elements(borderdata[0,*])-1L] = 0.
   if n_elements(bthick) LE 0 then bthick=2
   contour,borderdata,xt,yt,xstyle=5,ystyle=5,$
           position=[edge.left,edge.bottom,$
                     edge.left+xsize*zoomx, $
                     edge.bottom+ysize*zoomy], $
           xrange=[edge.left,edge.left+(xsize)*zoomx], $
           yrange=[edge.bottom,edge.bottom+(ysize)*zoomy], $
           levels=[0.01],c_color=bcolor,/device,/noerase,c_lab=0,$
           c_linestyle=[0],thick=bthick
endif

if n_elements(vertices) GT 0 then begin
   message,/info,'Field lines not fully implemented yet'
endif

; Draw vectors

if NOT keyword_set(novectors) then begin

   if NOT keyword_set(xyz) then begin
      u = -mag.b_trans*sin(mag.b_azim*!dtor)
      v = +mag.b_trans*cos(mag.b_azim*!dtor)
      angle = -rotation*!dtor
   endif else begin
      u = mag.bx
      v = mag.by
      if keyword_set(helio) then $
         angle = -rotation*!dtor $
      else $
         angle = mag.point.p - rotation*!dtor
   endelse

   if NOT keyword_set(quiet) then $
      message,/info,strcompress('Rotating vectors '+ $
                                string(angle*!radeg,format="(f15.1)")+ $
                                ' degrees')

   ; Since the magnetogram is rotated relative to the image, the y axis
   ; is not terrestrial north.  Thus, since the field azimuth is relative
   ; to T. North, the vectors must be rotated to the reference frame of 
   ; the image.  Further, if /xyz is set, Bx and By are relative to solar
   ; north rather than T. North, and the vectors must also be rotated by
   ; the solar p-angle, in this case.

   u1 = +u*cos(angle) - v*sin(angle)  ; Put vectors in the correct direction
   v1 = +u*sin(angle) + v*cos(angle)

   ; draw vectors

   if keyword_set(btskip) then begin
      su1 = size(u1)
      xxx = lindgen(su1[1],su1[2]) MOD su1[1]
      yyy = lindgen(su1[1],su1[2]) / su1[1]
      sss = where(xxx MOD long(btskip+1.5) NE 0. OR yyy MOD long(btskip+1.5) NE 0.)
      if sss[0] GE 0 then begin
         u1[sss] = 0.0  ; Not plotted since below missing
         v1[sss] = 0.0
      endif
      length = length * long(btskip+1.5)
   endif

   vectorfield,u1,v1,xt,yt,missing=missing>0.001,length=length, $
               maxT=btmax, color=vcolor, /noerase, $
               arrows=arrows, g1000=g1000

   ; Draw and label 1000-Gauss reference line segment for transverse field

   if NOT keyword_set(nowords) then begin
      xout = edge.left
      plots,[0,g1000]+xout,[yout,yout]+0.25*ychsize,color=acolor,/device
      xyouts, g1000+xout+0.5*xchsize, yout, '1000 G', color=acolor, $
              charsize=chsize,/device
   endif
endif

; Redraw axes for labels, titles etc.

xpos = indgen(xsize)*pixsize*float(xsize+1)/float(xsize)+x0
ypos = indgen(ysize)*pixsize*float(ysize+1)/float(ysize)+y0

xmin = min(xsize*pixsize, max=xmax)  ; Set axis ranges
xran = [xmin, xmax]
ymin = min(ysize*pixsize, max=ymax)
yran = [ymin, ymax]

if keyword_set(nolabels) then begin
   xtickname = replicate(' ',30)
   ytickname = replicate(' ',30)
endif else begin
   xtickname = ''
   ytickname = ''
endelse

contour,image,xpos,ypos, $
        xstyle=1+4*keyword_set(noaxes),ystyle=1+4*keyword_set(noaxes),$
        position=[edge.left,edge.bottom,$
                  edge.left+xsize*zoomx, $
                  edge.bottom+ysize*zoomy], $
        xrange=xran,yrange=yran,title=title,xtitle=xtitle,ytitle=ytitle, $
        /device,/noerase,/nodata,/noclip,color=acolor, $
        xtickname=xtickname,ytickname=ytickname
        


return
end

