
pro wbfits,data,filnam,header=inhead,comment=comment

;+
; NAME:
;        WBFITS
; PURPOSE:
;        Writes a magnetic field structure into a FITS file.
;        A simple FITS header may be created automatically, but it is better
;        to pass a more sophisticated header to be modified.
; CATEGORY:
;        Input/Output
; CALLING SEQUENCE:
;        wbfits, magnetic_structure,filename,header=header,comment=comment
; INPUTS:
;        magnetic_structure = data structure.
;        filename = string containing the file name.
; OPTIONAL (KEYWORD) INPUT PARAMETERS:
;        header = String array containing the header lines (to be modified)
;        comment = string. A FITS keyword parameter "COMMENT string" is added
;                   to the header.
; OUTPUTS:
;        No explicit outputs.
; COMMON BLOCKS:
;        None.
; SIDE EFFECTS:
;        A file is created, or overwritten.
; RESTRICTIONS:
;        The data is converted to integer data before writing the FITS file.
;        The magnetic data are scaled by a power of 10 to minimize the data 
;        loss.
; EXAMPLE:
;
;    IDL> mag = current('B910927.2238',header=header)
;    IDL> wbfits,mag,'file_name',header=header
;
; MODIFICATION HISTORY:
;        TRM, Oct, 1991.
;-


; open FITS file

get_lun,unit
openw,unit,filnam

bitpix = 16    ; The number of bits in the integer data

; create header if necessary

s=size(inhead)
createh=0
if s(n_elements(s)-2) EQ 0 then begin
   header = ['END']
   print,'Creating a simple header ...'
   createh=1
endif $
else header=inhead   ; Protect the input from change

; Check that there are no blank lines in the header

nhead = n_elements(header)
i = nhead-1
REPEAT begin
   bhead = byte(header(i))
   blank = where(bhead EQ 32B,count)    ; 32B is a space
   if count EQ n_elements(bhead) then begin
      temp = strarr(nhead-1)
      if i NE 0 then temp(0:i-1) = header(0:i-1)
      if i NE nhead-1 then temp(i:nhead-2) = header(i+1,nhead-1)
      header = temp
      nhead = nhead - 1
      i = nhead
   endif
   i = i - 1
endrep UNTIL i LT 0

; Update the header

; SIMPLE
check = wkey(header,'SIMPLE','LOGICAL','T')
if NOT check then check = wkey(header,'SIMPLE','LOGICAL','T',1,'','END')

; BITPIX
check = wkey(header,'BITPIX','INTEGER',bitpix)
if NOT check then check = wkey(header,'BITPIX','INTEGER',bitpix,1,'','END')

; NAXIS
check = wkey(header,'NAXIS','INTEGER',3)
if NOT check then check = wkey(header,'NAXIS','INTEGER',3,1, $
                               'NUMBER OF AXES','END')

nx = n_elements(data.b_long(*,0))
ny = n_elements(data.b_long(0,*))
ntotal = nx*ny

; NAXIS1
check = wkey(header,'NAXIS1','INTEGER',nx)
if NOT check then check = wkey(header,'NAXIS1','INTEGER',nx,1, $
                               '# OF POINTS IN RASTER FAST AXIS','END')
; NAXIS2
check = wkey(header,'NAXIS2','INTEGER',ny)
if NOT check then check = wkey(header,'NAXIS2','INTEGER',ny,1, $
                               '# OF POINTS IN RASTER SLOW AXIS','END')

ntags = n_tags(data)
tags = [-1]
for i=0,ntags-1 do begin
   s = size(data.(i))
   if s(0) EQ 2 then begin   ; If 2D array
      if s(n_elements(s)-1) EQ ntotal then tags = [tags,i]  ; If correct size
   endif
endfor
if n_elements(tags) LE 1 then begin
   print,'ERROR: wbfits: no array data in the data structure!'
   close, unit
   return
endif $
else begin
   tags = tags(1:(n_elements(tags)-1))   ; List of tag numbers to be output
endelse
ntags = n_elements(tags)

; NAXIS3
check = wkey(header,'NAXIS3','INTEGER',ntags)
if NOT check then check = wkey(header,'NAXIS3','INTEGER',ntags,1, $
                               '# OF STACKED IMAGES','END')

names = tag_names(data)   ; Get the tag names in the magnetic field structure

; Do the IMAGEx and ISCALEx keywords

for i=1,ntags do begin
   barray = iscale(data.(tags(i-1)),4,scale) ; This just gets scale ... wasteful
   name = strupcase(strcompress(names(tags(i-1)),/remove_all)) ; Tag name
   CASE name OF
      'B_LONG':  kcomment="LONGITUDINAL MAGNETIC FIELD"   ; Gets header comment
      'B_TRANS':  kcomment="TRANSVERSE MAGNETIC FIELD"
      'B_AZIM':  kcomment="AZIMUTH OF MAGNETIC FIELD"
      'B_TOTAL':  kcomment="TOTAL MAGNETIC FIELD"
      'BX':  kcomment="HORIZONTAL MAGNETIC FIELD"
      'BY':  kcomment="HORIZONTAL MAGNETIC FIELD"
      'BZ':  kcomment="VERTICAL MAGNETIC FIELD"
      'B_PXD':  kcomment="HORIZONTAL PIXEL POSITIONS"
      'B_PYD':  kcomment="HORIZONTAL PIXEL POSITIONS"
      'JX':  kcomment="HORIZ. ELECTRIC CURRENT (FORCE-FREE)"
      'JY':  kcomment="HORIZ. ELECTRIC CURRENT (FORCE-FREE)"
      'JZ':  kcomment="VERTICAL ELECTRIC CURRENT"
      'J_TOTAL':  kcomment="TOTAL FORCE-FREE CURRENT"
      'JXY_ALPHA': kcomment="FORCE-FREE ALPHA FOR JX,JY"
      'J_PXD':  kcomment="HORIZONTAL PIXEL POSITIONS"
      'J_PYD':  kcomment="HORIZONTAL PIXEL POSITIONS"
      'I_CONT':  kcomment="CONTINUUM INTENSITY"
      'I_CENT':  kcomment="LINE CENTRAL INTENSITY"
      'DOPPLER':  kcomment="DOPPLER WIDTH (mA)"
      'VELOCITY':  kcomment="VELOCITY"
      'LATITUDE':  kcomment="LATITUDE (DEG)"
      'CMD':  kcomment="CENTRAL MERIDEAN DISTANCE (DEG)"
      'B_FILL':  kcomment="MAGNETIC FILLING FACTOR"
      'UNNO': kcomment='1: L&S, 0: INTEGRAL REDUCTION'
      ELSE:  kcomment=""
   ENDCASE

   num_str = strcompress(string(i),/remove_all)

   check = wkey(header,'IMAGE'+num_str, 'CHAR',name)
   if NOT check then check = wkey(header,'IMAGE'+num_str,'CHAR',name,4, $
                                  kcomment,'COMMENT' )
   check = wkey(header,'ISCALE'+num_str,'REAL',scale )
   if NOT check then check = wkey(header,'ISCALE'+num_str,'REAL', $
                                scale,1,'SCALE FACTOR FOR IMAGE','CTYPE1' )
endfor

; Pointing information

point = tag_index(data,'POINT')
if point GE 0 then begin
   nptags = n_tags(data.(point))
   pnames = tag_names(data.(point))
   for i=0,nptags-1 do begin
      pname = strupcase(strcompress(pnames(i),/remove_all)) ; Tag name
      CASE pname OF
         'B0': begin
                  check = wkey(header,'B0','REAL',data.(point).(i)*!radeg)
                  if NOT check then check = wkey(header,'B0','REAL', $
                                                data.(point).(i)*!radeg,1, $
                                                'COMPUTED B0 ANGLES, DEGREES', $
                                                'END')
               end
         'P':  begin
                  check = wkey(header,'P','REAL',data.(point).(i)*!radeg)
                  if NOT check then check = wkey(header,'P','REAL', $
                                                data.(point).(i)*!radeg,1, $
                                                'COMPUTED P ANGLES, DEGREES', $
                                                'END')
               end
         'CMD':  begin
                  check = wkey(header,'POS-CMD','REAL',data.(point).(i)*!radeg)
                  if NOT check then check = wkey(header,'POS-CMD','REAL', $
                                                data.(point).(i)*!radeg,1, $
                                                'TELESCOPE MERIDIAN, DEG', $
                                                'END')
               end
         'LAT':  begin
                  check = wkey(header,'POS-LAT','REAL',data.(point).(i)*!radeg)
                  if NOT check then check = wkey(header,'POS-LAT','REAL', $
                                                data.(point).(i)*!radeg,1, $
                                                'TELESCOPE LATITUDE, DEG', $
                                                'END')
               end
         'RADIUS':  begin
                  check = wkey(header,'RADIUS','REAL',data.(point).(i))
                  if NOT check then check = wkey(header,'RADIUS','REAL', $
                                                data.(point).(i),1, $
                                                'SOLAR RADIUS, ARCSEC', $
                                                'END')
               end
         'PIX_SIZE':  begin
                  pix_size = data.(point).(i)
                  if n_elements(pix_size) EQ 1 then $
                     pix_size=[pix_size(0),pix_size(0)]
                  if inheader(header,'CDELT1') LT 0 then begin
                     ; Don't overwrite CDELT with ABS value if the correct
                     ; one already exists
                     check = wkey(header,'CDELT1','REAL',pix_size(0))
                     if NOT check then check = wkey(header,'CDELT1','REAL', $
                                                    pix_size(0),1, $
                                      'ABS OF PIXEL SPACING (ARCSEC): AXIS 1', $
                                                   'END')
                  endif
                  if inheader(header,'CDELT2') LT 0 then begin
                     check = wkey(header,'CDELT2','REAL',pix_size(1))
                     if NOT check then check = wkey(header,'CDELT2','REAL', $
                                                    pix_size(1),1, $
                                      'ABS OF PIXEL SPACING (ARCSEC): AXIS 2', $
                                                   'END')
                  endif
               end
      endcase
   endfor
endif

; COMMENT
for i=0,n_tags(data)-1L do begin
   if strupcase(strmid(strcompress(names(i),/remove_all),0,7)) EQ 'COMMENT' then begin
      sz = size(data.(i))
      if sz(sz(0)+1) EQ 7 then $
         check = wkey(header,'COMMENT','COMMENT',data.(i),1,' ','END')
   endif
endfor
if n_elements(comment) eq 1 then begin
   check = wkey(header,'COMMENT','COMMENT',comment,1,' ','END')
endif

; HISTORY
check = wkey(header,'HISTORY','COMMENT','wbfits: '+getenv('USER')+' '+systime(),1,' ','END')

; If a simple header is being created, add a comment about ISCALE values

if createh then begin
   check = wkey(header,'COMMENT','COMMENT', $
                'THE IMAGE VALUES MUST BE MULTIPLIED BY THE ISCALE VALUE', $
                1,' ','IMAGE1')
endif

; add blank lines to fill last FITS header record

for i=0,n_elements(header)-1 do begin
   hlen = strlen(header(i))
   if hlen LT 80 then begin
      for j=hlen,79 do header(i)=header(i)+' '
   endif $
   else if hlen GT 80 then begin
      header(i) = strmid(header(i),0,80)
   endif
endfor
i = 36 - (n_elements(header) mod 36)
blank = string(bytarr(80) + 32b)
if i NE 0 then begin
   for j=1,i do header=[header,blank]
endif

; write the stuff

writeu,unit,byte(header)

for i=0,ntags-1 do begin
   barray = iscale(data.(tags(i)),4,scale)
   writeu,unit,barray
endfor

; add null data to fill last FITS data record

i = ((ntotal*ntags)*(bitpix/8)) mod 2880
i = (2880-i) mod 2880
if i gt 0 then begin
   datb = bytarr(i)
   writeu,unit,datb
endif

close,unit
free_lun,unit
return
end

