
function wkey,header,keyword,ktype,value,num,kcomment,addbefore

;+
; NAME:
;        wkey
; PURPOSE:
;        Writes (or overwrites) a FITS value into a FITS header.
; CATEGORY:
;        FITS
; CALLING SEQUENCE:
;        result = wkey(header,keyword,ktype,value,num,kcomment,addbefore)
; INPUTS:
;        header = string array containing lines of FITS header
;	 keyword = string containing the keyword to write
;        ktype   = string containing type of keyword 
;		   (INTEGER,REAL,CHAR,LOGICAL,COMMENT)
;        value = value to be put into the FITS header
;        num = Index for multiple keywords like HISTORY and COMMENT, eg if
;              num is 2, add before the second occurance of the keyword.
;        kcomment = string containing the comment for an added header line
;        addbefore = string containing the keyword before which the input
;                   keyword is written.  If addbefore is NOT included,
;                   the routine will attempt to overwrite an existing keyword.
; OPTIONAL (KEYWORD) INPUT PARAMETERS:
;	 None
; OUTPUTS:
;        result = 1 if successful, 0 if not successful
; OPTIONAL (KEYWORD) OUTPUT PARAMETERS:
;        None
; COMMON BLOCKS:
;        None.
; SIDE EFFECTS:
;        None.
; RESTRICTIONS:
;        None.
; MODIFICATION HISTORY:
;        TRM, Oct, 1991
;-


; Create the header if it doesn't exist

s = size(header)
if s(n_elements(s)-2) EQ 0 then begin
   header=[string(replicate(32B,80))]
   strput,header,'END',0
   print,'wkey: Creating header ...'
endif

; The last keyword must be END.  If it's not, add END to tail of header.
; If, on the other hand END is not the last entry but is in the header, there
; is extraneous stuff in header.

nhead = n_elements(header)

end_index = inheader(header,'END')
if end_index GE 0 then begin
   if end_index NE nhead-1  then begin  ; There is junk after END
      print,'WARNING: wkey: Ignoring extraneous header entries after "END"'
      header = header(0:end_index)
      nhead = n_elements(header)
   endif
endif $
else begin    ; There is no "END" keyword in the header - add one
   temp = strarr(nhead+1)
   temp(0:nhead-1) = header(*)
   temp1=string(replicate(32B,80))
   strput,temp1,'END',0
   temp(nhead) = temp1
   header = temp
   nhead = nhead + 1
   print,'Warning: wkey: Added END to tail of header. Tail = ', $
        strmid(header(nhead-1),0,8)
endelse

if n_elements(num) EQ 0 then num = 1

count = 0
found = 0

; Set up keyword for search, kcomment for the update of the header
s = size(keyword)
if s(n_elements(s)-2) NE 7 then begin
   print,'ERROR: wkey: keyword must be a string'
   return,0
endif
s = size(kcomment)
if s(n_elements(s)-2) NE 7 then kcomment=' '
if strlen(kcomment) GT 38 then kcomment=strmid(kcomment,0,38)

; Make sure that keyword has length 8 as required by the FITS standard

if strlen(keyword) LT 8 then begin
  for i=1,8-strlen(keyword) do keyword=keyword+' '
endif $
else if strlen(keyword) GT 8 then keyword=strmid(keyword,0,8)

; If addbefore exists, then add a line to the FITS header.  Otherwise search
; for a line in the header to update

s = size(addbefore)
if s(n_elements(s)-2) EQ 7 then begin
   search=strupcase(strcompress(addbefore,/remove_all))
   add = 1
endif $
else begin
   search=strupcase(strcompress(keyword,/remove_all))
   add = 0
endelse

; Loop through the header 

for i=0,nhead-1 do begin

   ; ikey is the current keyword 

   ikey = strupcase(strcompress(strmid(header(i),0,8),/remove_all))

   ; Are we at the end of the header?  If so, fitsend = true

   if ((strpos('END',ikey,0) GE 0) AND (strlen(ikey) EQ 3)) then fitsend=1 $
   else fitsend = 0

   ; Have we found the keyword we are looking for?

   if ((strpos(search,ikey,0) EQ 0) AND (strlen(search) EQ strlen(ikey))) OR $
      (fitsend) then begin

     count = count + 1   ; Keep track of the number of times keyword is found

     ; If the search was for a header line to update, but it wasn't found,
     ; return 0 to indicate failure.

     if fitsend AND  $
        ((strpos('END',search) LT 0) OR (strlen(search) NE 3)) AND $
        (add EQ 0) AND (NOT found) then begin
        ;print,'WARNING: wkey: Couldnt find ',search
        return,0
     endif

     ; if we found the right place, proceed with the update or addition to 
     ; the header.

     if (num EQ count) OR (fitsend AND (NOT found) AND (add EQ 1)) then begin

        ; Update (add=0) or add to the header (add=1) depending on what
        ; type of data we are dealing with.

        CASE strcompress(ktype,/remove_all) OF
        'INTEGER': begin
           if add then begin
              temp=strarr(nhead+1)
              if i NE 0 then temp(0:i-1)=header(0:i-1)
              temp1=string(replicate(32B,80))
              strput,temp1,keyword,0
              strput,temp1,'=',8
              strput,temp1,'/',40
              strput,temp1,kcomment,42
              strput,temp1,string(value,format="(i20)"),10
              temp(i) = temp1
              temp(i+1:nhead)=header(i:nhead-1)
              header=temp
           endif $
           else begin
              temp = header(i)
              strput,temp,string(value,format="(i20)"),10
              header(i) = temp
              hlen = strlen(header(i))
              if hlen LT 80 then begin
                 for j=hlen,79 do header(i)=header(i)+' '
              endif $
              else if hlen GT 80 then begin
                 header(i) = strmid(header(i),0,80)
              endif
           endelse
           found = 1
         end
        'REAL': begin
           if add then begin
              temp=strarr(nhead+1)
              if i NE 0 then temp(0:i-1)=header(0:i-1)
              temp1=string(replicate(32B,80))
              strput,temp1,keyword,0
              strput,temp1,'=',8
              strput,temp1,'/',40
              strput,temp1,kcomment,42
              strput,temp1,string(value,format="(f20.4)"),10
              temp(i) = temp1
              temp(i+1:nhead)=header(i:nhead-1)
              header=temp
           endif $
           else begin
              temp = header(i)
              strput,temp,string(value,format="(f20.4)"),10
              header(i) = temp
              hlen = strlen(header(i))
              if hlen LT 80 then begin
                 for j=hlen,79 do header(i)=header(i)+' '
              endif $
              else if hlen GT 80 then begin
                 header(i) = strmid(header(i),0,80)
              endif
           endelse
           found=1
         end
        'CHAR':  begin
           if strlen(value) LT 8 then begin  ;CHAR's length is supposed to >= 8
              for j=1,8-strlen(value) do begin 
                 value=value+' ' 
              endfor
           endif
           if add then begin
              temp=strarr(nhead+1)
              if i NE 0 then temp(0:i-1)=header(0:i-1)
              temp1=string(replicate(32B,80))
              strput,temp1,keyword,0
              strput,temp1,'=',8
              strput,temp1,'/',40
              strput,temp1,kcomment,42
              if strlen(value) GT 18 then value=strmid(value,0,18)
              strput,temp1,"'"+value+"'",10
              temp(i) = temp1
              temp(i+1:nhead)=header(i:nhead-1)
              header=temp
           endif $
           else begin
              temp = header(i)
              temp1=string(replicate(32B,20))
              strput,temp,string(temp1,format="(a20)"),10
              if strlen(value) GT 18 then value=strmid(value,0,18)
              strput,temp,"'"+value+"'",10
              header(i) = temp
              hlen = strlen(header(i))
              if hlen LT 80 then begin
                 for j=hlen,79 do header(i)=header(i)+' '
              endif $
              else if hlen GT 80 then begin
                 header(i) = strmid(header(i),0,80)
              endif
           endelse
           found=1
         end
        'LOGICAL': begin
           CASE 1 OF
           (value EQ 'T') OR (value EQ 'F'): begin
              if add then begin
                 temp=strarr(nhead+1)
                 if i NE 0 then temp(0:i-1)=header(0:i-1)
                 temp1=string(replicate(32B,80))
                 strput,temp1,keyword,0
                 strput,temp1,'=',8
                 strput,temp1,'/',40
                 strput,temp1,kcomment,42
                 strput,temp1,string(value,format="(a20)"),10
                 temp(i) = temp1
                 temp(i+1:nhead)=header(i:nhead-1)
                 header=temp
              endif $
              else begin
                 temp = header(i)
                 strput,temp,string(value,format="(a20)"),10
                 header(i) = temp
                 hlen = strlen(header(i))
                 if hlen LT 80 then begin
                    for j=hlen,79 do header(i)=header(i)+' '
                 endif $
                 else if hlen GT 80 then begin
                    header(i) = strmid(header(i),0,80)
                 endif
              endelse
            end
           ELSE: print,'WARNING: wkey: Invalid LOGICAL encountered',value
           ENDCASE
           found=1
         end
        'COMMENT': begin
           if add then begin
              temp=strarr(nhead+1)
              if i NE 0 then temp(0:i-1)=header(0:i-1)
              temp1=string(replicate(32B,80))
              strput,temp1,keyword,0
              strput,temp1,'=',8
              if strlen(value) GT 70 then value=strmid(value,0,70)
              strput,temp1,value,10
              temp(i) = temp1
              temp(i+1:nhead)=header(i:nhead-1)
              header=temp
           endif $
           else begin
              temp = header(i)
              temp1=string(replicate(32B,70))
              strput,temp,string(temp1,format="(a70)"),10
              strput,temp,value,10
              header(i) = temp
              hlen = strlen(header(i))
              if hlen LT 80 then begin
                 for j=hlen,79 do header(i)=header(i)+' '
              endif $
              else if hlen GT 80 then begin
                 header(i) = strmid(header(i),0,80)
              endif
           endelse
           found=1
         end
        ELSE: begin
            print,'ERROR: wkey: Invalid keyword type encountered',ktype
            return,0
         end
        endcase
     endif
   endif
endfor

if (NOT found) then print,'WARNING: wkey: keyword ',search,' not found'

return,found

end

